#include "dump.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

typedef struct _SourceInfo SourceInfo;
typedef struct _DumpInfo DumpInfo;


struct _SourceInfo {
	int source_id;
	int num_args;
	GdamArg* args;
	char* name;
	char* type;
	DumpInfo* dump_info;
	int num_children;
	int* children;
};

typedef void (*DumpFunc)(SourceInfo* source_info);

struct _DumpInfo {
	GdamChannel* channel;
	CliInfo* cli_info;
	GSList* pending_source_ids;
	gboolean* mark_when_done;
	SourceInfo* root;
	GHashTable* source_infos;
	GHashTable* type_counts;
	DumpFunc dump_func;
	gboolean only_vital;
};

static const char* print_script(DumpInfo* info, int source_id);
void dump(DumpInfo* info, int source_id);


static int  alloc_type_id(DumpInfo* info, const char* name)
{
  gpointer rv = g_hash_table_lookup (info->type_counts, name);
  gpointer inc;
  int id;
  if (rv == 0)
    name = g_strdup (name);
  id = GPOINTER_TO_INT (rv) + 1;
  inc = GINT_TO_POINTER (id);
  g_hash_table_insert (info->type_counts, (gpointer) name, (gpointer) inc);
  return id;
}

typedef void (*PrintFunc)(SourceInfo* source_info, GdamArg* arg);

typedef struct _SpecialPrinter SpecialPrinter;
struct _SpecialPrinter {
	const char* type;
	const char* argname;
	PrintFunc printer;
};

static void print_subsource(SourceInfo* info, GdamArg* arg)
{
  const char *name = print_script (info->dump_info, GDAM_ARG_INT (arg));
  printf ("set $%s %s $%s\n", info->name, arg->name, name);
}
static void print_time_source_array(SourceInfo* source_info, GdamArg* arg)
{
  gint64 *array = GDAM_ARG_INT64_ARRAY (arg).array;
  int length = GDAM_ARG_INT64_ARRAY (arg).length;
  DumpInfo *dump_info = source_info->dump_info;
  int i;
  GString *string;
  string = g_string_new ("");
  g_string_sprintf (string, "set $%s %s [", source_info->name, arg->name);
  for (i = 0; i < length; i++)
    {
      if (i % 2 == 0)
	{
	  g_string_sprintfa (string, "%lldLL ", array[i]);
	}
      else
	{
	  int source_id = (int) array[i];
	  SourceInfo *sub_info;
	  const char *name;
	  sub_info = g_hash_table_lookup
	    (dump_info->source_infos, GINT_TO_POINTER (source_id));
	  print_script (dump_info, source_id);
	  g_return_if_fail (sub_info != NULL);
	  name = sub_info->name;
	  if (i < length - 1)
	    g_string_sprintfa (string, "$%s ", name);
	  else
	    g_string_sprintfa (string, "$%s", name);
	}
    }
  if (length > 0)
    g_string_sprintfa (string, "]");
  else
    g_string_sprintfa (string, "]LL");
  printf ("%s\n", string->str);
  g_string_free (string, TRUE);
}

static SpecialPrinter printer_cases[] = {
	{ NULL, "filtered", print_subsource },
	{ NULL, "vibrato", print_subsource },
	{ NULL, "modulator", print_subsource },
	{ "GdamMix", "mixed", print_time_source_array },
};

#define NUM_PRINTER_CASES (sizeof(printer_cases)/sizeof(printer_cases[0]))

static void print_child_arg(SourceInfo* info, GdamArg* arg)
{
  char *value;
  guint i;
  for (i = 0; i < NUM_PRINTER_CASES; i++)
    {
      if (printer_cases[i].type != NULL
	  && strcmp (printer_cases[i].type, info->type) != 0)
	continue;
      if (strcmp (printer_cases[i].argname, arg->name) != 0)
	continue;
      (*printer_cases[i].printer) (info, arg);
      return;
    }

  value = gdam_arg_val_as_string (arg);
  printf ("set $%s %s %s\n", info->name, arg->name, value);
  g_free (value);
}


static char* destudly_cap(const char* name)
{
  const char *at;
  int length = 0;
  char *rv;
  char *out;
  if (memcmp (name, "Gdam", 4) == 0)
    name += 4;
  for (at = name; *at; at++)
    {
      if (at != name && isupper (*at))
	length += 2;
      else
	length += 1;
    }
  rv = g_new (char, length + 1);
  out = rv;
  for (at = name; *at; at++)
    {
      if (at != name && isupper (*at))
	*out++ = '_';
      *out++ = tolower (*at);
    }
  *out = 0;
  return rv;
}

static void alloc_name(SourceInfo* source)
{
  int type_seqno;
  char *destudly_capped;
  type_seqno = alloc_type_id (source->dump_info, source->type);
  destudly_capped = destudly_cap (source->type);
  source->name = g_strdup_printf ("%s%d", destudly_capped, type_seqno);
  g_free (destudly_capped);
}
static const char* print_script(DumpInfo* info, int source_id)
{
  SourceInfo *source;
  int i;
  source = g_hash_table_lookup (info->source_infos,
				GINT_TO_POINTER (source_id));
  if (source->name != NULL)
    return source->name;
  alloc_name (source);
  printf ("source %s %s\n", source->name, source->type);
  for (i = 0; i < source->num_args; i++)
    print_child_arg (source, source->args + i);
  return source->name;
}
static void do_dump(SourceInfo* source_info)
{
  print_script (source_info->dump_info, source_info->source_id);
}

/*
 * ls -R
 */
static void indent(int depth)
{
  while (depth-- > 0)
    fputc (' ', stdout);
}
static const char* ls_R(DumpInfo* info, int source_id, int depth)
{
  SourceInfo *source;
  int i;
  source = g_hash_table_lookup (info->source_infos,
				GINT_TO_POINTER (source_id));
  if (source->name != NULL)
    return source->name;
  alloc_name (source);
  indent (depth);
  printf ("source %d, %s\n", source_id, source->type);
  for (i = 0; i < source->num_args; i++)
    {
      char *arg = gdam_arg_val_as_string (source->args + i);
      indent (depth + 2);
      printf ("%d: %s => %s\n", i, source->args[i].name, arg);
    }
  for (i = 0; i < source->num_children; i++)
    {
      ls_R (info, source->children[i], depth + 4);
    }
  return source->name;
}
static void do_ls_R(SourceInfo* source_info)
{
  ls_R (source_info->dump_info, source_info->source_id, 0);
}

static void done_getting_data(DumpInfo* dump_info)
{
  if (dump_info->root == NULL)
    g_message ("no root node");
  else
    {
      (*dump_info->dump_func) (dump_info->root);
    }
}

static void
on_dump_get_type_arg(int num_args, GdamArg* args, gpointer user_data)
{
  SourceInfo *source_info = user_data;
  const char *type;
  g_assert (num_args == 1);
  type = GDAM_ARG_STRING (args);
  source_info->type = g_strdup (type);
}

static void
on_dump_didnt_get_type_arg(gpointer user_data)
{
  g_assert_not_reached ();
}

static void
on_dump_get_subsources_arg(int num_args, GdamArg* args, gpointer user_data)
{
  SourceInfo *source_info = user_data;
  int *subsources;
  int num, i;
  g_assert (num_args == 1);
  subsources = GDAM_ARG_INT_ARRAY (args).array;
  num = GDAM_ARG_INT_ARRAY (args).length;
  source_info->num_children = num;
  source_info->children = g_memdup (subsources, sizeof (int) * num);
  for (i = 0; i < num; i++)
    dump (source_info->dump_info, subsources[i]);
}

static void
on_dump_didnt_get_subsources_arg(gpointer user_data)
{
  g_assert_not_reached ();
}

static void on_dump_get_args(int num_args, GdamArg* args, gpointer user_data)
{
  SourceInfo *source_info = user_data;
  DumpInfo *dump_info = source_info->dump_info;
  source_info->num_args = num_args;
  source_info->args = gdam_arg_copy_array (num_args, args);
  dump_info->pending_source_ids
    = g_slist_remove (dump_info->pending_source_ids, source_info);
  //g_message("got %d arguments to %s", num_args, source_info->type);
  if (dump_info->pending_source_ids == NULL)
    {
      done_getting_data (dump_info);
      if (dump_info->mark_when_done != NULL)
	*dump_info->mark_when_done = TRUE;
    }
}
static void on_dump_get_args_failed(gpointer user_data)
{
  SourceInfo *source_info = user_data;
  DumpInfo *dump_info = source_info->dump_info;
  g_warning ("failed to get args for %d", source_info->source_id);
  dump_info->pending_source_ids
    = g_slist_remove (dump_info->pending_source_ids, source_info);
  if (dump_info->pending_source_ids == NULL)
    {
      done_getting_data (dump_info);
      if (dump_info->mark_when_done != NULL)
	*dump_info->mark_when_done = TRUE;
    }
}

void dump(DumpInfo* info, int source_id)
{
  SourceInfo *source_info;
  const char *type_string = "type";
  const char *subsources_string = "subsources";
  source_info = g_new (SourceInfo, 1);
  source_info->source_id = source_id;
  source_info->num_args = 0;
  source_info->args = 0;
  source_info->name = NULL;
  source_info->type = NULL;
  source_info->dump_info = info;
  info->pending_source_ids
    = g_slist_prepend (info->pending_source_ids, source_info);
  g_hash_table_insert (info->source_infos,
		       GINT_TO_POINTER (source_id), source_info);
  if (info->root == NULL)
    info->root = source_info;
  gdam_channel_append_request
    (info->channel,
     gdam_request_get_args
     (source_id,
      1, &type_string,
      on_dump_get_type_arg, on_dump_didnt_get_type_arg, source_info));
  gdam_channel_append_request
    (info->channel,
     gdam_request_get_args
     (source_id,
      1, &subsources_string,
      on_dump_get_subsources_arg,
      on_dump_didnt_get_subsources_arg, source_info));
  //g_message("appending request for %d", source_id);
  gdam_channel_append_request
    (info->channel,
     gdam_request_get_all_args
     (source_id,
      TRUE, on_dump_get_args, on_dump_get_args_failed, source_info));
}


static DumpInfo* dump_info_add(CliInfo* cli_info,
                               int      source_id,
			       DumpFunc do_dump_func,
			       gboolean only_vital)
{
  DumpInfo *dump_info;

  dump_info = g_new (DumpInfo, 1);
  dump_info->channel = cli_info->channel;
  dump_info->cli_info = cli_info;
  dump_info->source_infos = g_hash_table_new (NULL, NULL);
  dump_info->type_counts = g_hash_table_new (g_str_hash, g_str_equal);

  dump_info->pending_source_ids = NULL;
  dump_info->mark_when_done = cli_info->is_done_processing;
  dump_info->root = NULL;
  dump_info->only_vital = only_vital;
  dump (dump_info, source_id);
  *dump_info->mark_when_done = FALSE;
  dump_info->dump_func = do_dump_func;
  return dump_info;
}
void cli_do_ls_R(CliInfo* info, int source_id, gboolean only_vital)
{
  DumpInfo *dump_info;
  dump_info = dump_info_add (info, source_id, do_ls_R, only_vital);
}
gboolean handle_dump_command(LineTokenizer* tokenizer,
                             CliInfo*       info,
		             GdamRequest**  request)
{
  const char *num = NULL;
  int source_id;
  while (num == NULL)
    {
      char *arg = line_tokenizer_get_word (tokenizer);
      if (arg == NULL)
	{
	  fprintf (stderr, "no source id found...\n");
	  return FALSE;
	}
      num = arg;
    }
  *request = NULL;
  source_id = (int) strtol (num, NULL, 10);
  dump_info_add (info, source_id, do_dump, TRUE);
  return TRUE;
}
