/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "linetokenizer.h"
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <common/gdamutil.h>




LineTokenizer* line_tokenizer_new        (const char*      line)
{
  LineTokenizer *rv = g_new (LineTokenizer, 1);
  rv->line = line;
  rv->at = line;
  return rv;
}

#define SKIP_WHITESPACE(ptr)				\
	while (*(ptr) && isspace(*(ptr))) (ptr)++
#define SKIP_NON_WHITESPACE(ptr)			\
	while (*(ptr) && !isspace(*(ptr))) (ptr)++

char*          line_tokenizer_get_word   (LineTokenizer*   tokenizer)
{
  const char *at = tokenizer->at;
  const char *start;
  char *rv;
  SKIP_WHITESPACE (at);
  start = at;
  SKIP_NON_WHITESPACE (at);
  if (start == at)
    {
      tokenizer->at = at;
      return NULL;
    }
  rv = g_new (char, at - start + 1);
  memcpy (rv, start, at - start);
  rv[at - start] = 0;
  tokenizer->at = at;
  if (rv[0] == '$' && tokenizer->variables)
    {
      char *var;
      var = (char *) g_hash_table_lookup (tokenizer->variables, rv + 1);
      if (var)
	{
	  g_free (rv);
	  rv = g_strdup (var);
	}
    }
  return rv;
}
/* Parse a double-quoted string, starting at the first double-quote. */
static char*    parse_double_quoted(LineTokenizer* tokenizer)
{
  /* TODO: escape sequences */
  const char *at = tokenizer->at;
  int num_chars = 0;
  char *rv;
  g_assert (*at == '"');
  at++;
  while (*at != '\0' && *at != '"')
    {
      if (*at == '\\')
	{
	  at += 2;
	  num_chars++;
	}
      else
	{
	  at++;
	  num_chars++;
	}
    }
  rv = g_new (char, num_chars + 1);
  at = tokenizer->at;
  num_chars = 0;
  at++;
  while (*at != '\0' && *at != '"')
    {
      if (*at == '\\')
	{
	  at++;
	  if (*at == '\\')
	    {
	      rv[num_chars++] = '\\';
	      at++;
	    }
	  else if (*at == '"')
	    {
	      rv[num_chars++] = '"';
	      at++;
	    }
	  else
	    {
	      rv[num_chars++] = *at++;
	    }
	}
      else
	{
	  rv[num_chars++] = *at++;
	}
    }
  rv[num_chars] = '\0';
  if (*at == '"')
    at++;
  tokenizer->at = at;
  return rv;
}

static gboolean parse_number_arg_value(LineTokenizer* tokenizer, GdamArg* arg)
{
  char *word = line_tokenizer_get_word (tokenizer);
  char *endp;
  if (strchr (word, '.'))
    {
      arg->type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (arg) = strtod (word, &endp);
      if (endp == word)
	{
	  gdam_debug ("error parsing double (in line tokenizer)");
	  g_free (word);
	  return FALSE;
	}
    }
  else if (strchr (word, 'L'))
    {
      arg->type = GDAM_ARG_TYPE_INT64;
      GDAM_ARG_INT64 (arg) = gdam_strtoll (word, &endp, 10);
      if (endp == word || (*endp && *endp != 'L'))
	{
	  gdam_debug ("error parsing int64 (in line tokenizer)");
	  g_free (word);
	  return FALSE;
	}
    }
  else
    {
      arg->type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (arg) = (int) strtol (word, &endp, 10);
      if (endp == word)
	{
	  gdam_debug ("error parsing int (in line tokenizer)");
	  g_free (word);
	  return FALSE;
	}
    }
  g_free (word);
  return TRUE;
}
static void tokenizer_skip_whitespace(LineTokenizer* tokenizer)
{
  const char *at = tokenizer->at;
  SKIP_WHITESPACE (at);
  tokenizer->at = at;
}

	

/* Starting at the first double-quote, parse the string array. */
static gboolean parse_string_array(LineTokenizer* tokenizer, GdamArg* arg)
{
  int num_str = 0, max_str = 16;
  char **str_array;
  str_array = g_new (char *, max_str);
  while (*tokenizer->at && *tokenizer->at != ']')
    {
      char *str;
      if (*tokenizer->at != '"')
	{
	  g_warning ("unexpected character %c in string array"
		     " [index %d]", *tokenizer->at, num_str);
	  goto fail;
	}
      str = parse_double_quoted (tokenizer);
      tokenizer_skip_whitespace (tokenizer);
      if (!str)
	{
	  g_warning ("error parsing substring %d", num_str);
	  goto fail;
	}
      if (num_str == max_str)
	{
	  max_str *= 2;
	  str_array = g_renew (char *, str_array, max_str);
	}
      str_array[num_str++] = str;
    }
  if (num_str < max_str)
    str_array = g_renew (char *, str_array, num_str);
  arg->type = GDAM_ARG_TYPE_STRING_ARRAY;
  GDAM_ARG_STRING_ARRAY (arg).array = str_array;
  GDAM_ARG_STRING_ARRAY (arg).length = num_str;
  return TRUE;

fail:
  /* Clean up allocated memory and return FALSE. */
  max_str = num_str;
  for (num_str = 0; num_str < max_str; num_str++)
    g_free (str_array[num_str]);
  g_free (str_array);
  return FALSE;
}

static void int64_copy(void* out, const GdamArg* in)
{
  g_assert (in->type == GDAM_ARG_TYPE_INT64);
  memcpy (out, &GDAM_ARG_INT64 (in), sizeof (gint64));
}
static void int_copy(void* out, const GdamArg* in)
{
  g_assert (in->type == GDAM_ARG_TYPE_INT);
  memcpy (out, &GDAM_ARG_INT (in), sizeof (gint));
}
static void double_copy(void* out, const GdamArg* in)
{
  g_assert (in->type == GDAM_ARG_TYPE_DOUBLE);
  memcpy (out, &GDAM_ARG_DOUBLE (in), sizeof (gdouble));
}


/* Starting at the first number, parse the number array. */
static gboolean parse_number_array(LineTokenizer* tokenizer, GdamArg* arg)
{
  GdamArg first;
  void (*copier_func) (void *, const GdamArg *) = NULL;
  void *buf;
  int size = 0;
  GArray *array;
  if (!parse_number_arg_value (tokenizer, &first))
    return FALSE;
  switch (first.type)
    {
    case GDAM_ARG_TYPE_INT:
      arg->type = GDAM_ARG_TYPE_INT_ARRAY;
      copier_func = int_copy;
      size = sizeof (gint);
      break;
    case GDAM_ARG_TYPE_DOUBLE:
      arg->type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
      copier_func = double_copy;
      size = sizeof (gdouble);
      break;
    case GDAM_ARG_TYPE_INT64:
      arg->type = GDAM_ARG_TYPE_INT64_ARRAY;
      copier_func = int64_copy;
      size = sizeof (gint64);
      break;
    default:
      g_assert_not_reached ();
    }
  array = g_array_new (FALSE, FALSE, size);
  buf = alloca (size);
  (*copier_func) (buf, &first);
  array = g_array_append_vals (array, buf, 1);
  tokenizer_skip_whitespace (tokenizer);
  while (*tokenizer->at != '\0' && *tokenizer->at != ']')
    {
      GdamArg next_arg;
      if (!parse_number_arg_value (tokenizer, &next_arg))
	{
	  g_array_free (array, TRUE);
	  return FALSE;
	}
      if (next_arg.type != first.type)
	{
	  g_warning ("type mismatch parsing array");
	  g_array_free (array, TRUE);
	  return FALSE;
	}
      (*copier_func) (buf, &next_arg);
      array = g_array_append_vals (array, buf, 1);
      tokenizer_skip_whitespace (tokenizer);
    }
	/** XXX:  this assumes that int_array/int64_array etc 
	 **       are all the same order etc... */
  GDAM_ARG_INT_ARRAY (arg).array = (int *) (array->data);
  GDAM_ARG_INT_ARRAY (arg).length = array->len;
  g_array_free (array, FALSE);
  return TRUE;
}

/* Parse any kind of argument, filling in its type, name, and value. */
gboolean       line_tokenizer_tokenize   (LineTokenizer*   tokenizer,
					  ArgHints*        arg_hints,
                                          GdamArg*         arg)
{
  const char *at;
  int type_hint;
  arg->name = line_tokenizer_get_word (tokenizer);
  if (!arg->name)
    return FALSE;
  tokenizer_skip_whitespace (tokenizer);
  at = tokenizer->at;

  if (arg_hints == NULL)
    type_hint = 0;
  else
    type_hint = arg_hints_lookup (arg_hints, arg->name);

  if (type_hint != 0)
    {
      int n_chars;
      arg->type = type_hint;
      n_chars = gdam_arg_parse_val_from_string (arg, at);
      if (n_chars <= 0)
	{
	  return FALSE;
	}
      tokenizer->at += n_chars;
      return TRUE;
    }

  /* Get value. */
  if (*at == '[')
    {
      at++;
      SKIP_WHITESPACE (at);
      if (*at == '"')
	{
	  /* String array. */
	  tokenizer->at = at;
	  return parse_string_array (tokenizer, arg);
	}
      else if (isdigit (*at) || *at == '.' || *at == '+' || *at == '-')
	{
	  /* Double/int/int64 array. */
	  tokenizer->at = at;
	  return parse_number_array (tokenizer, arg);
	}
    }
  else if (isdigit (*at) || *at == '.' || *at == '+' || *at == '-')
    {
      if (!parse_number_arg_value (tokenizer, arg))
	{
	  g_free (arg->name);
	  arg->name = NULL;
	  return FALSE;
	}
    }
  else if (*at == '"')
    {
      char *str = parse_double_quoted (tokenizer);
      if (!str)
	{
	  g_free (arg->name);
	  arg->name = NULL;
	  return FALSE;
	}
      arg->type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (arg) = str;
    }
  else if (*at != '\0')
    {
      char *str = line_tokenizer_get_word (tokenizer);
      arg->type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (arg) = str;
    }
  else
    {
      g_warning ("expected argument, not `%c'", *at);
      return FALSE;
    }
  return TRUE;
}

void           line_tokenizer_destroy    (LineTokenizer*   tokenizer)
{
  g_free (tokenizer);
}
