#include "gdamarg.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "gdamutil.h"

/*
 * Implementation of gdam_arg_get_default_validator.
 */

typedef struct _DoubleValidator DoubleValidator;
struct _DoubleValidator {
	gboolean		has_min_value;
	gboolean		has_max_value;
	gdouble			min_value;
	gdouble			max_value;
};

static gboolean double_validator(GtkObject*    object,
                                 GdamArg*      arg,
				 gpointer      v_data)
{
  DoubleValidator *validation_data = v_data;
  gdouble value;
  g_return_val_if_fail (arg->type == GDAM_ARG_TYPE_DOUBLE, FALSE);
  value = GDAM_ARG_DOUBLE (arg);
  if (!finite (value))
    return FALSE;
  if (validation_data->has_min_value)
    if (value < validation_data->min_value)
      return FALSE;
  if (validation_data->has_max_value)
    if (value > validation_data->max_value)
      return FALSE;
  return TRUE;
}

static gboolean double_array_validator(GtkObject*    object,
                                       GdamArg*      arg,
				       gpointer      v_data)
{
  DoubleValidator *validation_data = v_data;
  gdouble *values;
  int num_values;
  int i;
  g_return_val_if_fail (arg->type == GDAM_ARG_TYPE_DOUBLE_ARRAY, FALSE);
  values = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  num_values = GDAM_ARG_DOUBLE_ARRAY (arg).length;
  for (i = 0; i < num_values; i++)
    {
      gdouble value = values[i];
      if (!finite (value))
	return FALSE;
    }
  if (validation_data->has_min_value)
    for (i = 0; i < num_values; i++)
      if (values[i] < validation_data->min_value)
	return FALSE;
  if (validation_data->has_max_value)
    for (i = 0; i < num_values; i++)
      if (values[i] > validation_data->max_value)
	return FALSE;
  return TRUE;
}


typedef struct _IntValidator IntValidator;
struct _IntValidator {
	gboolean		has_min_value;
	gboolean		has_max_value;
	gint			min_value;
	gint			max_value;
};

static gboolean int_validator(GtkObject*    object,
                              GdamArg*      arg,
			      gpointer      v_data)
{
  IntValidator *validation_data = v_data;
  gint value;
  g_return_val_if_fail (arg->type == GDAM_ARG_TYPE_INT, FALSE);
  value = GDAM_ARG_INT (arg);
  if (validation_data->has_min_value)
    if (value < validation_data->min_value)
      return FALSE;
  if (validation_data->has_max_value)
    if (value > validation_data->max_value)
      return FALSE;
  return TRUE;
}

static gboolean int_array_validator(GtkObject*    object,
                                    GdamArg*      arg,
				    gpointer      v_data)
{
  IntValidator *validation_data = v_data;
  gint *values;
  int num_values;
  int i;
  g_return_val_if_fail (arg->type == GDAM_ARG_TYPE_INT_ARRAY, FALSE);
  values = GDAM_ARG_INT_ARRAY (arg).array;
  num_values = GDAM_ARG_INT_ARRAY (arg).length;
  for (i = 0; i < num_values; i++)
    {
      int value = values[i];
      if (!finite (value))
	return FALSE;
    }
  if (validation_data->has_min_value)
    for (i = 0; i < num_values; i++)
      if (values[i] < validation_data->min_value)
	return FALSE;
  if (validation_data->has_max_value)
    for (i = 0; i < num_values; i++)
      if (values[i] > validation_data->max_value)
	return FALSE;
  return TRUE;
}

static gboolean boolean_validator (GtkObject*    object,
                                   GdamArg*      arg,
				   gpointer      v_data)
{
  int v = GDAM_ARG_INT (arg);
  return v == 0 || v == 1;
}
static gboolean boolean_array_validator(GtkObject* object,
                                        GdamArg*   arg,
					gpointer   v_data)
{
  int *array = GDAM_ARG_INT_ARRAY (arg).array;
  int length = GDAM_ARG_INT_ARRAY (arg).length;
  int i;
  for (i = 0; i < length; i++)
    if (array[i] != 0 && array[i] != 1)
      return FALSE;
  return TRUE;
}

typedef struct _StringValidator StringValidator;
struct _StringValidator {
	gboolean		must_exist;
};

static gboolean string_validator (GtkObject*    object,
                                  GdamArg*      arg,
				  gpointer      v_data)
{
  char *c = GDAM_ARG_STRING (arg);
  if (c == NULL)
    return FALSE;
  return TRUE;
}


typedef struct _Int64Validator Int64Validator;
struct _Int64Validator {
	gboolean		has_min_value;
	gboolean		has_max_value;
	gint64			min_value;
	gint64			max_value;
};

static gboolean int64_validator(GtkObject*    object,
                              GdamArg*      arg,
			      gpointer      v_data)
{
  Int64Validator *validation_data = v_data;
  gint64 value;
  g_return_val_if_fail (arg->type == GDAM_ARG_TYPE_INT64, FALSE);
  value = GDAM_ARG_INT64 (arg);
  if (validation_data->has_min_value)
    if (value < validation_data->min_value)
      return FALSE;
  if (validation_data->has_max_value)
    if (value > validation_data->max_value)
      return FALSE;
  return TRUE;
}

static gboolean int64_array_validator(GtkObject*    object,
                                    GdamArg*      arg,
				    gpointer      v_data)
{
  Int64Validator *validation_data = v_data;
  gint64 *values;
  int num_values;
  gint i;
  g_return_val_if_fail (arg->type == GDAM_ARG_TYPE_INT64_ARRAY, FALSE);
  values = GDAM_ARG_INT64_ARRAY (arg).array;
  num_values = GDAM_ARG_INT64_ARRAY (arg).length;
  for (i = 0; i < num_values; i++)
    {
      gint64 value = values[i];
      if (!finite (value))
	return FALSE;
    }
  if (validation_data->has_min_value)
    for (i = 0; i < num_values; i++)
      if (values[i] < validation_data->min_value)
	return FALSE;
  if (validation_data->has_max_value)
    for (i = 0; i < num_values; i++)
      if (values[i] > validation_data->max_value)
	return FALSE;
  return TRUE;
}


gboolean  gdam_arg_get_default_validator(GdamArgType      type,
                                         const char*      validation_string,
                                         GdamArgValidator* validator_out,
                                         gpointer*       validation_data_out)
{
  *validator_out = NULL;
  switch (type)
    {
    case GDAM_ARG_TYPE_DOUBLE_ARRAY:
    case GDAM_ARG_TYPE_DOUBLE:
      {
	char **min_max;
	DoubleValidator *v_data;
	min_max = g_strsplit (validation_string, ",", 2);
	if (min_max[0] == NULL)
	  return FALSE;
	v_data = g_new0 (DoubleValidator, 1);
	if (min_max[0][0] == '\0')
	  {
	    v_data->has_min_value = FALSE;
	  }
	else
	  {
	    v_data->has_min_value = TRUE;
	    v_data->min_value = strtod (min_max[0], NULL);
	  }
	if (min_max[1] == NULL || min_max[1][0] == '\0')
	  {
	    v_data->has_max_value = FALSE;
	  }
	else
	  {
	    v_data->has_max_value = TRUE;
	    v_data->max_value = strtod (min_max[1], NULL);
	  }
	*validation_data_out = v_data;
	if (type == GDAM_ARG_TYPE_DOUBLE)
	  *validator_out = double_validator;
	else
	  *validator_out = double_array_validator;
	g_strfreev (min_max);
	return TRUE;
      }
    case GDAM_ARG_TYPE_INT_ARRAY:
    case GDAM_ARG_TYPE_INT:
      {
	char **min_max;
	IntValidator *v_data;
	if (strcmp (validation_string, "boolean") == 0)
	  {
	    if (type == GDAM_ARG_TYPE_INT)
	      *validator_out = boolean_validator;
	    else
	      *validator_out = boolean_array_validator;
	    *validation_data_out = NULL;
	    return TRUE;
	  }
	min_max = g_strsplit (validation_string, ",", 2);
	if (min_max[0] == NULL)
	  return FALSE;
	v_data = g_new0 (IntValidator, 1);
	if (min_max[0][0] == '\0')
	  {
	    v_data->has_min_value = FALSE;
	  }
	else
	  {
	    v_data->has_min_value = TRUE;
	    v_data->min_value = atoi (min_max[0]);
	  }
	if (min_max[1] == NULL || min_max[1][0] == '\0')
	  {
	    v_data->has_max_value = FALSE;
	  }
	else
	  {
	    v_data->has_max_value = TRUE;
	    v_data->max_value = atoi (min_max[1]);
	  }
	*validation_data_out = v_data;
	if (type == GDAM_ARG_TYPE_INT)
	  *validator_out = int_validator;
	else
	  *validator_out = int_array_validator;
	g_strfreev (min_max);
	return TRUE;
      }
    case GDAM_ARG_TYPE_INT64_ARRAY:
    case GDAM_ARG_TYPE_INT64:
      {
	char **min_max;
	Int64Validator *v_data;
	min_max = g_strsplit (validation_string, ",", 2);
	if (min_max[0] == NULL)
	  return FALSE;
	v_data = g_new0 (Int64Validator, 1);
	if (min_max[0][0] == '\0')
	  {
	    v_data->has_min_value = FALSE;
	  }
	else
	  {
	    v_data->has_min_value = TRUE;
	    v_data->min_value = gdam_strtoll (min_max[0], NULL, 0);
	  }
	if (min_max[1] == NULL || min_max[1][0] == '\0')
	  {
	    v_data->has_max_value = FALSE;
	  }
	else
	  {
	    v_data->has_max_value = TRUE;
	    v_data->max_value = gdam_strtoll (min_max[1], NULL, 0);
	  }
	*validation_data_out = v_data;
	if (type == GDAM_ARG_TYPE_INT64)
	  *validator_out = int64_validator;
	else
	  *validator_out = int64_array_validator;
	g_strfreev (min_max);
	return TRUE;
      }
    case GDAM_ARG_TYPE_STRING:
      {
	StringValidator *v_data;
	v_data = g_new0 (StringValidator, 1);
	*validation_data_out = v_data;
	*validator_out = string_validator;
	return TRUE;
      }
    default:
      g_warning ("many types don't have validations yet...");
      return FALSE;
    }
}

