/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdammodule.h"
#include "gdamdebug.h"
#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

static gboolean read_dependencies(const char* filename);

GModule* gdam_module_self()
{
  static GModule *self_mod;
  static gboolean inited = FALSE;
  if (!inited)
    {
      inited = TRUE;
      self_mod = g_module_open (NULL, G_MODULE_BIND_LAZY);
    }
  return self_mod;
}

static gboolean inited_paths = FALSE;
static char* paths[] = {
	"home",
	"./plugins-binary",
#ifdef GDAM_PLUGIN_DIRECTORY
	GDAM_PLUGIN_DIRECTORY,
#endif
	NULL
};

static void maybe_init_paths()
{
  if (!inited_paths)
    {
      char *home = g_get_home_dir ();
      char **paths_at = paths;
      while (*paths_at)
	{
	  if (strcmp (*paths_at, "home") == 0)
	    {
	      *paths_at = g_strdup_printf ("%s/.gdam/plugins", home);
	      break;
	    }
	  paths_at++;
	}
      inited_paths = TRUE;
      /* XXX: should we free home??? */
    }
}
static GModule*
gdam_get_module_by_name_real(const char* slashname)
{
  char **at = paths;
  maybe_init_paths ();
  while (*at)
    {
      char *module_name;
      GModule *module;
      char *err_msg;
      module_name = g_strdup_printf ("%s/%s.so", *at, slashname);
      gdam_debug ("get_gdam_module: try %s", module_name);
      module = g_module_open (module_name, 0);
      if (module)
	{
	  gdam_debug ("get_gdam_module: opened %s", module_name);
	  return module;
	}
      err_msg = g_module_error ();
      gdam_debug ("failed to open %s: %s", module_name, err_msg);
      at++;
      g_free (module_name);
    }
  return NULL;
}

static GHashTable* dependencies = NULL;
static char* dependency_filename = NULL;
static int last_dependency_time = -1;

static void kill_entries(gpointer key, gpointer value, gpointer unused)
{
  g_free (key);
  g_free (value);
}

static void clear_dependencies()
{
  if (dependencies == NULL)
    {
      dependencies = g_hash_table_new (g_str_hash, g_str_equal);
      return;
    }
  g_hash_table_foreach (dependencies, kill_entries, NULL);
  g_hash_table_destroy (dependencies);
  dependencies = g_hash_table_new (g_str_hash, g_str_equal);
}

static gboolean find_time(const char* filename, int* t)
{
  struct stat sbuf;
  if (stat (filename, &sbuf) < 0)
    return FALSE;
  *t = (int) (sbuf.st_mtime);
  return TRUE;
}

static gboolean read_dependencies(const char* filename)
{
  FILE *file = fopen (filename, "r");
  char buf[8192];
  int t;
  if (!file)
    return FALSE;
  gdam_debug ("reading dependencies from %s", filename);
  clear_dependencies ();
  dependency_filename = g_strdup (filename);
  while (fgets (buf, sizeof (buf), file))
    {
      char *nl = strchr (buf, '\n');
      char **strs;
      char *colon = strchr (buf, ':');
      char *dependent;

      if (nl != NULL)
	*nl = '\0';
      g_assert (colon != NULL);
      dependent = g_new (char, colon - buf + 1);
      memcpy (dependent, buf, colon - buf);
      dependent[colon - buf] = '\0';
      g_assert (colon[1] == ' ');
      strs = g_strsplit (colon + 2, " ", 1000);
      g_assert (strs != NULL);
      g_hash_table_insert (dependencies, dependent, strs);
    }
  fclose (file);
  if (!find_time (dependency_filename, &t))
    g_warning ("error stating file for file time");
  return TRUE;
}
static void maybe_update_dependencies()
{
  char **at;
  int t;			/* file modification time */
  maybe_init_paths ();
  at = paths;
  if (dependency_filename == NULL)
    {
      for (at = paths; *at; at++)
	{
	  char *filename;
	  filename = g_strdup_printf ("%s/dependencies", *at);
	  if (read_dependencies (filename))
	    {
	      dependency_filename = filename;
	      g_assert (dependencies != NULL);
	      return;
	    }
	  g_free (filename);
	}
      return;
    }

  if (!find_time (dependency_filename, &t))
    return;

  if (t != last_dependency_time)
    {
      if (!read_dependencies (dependency_filename))
	{
	  g_warning ("sorry -- couldn't read module dependency file"
		     " (%s)", dependency_filename);
	}
    }
  return;
}

static GHashTable* module_cache = NULL;
static GHashTable* loading_table;

static GModule* gdam_get_module_by_slashname(const char* slashname,
					     const char* family,
                                             const char* type)
{
  GModule *rv;
  char **dep;
  char *name;
  static int depth = 0;

  depth++;
  if (module_cache == NULL)
    {
      module_cache = g_hash_table_new (g_str_hash, g_str_equal);
      loading_table = g_hash_table_new (g_str_hash, g_str_equal);
    }
  rv = (GModule *) g_hash_table_lookup (module_cache, slashname);
  if (rv != NULL)
    {
      depth--;
      return rv;
    }

  /* Do security checks:  don't try to load modules
   * whose family or type contain slashes.
   * 
   * XXX: seems unnecessary to check `family' actually
   */
  /*  CHANGED by geoff 6-5-00 - sense of NULL checks backwards!  */
  if (type != NULL)
    {
      const char *at;
      for (at = type; *at != '\0'; at++)
	{
	  if (!isalnum (*at) && *at != '_')
	    return NULL;
	}
    }
  if (family != NULL)
    {
      const char *at;
      for (at = family; *at != '\0'; at++)
	{
	  if (!isalnum (*at) && *at != '_')
	    return NULL;
	}
    }

  if (depth == 1)
    maybe_update_dependencies ();

  /* Check for circular module dependencies. */
  if (g_hash_table_lookup (loading_table, slashname))
    {
      g_warning ("circular dependency detected... ");
      return NULL;
    }
  name = g_strdup (slashname);
  g_hash_table_insert (loading_table, name, GINT_TO_POINTER (1));

  /* Get the modules this one depends on... */
  dep = (char **) g_hash_table_lookup (dependencies, slashname);
  if (dep != NULL)
    {
      while (*dep)
	{
	  gdam_debug ("dependency: %s on %s", *dep, slashname);
	  if (g_hash_table_lookup (module_cache, *dep) == NULL)
	    if (!gdam_get_module_by_slashname (*dep, NULL, NULL))
	      {
		g_warning ("error loading dependency %s", *dep);
		g_hash_table_remove (loading_table, name);
		return NULL;
	      }
	  dep++;
	}
    }
  g_hash_table_remove (loading_table, name);
  rv = gdam_get_module_by_name_real (slashname);
  if (rv == NULL)
    {
      g_warning ("error loading %s", name);
      return NULL;
    }
  g_hash_table_insert (module_cache, name, rv);
  depth--;
  return rv;
}
GModule* gdam_module_by_name(const char* family, const char* type)
{
  char *name;
  GModule *rv;


  name = g_strdup_printf ("%s/%s", family, type);
  g_strdown (name);
  rv = gdam_get_module_by_slashname (name, family, type);
  g_free (name);

  /* Just print a warning this way on failure.... */
  if (rv == NULL)
    g_warning ("failed to load module %s", name);
  g_return_val_if_fail (rv != NULL, NULL);

  return rv;
}
