/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuffermanager.h"
#include <common/gdam-protocol.h>
#include <gtk/gtksignal.h>

#ifndef G_N_ELEMENTS
#define G_N_ELEMENTS(arr)     ((int)((sizeof(arr)) / sizeof((arr)[0])))
#endif

static guint default_id = 2;

static GHashTable* hashtable = NULL;
GHashTable* gdam_buffer_manager_buffers = NULL;
GHashTable* gdam_buffers_by_desc = NULL;

/* This channel is used for obtaining the sine wave, triangle wave, etc. */
static GdamChannel*	global_channel = NULL;

/* common buffers  <===>  buffer ids */
static GHashTable* common_id_from_name = NULL;
static GPtrArray*  common_names        = NULL;

typedef struct _GdamBufferConnection GdamBufferConnection;
static void remove_connection(GdamBufferConnection* connection);
static void notify_buffer_connections(GdamBufferManagerInfo* info);

static void deliver_unref_buffer(guint buffer_id)
{
  g_return_if_fail (global_channel != NULL);
  gdam_channel_append_request (global_channel,
			       gdam_request_unref (GDAM_PROTOCOL_BUFFER_ARG,
						   buffer_id,
						   NULL, NULL, NULL));
}

/* model == NULL is used in these functions to represent the
 * global count, ie which buffers are exposed to the world. */

static void ref_id(guint id, GdamModel* model)
{
  GHashTable *per_id;
  guint ref_count;
  if (hashtable == NULL)
    hashtable = g_hash_table_new (NULL, NULL);
  per_id = (GHashTable *) g_hash_table_lookup (hashtable,
					       GUINT_TO_POINTER (id));
  if (per_id == NULL)
    {
      per_id = g_hash_table_new (NULL, NULL);
      g_hash_table_insert (hashtable, GUINT_TO_POINTER (id), per_id);
    }
  ref_count = GPOINTER_TO_INT (g_hash_table_lookup (per_id, model));
  ref_count++;
  g_hash_table_insert (per_id, model, GINT_TO_POINTER (ref_count));

  gdam_debug ("ref_buffer_id[%d] by %s, new ref count %d", id,
	      model ? gtk_type_name (GTK_OBJECT (model)->klass->
				     type) : "NULL", ref_count);
}

static void free_id(guint id)
{
  GdamBufferManagerInfo *info = NULL;
  gpointer id_ptr = GUINT_TO_POINTER (id);
  if (gdam_buffer_manager_buffers != NULL)
    {
      info = g_hash_table_lookup (gdam_buffer_manager_buffers, id_ptr);
      if (info != NULL && info->is_permanent)
	return;
    }

  if (hashtable != NULL)
    {
      GHashTable *per_id;
      per_id = g_hash_table_lookup (hashtable, id_ptr);
      if (per_id != NULL)
	{
	  g_hash_table_remove (hashtable, id_ptr);
	  g_hash_table_destroy (per_id);
	}
    }
  if (info != NULL)
    {
      g_slist_foreach (info->connections, (GFunc) remove_connection, NULL);
      g_slist_free (info->connections);
      g_free (info);
      /* XXX: is beat info a memory leak? */
    }
  deliver_unref_buffer (id);
}

static void unref_id(guint id, GdamModel* model)
{
  GHashTable *per_id;
  guint ref_count;
  g_return_if_fail (hashtable != NULL);
  per_id = (GHashTable *) g_hash_table_lookup (hashtable,
					       GUINT_TO_POINTER (id));
  g_return_if_fail (per_id != NULL);
  ref_count = GPOINTER_TO_INT (g_hash_table_lookup (per_id, model));
  ref_count--;
  if (ref_count == 0)
    {
      g_hash_table_remove (per_id, model);
      if (g_hash_table_size (per_id) == 0
	  && g_hash_table_lookup (gdam_buffer_manager_buffers,
				  GUINT_TO_POINTER (id)) == NULL)
	{
	  free_id (id);
	}
    }
  else
    {
      g_hash_table_insert (per_id, model, GINT_TO_POINTER (ref_count));
    }
  gdam_debug ("unref_buffer_id[%d] by %s, new ref count %d", id,
	      model ? gtk_type_name (GTK_OBJECT (model)->klass->
				     type) : "NULL", ref_count);
}

GSList* subscribers = NULL;

static void emit_one_buffer_added(GdamBufferManagerNotifier* notifier,
                                  GdamBufferManagerInfo*     info)
{
  if (notifier->buffer_add_callback)
    (notifier->buffer_add_callback) (info, notifier->user_data);
}
static void emit_one_buffer_removed(GdamBufferManagerNotifier* notifier,
                                    GdamBufferManagerInfo*     info)
{
  if (notifier->buffer_remove_callback)
    (notifier->buffer_remove_callback) (info, notifier->user_data);
}

static void emit_buffer_added(GdamBufferManagerInfo* info)
{
  g_slist_foreach (subscribers, (GFunc) emit_one_buffer_added, info);
}
static void emit_buffer_removed(GdamBufferManagerInfo* info)
{
  g_slist_foreach (subscribers, (GFunc) emit_one_buffer_removed, info);
}

void gdam_buffer_set_default(guint buffer_id)
{
  default_id = buffer_id;
}
guint gdam_buffer_get_default()
{
  return default_id;
}
guint gdam_buffer_manager_obtain_id(GdamModel* model)
{
  gdam_debug ("gdam_buffer_manager_obtain_id");
  ref_id (default_id, model);
  return default_id;
}

gboolean  gdam_buffer_manager_is_readonly(guint buffer_id)
{
  GdamBufferManagerInfo *info;
  info = gdam_buffer_manager_info_by_id (buffer_id);
  return info == NULL ? FALSE : info->is_readonly;
}

GdamBufferManagerInfo*  gdam_buffer_manager_info_by_id(guint buffer_id)
{
  if (gdam_buffer_manager_buffers == NULL)
    return NULL;
  return g_hash_table_lookup (gdam_buffer_manager_buffers,
			      GUINT_TO_POINTER (buffer_id));
}

GdamBufferManagerInfo*  gdam_buffer_manager_info_by_desc(const char* desc)
{
  if (gdam_buffers_by_desc == NULL)
    return NULL;
  return g_hash_table_lookup (gdam_buffers_by_desc, desc);
}

void gdam_buffer_manager_unref(guint buffer_id, GdamModel* model)
{
  unref_id (buffer_id, model);
}
void gdam_buffer_manager_ref(guint buffer_id, GdamModel* model)
{
  ref_id (buffer_id, model);
}

static void query_info(GdamBufferManagerInfo* info);

void gdam_buffer_manager_refresh(guint   buffer_id)
{
  GdamBufferManagerInfo *info;
  info = gdam_buffer_manager_info_by_id (buffer_id);
  if (info != NULL)
    query_info (info);
}

static void got_buffer_args(int num_args, GdamArg* args, gpointer user_data)
{
  GdamBufferManagerInfo *info = user_data;
  g_return_if_fail (num_args == 1);
  g_return_if_fail (strcmp (args[0].name, "length") == 0);
  g_return_if_fail (args[0].type == GDAM_ARG_TYPE_INT64);
  info->length = GDAM_ARG_INT64 (args + 0);
  emit_buffer_added (info);
  notify_buffer_connections (info);
  unref_id (info->buffer_id, NULL);
  info->is_querying = FALSE;
  if (info->redo_query)
    {
      info->redo_query = FALSE;
      query_info (info);
    }
}
static void failed_to_get_buffer_args(gpointer user_data)
{
  GdamBufferManagerInfo *info = user_data;
  g_warning ("couldn't get length for buffer %d", info->buffer_id);
  unref_id (info->buffer_id, NULL);
  info->is_querying = FALSE;
}

static void query_info(GdamBufferManagerInfo* info)
{
  const char *names[] = { "length" };
  GdamRequest *request;
  ref_id (info->buffer_id, NULL);
  if (info->is_querying)
    {
      info->redo_query = TRUE;
      return;
    }
  info->is_querying = TRUE;
  info->redo_query = FALSE;
  request = gdam_request_get_gen_args (GDAM_PROTOCOL_BUFFER_ARG,
				       info->buffer_id,
				       G_N_ELEMENTS (names),
				       names,
				       got_buffer_args,
				       failed_to_get_buffer_args, info);
  gdam_channel_append_request (info->channel, request);
}
void gdam_buffer_manager_add_buffer(GdamBufferManagerInfo* info)
{
  if (gdam_buffer_manager_buffers == NULL)
    {
      gdam_buffer_manager_buffers = g_hash_table_new (NULL, NULL);
      gdam_buffers_by_desc = g_hash_table_new (g_str_hash, g_str_equal);
    }


  g_return_if_fail (g_hash_table_lookup (gdam_buffer_manager_buffers,
					 GINT_TO_POINTER (info->buffer_id)) ==
		    NULL);

  info->is_querying = info->redo_query = FALSE;

  query_info (info);
  g_hash_table_insert (gdam_buffer_manager_buffers,
		       GINT_TO_POINTER (info->buffer_id), info);
  g_hash_table_insert (gdam_buffers_by_desc, info->description, info);

}

static inline void destroy_info(GdamBufferManagerInfo* info)
{
  g_free (info->description);
  g_slist_foreach (info->connections, (GFunc) remove_connection, NULL);
  g_slist_free (info->connections);
  g_free (info);
}
void gdam_buffer_manager_soft_destroy(guint buffer_id)
{
  GHashTable *per_id;
  GdamBufferManagerInfo *info = (GdamBufferManagerInfo *)
    g_hash_table_lookup (gdam_buffer_manager_buffers,
			 GUINT_TO_POINTER (buffer_id));
  g_return_if_fail (gdam_buffer_manager_buffers != NULL);
  if (info)
    {
      if (info->is_permanent)
	return;
      emit_buffer_removed (info);
      destroy_info (info);
    }

  g_hash_table_remove (gdam_buffers_by_desc, info->description);
  g_hash_table_remove (gdam_buffer_manager_buffers,
		       GINT_TO_POINTER (buffer_id));

  per_id = g_hash_table_lookup (hashtable, GUINT_TO_POINTER (buffer_id));
  if (per_id == NULL || g_hash_table_size (per_id) == 0)
    {
      deliver_unref_buffer (buffer_id);
    }
}

static void add_to_array              (GdamModel*     model,
                                       gpointer       ref_count,
				       GdamModel***   models)
{
  GdamModel **at = (*models);
  *at++ = model;
  (*models) = at;
}

void           gdam_buffer_manager_hard_destroy(guint buffer_id)
{
  GHashTable *table;
  GdamBufferManagerInfo *info = (GdamBufferManagerInfo *)
    g_hash_table_lookup (gdam_buffer_manager_buffers,
			 GUINT_TO_POINTER (buffer_id));
  if (info != NULL)
    {
      if (info->is_permanent)
	return;
      emit_buffer_removed (info);
      g_hash_table_remove (gdam_buffers_by_desc, info->description);
      g_hash_table_remove (gdam_buffer_manager_buffers,
			   GUINT_TO_POINTER (buffer_id));
      destroy_info (info);
    }
  table = (GHashTable *) g_hash_table_lookup (hashtable,
					      GINT_TO_POINTER (buffer_id));
  g_return_if_fail (table != NULL);
  if (g_hash_table_size (table) > 0)
    {
      GdamModel **model_at;
      GdamModel **models;
      models = alloca (sizeof (GdamModel *) * g_hash_table_size (table));
      model_at = models;
      g_hash_table_foreach (table, (GHFunc) add_to_array, &model_at);
      while (models < model_at)
	{
	  gtk_signal_emit_by_name (GTK_OBJECT (*models),
				   "detach_from_buffer",
				   GPOINTER_TO_INT (buffer_id));
	  models++;
	}
    }
  deliver_unref_buffer (buffer_id);
}

static void initial_subscribe_notify(gpointer key,
                                     GdamBufferManagerInfo* info,
                                     GdamBufferManagerNotifier* notifier)
{
  if (notifier->buffer_add_callback)
    (*notifier->buffer_add_callback) (info, notifier->user_data);
}

void gdam_buffer_manager_subscribe   (GdamBufferManagerNotifier* notifier)
{
  subscribers = g_slist_prepend (subscribers, notifier);
  if (gdam_buffer_manager_buffers != NULL)
    g_hash_table_foreach (gdam_buffer_manager_buffers,
			  (GHFunc) initial_subscribe_notify, notifier);
}
void gdam_buffer_manager_unsubscribe (GdamBufferManagerNotifier* notifier)
{
  subscribers = g_slist_remove (subscribers, notifier);
  g_free (notifier);
}

static void got_named_buffers(int num_bufs, char** bufs, gpointer unused)
{
  int i;
  if (common_id_from_name == NULL)
    {
      common_id_from_name = g_hash_table_new (g_str_hash, g_str_equal);
      common_names = g_ptr_array_new ();
      g_ptr_array_add (common_names, NULL);
    }
  for (i = 0; i < num_bufs; i++)
    {
      GdamBufferManagerInfo *info;
      char *copy = g_strdup (bufs[i]);

      g_hash_table_insert (common_id_from_name,
			   copy, GINT_TO_POINTER (i + 1));
      g_ptr_array_add (common_names, copy);

      info = g_new0 (GdamBufferManagerInfo, 1);
      info->buffer_id = i + 1;
      info->channel = global_channel;
      info->description = copy;
      info->length = 0;
      info->is_permanent = 1;
      info->is_readonly = 1;
      gdam_buffer_manager_add_buffer (info);
    }
}
static void didnt_get_named_buffers(gpointer id_slot)
{
  g_error ("couldn't get named wavetables, upgrade your gdam server");
}


void gdam_buffer_manager_init(GdamChannel* channel)
{
  global_channel = channel;
  gdam_channel_append_request
    (global_channel,
     gdam_request_get_named_buffers
     (got_named_buffers, didnt_get_named_buffers, NULL));
}


guint          gdam_buffer_manager_get_named(const char* name)
{
  g_return_val_if_fail (common_id_from_name != NULL, 0);
  return GPOINTER_TO_UINT (g_hash_table_lookup (common_id_from_name, name));
}

GPtrArray*     gdam_buffer_manager_get_names()
{
  return common_names;
}


/*
 *                  gdam_buffer_manager_start_caching
 *
 * Easier interface to cache a buffer.
 */

typedef struct _CachingInfo CachingInfo;
struct _CachingInfo {
	int buffer_id, finished_caching_event_id;
	char* description;
	void (*cached)(int buffer_id, gpointer user_data);
	void (*cache_failed)(gpointer user_data);
	gboolean add_to_buffer_manager;
	gpointer user_data;
	GdamChannel* channel;
};


static void finished_caching(gpointer user_data)
{
  CachingInfo *cache_info = (CachingInfo *) user_data;
  GdamBufferManagerInfo *info;
  gdam_debug ("buffer-manager: finished_caching");
  info = g_new0 (GdamBufferManagerInfo, 1);
  info->buffer_id = cache_info->buffer_id;
  info->description = cache_info->description;
  info->channel = cache_info->channel;
  if (cache_info->add_to_buffer_manager)
    gdam_buffer_manager_add_buffer (info);
  if (cache_info->cached)
    (*cache_info->cached) (cache_info->buffer_id, cache_info->user_data);
  g_free (cache_info);
}

static void started_caching(int buffer_id,
                            int connection_id,
			    gpointer user_data)
{
  CachingInfo *cache_info = (CachingInfo *) user_data;
  cache_info->buffer_id = buffer_id;
  gdam_debug ("buffer_manager: started_caching");
  gdam_channel_register_handler (cache_info->channel,
				 gdam_channel_simple_handler_new
				 (connection_id, TRUE, finished_caching,
				  cache_info));
}

static void failed_to_start_caching(gpointer user_data)
{
  CachingInfo *cache_info = (CachingInfo *) user_data;
  g_warning ("failed to start caching %s", cache_info->description);
  g_free (cache_info->description);
  g_free (cache_info);
}
void           gdam_buffer_manager_start_caching
				(GdamRequestBuilder*   request_builder,
				 int                   local_src_id,
				 GdamChannel*          channel,
				 gboolean              use_buffer_manager,
				 const char*           description,
			 void (*cached)(int buffer_id, gpointer user_data),
		             void (*cache_failed)(gpointer user_data),
				                 gpointer user_data)
{
  CachingInfo *cache_info;
  gdam_debug ("buffer_manager: start_caching");
  cache_info = g_new0 (CachingInfo, 1);
  cache_info->channel = channel;
  cache_info->description = g_strdup (description);
  cache_info->cached = cached;
  cache_info->cache_failed = cache_failed;
  cache_info->user_data = user_data;
  cache_info->add_to_buffer_manager = use_buffer_manager;
  cache_info->buffer_id = 0;
  cache_info->finished_caching_event_id = 0;

  gdam_request_builder_append_generic (request_builder,
				       gdam_request_start_caching
				       (local_src_id, started_caching,
					failed_to_start_caching, cache_info));
  gdam_channel_append_request (channel,
			       gdam_request_builder_package
			       (request_builder));
  gdam_request_builder_destroy (request_builder);
}

struct _GdamBufferConnection {
	int connection_id;
	GdamBufferChangedFunc func;
	gpointer func_data;
	GdamBufferManagerInfo* info;
};
static GHashTable* connection_table = NULL;

static int alloc_connection_id()
{
	static int last_id = 0;
	return ++last_id;
}

int            gdam_buffer_manager_connect     (int               buffer_id,
                                                GdamBufferChangedFunc func,
						gpointer          func_data)
{
  GdamBufferManagerInfo *info;
  GdamBufferConnection *connection;
  info = gdam_buffer_manager_info_by_id (buffer_id);
  g_return_val_if_fail (info != NULL, 0);
  connection = g_new (GdamBufferConnection, 1);
  connection->connection_id = alloc_connection_id ();
  connection->func = func;
  connection->func_data = func_data;
  connection->info = info;
  info->connections = g_slist_prepend (info->connections, connection);
  if (connection_table == NULL)
    connection_table = g_hash_table_new (NULL, NULL);
  g_hash_table_insert (connection_table,
		       GINT_TO_POINTER (buffer_id), connection);
  return connection->connection_id;
}

void           gdam_buffer_manager_disconnect  (int               conn_id)
{
  GdamBufferManagerInfo *info;
  GdamBufferConnection *connection;
  g_return_if_fail (connection_table != NULL);
  connection = g_hash_table_lookup (connection_table,
				    GINT_TO_POINTER (conn_id));
  g_return_if_fail (connection != NULL);
  info = connection->info;
  info->connections = g_slist_remove (info->connections, connection);
  g_hash_table_remove (connection_table, GINT_TO_POINTER (conn_id));
  g_free (connection);
}

void           gdam_buffer_manager_reconnect  (int               conn_id,
                                               int               new_buffer_id)
{
  GdamBufferManagerInfo *old_info;
  GdamBufferManagerInfo *new_info;
  GdamBufferConnection *connection;
  g_return_if_fail (connection_table != NULL);
  connection = g_hash_table_lookup (connection_table,
				    GINT_TO_POINTER (conn_id));
  g_return_if_fail (connection != NULL);
  old_info = connection->info;
  new_info = gdam_buffer_manager_info_by_id (new_buffer_id);
  if (old_info == new_info)
    return;
  old_info->connections = g_slist_remove (old_info->connections, connection);
  new_info->connections = g_slist_prepend (new_info->connections, connection);
  connection->info = new_info;
}
static void info_changed(GdamBufferConnection* connection,
                         GdamBufferManagerInfo* info)
{
  if (connection->func == NULL)
    return;
  (*connection->func) (info, connection->func_data);
}
static void notify_buffer_connections(GdamBufferManagerInfo* info)
{
  g_slist_foreach (info->connections, (GFunc) info_changed, info);
}


/* Called from BufferManagerInfo's finalize method */
static void remove_connection(GdamBufferConnection* connection)
{
  g_hash_table_remove (connection_table,
		       GINT_TO_POINTER (connection->connection_id));
  g_free (connection);
}
