/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuilder.h"
#include <common/gdam-protocol.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <stdlib.h>


static void gdam_builder_destroy(GtkObject* object);

static void gdam_builder_add_xml  (GdamModel*        model,
                                   xmlNode*          parent,
				   GdamXmlInfo*      xml_info);
static void gdam_builder_parse_xml(GdamModel*        model,
                                   xmlNode*          node,
				   GdamXmlInfo*      xml_info);


/*
 *  ____        _ _     _           ____
 * | __ ) _   _(_) | __| | ___ _ __|  _ \ _   _ _ __ ___  _ __ ___  _   _
 * |  _ \| | | | | |/ _` |/ _ \ '__| | | | | | | '_ ` _ \| '_ ` _ \| | | |
 * | |_) | |_| | | | (_| |  __/ |  | |_| | |_| | | | | | | | | | | | |_| |
 * |____/ \__,_|_|_|\__,_|\___|_|  |____/ \__,_|_| |_| |_|_| |_| |_|\__, |
 *                                                                  |___/
 */

typedef struct _GdamBuilderDummy GdamBuilderDummy;
typedef struct _GdamBuilderDummyClass GdamBuilderDummyClass;

gboolean gdam_builder_verify(GdamBuilder* builder);
static void gdam_builder_new_request(GdamModel*   builder,
                                     GdamModel*   dummy_filter,
				     GdamPlayInfo*    play_info,
				     int              dummy_gain_id,
				     GdamRequestBuilder*  request);
static int gdam_builder_dummy_append_start_request(
				GdamFilter*    filter,
				int            source_id,
				GdamRequestBuilder* request_builder);
static GtkType gdam_builder_dummy_get_type();
#define GDAM_TYPE_BUILDER_DUMMY			(gdam_builder_dummy_get_type ())
#define GDAM_BUILDER_DUMMY(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_BUILDER_DUMMY, GdamBuilderDummy))

struct _GdamBuilderDummyClass {
	GdamFilterClass		builder_dummy_class;
};
struct _GdamBuilderDummy {
	GdamFilter			builder_dummy;
	GdamBuilder*			builder;
};

static void new_dummy_out_gain_source_id(int id, gpointer comp)
{
  GdamBuilder *builder = GDAM_BUILDER (comp);
  GdamModel *model = GDAM_MODEL (builder->dummy_gain_out);
  gdam_model_set_source_id (model, id);
  /*gdam_debug("new_dummy_out_gain_source_id: set id=%d", id); */
  gdam_model_set_source_id (GDAM_MODEL (builder), id);
}

static void new_dummy_in_gain_source_id(int id, gpointer comp)
{
  GdamBuilder *builder = GDAM_BUILDER (comp);
  GdamModel *model = GDAM_MODEL (builder->dummy_gain_in);
  /*gdam_debug("new_dummy_in_gain_source_id: set id=%d", id); */
  gdam_model_set_source_id (model, id);
}

static void dummy_failure_func(gpointer dummy)
{
  g_error ("error creating dummy gain in builder");
}

static int gdam_builder_dummy_append_start_request(
				GdamFilter*    filter,
				int            source_id,
				GdamRequestBuilder* request_builder)
{
  GdamBuilder *builder = GDAM_BUILDER_DUMMY (filter)->builder;
  GdamRequestGotIntFunc success_func;
  GdamModel *model = GDAM_MODEL (filter);
  int local_id;

  if (builder->dummy_gain_out == filter)
    success_func = new_dummy_out_gain_source_id;
  else
    success_func = new_dummy_in_gain_source_id;
  local_id = gdam_channel_new_local_id (model->channel);
  model->source_id = local_id;
  return gdam_request_builder_append_create (request_builder,
					     gdam_request_create_source
					     ("GdasGain", local_id,
					      success_func,
					      dummy_failure_func, builder, 1,
					      "filtered", GDAM_ARG_TYPE_INT,
					      source_id));
}


static void dummy_on_child_set_beat_info(GdamFilter*   filter,
                                         const GdamBeatInfo* beat_info)
{
  GdamBuilder *builder = GDAM_BUILDER_DUMMY (filter)->builder;
  GdamModel *model = GDAM_MODEL (filter);

  gdam_model_set_beat_info (model, beat_info);
  if (filter == builder->dummy_gain_out)
    gdam_model_set_beat_info (GDAM_MODEL (builder), beat_info);
}

static GtkObjectClass* parent_class;
static void gdam_builder_dummy_init(GdamBuilderDummy* builder_dummy)
{
}
static void gdam_builder_dummy_class_init(GdamFilterClass* klass)
{
  klass->on_child_set_beat_info = dummy_on_child_set_beat_info;
  klass->append_start_request = gdam_builder_dummy_append_start_request;
}

static GtkType gdam_builder_dummy_get_type()
{
  static GtkType builder_dummy_type = 0;
  if (!builder_dummy_type)
    {
      static const GtkTypeInfo builder_dummy_info = {
	"GdamBuilderDummy",
	sizeof (GdamBuilderDummy),
	sizeof (GdamBuilderDummyClass),
	(GtkClassInitFunc) gdam_builder_dummy_class_init,
	(GtkObjectInitFunc) gdam_builder_dummy_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      builder_dummy_type = gtk_type_unique (parent, &builder_dummy_info);
    }
  return builder_dummy_type;
}

static GdamBuilderDummy* gdam_builder_dummy_new(GdamBuilder* builder)
{
  GdamBuilderDummy *dummy;
  dummy = GDAM_BUILDER_DUMMY (gtk_type_new (GDAM_TYPE_BUILDER_DUMMY));
  dummy->builder = builder;
  return dummy;
}

/*
 *  ____        _ _     _
 * | __ ) _   _(_) | __| | ___ _ __
 * |  _ \| | | | | |/ _` |/ _ \ '__|
 * | |_) | |_| | | | (_| |  __/ |
 * |____/ \__,_|_|_|\__,_|\___|_|
 * 
 */
static void allocate_dummy_gains(GdamBuilder* builder)
{
  GdamFilter *in = (GdamFilter *) gdam_builder_dummy_new (builder);
  GdamFilter *out = (GdamFilter *) gdam_builder_dummy_new (builder);
  GdamChannel *channel = GDAM_MODEL (builder)->channel;
  builder->dummy_gain_in = in;
  builder->dummy_gain_out = out;
  GDAM_MODEL (in)->channel = channel;
  GDAM_MODEL (out)->channel = channel;

  gtk_object_ref ((GtkObject *) builder->dummy_gain_in);
  gtk_object_sink ((GtkObject *) builder->dummy_gain_in);
  gtk_object_ref ((GtkObject *) builder->dummy_gain_out);
  gtk_object_sink ((GtkObject *) builder->dummy_gain_out);

  gdam_filter_set_child (out, GDAM_MODEL (in));

  GDAM_MODEL (out)->parent = GDAM_MODEL (builder);

#if 0				/* for now, we just do this through
				 * builder-dummies set-source callback.... */
  gtk_signal_connect_object (
			     (GtkObject *) builder->dummy_gain_out,
			     "set_source",
			     GTK_SIGNAL_FUNC (gdam_model_set_source_id),
			     (GtkObject *) builder);
#endif
}

static void gdam_builder_add_xml(GdamModel*         model,
                                 xmlNode*           parent,
				 GdamXmlInfo*       xml_info)
{
  GList *filter_list;
  GdamBuilder *builder;
  GHashTable *disabled_filters;
  GHashTable *saveable_filters;
  GdamModel *component;

  builder = GDAM_BUILDER (model);
  disabled_filters = builder->disabled_filters;
  saveable_filters = builder->saveable_filters;

  for (filter_list = builder->filters_in;
       filter_list != NULL; filter_list = filter_list->next)
    {
      GdamModel *model = GDAM_MODEL (filter_list->data);
      const char *str;
      xmlNode *node;
      xmlNode *stack_node;
      node = gdam_model_to_xml (model, xml_info);
      stack_node = xmlNewNode (NULL, "filter");

      if (g_hash_table_lookup (disabled_filters, model) == NULL)
	str = "1";
      else
	str = "0";
      xmlNewProp (stack_node, "enabled", str);

      if (g_hash_table_lookup (saveable_filters, model) != NULL)
	str = "1";
      else
	str = "0";
      xmlNewProp (stack_node, "saveable", str);

      xmlAddChild (stack_node, node);
      xmlAddChild (parent, stack_node);
    }

  component = builder->component;
  if (component != NULL
      && gdam_xml_info_lookup_by_object (xml_info, component) == 0)
    {
      xmlNode *node;
      xmlNode *stack_node;
      node = gdam_model_to_xml (component, xml_info);
      stack_node = xmlNewNode (NULL, "component");
      xmlAddChild (stack_node, node);
      xmlAddChild (parent, stack_node);
    }
}

static void gdam_builder_parse_xml(GdamModel*     model,
                                   xmlNode*       node,
				   GdamXmlInfo*   xml_info)
{
  GdamModel *submodel;
  GdamBuilder *builder;
  gboolean enabled = TRUE;
  gboolean saveable = TRUE;
  char *enabled_str;
  char *saveable_str;
  builder = GDAM_BUILDER (model);
  if (strcmp (node->name, "filter") == 0)
    {
      if (node->childs == NULL || node->childs->next != NULL)
	{
	  g_warning ("tag <filter> must have exactly one subtag");
	  return;
	}
      submodel = gdam_model_from_xml (node->childs, model->channel, xml_info);
      if (!GDAM_IS_FILTER (submodel))
	{
	  g_warning ("tag <filter> must have a " "GdamFilter under it");
	  gtk_object_sink (GTK_OBJECT (submodel));
	  return;
	}
      enabled_str = xmlGetProp (node, "enabled");
      if (enabled_str != NULL)
	enabled = atoi (enabled_str);
      saveable_str = xmlGetProp (node, "saveable");
      if (saveable_str != NULL)
	saveable = atoi (saveable_str);
      gdam_builder_append_filter (GDAM_BUILDER (model),
				  GDAM_FILTER (submodel), enabled, saveable);
      g_assert (gdam_builder_verify (builder));
      return;
    }

  if (strcmp (node->name, "component") == 0)
    {
      if (node->childs == NULL || node->childs->next != NULL)
	{
	  g_warning ("tag <filter> must have exactly one subtag");
	  return;
	}
      submodel = gdam_model_from_xml (node->childs, model->channel, xml_info);
      gdam_builder_set_component (builder, submodel);
      g_assert (gdam_builder_verify (builder));
      return;
    }

  g_warning ("node <%s> unhandler in GdamBuilder", node->name);
}

static void gdam_builder_init(GdamBuilder* builder)
{
  builder->disabled_filters = g_hash_table_new (NULL, NULL);
  builder->saveable_filters = g_hash_table_new (NULL, NULL);
}
static void gdam_builder_class_init(GdamBuilderClass* builder_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (builder_class);
  GdamModelClass *model_class;
  model_class = GDAM_MODEL_CLASS (builder_class);
  object_class->destroy = gdam_builder_destroy;
  model_class->new_request = gdam_builder_new_request;
  model_class->add_xml = gdam_builder_add_xml;
  model_class->parse_xml = gdam_builder_parse_xml;
}

GtkType gdam_builder_get_type()
{
  static GtkType builder_type = 0;
  if (!builder_type)
    {
      static const GtkTypeInfo builder_info = {
	"GdamBuilder",
	sizeof (GdamBuilder),
	sizeof (GdamBuilderClass),
	(GtkClassInitFunc) gdam_builder_class_init,
	(GtkObjectInitFunc) gdam_builder_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_MODEL;
      parent_class = gtk_type_class (parent);
      builder_type = gtk_type_unique (parent, &builder_info);
    }
  return builder_type;
}

static GdamRequest* set_filtered_request(int filter_id, int child_id)
{
  GdamArg arg;
  arg.name = "filtered";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = child_id;
  return gdam_request_set_args (filter_id, 1, &arg, NULL, NULL, NULL);
}

static void append_set_filtered_request(int filter_id,
                                        int child_id,
					GdamRequestBuilder* request_builder)
{
  gdam_request_builder_append_generic (request_builder,
				       set_filtered_request (filter_id,
							     child_id));
}

static GList* find_last_enabled(GdamBuilder* builder)
{
  GList *at = builder->filters_out;
  while (at != NULL
	 && g_hash_table_lookup (builder->disabled_filters, at->data))
    at = at->prev;
  return at;
}

static GdamFilter* find_next_enabled_filter(GdamBuilder* builder, GList* at)
{
  for (; at != NULL; at = at->next)
    {
      if (g_hash_table_lookup (builder->disabled_filters, at->data) == NULL)
	break;
    }
  if (at)
    return GDAM_FILTER (at->data);
  else
    return builder->dummy_gain_out;
}

static GdamFilter* find_prev_enabled_filter(GdamBuilder* builder, GList* at)
{
  for (; at != NULL; at = at->prev)
    {
      if (g_hash_table_lookup (builder->disabled_filters, at->data) == NULL)
	break;
    }
  if (at)
    return GDAM_FILTER (at->data);
  else
    return builder->dummy_gain_in;
}


void       gdam_builder_append_filter  (GdamBuilder*      builder,
                                        GdamFilter*       filter,
					gboolean          enabled,
					gboolean	  saveable)
{
  int new_source_id = GDAM_MODEL (filter)->source_id;
  GdamChannel *channel;
  GList *us;
  GList *last_enabled = find_last_enabled (builder);

  if (builder->dummy_gain_in == 0)
    {
      allocate_dummy_gains (builder);
    }

  if (saveable)
    {
      g_hash_table_insert (builder->saveable_filters,
			   filter, GINT_TO_POINTER (1));
    }

  /* Find the new child of `filter'. */
  /* Adjust filter-list/parent/child pointers */
  if (enabled)
    {
      GdamFilter *parent = builder->dummy_gain_out;
      GdamModel *child;
      if (last_enabled)
	child = last_enabled->data;
      else
	child = (GdamModel *) builder->dummy_gain_in;
      gdam_filter_disconnect (parent);
      parent->child = (GdamModel *) filter;
      filter->child = (GdamModel *) child;

      ((GdamModel *) filter)->parent = (GdamModel *) parent;
      ((GdamModel *) child)->parent = (GdamModel *) filter;

      gdam_filter_connect (parent);
      gdam_filter_connect (filter);
      gdam_model_set_beat_info (GDAM_MODEL (child),
				&GDAM_MODEL (child)->beat_info);
    }
  else
    {
      g_hash_table_insert (builder->disabled_filters,
			   filter, GINT_TO_POINTER (1));
    }

  g_return_if_fail (filter != NULL);

  channel = GDAM_MODEL (builder)->channel;

  if (enabled && builder->component)
    {
      int parent_id, child_id;
      GdamModel *child;
      if (last_enabled)
	{
	  child = GDAM_MODEL (last_enabled->data);
	}
      else
	{
	  child = GDAM_MODEL (builder->dummy_gain_in);
	  gdam_debug ("append_filter: no filters... using dummy");
	}
      child_id = child->source_id;
      parent_id = GDAM_MODEL (builder->dummy_gain_out)->source_id;
      gdam_debug ("append_filter: parent_id=%d, child_id=%d, "
		  "new_source_id=%d", parent_id, child_id, new_source_id);
      if (parent_id != 0 && child_id != 0)
	{
	  GdamRequestBuilder *request_builder;
	  request_builder = gdam_request_builder_new ();

	  if (child_id != 0)
	    {
	      if (new_source_id == 0)
		new_source_id =
		  gdam_filter_append_start_request (filter,
						    child_id,
						    request_builder);
	      else
		append_set_filtered_request (new_source_id,
					     child_id, request_builder);
	    }
	  append_set_filtered_request (parent_id,
				       new_source_id, request_builder);
	  gdam_channel_append_request (channel,
				       gdam_request_builder_package
				       (request_builder));
	  gdam_request_builder_destroy (request_builder);
	}
    }

  /* Append filter to list. */
  builder->filters_out = g_list_append (builder->filters_out, filter);
  if (builder->filters_in == NULL)
    builder->filters_in = builder->filters_out;
  else
    builder->filters_out = builder->filters_out->next;
  us = builder->filters_out;

  gtk_object_ref ((GtkObject *) filter);
  gtk_object_sink ((GtkObject *) filter);

  g_return_if_fail (gdam_builder_verify (builder));
}

/*
 * `next' is the *parent*, since forward in the list is
 * towards the output.
 */
void       gdam_builder_move_filter    (GdamBuilder*      builder,
                                        GdamFilter*       filter,
					int               new_index)
{
  GdamFilter *old_parent = NULL, *new_parent = NULL;
  GdamFilter *old_child = NULL, *new_child = NULL;
  GList *old_next_list;
  GList *this_list;
  GList *new_prev_list;
  GList *new_next_list;
  gboolean enabled;
  if (g_hash_table_lookup (builder->disabled_filters, filter) == NULL)
    enabled = TRUE;
  else
    enabled = FALSE;

  /* Find the new/old parents/children. */
  this_list = g_list_find (builder->filters_in, filter);
  g_return_if_fail (this_list != NULL);
  /* (Remove this_list) */
  if (this_list->next)
    this_list->next->prev = this_list->prev;
  else
    builder->filters_out = this_list->prev;
  if (this_list->prev)
    this_list->prev->next = this_list->next;
  else
    builder->filters_in = this_list->next;

  old_next_list = this_list->next;
  if (new_index == 0)
    {
      new_prev_list = NULL;
      new_next_list = builder->filters_in;
    }
  else
    {
      new_prev_list = g_list_nth (builder->filters_in, new_index - 1);
      new_next_list = new_prev_list->next;
      g_return_if_fail (new_prev_list != NULL);
    }
  if (new_prev_list == this_list || new_next_list == this_list)
    return;

  if (enabled)
    {
      old_child = GDAM_FILTER (filter->child);
      old_parent = GDAM_FILTER (GDAM_MODEL (filter)->parent);
      new_child = find_prev_enabled_filter (builder, new_prev_list);
      new_parent = find_next_enabled_filter (builder, new_next_list);

      /* Disconnect the children. */
      gdam_filter_disconnect (old_parent);
      gdam_filter_disconnect (new_parent);
      gdam_filter_disconnect (filter);
    }

  /* Shimmy around a million next and prev pointers. */
  /* (Add this_list) */
  this_list->prev = new_prev_list;
  this_list->next = new_next_list;
  if (this_list->next)
    this_list->next->prev = this_list;
  else
    builder->filters_out = this_list;
  if (this_list->prev)
    this_list->prev->next = this_list;
  else
    {
      builder->filters_in = this_list;
    }

  if (enabled)
    {
      GdamRequest **requests;
      int old_parent_id, new_parent_id, our_source_id;
      GdamArg arg;
      /* (Now do the exact same thing with
       *  parent/child pointers, except that you are
       *  guaranteed of the presence of sentinels.) */
      new_parent->child = (GdamModel *) filter;
      GDAM_MODEL (new_child)->parent = (GdamModel *) filter;
      old_parent->child = (GdamModel *) old_child;
      GDAM_MODEL (old_child)->parent = (GdamModel *) old_parent;
      filter->child = (GdamModel *) new_child;
      GDAM_MODEL (filter)->parent = (GdamModel *) new_parent;

      /* Append a package with two set "filtered" requests to 
       * rearrange things. */
      requests = g_new (GdamRequest *, 5);
      old_parent_id = ((GdamModel *) old_parent)->source_id;
      new_parent_id = ((GdamModel *) new_parent)->source_id;
      our_source_id = ((GdamModel *) filter)->source_id;
      arg.name = "filtered";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = old_parent->child->source_id;
      requests[0] = gdam_request_ref (GDAM_PROTOCOL_SOURCE_ARG,
				      our_source_id, NULL, NULL, NULL);
      requests[1] = gdam_request_set_args (old_parent_id,
					   1, &arg, NULL, NULL, NULL);
      our_source_id = ((GdamModel *) filter)->source_id;
      GDAM_ARG_INT (&arg) = ((GdamModel *) new_child)->source_id;
      requests[2] = gdam_request_set_args (our_source_id,
					   1, &arg, NULL, NULL, NULL);
      GDAM_ARG_INT (&arg) = our_source_id;
      requests[3] = gdam_request_set_args (new_parent_id,
					   1, &arg, NULL, NULL, NULL);
      requests[4] = gdam_request_unref (GDAM_PROTOCOL_SOURCE_ARG,
					our_source_id, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (builder)->channel,
				   gdam_request_package_v (5, requests,
							   NULL, NULL, NULL));

      /* Reconnect the various children. */
      gdam_filter_connect (old_parent);
      gdam_filter_connect (new_parent);
      gdam_filter_connect (filter);
    }
  g_return_if_fail (gdam_builder_verify (builder));
}

void gdam_builder_disable_filter(GdamBuilder* builder, GdamFilter* filter)
{
  GdamFilter *parent, *child;
  GdamRequestBuilder *request_builder;
  gint parent_id, child_id, our_id;

  g_return_if_fail (builder->dummy_gain_in != filter);
  g_return_if_fail (builder->dummy_gain_out != filter);
  g_return_if_fail (g_hash_table_lookup (builder->disabled_filters,
					 filter) == NULL);

  parent = GDAM_FILTER (GDAM_MODEL (filter)->parent);
  child = GDAM_FILTER (filter->child);
  g_return_if_fail (parent != NULL);
  g_return_if_fail (child != NULL);

  g_hash_table_insert (builder->disabled_filters,
		       filter, GINT_TO_POINTER (1));
  our_id = ((GdamModel *) filter)->source_id;
  if (our_id != 0)
    {
      parent_id = ((GdamModel *) parent)->source_id;
      child_id = ((GdamModel *) child)->source_id;
      request_builder = gdam_request_builder_new ();
      gdam_request_builder_append_generic (request_builder,
					   gdam_request_ref
					   (GDAM_PROTOCOL_SOURCE_ARG, our_id,
					    NULL, NULL, NULL));
      append_set_filtered_request (parent_id, child_id, request_builder);
      gdam_channel_append_request (GDAM_MODEL (filter)->channel,
				   gdam_request_builder_package
				   (request_builder));
      gdam_request_builder_destroy (request_builder);
    }

  /* Disconnect these filters. */
  gdam_filter_disconnect (filter);
  gdam_filter_disconnect (parent);

  filter->child = NULL;
  GDAM_MODEL (filter)->parent = NULL;
  GDAM_MODEL (child)->parent = (GdamModel *) parent;
  parent->child = (GdamModel *) child;

  gdam_filter_connect (parent);
  g_return_if_fail (gdam_builder_verify (builder));
}

void gdam_builder_enable_filter(GdamBuilder* builder, GdamFilter* filter)
{
  GdamFilter *parent, *child;
  GList *this_list;
  GdamRequestBuilder *request_builder = NULL;
  gint parent_id, child_id, our_id;

  g_return_if_fail (g_hash_table_lookup (builder->disabled_filters,
					 filter) == GINT_TO_POINTER (1));
  g_hash_table_remove (builder->disabled_filters, filter);

  this_list = g_list_find (builder->filters_in, filter);
  parent = find_next_enabled_filter (builder, this_list->next);
  child = find_prev_enabled_filter (builder, this_list->prev);
  g_return_if_fail (parent != NULL);
  g_return_if_fail (child != NULL);

  our_id = ((GdamModel *) filter)->source_id;
  parent_id = ((GdamModel *) parent)->source_id;
  child_id = ((GdamModel *) child)->source_id;
  gdam_debug ("Builder: enabling %p, source id %d, parent_id = %d",
	      filter, our_id, parent_id);
  if (our_id != 0)
    {
      request_builder = gdam_request_builder_new ();
      append_set_filtered_request (our_id, child_id, request_builder);
      append_set_filtered_request (parent_id, our_id, request_builder);
      gdam_request_builder_append_generic (request_builder,
					   gdam_request_unref
					   (GDAM_PROTOCOL_SOURCE_ARG, our_id,
					    NULL, NULL, NULL));
    }
  else if (parent_id != 0)
    {
      request_builder = gdam_request_builder_new ();
      our_id = gdam_filter_append_start_request (filter, child_id,
						 request_builder);
      append_set_filtered_request (parent_id, our_id, request_builder);
    }
  if (request_builder)
    {
      gdam_channel_append_request (GDAM_MODEL (filter)->channel,
				   gdam_request_builder_package
				   (request_builder));
      gdam_request_builder_destroy (request_builder);
    }

  /* Disconnect these filters. */
  gdam_filter_disconnect (parent);

  GDAM_MODEL (filter)->parent = (GdamModel *) parent;
  GDAM_MODEL (child)->parent = (GdamModel *) filter;
  filter->child = (GdamModel *) child;
  parent->child = (GdamModel *) filter;

  gdam_filter_connect (parent);
  gdam_filter_connect (filter);
  g_return_if_fail (gdam_builder_verify (builder));
}

void       gdam_builder_remove_filter  (GdamBuilder*      builder,
                                        GdamFilter*       to_remove)
{
  GList *list = g_list_find (builder->filters_in, to_remove);
  GdamChannel *channel = GDAM_MODEL (builder)->channel;

  g_return_if_fail (list != NULL);
  if (builder->filters_out != NULL && builder->filters_out->data == to_remove)
    builder->filters_out = builder->filters_out->prev;
  builder->filters_in = g_list_remove (builder->filters_in, to_remove);
  if (g_hash_table_lookup (builder->disabled_filters, to_remove) == NULL)
    {
      GdamFilter *next, *prev;
      int next_id, prev_id;
      /* Adjust the lists... */
      next = GDAM_FILTER (GDAM_MODEL (to_remove)->parent);
      prev = GDAM_FILTER (to_remove->child);
      gdam_filter_disconnect (next);
      gdam_filter_disconnect (to_remove);
      next->child = (GdamModel *) prev;
      next_id = GDAM_MODEL (next)->source_id;
      prev_id = GDAM_MODEL (prev)->source_id;
      if (next_id && prev_id)
	{
	  GdamArg arg;
	  arg.name = "filtered";
	  arg.type = GDAM_ARG_TYPE_INT;
	  GDAM_ARG_INT (&arg) = prev_id;
	  gdam_channel_append_request (channel,
				       gdam_request_set_args (next_id,
							      1,
							      &arg,
							      NULL,
							      NULL, NULL));
	}
      ((GdamModel *) prev)->parent = (GdamModel *) next;
      gdam_filter_connect (next);
    }
  else
    {
      int source_id = GDAM_MODEL (to_remove)->source_id;
      if (source_id > 0)
	{
	  gdam_channel_append_request (channel,
				       gdam_request_unref
				       (GDAM_PROTOCOL_SOURCE_ARG, source_id,
					NULL, NULL, NULL));
	}
      g_hash_table_remove (builder->disabled_filters, to_remove);
    }

  if (g_hash_table_lookup (builder->saveable_filters, to_remove) != NULL)
    {
      g_hash_table_remove (builder->saveable_filters, to_remove);
    }

  to_remove->child = NULL;
  GDAM_MODEL (to_remove)->parent = NULL;
  gtk_object_unref ((GtkObject *) to_remove);
  g_return_if_fail (gdam_builder_verify (builder));
}

GdamModel* gdam_builder_get_component(GdamBuilder* builder)
{
  return builder->component;
}
void gdam_builder_set_component(GdamBuilder* builder,
                                GdamModel* component)
{
  if (builder->component == component)
    return;

  gtk_object_ref ((GtkObject *) component);
  gtk_object_sink ((GtkObject *) component);
  if (builder->dummy_gain_in == NULL)
    {
      allocate_dummy_gains (builder);
    }
  if (builder->component)
    {
      gtk_object_unref (GTK_OBJECT (builder->component));
      builder->component = NULL;
    }
  gdam_filter_set_child (builder->dummy_gain_in, component);
  builder->component = component;
  g_return_if_fail (builder->component->parent != NULL);
  g_return_if_fail (gdam_builder_verify (builder));
}
static void gdam_builder_destroy(GtkObject* object)
{
  GdamBuilder *builder = GDAM_BUILDER (object);
  if (builder->component != NULL)
    {
      GtkObject *object = GTK_OBJECT (builder->component);
      builder->component = NULL;
      gtk_object_unref (object);
    }
  if (builder->dummy_gain_in)
    {
      gtk_object_unref (GTK_OBJECT (builder->dummy_gain_in));
      gtk_object_unref (GTK_OBJECT (builder->dummy_gain_out));
      builder->dummy_gain_in = NULL;
      builder->dummy_gain_out = NULL;
    }
  if (parent_class)
    (*parent_class->destroy) (object);
}

gboolean gdam_builder_verify(GdamBuilder* builder)
{
  GdamFilter *child = builder->dummy_gain_in;
  GList *at;
  for (at = builder->filters_in; at != NULL; at = at->next)
    {
      GdamFilter *filter = GDAM_FILTER (at->data);
      g_return_val_if_fail (filter != NULL, FALSE);
      if (at->next == NULL)
	g_return_val_if_fail (at == builder->filters_out, FALSE);
      if (at->prev == NULL)
	g_return_val_if_fail (at == builder->filters_in, FALSE);
      else
	g_return_val_if_fail (at->prev->next == at, FALSE);
      if (g_hash_table_lookup (builder->disabled_filters, filter))
	{
	  g_return_val_if_fail (filter->child == NULL, FALSE);
	  g_return_val_if_fail (GDAM_MODEL (filter)->parent == NULL, FALSE);
	}
      else
	{
	  g_return_val_if_fail (GDAM_MODEL (child)->parent
				== GDAM_MODEL (filter), FALSE);
	  g_return_val_if_fail (filter->child == (GdamModel *) child, FALSE);
	  child = filter;
	}
    }
  g_return_val_if_fail (builder->dummy_gain_out->child
			== (GdamModel *) child, FALSE);
  g_return_val_if_fail (GDAM_MODEL (child)->parent
			== GDAM_MODEL (builder->dummy_gain_out), FALSE);
  return TRUE;
}

static void         
gdam_builder_new_request(GdamModel*         builder,
                         GdamModel*         dummy_filter,
			 GdamPlayInfo*          play_info,
                         int                    dummy_gain_id,
			 GdamRequestBuilder*    request_builder)
{
  GdamBuilder *b = GDAM_BUILDER (builder);
  g_return_if_fail (b->dummy_gain_out == GDAM_FILTER (dummy_filter));

  gdam_model_new_child_request (builder->parent,
				builder,
				play_info, dummy_gain_id, request_builder);
}
