/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamfilter.h"
#include <gtk/gtksignal.h>

/* 
 *    ____ _                 ___       _ _   
 *   / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 *  | |   | |/ _` / __/ __|  | || '_ \| | __|
 *  | |___| | (_| \__ \__ \  | || | | | | |_ 
 *   \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */

/* Functions for filter class */
static gboolean gdam_filter_attach_child(GdamFilter* filter, 
					 GdamModel*  model);
static gboolean gdam_filter_detach_child(GdamFilter* filter, 
					 GdamModel*  model);
static void filter_on_child_destroy(GdamFilter* filter);
static void gdam_filter_on_child_set_source(GdamFilter* filter,
			                    guint new_source_id);
static void gdam_filter_on_child_set_beat_info(GdamFilter* filter,
			                       const GdamBeatInfo* beat_info);
static void gdam_filter_destroy       (GtkObject*    object);
static void gdam_filter_real_connect(GdamFilter* filter);
static void gdam_filter_real_disconnect(GdamFilter* filter);
static void gdam_filter_new_request(GdamModel*      filter,
                                    GdamModel*      child,
				    GdamPlayInfo*       play_info,
				    int                 local_src_id,
				    GdamRequestBuilder* builder);
#if 0
static void gdam_filter_real_on_child_set_source(GdamFilter*     filter,
	                                         guint       source_id);
#endif
static int gdam_filter_real_append_start_request (GdamFilter*     filter,
	                                          int             child_id,
						  GdamRequestBuilder*);

static gdouble gdam_filter_get_length(GdamModel* model)
{
  GdamFilter *filter = GDAM_FILTER (model);
  GdamModel *child = filter->child;
  GdamModelClass *klass;
  if (!child)
    return -1;
  klass = GDAM_MODEL_CLASS (GTK_OBJECT (child)->klass);
  if (klass->get_length != NULL)
    return (*klass->get_length) (child);
  return -1;
}

static GtkObjectClass* parent_class = NULL;
static void gdam_filter_class_init(GdamFilterClass* klass)
{
  klass->attach_child = gdam_filter_attach_child;
  klass->detach_child = gdam_filter_detach_child;
  klass->connect = gdam_filter_real_connect;
  klass->disconnect = gdam_filter_real_disconnect;
  klass->append_start_request = gdam_filter_real_append_start_request;
  klass->on_child_set_source = NULL;
  klass->on_child_set_beat_info = gdam_filter_on_child_set_beat_info;
  GDAM_MODEL_CLASS (klass)->new_request = gdam_filter_new_request;
  GDAM_MODEL_CLASS (klass)->get_length = gdam_filter_get_length;
  GTK_OBJECT_CLASS (klass)->destroy = gdam_filter_destroy;
}

static void gdam_filter_init(GdamFilter* echo)
{
}

GtkType gdam_filter_get_type()
{
  static GtkType filter_type = 0;
  if (!filter_type)
    {
      static const GtkTypeInfo echo_info = {
	"GdamFilter",
	sizeof (GdamFilter),
	sizeof (GdamFilterClass),
	(GtkClassInitFunc) gdam_filter_class_init,
	(GtkObjectInitFunc) gdam_filter_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAM_TYPE_MODEL);
      filter_type = gtk_type_unique (GDAM_TYPE_MODEL, &echo_info);
    }
  return filter_type;
}


/*
 *     ____      _ _ _                _        
 *    / ___|__ _| | | |__   __ _  ___| | _____ 
 *   | |   / _` | | | '_ \ / _` |/ __| |/ / __|
 *   | |__| (_| | | | |_) | (_| | (__|   <\__ \
 *    \____\__,_|_|_|_.__/ \__,_|\___|_|\_\___/
 */

static void filter_on_child_destroy(GdamFilter* filter)
{
  GdamFilterClass *filter_class;
  GdamModel *child = filter->child;
  filter_class = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);
  (*filter_class->detach_child) (filter, child);
  filter->child = NULL;
}

static void handle_filter_child_set_beat_info(GdamFilter* filter,
			                  const GdamBeatInfo*  bi)
{
  GdamFilterClass *filter_class;
  filter_class = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);
  (*filter_class->on_child_set_beat_info) (filter, bi);
}

static void gdam_filter_on_child_set_source(GdamFilter* filter,
			        guint new_source_id)
{
  void (*set_source_func) (GdamFilter *, guint);
  GdamFilterClass *klass;
  klass = GDAM_FILTER_CLASS (((GtkObject *) filter)->klass);
  set_source_func = klass->on_child_set_source;
  if (set_source_func)
    (*set_source_func) (filter, new_source_id);
}

static int gdam_filter_real_append_start_request (GdamFilter*     filter,
	                                          int             child_id,
						  GdamRequestBuilder* builder)
{
  g_warning ("gdam_filter_real_append_start_request: "
	     "never should be called");
  return 0;
}

static gboolean gdam_filter_detach_child(GdamFilter* filter,
					 GdamModel*  model)
{
  if (!filter)
    return FALSE;

  if (filter->child != model)
    {
      gdam_debug ("Asked to detach child that isn't ours.");
      return FALSE;
    }

  /* Remove references to old child */
  gdam_filter_disconnect (filter);
  if (filter->child)
    {
      GtkObject *child = GTK_OBJECT (filter->child);
      /* XXX: is there a better way to resolve this problem? */
      if (child->ref_count > 1)
	filter->child->parent = NULL;
      gtk_object_unref (GTK_OBJECT (filter->child));
    }

  filter->child = NULL;

  return TRUE;
}

static void gdam_filter_real_disconnect(GdamFilter* filter)
{
  if (filter->destroy_connect_id != 0)
    {
      gtk_signal_disconnect (GTK_OBJECT (filter->child),
			     filter->destroy_connect_id);
      gtk_signal_disconnect (GTK_OBJECT (filter->child),
			     filter->set_source_connect_id);
      gtk_signal_disconnect (GTK_OBJECT (filter->child),
			     filter->set_beat_info_connect_id);
      filter->destroy_connect_id = 0;
      filter->set_source_connect_id = 0;
      filter->set_beat_info_connect_id = 0;
    }
}

static void gdam_filter_real_connect(GdamFilter* filter)
{
  GdamFilterClass *filter_class;
  filter_class = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);

  if (filter->destroy_connect_id != 0)
    gdam_filter_real_disconnect (filter);

  /* Trap our child's destroy signal so we know when it is deleted */
  filter->destroy_connect_id =
    gtk_signal_connect_object (GTK_OBJECT (filter->child), "destroy",
			       GTK_SIGNAL_FUNC (filter_on_child_destroy),
			       GTK_OBJECT (filter));

  /* Trap our child's set_source signal so we know when it is changed */
  filter->set_source_connect_id =
    gtk_signal_connect_object (GTK_OBJECT (filter->child), "set_source",
			       GTK_SIGNAL_FUNC
			       (gdam_filter_on_child_set_source),
			       GTK_OBJECT (filter));

  /* Trap our child's set_beat_info signal */
  filter->set_beat_info_connect_id =
    gtk_signal_connect_object (GTK_OBJECT (filter->child), "set_beat_info",
			       GTK_SIGNAL_FUNC
			       (handle_filter_child_set_beat_info),
			       GTK_OBJECT (filter));
}

#if 0
static void gdam_filter_real_on_child_set_source(GdamFilter*     filter,
	                                         guint       source_id)
{
  GdamRequest *request;
  GdamFilterClass *filter_class;
  if (source_id == 0)
    {
      gdam_model_set_source_id (GDAM_MODEL (filter), 0);
      return;
    }
  filter_class = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);
  request = (*filter_class->build_source_request) (filter, source_id);
  g_assert (request != NULL);
  gdam_channel_append_request (GDAM_MODEL (filter)->channel, request);
}
#endif

static void set_args_failed(GdamFilter* filter)
{
  /* TODO: rebuild filter in this case. */
  gtk_object_unref (GTK_OBJECT (filter));
}

static void gdam_filter_new_request(GdamModel*      filter,
                                    GdamModel*      child,
				    GdamPlayInfo*       play_info,
				    int                 local_src_id,
				    GdamRequestBuilder* builder)
{
  if (local_src_id == 0)
    {
      if (filter->source_id != 0)
	{
	  GdamRequest *set;
	  GdamArg arg;
	  gdam_filter_append_stop_request (GDAM_FILTER (filter), builder);
	  arg.type = GDAM_ARG_TYPE_INT;
	  arg.name = "filtered";
	  GDAM_ARG_INT (&arg) = 0;
	  gtk_object_ref (GTK_OBJECT (filter));
	  set = gdam_request_set_args (filter->source_id,
				       1, &arg, NULL, NULL, NULL);
	  gdam_request_builder_append_generic (builder, set);

	}
      gdam_model_new_child_request (filter->parent,
				    filter, play_info, 0, builder);
      gdam_model_set_source_id (filter, 0);
      return;
    }
  if (filter->source_id != 0)
    {
      GdamRequest *set, *package;
      GdamArg arg;
      arg.type = GDAM_ARG_TYPE_INT;
      arg.name = "filtered";
      GDAM_ARG_INT (&arg) = local_src_id;
      gtk_object_ref (GTK_OBJECT (filter));
      set = gdam_request_set_args (filter->source_id,
				   1, &arg,
				   (GHookFunc) gtk_object_unref,
				   (GHookFunc) set_args_failed, filter);
      gdam_request_builder_append_generic (builder, set);
      package = gdam_request_builder_package (builder);
      gdam_channel_append_request (filter->channel, package);
    }
  else
    {
      int id;
      /* Build filter. */
      id = gdam_filter_append_start_request (GDAM_FILTER (filter),
					     local_src_id, builder);
      gdam_model_new_child_request (filter->parent,
				    filter, play_info, id, builder);
    }
}

void gdam_filter_disconnect(GdamFilter* filter)
{
  GdamFilterClass *klass;
  klass = GDAM_FILTER_CLASS (((GtkObject *) filter)->klass);
  (*klass->disconnect) (filter);
}
void gdam_filter_connect(GdamFilter* filter)
{
  GdamFilterClass *klass;
  klass = GDAM_FILTER_CLASS (((GtkObject *) filter)->klass);
  (*klass->connect) (filter);
}

static gboolean gdam_filter_attach_child(GdamFilter* filter, 
				      	 GdamModel*  model)
{

  float comp_bpm;
  if (!model)
    return FALSE;

  filter->child = model;
  gtk_object_ref (GTK_OBJECT (filter->child));
  gtk_object_sink (GTK_OBJECT (filter->child));
  gdam_filter_connect (filter);
  model->parent = GDAM_MODEL (filter);
  comp_bpm = gdam_beat_info_get_bpm (&model->beat_info, 0);
  if (comp_bpm > 0.0)
    {
      gdam_model_set_beat_info (model, &model->beat_info);
    }
  return TRUE;
}
static void gdam_filter_destroy       (GtkObject*    object)
{
  gdam_filter_set_child (GDAM_FILTER (object), NULL);
  (*parent_class->destroy) (object);
}

GdamModel* gdam_filter_get_child(GdamFilter* filter)
{
  return filter->child;
}

void gdam_filter_set_child    (GdamFilter* filter, 
			       GdamModel* model)
{
  GdamFilterClass *klass;
  if (filter->child == model)
    return;

  klass = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);
  if (filter->child)
    {
      (*klass->detach_child) (filter, filter->child);
      g_assert (filter->child == NULL);
    }
  if (model)
    (*klass->attach_child) (filter, model);
}



static void
gdam_filter_on_child_set_beat_info(GdamFilter*         filter,
			           const GdamBeatInfo* beat_info)
{
  gdam_model_set_beat_info (GDAM_MODEL (filter), beat_info);
}

int          gdam_filter_append_start_request(GdamFilter*           filter,
                                              int                   child_id,
					      GdamRequestBuilder*   builder)
{
  GdamFilterClass *filter_class;
  filter_class = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);
  gdam_debug ("gdam_filter_append_start_request: %s",
	      gtk_type_name (GTK_OBJECT (filter)->klass->type));
  return (*filter_class->append_start_request) (filter, child_id, builder);
}
void         gdam_filter_append_stop_request (GdamFilter*           filter,
                                              GdamRequestBuilder*   builder)
{
  GdamFilterClass *filter_class;
  filter_class = GDAM_FILTER_CLASS (GTK_OBJECT (filter)->klass);
  if (filter_class->append_stop_request == NULL)
    return;
  (*filter_class->append_stop_request) (filter, builder);
}
