/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdaminputdevice.h"
#include <gtk/gtksignal.h>
#include <common/gdamsearchpath.h>
#include <common/gdamplugin.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

static GtkObjectClass* parent_class = NULL;

typedef struct _GdamInputDeviceConfiguration GdamInputDeviceConfiguration;
struct _GdamInputDeviceConfiguration {
	char*			device_name;
	char*			patch_file;
	char*			device_type;
	GdamInputDevice*	input_device;
};
static void gdam_input_device_real_event(GdamInputDevice*  input_device, 
	                                 GdamInputEvent*   event,
	                                 GdamInputTrap*    input_trap);
static void gdam_input_device_real_new_trap(GdamInputDevice*  input_device, 
	                                    GdamInputTrap*    input_trap);

static GHashTable* config_table = NULL;
static GHashTable* config_type_table = NULL;

enum {
	NEW_TRAP,
	EVENT,
	LAST_SIGNAL
};

static guint input_signals[LAST_SIGNAL] = { 0 };

typedef struct _GdamInputHook GdamInputHook;

static void gdam_input_device_init(GdamInputDevice* input_device)
{
}
static void gdam_input_device_class_init(GdamInputDeviceClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  input_signals[NEW_TRAP]
    = gtk_signal_new ("new-trap",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamInputDeviceClass,
					 new_trap),
		      gtk_marshal_NONE__POINTER,
		      GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  input_signals[EVENT]
    = gtk_signal_new ("event",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamInputDeviceClass,
					 event),
		      gtk_marshal_NONE__POINTER_POINTER,
		      GTK_TYPE_NONE, 2, GTK_TYPE_POINTER, GTK_TYPE_POINTER);
  gtk_object_class_add_signals (object_class, input_signals, LAST_SIGNAL);
  klass->event = gdam_input_device_real_event;
  klass->new_trap = gdam_input_device_real_new_trap;
}

/* Map from GdamInputEvent*  =>  GdamInputTrap*  */
static GHashTable* trap_table = NULL;
static GHashTable* trap_id_table = NULL;
static guint       last_trap_id = 0;


GdamInputTrap*   gdam_input_trap_find        (GdamInputEvent*    input_event)
{
  if (trap_table == NULL)
    return NULL;

  return g_hash_table_lookup (trap_table, input_event);
}

GdamInputTrap*   gdam_input_trap_find_or_create(GdamInputEvent*    input_event)
{
  GdamInputTrap *trap;
  GdamInputDevice *device = input_event->input_device;
  if (trap_table == NULL)
    {
      trap_table = g_hash_table_new (
				     (GHashFunc) gdam_input_event_hash,
				     (GCompareFunc) gdam_input_event_equal);
      trap_id_table = g_hash_table_new (NULL, NULL);
    }

  trap = g_hash_table_lookup (trap_table, input_event);
  if (trap != NULL)
    return trap;

  trap = g_new0 (GdamInputTrap, 1);
  trap->event = gdam_input_event_canonicalize (input_event);
  if (trap->event == NULL)
    {
      g_free (trap);
      g_warning ("error canonicalized input-event");
      return NULL;
    }
  trap->hooks = NULL;
  trap->trap_id = ++last_trap_id;
  g_hash_table_insert (trap_id_table, GUINT_TO_POINTER (last_trap_id), trap);
  g_hash_table_insert (trap_table, trap->event, trap);
  device->traps = g_slist_prepend (device->traps, trap);
  gtk_signal_emit (GTK_OBJECT (input_event->input_device),
		   input_signals[NEW_TRAP], trap);
  return trap;
}

GdamInputTrap* gdam_input_trap_get_by_id(guint id)
{
  if (trap_id_table == NULL)
    return NULL;
  return g_hash_table_lookup (trap_id_table, GUINT_TO_POINTER (id));
}

struct _GdamInputHook {
	GdamInputTrap*		trap;
	GdamInputHookFunc       hook_func;
	GDestroyNotify          destroy;
	gpointer                user_data;
	GtkObject*              alive_object;
	guint			alive_object_destroy;
	guint			hook_id;
	gint 			block_count;
};

static GHashTable* hook_id_table = NULL;
static guint       last_hook_id  = 0;

static guint gdam_input_hook_alloc_id(GdamInputHook* hook)
{
  guint id;
  if (hook_id_table == NULL)
    hook_id_table = g_hash_table_new (NULL, NULL);
  id = ++last_hook_id;
  hook->hook_id = id;
  g_hash_table_insert (hook_id_table, GINT_TO_POINTER (id), hook);
  return id;
}
static GdamInputHook* gdam_input_hook_find_by_id(guint hook)
{
  return g_hash_table_lookup (hook_id_table, GINT_TO_POINTER (hook));
}

static void gdam_input_hook_destroy(GdamInputHook* input_hook)
{
  if (input_hook->trap != NULL)
    {
      GdamInputTrap *trap = input_hook->trap;
      trap->hooks = g_slist_remove (trap->hooks, input_hook);
    }
  if (input_hook->alive_object != NULL)
    gtk_signal_disconnect (input_hook->alive_object,
			   input_hook->alive_object_destroy);
  if (input_hook->destroy != NULL)
    (*input_hook->destroy) (input_hook->user_data);
  g_hash_table_remove (hook_id_table, GINT_TO_POINTER (input_hook->hook_id));
  g_free (input_hook);
}

static void alive_object_destroyed(GtkObject* object, GdamInputHook* hook)
{
  hook->alive_object = NULL;
  hook->alive_object_destroy = 0;
  gdam_input_hook_destroy (hook);
}

guint            gdam_input_hook_connect     (GdamInputTrap*     trap,
				              GdamInputHookFunc  hook_func,
				              GDestroyNotify     destroy,
				              gpointer           user_data,
				              GtkObject*         alive_object)
{
  GdamInputHook *hook;
  hook = g_new (GdamInputHook, 1);
  hook->trap = trap;
  hook->hook_func = hook_func;
  hook->destroy = destroy;
  hook->user_data = user_data;
  hook->block_count = 0;
  hook->alive_object = alive_object;
  hook->alive_object_destroy
    = gtk_signal_connect (alive_object,
			  "destroy",
			  GTK_SIGNAL_FUNC (alive_object_destroyed), hook);
  trap->hooks = g_slist_append (trap->hooks, hook);
  return gdam_input_hook_alloc_id (hook);
}

void             gdam_input_hook_disconnect  (guint              hook_id)
{
  GdamInputHook *hook;
  hook = gdam_input_hook_find_by_id (hook_id);

  if (hook == NULL)
    return;
  gdam_input_hook_destroy (hook);
}
void             gdam_input_hook_block       (guint              hook_id)
{
  GdamInputHook *hook;
  hook = gdam_input_hook_find_by_id (hook_id);
  g_return_if_fail (hook != NULL);
  hook->block_count++;
}
void             gdam_input_hook_unblock     (guint              hook_id)
{
  GdamInputHook *hook;
  hook = gdam_input_hook_find_by_id (hook_id);
  g_return_if_fail (hook != NULL);
  g_return_if_fail (hook->block_count > 0);
  hook->block_count--;

}

/*
 * The callstack for emit is basically:
 *
 *       gdam_input_trap_emit
 *            GdamInputEvent::event
 *                  hook_funcs
 *
 * ie everything goes through the virtual `event',
 * so that perhaps derivers can act as event filters.
 * Also allows use gtk_signal_emission_stop.
 */
static void gdam_input_device_real_event(GdamInputDevice*  input_device, 
	                                 GdamInputEvent*   event,
	                                 GdamInputTrap*    input_trap)
{
  GSList *list;
  for (list = input_trap->hooks; list != NULL; list = list->next)
    {
      GdamInputHook *hook;
      hook = list->data;
      if (hook->block_count == 0)
	(*hook->hook_func) (event, hook->user_data);
    }
}
void             gdam_input_trap_emit        (GdamInputTrap*     input_trap,
                                              GdamInputEvent*    input_event)
{

  if (input_trap == NULL)
    input_trap = gdam_input_trap_find_or_create (input_event);
  g_return_if_fail (input_trap != NULL);
  gtk_signal_emit (GTK_OBJECT (input_event->input_device),
		   input_signals[EVENT], input_event, input_trap);
}

static void gdam_input_device_real_new_trap(GdamInputDevice*  input_device, 
	                                    GdamInputTrap*    input_trap)
{
  gdam_debug ("gdam_input_device_real_new_trap [id %d]", input_trap->trap_id);
}

GtkType gdam_input_device_get_type()
{
  static GtkType input_device_type = 0;
  if (!input_device_type)
    {
      static const GtkTypeInfo input_device_info = {
	"GdamInputDevice",
	sizeof (GdamInputDevice),
	sizeof (GdamInputDeviceClass),
	(GtkClassInitFunc) gdam_input_device_class_init,
	(GtkObjectInitFunc) gdam_input_device_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      parent_class = gtk_type_class (parent);
      input_device_type = gtk_type_unique (parent, &input_device_info);

      gdam_input_event_get_type ();
      gdam_input_event_range_get_type ();
      gdam_input_event_int_get_type ();
      gdam_input_event_string_get_type ();

    }

  return input_device_type;
}

/*
 * Global default trap.
 */
static GdamInputTrap* default_trap = NULL;
void gdam_input_trap_set_default(GdamInputTrap* trap)
{
  default_trap = trap;
}
GdamInputTrap* gdam_input_trap_get_default()
{
  return default_trap;
}


/*
 * Naming the traps.
 */
void             gdam_input_trap_set_name    (GdamInputTrap*     trap,
                                              const char*        name)
{
  if (trap->name)
    g_free (trap->name);
  trap->name = g_strdup (name);
  trap->name_is_generated = FALSE;
}
const char*      gdam_input_trap_get_name    (GdamInputTrap*     trap)
{
  if (trap->name == NULL)
    {
      trap->name = gdam_input_event_stringify (trap->event);
      trap->name_is_generated = TRUE;
    }
  return trap->name;
}

/*
 * Iterating the Traps.
 */

typedef struct _IterateData IterateData;
struct _IterateData {
	gpointer iterate_data;
	GdamInputTrapIterator iterate_func;
};

static void for_one_device(const char*       device_name,
                            GdamInputDeviceConfiguration* configuration,
			    IterateData*      data)
{
  GdamInputDevice *device = configuration->input_device;
  if (device != NULL)
    g_slist_foreach (device->traps,
		     (GFunc) data->iterate_func, data->iterate_data);
}

void             gdam_input_trap_forall      (GdamInputDevice*   device,
					      GdamInputTrapIterator iter_func,
					      gpointer           iter_data)
{
  IterateData iterate_data;
  iterate_data.iterate_func = iter_func;
  iterate_data.iterate_data = iter_data;
  if (device == NULL)
    {
      g_hash_table_foreach (config_table,
			    (GHFunc) for_one_device, &iterate_data);
    }
  else if (device->traps != NULL)
    {
      g_slist_foreach (device->traps, (GFunc) iter_func, iter_data);
    }
}

/*=====================================================================
 * Setting arguments when a event is tripped.
 */

typedef struct _GdamInputArg GdamInputArg;
struct _GdamInputArg {
	GdamInputTrap*	trap;
	GdamArg		arg;

	GtkObject*      object;

	guint		connect_id;

	/* For range arguments. */
	gdouble		min_value, max_value;
};

static void gdam_input_arg_free(GdamInputArg* arg)
{
  gdam_arg_destruct (1, &arg->arg);
  g_free (arg);
}

/* 
 * `event_type' must be a GdamInputEventRange.
 */
static void on_range_set_arg(GdamInputEvent*     input_event,
			     gpointer            user_data)
{
  GdamInputEventRange *range;
  GdamInputArg *input_arg = user_data;
  gdouble value;
  range = (GdamInputEventRange *) input_event;
  value = (range->value - range->min_value)
    / (range->max_value - range->min_value)
    * (input_arg->max_value - input_arg->min_value) + input_arg->min_value;
  GDAM_ARG_DOUBLE (&input_arg->arg) = value;
  gdam_arg_set_object_arg (input_arg->object, &input_arg->arg);
}

guint gdam_input_hook_connect_range(GdamInputTrap*   trap,
                                    GtkObject*       object,
				    const char*      arg_name,
				    gdouble          min_value,
				    gdouble          max_value)
{
  GdamInputArg *input_arg;

  if (!gdam_input_event_type_is_a (trap->event->event_type,
				   gdam_input_event_range_get_type ()))
    {
      g_warning ("only GdamInputEventRange's (not %s) may be connected"
		 " through %s",
		 gdam_input_event_type_name (trap->event->event_type),
		 __FUNCTION__);
      return 0;
    }

  input_arg = g_new (GdamInputArg, 1);
  input_arg->trap = trap;
  input_arg->arg.type = GDAM_ARG_TYPE_DOUBLE;
  input_arg->arg.name = g_strdup (arg_name);
  GDAM_ARG_DOUBLE (&input_arg->arg) = min_value;
  input_arg->object = object;
  input_arg->min_value = min_value;
  input_arg->max_value = max_value;
  input_arg->connect_id = gdam_input_hook_connect
    (trap,
     on_range_set_arg,
     (GDestroyNotify) gdam_input_arg_free, input_arg, object);
  return input_arg->connect_id;
}


/*
 * Setting constant arguments when a GdamInputEvent is trapped.
 */
static void on_const_set_arg(GdamInputEvent*     input_event,
			     gpointer            user_data)
{
  GdamInputArg *input_arg = user_data;
  gdam_arg_set_object_arg (input_arg->object, &input_arg->arg);
}

guint gdam_input_hook_connect_const(GdamInputTrap*   trap,
                                    GtkObject*       object,
				    GdamArg*         arg)
{
  GdamInputArg *input_arg;

  input_arg = g_new (GdamInputArg, 1);
  input_arg->trap = trap;
  input_arg->arg = *arg;
  input_arg->object = object;
  input_arg->connect_id = gdam_input_hook_connect
    (trap,
     on_const_set_arg,
     (GDestroyNotify) gdam_input_arg_free, input_arg, object);
  return input_arg->connect_id;
}


/*
 *      ``Master Patch File''
 *
 * Store collections of triplets:
 *         device_name		patch_file	Device
 *
 * So that the devices can be opened on demand.
 */
void gdam_input_device_configure(const char* device_name,
                                 const char* patch_file,
				 const char* device_type)
{
  GdamInputDeviceConfiguration *configuration;
  g_return_if_fail (device_name != NULL);
  g_return_if_fail (device_type != NULL);
  configuration = g_new (GdamInputDeviceConfiguration, 1);
  configuration->device_name = g_strdup (device_name);
  configuration->patch_file = g_strdup (patch_file);
  configuration->device_type = g_strdup (device_type);
  configuration->input_device = NULL;
  if (config_table == NULL)
    {
      config_table = g_hash_table_new (g_str_hash, g_str_equal);
      config_type_table = g_hash_table_new (g_str_hash, g_str_equal);
    }
  g_hash_table_insert (config_table,
		       configuration->device_name, configuration);
  g_hash_table_insert (config_type_table,
		       configuration->device_type, configuration);
}

/* Read input.devices:
 *
 *         /dev/midi???        patch-file-name
 */
gboolean gdam_input_device_load_database(const char* filename)
{
  FILE *fp;
  gboolean must_free_filename = FALSE;
  char buf[4096];
  int lineno = 0;

  /* Find ~/.gdam/midi.devices */
  if (filename == NULL)
    {
      filename = gdam_search_path ("input.devices",
				   NULL, GDAM_SEARCH_PATH_FIND_FILE);
      must_free_filename = TRUE;
    }

  if (filename == NULL)
    {
      g_message ("no input.devices file -- "
		 "no input patch information files loaded");
      return FALSE;
    }

  fp = fopen (filename, "r");
  if (fp == NULL)
    {
      g_warning ("error opening input device database file %s: %s",
		 filename, strerror (errno));
      return FALSE;
    }
  while (fgets (buf, sizeof (buf), fp))
    {
      const char *start_dev_name;
      const char *start_path_fname;
      const char *tmp;
      char *module_name;
      char *path_fname;
      char *dev_name;
      ++lineno;
      start_dev_name = buf;
      while (*start_dev_name != '\0' && isspace (*start_dev_name))
	start_dev_name++;
      if (*start_dev_name == '\0' || *start_dev_name == '#')
	continue;
      tmp = start_dev_name;
      while (*tmp != '\0' && !isspace (*tmp))
	tmp++;
      if (tmp == start_dev_name || *tmp == '\0')
	{
	  g_warning ("too few arguments on line %d", lineno);
	  continue;
	}
      dev_name = alloca ((tmp - start_dev_name) + 1);
      memcpy (dev_name, start_dev_name, tmp - start_dev_name);
      dev_name[tmp - start_dev_name] = '\0';

      start_path_fname = tmp;
      while (*start_path_fname != '\0' && isspace (*start_path_fname))
	start_path_fname++;
      tmp = start_path_fname;
      while (*tmp != '\0' && !isspace (*tmp))
	tmp++;
      if (tmp == start_path_fname)
	{
	  g_warning ("no path on line %d", lineno);
	  continue;
	}
      path_fname = alloca ((tmp - start_path_fname) + 1);
      memcpy (path_fname, start_path_fname, tmp - start_path_fname);
      path_fname[tmp - start_path_fname] = '\0';

      while (*tmp != '\0' && isspace (*tmp))
	tmp++;

      if (*tmp)
	{
	  const char *end = tmp;
	  while (*end != '\0' && !isspace (*end))
	    end++;
	  module_name = alloca (end + 1 - tmp);
	  memcpy (module_name, tmp, end - tmp);
	  module_name[end - tmp] = '\0';
	}
      else
	{
	  module_name = NULL;
	  g_warning ("line %d in %s didn't have "
		     "device_type for %s", lineno, filename, dev_name);
	  continue;
	}
      gdam_input_device_configure (dev_name, path_fname, module_name);
    }

  fclose (fp);
  return TRUE;
}

GdamInputDevice* gdam_input_device_get_by_name(const char* device_name)
{
  GdamInputDeviceConfiguration *configuration;
  GdamInputDevice *input_device;
  GtkType type;
  if (config_table == NULL)
    return NULL;
  configuration = g_hash_table_lookup (config_table, device_name);
  if (configuration == NULL)
    return NULL;
  if (configuration->input_device != NULL)
    return configuration->input_device;
  type = gdam_plugin_get_type_by_name ("model", configuration->device_type);
  if (type == 0)
    {
      g_warning ("couldn't load input device module %s",
		 configuration->device_type);
      return NULL;
    }
  if (!gtk_type_is_a (type, GDAM_TYPE_INPUT_DEVICE))
    {
      g_warning ("%s was not an input device", configuration->device_type);
      return NULL;
    }
  input_device = (GdamInputDevice *) gtk_type_new (type);
  if (!gdam_input_device_open (input_device, configuration->device_name))
    {
      g_warning ("error opening %s", configuration->device_name);
      gtk_object_sink ((GtkObject *) input_device);
      return NULL;
    }
  configuration->input_device = input_device;
  gtk_object_ref ((GtkObject *) configuration->input_device);
  gtk_object_sink ((GtkObject *) configuration->input_device);
  return input_device;
}

GdamInputDevice* gdam_input_device_get_by_type(const char* type_name)
{
  GdamInputDeviceConfiguration *configuration;
  if (config_type_table == NULL)
    return NULL;
  configuration = g_hash_table_lookup (config_type_table, type_name);
  if (configuration == NULL)
    return NULL;
  return gdam_input_device_get_by_name (configuration->device_name);
}

gboolean gdam_input_device_open(GdamInputDevice* input_device,
                                const char*      device_name)
{
  GdamInputDeviceClass *klass;
  klass = GDAM_INPUT_DEVICE_CLASS (GTK_OBJECT (input_device)->klass);
  g_free (input_device->device_name);
  input_device->device_name = g_strdup (device_name);
  g_return_val_if_fail (klass->open != NULL, FALSE);
  return (*klass->open) (input_device);
}
