/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducermp3.h"
#include "gdamsongdatabase.h"
#include <gtk/gtksignal.h>
#include <string.h>

int gdam_producer_mp3_default_buffer_size = 0;

static void     gdam_producer_mp3_real_set_filename
					      (GdamProducerMp3* mp3,
					       const char*      filename);
static int      gdam_producer_mp3_play_request(GdamProducer*    producer,
					       GdamPlayInfo*    play_info,
					       int              source_id,
					       GdamRequestBuilder* builder);
static void     gdam_producer_mp3_destroy     (GtkObject*       object);

static GtkObjectClass* parent_class = NULL;

static void gdam_producer_mp3_arg_set_filename(GtkObject* object, GdamArg* arg)
{
  gdam_producer_mp3_real_set_filename (GDAM_PRODUCER_MP3 (object),
				       GDAM_ARG_STRING (arg));
}
static void gdam_producer_mp3_arg_get_filename(GtkObject* object, GdamArg* arg)
{
  GdamProducerMp3 *producer_mp3 = GDAM_PRODUCER_MP3 (object);
  GDAM_ARG_STRING (arg) = g_strdup (producer_mp3->filename);
}

static void gdam_producer_mp3_class_init(GdamProducerMp3Class* klass)
{
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  producer_class->play_request = gdam_producer_mp3_play_request;
  GTK_OBJECT_CLASS (klass)->destroy = gdam_producer_mp3_destroy;
  parent_class = gtk_type_class (GDAM_TYPE_PRODUCER);
  gdam_arg_register (object_class,
		     "filename",
		     GDAM_ARG_TYPE_STRING,
		     gdam_producer_mp3_arg_set_filename,
		     gdam_producer_mp3_arg_get_filename,
		     NULL, NULL, GDAM_ARG_VITAL);
}

static void gdam_producer_mp3_init(GdamProducerMp3* producer_mp3)
{
  producer_mp3->fade_envelope = 64;
  producer_mp3->buffer_size = gdam_producer_mp3_default_buffer_size;
}

GtkType gdam_producer_mp3_get_type()
{
  static GtkType mp3_type = 0;
  if (!mp3_type)
    {
      static const GtkTypeInfo mp3_info = {
	"GdamProducerMp3",
	sizeof (GdamProducerMp3),
	sizeof (GdamProducerMp3Class),
	(GtkClassInitFunc) gdam_producer_mp3_class_init,
	(GtkObjectInitFunc) gdam_producer_mp3_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      mp3_type = gtk_type_unique (GDAM_TYPE_PRODUCER, &mp3_info);
    }
  return mp3_type;
}

GdamProducerMp3* gdam_producer_mp3_new(GdamChannel*	channel)
{
  GdamProducerMp3 *rv;
  rv = GDAM_PRODUCER_MP3 (gtk_type_new (GDAM_TYPE_PRODUCER_MP3));
  ((GdamModel *) rv)->channel = channel;
  return rv;
}

static int      gdam_producer_mp3_play_request(GdamProducer*       producer,
					       GdamPlayInfo*       play_info,
					       int                 source_id,
					       GdamRequestBuilder* builder)
{
  GdamProducerMp3 *mp3 = GDAM_PRODUCER_MP3 (producer);
  GdamChannel *channel;
  GdamModel *model;
  GdamRequest *request;
  int local_source_id;
  int num_args;
  GdamArg args[4];
  gdouble position = play_info ? play_info->position : 0.0;

  g_return_val_if_fail (mp3 != NULL, 0);
  if (!mp3->filename)
    {
      g_warning ("gdam_producer_mp3_play_request: no filename");
      return 0;
    }
  model = GDAM_MODEL (mp3);
  /*  TODO:  what if we have a source id??  Should we abort??  */
  channel = model->channel;
  g_return_val_if_fail (channel != NULL, 0);

  args[0].type = GDAM_ARG_TYPE_STRING;
  args[0].name = "filename";
  GDAM_ARG_STRING (args + 0) = mp3->filename;
  num_args = 1;

  args[1].type = GDAM_ARG_TYPE_INT64;
  args[1].name = "position";
  GDAM_ARG_INT64 (args + 1) = (gint64) (position * 44100.0);
  num_args++;
  if (mp3->buffer_size > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "buffer_size";
      GDAM_ARG_INT (args + num_args) = mp3->buffer_size;
      num_args++;
    }
  if (mp3->fade_envelope > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "fade_envelope";
      GDAM_ARG_INT (args + num_args) = mp3->fade_envelope;
      num_args++;
    }
  if (play_info->length > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT64;
      args[num_args].name = "length";
      GDAM_ARG_INT64 (args + num_args) =
	(gint64) (play_info->length * 44100.0);
      num_args++;
    }
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source_v ("GdasSeekableMp3",
					  model->source_id,
					  num_args,
					  args,
					  gdam_model_got_subsource,
					  gdam_model_didnt_get_subsource,
					  mp3);

  gdam_debug ("Creating producing mp3 with position %f", position);
  gtk_object_ref (GTK_OBJECT (mp3));

  local_source_id = gdam_request_builder_append_create (builder, request);
  return local_source_id;
}

static void gdam_producer_mp3_real_set_filename(GdamProducerMp3* mp3,
                                                const char*      filename)
{
  GdamSongDatabase *database = gdam_song_database_system ();
  GdamSongInfo *info = NULL;
  GdamBeatInfo *beat_info;
  gboolean was_playing = FALSE;
  gdouble position = 0;
  /* TODO: might this permit multiple requests? */
  if (GDAM_MODEL (mp3)->source_id != 0)
    {
      was_playing = TRUE;
      gdam_producer_stop (GDAM_PRODUCER (mp3));
    }
  if (mp3->filename)
    g_free (mp3->filename);
  mp3->filename = g_strdup (filename);
  if (database)
    info = gdam_song_database_find (database, filename);
  if (info)
    {
      gdam_producer_update_length (GDAM_PRODUCER (mp3),
				   gdam_song_info_get_length (info));
      position =
	gdam_beat_info_get_first_beat (gdam_song_info_get_beat_info (info),
				       0);
      beat_info = gdam_beat_info_copy ((gdam_song_info_get_beat_info (info)));
    }
  else
    {
      beat_info = gdam_beat_info_new ();
      gdam_beat_info_set_first_beat (beat_info, 0);
      /* TODO: get rid of this */
      gdam_beat_info_set_bpm (beat_info, 90);
    }
  gdam_model_set_beat_info (GDAM_MODEL (mp3), beat_info);
  gdam_beat_info_destroy (beat_info);
  gdam_producer_set_position (GDAM_PRODUCER (mp3), position);
}

static void gdam_producer_mp3_destroy(GtkObject* object)
{
  GdamProducerMp3 *mp3 = GDAM_PRODUCER_MP3 (object);
  g_free (mp3->filename);
  (*parent_class->destroy) (object);
}

void gdam_producer_mp3_set_filename(GdamProducerMp3*  	producer_mp3,
				    const char* 	filename)
{
  GdamArg arg;
  arg.name = "filename";
  arg.type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (&arg) = (char *) filename;
  gdam_arg_set_object_arg (GTK_OBJECT (producer_mp3), &arg);
}

void gdam_producer_mp3_set_buffer_size(GdamProducerMp3* producer_mp3,
                                       int              buffer_size)
{
  int source_id;
  g_return_if_fail (producer_mp3 != NULL);
  source_id = GDAM_MODEL (producer_mp3)->source_id;
  if (source_id && buffer_size)
    {
      GdamArg *arg = g_new (GdamArg, 1);
      arg->name = g_strdup ("buffer_size");
      GDAM_ARG_INT (arg) = buffer_size;
      arg->type = GDAM_ARG_TYPE_INT;
      gdam_channel_append_request (GDAM_MODEL (producer_mp3)->channel,
				   gdam_request_set_args (source_id,
							  1,
							  arg,
							  NULL, NULL, NULL));
    }
  producer_mp3->buffer_size = buffer_size;
}
