/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamresample.h"
#include <common/gdamsample.h>
#include <gtk/gtksignal.h>
#include "gdammodel.h"
#include "gdamfilter.h"
#include "gdamutil.h"

static void gdam_resample_class_init(GdamResampleClass* klass);
static void              on_child_set_beat_info(GdamFilter*         filter,
						const GdamBeatInfo* child_bi);

static int gdam_resample_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);

static void gdam_resample_init(GdamResample* resample)
{
  resample->rate = 1.0;
}

GtkType gdam_resample_get_type()
{
  static GtkType resample_type = 0;
  if (!resample_type)
    {
      static const GtkTypeInfo resample_info = {
	"GdamResample",
	sizeof (GdamResample),
	sizeof (GdamResampleClass),
	(GtkClassInitFunc) gdam_resample_class_init,
	(GtkObjectInitFunc) gdam_resample_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      resample_type = gtk_type_unique (GDAM_TYPE_FILTER, &resample_info);
    }
  return resample_type;
}

static inline void maybe_dispense_request(GdamResample* resample)
{
  GdamArg arg;

  arg.name = "rate";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = resample->rate;

  gdam_model_set_args (GDAM_MODEL (resample), 0, 1, &arg);
}

static int gdam_resample_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamResample *resample = GDAM_RESAMPLE (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  float rate;
  rate = resample->rate;
  gtk_object_ref (GTK_OBJECT (resample));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasResample",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					resample,
					2,
					"rate",
					GDAM_ARG_TYPE_DOUBLE,
					rate,
					"filtered",
					GDAM_ARG_TYPE_INT, new_source_id);
  return gdam_request_builder_append_create (builder, request);
}

enum {
	TEMPO_CHANGED,
	LAST_SIGNAL 
};

static guint resample_signals[LAST_SIGNAL];
static void gdam_resample_arg_set_rate(GtkObject* object, GdamArg* arg);
static void gdam_resample_arg_get_rate(GtkObject* object, GdamArg* arg);

static void gdam_resample_class_init(GdamResampleClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  model_class->num_arg_sets = 1;
  filter_class->append_start_request = gdam_resample_append_start_request;
  filter_class->on_child_set_beat_info = on_child_set_beat_info;
  gdam_arg_register (object_class,
		     "rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_resample_arg_set_rate,
		     gdam_resample_arg_get_rate,
		     NULL, "0,100", GDAM_ARG_VITAL);
  resample_signals[TEMPO_CHANGED] =
    gtk_signal_new ("tempo_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamResampleClass, tempo_changed),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, resample_signals, LAST_SIGNAL);
}

static void gdam_resample_tempo_changed(GdamResample* resample, float newtempo)
{
  GtkObject *obj = GTK_OBJECT (resample);
  if (gtk_signal_n_emissions (obj, resample_signals[TEMPO_CHANGED]))
    return;
  gtk_signal_emit (obj, resample_signals[TEMPO_CHANGED], newtempo);
}

static void gdam_resample_arg_set_rate(GtkObject* object, GdamArg* arg)
{
  GdamResample *resample;
  GdamModel *child;
  GdamBeatInfo child_bi;
  float new_bpm;
  float new_first_beat;
  resample = GDAM_RESAMPLE (object);
  child = GDAM_FILTER (resample)->child;
  resample->rate = GDAM_ARG_DOUBLE (arg);
  g_return_if_fail (child != NULL);
  if (child != NULL)
    {
      GdamBeatInfo *new_beat;
      new_beat = gdam_beat_info_new ();
      child_bi = child->beat_info;
      new_bpm = gdam_beat_info_get_bpm (&child_bi, 0) * resample->rate;
      new_first_beat = gdam_beat_info_get_first_beat (&child_bi, 0)
	/ resample->rate;
      gdam_beat_info_set_bpm (new_beat, new_bpm);
      gdam_beat_info_set_first_beat (new_beat, new_first_beat);
      gdam_model_set_beat_info (GDAM_MODEL (resample), new_beat);
      gdam_resample_tempo_changed (resample, new_bpm);
      gdam_beat_info_destroy (new_beat);
    }
  maybe_dispense_request (resample);
}

static void gdam_resample_arg_get_rate(GtkObject* object, GdamArg* arg)
{
  GdamResample *resample;
  resample = GDAM_RESAMPLE (object);
  GDAM_ARG_DOUBLE (arg) = resample->rate;
}

void gdam_resample_set_rate(GdamResample* resample, gdouble rate)
{
  GdamArg arg;
  arg.name = "rate";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = rate;
  gdam_arg_set_object_arg (GTK_OBJECT (resample), &arg);
}

static void              on_child_set_beat_info(GdamFilter*         filter,
						const GdamBeatInfo* child_bi)
{
  GdamResample *resample = GDAM_RESAMPLE (filter);
  GdamBeatInfo *new_beat;
  float new_bpm;
  float new_first_beat;
  new_beat = gdam_beat_info_new ();
  new_bpm = gdam_beat_info_get_bpm (child_bi, 0) * resample->rate;
  new_first_beat = gdam_beat_info_get_first_beat (child_bi, 0)
    / resample->rate;
  gdam_beat_info_set_bpm (new_beat, new_bpm);
  gdam_beat_info_set_first_beat (new_beat, new_first_beat);
  gdam_model_set_beat_info (GDAM_MODEL (resample), new_beat);
  gdam_resample_tempo_changed (resample, new_bpm);
  gdam_beat_info_destroy (new_beat);
}
