/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencer.h"
#include "gdambuffermanager.h"
#include <common/gdamtildeexpand.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtksignal.h>

#define MAX_SPLIT 12

typedef struct _GdamSequencerLoadThread GdamSequencerLoadThread;
struct _GdamSequencerLoadThread {
	GdamSequencer*			sequencer;
	GdamChannel*			channel;
	GList*				elements;
	FILE*				fp;
	int				destroy_connect_id;

	char*				filename;
	GdamSequencerElement*		under_construction;
};

static void continue_processing_thread(GdamSequencerLoadThread* load_thread);
static void on_sequencer_destroy(GdamSequencer* sequencer,
                                 GdamSequencerLoadThread* thread)
{
  thread->destroy_connect_id = 0;
  thread->sequencer = NULL;
}

static void free_load_thread(GdamSequencerLoadThread* load_thread)
{
  if (load_thread->fp)
    fclose (load_thread->fp);
  g_list_foreach (load_thread->elements, (GFunc) g_free, NULL);
  g_list_free (load_thread->elements);

  if (load_thread->sequencer && load_thread->destroy_connect_id)
    {
      gtk_signal_disconnect (GTK_OBJECT (load_thread->sequencer),
			     load_thread->destroy_connect_id);
      load_thread->destroy_connect_id = 0;
      load_thread->sequencer = NULL;
    }
  g_free (load_thread);
}

static void finished_constructing_element(GdamSequencerLoadThread* thread)
{
  thread->elements = g_list_prepend (thread->elements,
				     thread->under_construction);
  thread->under_construction = NULL;
  g_free (thread->filename);
  thread->filename = NULL;
}


static void finished_caching(gpointer user_data)
{
  GdamSequencerLoadThread *load_thread;
  GdamBufferManagerInfo *info;
  load_thread = (GdamSequencerLoadThread *) user_data;

  info = g_new0 (GdamBufferManagerInfo, 1);
  info->description = g_strdup_printf ("Wav %s", load_thread->filename);
  info->buffer_id = load_thread->under_construction->buffer_id;
  info->channel = load_thread->channel;
  gdam_buffer_manager_add_buffer (info);

  if (load_thread->sequencer)
    {
      finished_constructing_element (load_thread);
      continue_processing_thread (load_thread);
    }
  else
    {
      free_load_thread (load_thread);
    }
}

static void started_caching(int buffer_id,
                            int connection_id,
			    gpointer user_data)
{
  GdamSequencerLoadThread *load_thread;
  load_thread = (GdamSequencerLoadThread *) user_data;

  load_thread->under_construction->buffer_id = buffer_id;
  gdam_debug ("sequencer-load: started_caching");
  gdam_channel_register_handler (load_thread->channel,
				 gdam_channel_simple_handler_new
				 (connection_id, TRUE, finished_caching,
				  load_thread));
}

static void failed_to_start_caching(gpointer user_data)
{
  GdamSequencerLoadThread *load_thread;
  load_thread = (GdamSequencerLoadThread *) user_data;

  gdam_debug ("sequencer-load: failed to start caching %s",
	      load_thread->filename);

  /* Continue processing??? */
  /*continue_processing_thread(load_thread); */

  /* Or not??? */
  free_load_thread (load_thread);
}


static void append_cache_request(GdamSequencerLoadThread* load_thread,
                                 char*                    filename)
{
  GdamRequestBuilder *builder = gdam_request_builder_new ();
  int local_id;
  GdamChannel *channel = load_thread->channel;

  g_assert (load_thread->filename == NULL);
  gdam_debug ("append_cache_request: loading %s", filename);

  load_thread->filename = filename;

  local_id = gdam_request_builder_append_create (builder,
						 gdam_request_create_source
						 ("GdasWav",
						  gdam_channel_new_local_id
						  (channel), NULL, NULL, NULL,
						  1, "filename",
						  GDAM_ARG_TYPE_STRING,
						  filename));
  gdam_request_builder_append_generic (builder,
				       gdam_request_start_caching (local_id,
								   started_caching,
								   failed_to_start_caching,
								   load_thread));

  gdam_channel_append_request (channel,
			       gdam_request_builder_package (builder));
  gdam_request_builder_destroy (builder);
}


static GdamSequencerElement* 
parse_element (GdamSequencerLoadThread* load_thread, char** filename_out)
{
  GdamSequencerElement *rv;
  GdamBufferManagerInfo *info;
  char buf[1024];
  FILE* fp;
  fp = load_thread->fp; 
  while (fgets (buf, sizeof (buf), fp) != NULL)
    {
      char *at = buf;
      char *split[128];
      char *description;
      char *filename;
      int num_split = 0;
      while (*at && isspace (*at))
	at++;
      /* Skip comment lines. */
      if (*at == 0 || *at == '#')
	continue;

      g_strchomp (at);

      /* Split the line into whitespace delimited pieces. */
      while (*at && num_split < MAX_SPLIT)
	{
	  char *end_word = at;
	  while (*end_word && !isspace (*end_word))
	    end_word++;
	  split[num_split++] = at;
	  if (*end_word == '\0')
	    break;
	  *end_word = 0;
	  at = end_word + 1;
	  while (*at && isspace (*at))
	    at++;
	}
      if (num_split == 2 && strcmp(split[0], "bpm:") == 0)
        {
	  GdamBeatInfo* new_beat_info;
	  float bpm = atof (split[1]);
	  new_beat_info = gdam_beat_info_copy 
			  (&GDAM_MODEL(load_thread->sequencer)->beat_info);
	  gdam_beat_info_set_bpm (new_beat_info, bpm);
	  gdam_model_set_beat_info (GDAM_MODEL(load_thread->sequencer), 
				    new_beat_info);
	  gdam_beat_info_destroy(new_beat_info);
	  continue;
        }
      else if (num_split == 2 && strcmp(split[0], "beats:") == 0)
        {
	  float beats = atoi (split[1]);
	  gdam_sequencer_set_num_beats(load_thread->sequencer, beats);
	  continue;
        }
      else if (num_split != 6 && num_split != 7)
	{
	  g_warning ("bad number of fields in line " "parsing sequencer");
	  continue;
	}

      rv = g_new (GdamSequencerElement, 1);
      filename = split[0];
      rv->beat_offset = atof (split[1]);
      rv->rate = atof (split[2]);
      rv->gain = atof (split[3]);
      rv->sample_offset = atoi (split[4]);
      rv->output_length = atoi (split[5]);
      rv->gain_factor = 1;
      rv->rate_factor = 1;
      if (num_split == 7)
	rv->pattern_number = atoi (split[6]);
      else
	rv->pattern_number = 0;

      description = g_strdup_printf ("Wav %s", filename);
      info = gdam_buffer_manager_info_by_desc (description);
      if (info != NULL)
	rv->buffer_id = info->buffer_id;
      else
	{
	  rv->buffer_id = 0;
	  *filename_out = g_strdup (filename);
	}
      g_free (description);
      return rv;

    }
  return NULL;
}

static void continue_processing_thread(GdamSequencerLoadThread* load_thread)
{
  char *filename;
  filename = NULL;
  g_assert (load_thread->under_construction == NULL);

  gdam_debug ("processing thread...");

  /* Check if the sequencer has been destroyed. */
  if (load_thread->sequencer == NULL)
    {
      free_load_thread (load_thread);
      return;
    }

  while ((load_thread->under_construction
	  = parse_element (load_thread, &filename)) != NULL)
    {
      if (load_thread->under_construction->buffer_id == 0)
	{
	  append_cache_request (load_thread, filename);
	  return;
	}
      finished_constructing_element (load_thread);
    }
  if (load_thread->elements)
    {
      gdam_sequencer_set_elements (load_thread->sequencer,
				   load_thread->elements);
      load_thread->elements = NULL;
    }
  free_load_thread (load_thread);
}

void                    gdam_sequencer_load       (GdamSequencer* sequencer,
                                                   const char*    filename)
{
  GdamSequencerLoadThread *loader;
  FILE *fp;
  char *expanded_filename = gdam_tilde_expand (filename);
  fp = fopen (expanded_filename, "r");
  g_free (expanded_filename);
  if (!fp)
    {
      g_warning ("gdam_sequencer_load: error loading %s", filename);
      return;
    }
  gdam_debug ("loading file %s", expanded_filename);
  loader = g_new (GdamSequencerLoadThread, 1);
  loader->fp = fp;
  loader->sequencer = sequencer;
  loader->channel = GDAM_MODEL (sequencer)->channel;
  loader->destroy_connect_id
    = gtk_signal_connect (GTK_OBJECT (sequencer),
			  "destroy",
			  GTK_SIGNAL_FUNC (on_sequencer_destroy), loader);
  loader->elements = NULL;
  loader->under_construction = NULL;
  loader->filename = NULL;
  continue_processing_thread (loader);
}

static void write_element(GdamSequencerElement* element, FILE* fp)
{
  GdamBufferManagerInfo *info;
  const char *filename;
  info = gdam_buffer_manager_info_by_id (element->buffer_id);
  g_return_if_fail (info != NULL);
  if (strncmp (info->description, "Wav ", 4) != 0)
    {
      g_warning ("sequence of nonwavetable ignored (cannot save it)");
      return;
    }
  filename = info->description + 4;
  fprintf (fp, "%s %4.4g %4.4g %4.4g %d %d %d\n",
	   filename,
	   element->beat_offset,
	   element->rate,
	   element->gain,
	   element->sample_offset,
	   element->output_length, element->pattern_number);
}
	

void                    gdam_sequencer_save       (GdamSequencer* sequencer,
                                                   const char*    filename)
{
  GdamBeatInfo* bi;
  FILE *fp;
  char *expanded_filename = gdam_tilde_expand (filename);
  fp = fopen (expanded_filename, "w");
  g_free (expanded_filename);
  if (!fp)
    {
      g_warning ("gdam_sequencer_save: error creating %s", filename);
      return;
    }
  bi = &(GDAM_MODEL (sequencer)->beat_info);
  fprintf (fp, "beats: %4.4g\n", sequencer->num_beats);
  fprintf (fp, "bpm: %4.4g\n", gdam_beat_info_get_bpm(bi, 0));
  g_list_foreach (sequencer->elements, (GFunc) write_element, fp);
  fclose (fp);
}
