/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsongdatabase.h"
#include "gdamutil.h"
#include <common/gdamtildeexpand.h>
#include <ctype.h>
#include "../common/gdamdebug.h"
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <gnome-xml/parser.h>

enum {
	CHANGED,
	SONG_INFO_CHANGED,
	LAST_SIGNAL
};

static guint database_signals[LAST_SIGNAL] = { 0 };


static void gdam_song_database_destroy(GtkObject* object);

static GdamSongDatabase* default_song_database = NULL;
static GtkObjectClass* parent_class = NULL;

static int compare_by_pointer(gpointer a, gpointer b)
{
  if (GPOINTER_TO_INT (a) < GPOINTER_TO_INT (b))
    return -1;
  if (GPOINTER_TO_INT (a) > GPOINTER_TO_INT (b))
    return +1;
  return 0;
}

static int gdam_song_info_by_order(GdamSongInfo* a, GdamSongInfo* b)
{
  int aorder = gdam_song_info_get_order (a);
  int border = gdam_song_info_get_order (b);
  if (aorder < border)
    return -1;
  if (aorder > border)
    return +1;
  return compare_by_pointer (a, b);
}

static int gdam_song_info_by_name(GdamSongInfo* a, GdamSongInfo* b)
{
  const char *afilename = gdam_song_info_get_filename (a);
  const char *bfilename = gdam_song_info_get_filename (b);
  int order = strcmp (afilename, bfilename);
  if (order == 0)
    return compare_by_pointer (a, b);
  return order;
}

static int gdam_song_info_by_artist(GdamSongInfo* a, GdamSongInfo* b)
{
  const GHashTable *a_props;
  const GHashTable *b_props;
  gpointer adata;
  gpointer bdata;
  a_props = gdam_song_info_get_properties (a);
  b_props = gdam_song_info_get_properties (b);
  adata = g_hash_table_lookup ((GHashTable *) a_props, "artist");
  bdata = g_hash_table_lookup ((GHashTable *) b_props, "artist");
  if (adata && !bdata)
    return 1;
  else if (!adata && bdata)
    return -1;
  else if (adata && bdata)
    return strcmp ((char *) adata, (char *) bdata);
  else
    return gdam_song_info_by_name (a, b);
}

static int gdam_song_info_by_album(GdamSongInfo* a, GdamSongInfo* b)
{
  const GHashTable *a_props;
  const GHashTable *b_props;
  gpointer adata;
  gpointer bdata;
  a_props = gdam_song_info_get_properties (a);
  b_props = gdam_song_info_get_properties (b);
  adata = g_hash_table_lookup ((GHashTable *) a_props, "album");
  bdata = g_hash_table_lookup ((GHashTable *) b_props, "album");
  if (adata && !bdata)
    return 1;
  else if (!adata && bdata)
    return -1;
  else if (adata && bdata)
    return strcmp ((char *) adata, (char *) bdata);
  else
    return gdam_song_info_by_name (a, b);
}

static int gdam_song_info_by_length(GdamSongInfo* a, GdamSongInfo* b)
{
  float length_a = gdam_song_info_get_length (a);
  float length_b = gdam_song_info_get_length (b);
  if (length_a < 0)
    length_a = 1000000;
  if (length_b < 0)
    length_b = 1000000;
  if (length_a < length_b)
    return -1;
  if (length_a > length_b)
    return +1;
  return compare_by_pointer (a, b);
}

static int gdam_song_info_by_bpm(GdamSongInfo* a, GdamSongInfo* b)
{
  float abpm = gdam_beat_info_get_bpm (gdam_song_info_get_beat_info (a), 0);
  float bbpm = gdam_beat_info_get_bpm (gdam_song_info_get_beat_info (b), 0);
  if (abpm < bbpm)
    return -1;
  if (abpm > bbpm)
    return +1;
  return gdam_song_info_by_name (a, b);
}

GdamSongDatabase*    gdam_song_database_new(const char* filename)
{
  GdamSongDatabase *database;
  database = GDAM_SONG_DATABASE (gtk_type_new (GDAM_TYPE_SONG_DATABASE));
  database->dbname = g_strdup (filename);
  return database;
}


void                 gdam_song_database_add     (GdamSongDatabase* db,
                                                 GdamSongInfo*     song_info)
{
  const char *filename;
  GdamSongInfo *si;
  filename = gdam_song_info_get_filename (song_info);
  g_return_if_fail (db != NULL);
  si = g_hash_table_lookup (db->name_lookup, filename);
  if (si == song_info)
    {
      gdam_debug ("already have %s in database", filename);
      return;
    }
  if (gdam_song_info_get_beat_info (song_info)->range_list != NULL)
    if (si != NULL)
      gdam_song_database_remove (db, filename);
  g_hash_table_insert (db->name_lookup, g_strdup (filename), song_info);
  g_tree_insert (db->by_order, song_info, song_info);
  g_tree_insert (db->by_name, song_info, song_info);
  g_tree_insert (db->by_length, song_info, song_info);
  g_tree_insert (db->by_bpm, song_info, song_info);
  if (db->frozen_count > 0)
    db->changed_while_frozen = TRUE;
  else
    {
      if (db->frozen_count > 0)
	db->changed_while_frozen = TRUE;
      else
	gtk_signal_emit (GTK_OBJECT (db), database_signals[CHANGED]);
    }
}

GdamSongInfo*    gdam_song_database_find(GdamSongDatabase* database,
					 const char* filename)
{
  g_return_val_if_fail(database != NULL, NULL);
  g_return_val_if_fail(database->name_lookup != NULL, NULL);
  return g_hash_table_lookup (database->name_lookup, filename);
}

void                 gdam_song_database_remove  (GdamSongDatabase* db,
                                                 const char*       filename)
{
  GdamSongInfo *to_kill;
  const char *doomed_fname;
  g_return_if_fail (filename != NULL);
  g_return_if_fail (db != NULL);
  gdam_debug ("gdam song database: removing %s", filename);
  to_kill = (GdamSongInfo *) g_hash_table_lookup (db->name_lookup, filename);
  doomed_fname = gdam_song_info_get_filename (to_kill);
  if (!to_kill)
    return;
  g_hash_table_remove (db->name_lookup, doomed_fname);
  g_tree_remove (db->by_order, to_kill);
  g_tree_remove (db->by_name, to_kill);
  g_tree_remove (db->by_length, to_kill);
  g_tree_remove (db->by_bpm, to_kill);
  gdam_song_info_unref (to_kill);
  if (db->frozen_count > 0)
    db->changed_while_frozen = TRUE;
  else
    {
      if (db->frozen_count > 0)
	db->changed_while_frozen = TRUE;
      else
	gtk_signal_emit (GTK_OBJECT (db), database_signals[CHANGED]);
    }
}

static void kill_songinfo(gpointer key, gpointer value, gpointer unused)
{
  gdam_song_info_unref ((GdamSongInfo *) value);
}

static void print_props(gpointer key, gpointer value, gpointer data)
{
  FILE *fp = (FILE *) data;
  fprintf (fp, "%s: %s\n", (char *) key, (char *) value);
}

static void print_range(gpointer data, gpointer udata)
{
  GdamBeatInfoRange *range = (GdamBeatInfoRange *) data;
  FILE *fp = (FILE *) udata;
  fprintf (fp, "range: %.8g:%.8g\n", range->bpm, range->first_beat);
}

typedef struct _PrintInfo 
{
  FILE *fp;
  GdamSongInfoDecideFunc decidefunc;
  gpointer decideptr;
}

PrintInfo;

static gint gdam_song_info_print(gpointer key, gpointer value, gpointer data)
{
  GdamSongInfo *song_info = (GdamSongInfo *) value;
  const GdamBeatInfo *beat_info;
  float first_beat;
  float bpm;
  float length;
  float gain;
  const char *filename;
  PrintInfo *printinfo = (PrintInfo *) data;
  FILE *fp = printinfo->fp;
  const GHashTable *prop_table;
  const GList *list;
  int num_ranges;
  g_return_val_if_fail (song_info != NULL, 0);
  beat_info = gdam_song_info_get_beat_info (song_info);
  first_beat = gdam_beat_info_get_first_beat (beat_info, 0);
  bpm = gdam_beat_info_get_bpm (beat_info, 0);
  length = gdam_song_info_get_length (song_info);
  gain = gdam_song_info_get_gain (song_info);
  filename = gdam_song_info_get_filename (song_info);

  if (!filename)
    g_warning ("NULL filename in song_info_print in songdb");

  if (printinfo->decidefunc)
    {
      gboolean rv;
      rv = (*printinfo->decidefunc) (song_info, printinfo->decideptr);
      if (!rv)
	return 0;
    }

  fprintf (fp, "file: %s\n", filename ? filename : "");
  if (first_beat >= 0.0)
    fprintf (fp, "firstbeat: %3.6f\n", first_beat);
  if (length >= 0.0)
    fprintf (fp, "length: %3.6f\n", length);
  if (bpm >= 0.0)
    fprintf (fp, "bpm: %3.6f\n", bpm);
  if (gain >= 0.0)
    fprintf (fp, "gain: %3.6f\n", gain);

  prop_table = gdam_song_info_get_properties (song_info);
  if (prop_table)
    {
      g_hash_table_foreach ((GHashTable *) prop_table, print_props, fp);
    }
  list = gdam_song_info_get_loops (song_info);
  for (list = list; list != NULL; list = list->next)
    {
      GdamLoopInfo *li = (GdamLoopInfo *) list->data;
      fprintf (fp, "loop: %s:%3.6f:%3.6f\n", li->name, li->start, li->length);
    }
  num_ranges = g_list_length (beat_info->range_list);
  if (num_ranges > 0)
    g_list_foreach (beat_info->range_list, print_range, fp);

  fprintf (fp, "\n");
  return 0;
}

static void          gdam_song_database_write_fp_ext(GdamSongDatabase* db,
                                                     FILE* fp,
						     GdamSongInfoDecideFunc df,
						     gpointer ptr)
{
  PrintInfo pi;
  g_return_if_fail (db != NULL);
  g_return_if_fail (fp != NULL);
  g_return_if_fail (db->by_order != NULL);
  pi.fp = fp;
  pi.decidefunc = df;
  pi.decideptr = ptr;
  g_tree_traverse (db->by_order,
		   (GTraverseFunc) gdam_song_info_print, G_IN_ORDER, &pi);
}

void                 gdam_song_database_write_fp(GdamSongDatabase* db,
                                                 FILE* fp)
{
  gdam_song_database_write_fp_ext (db, fp, NULL, NULL);
}

static gint gdam_song_info_add_xml(gpointer key, gpointer value, gpointer data)
{
  GdamSongInfo *song_info = (GdamSongInfo *) value;
  xmlNode *root = data;
  g_return_val_if_fail (data != NULL, 0);
  gdam_song_info_to_xml (song_info, root);
  return 0;
}
static void          gdam_song_database_write_xml(GdamSongDatabase* db,
                                                  const char* fname)
{
  xmlDoc *document;
  xmlNode *root;
  g_return_if_fail (db != NULL);
  g_return_if_fail (db->by_order != NULL);
  document = xmlNewDoc (XML_DEFAULT_VERSION);
  root = xmlNewDocNode (document, NULL, "Gdam-Song-Database", NULL);
#if 0				/* only in libxml >= 1.8 */
  xmlDocSetRootElement (document, root);
#else
  document->root = root;
#endif
  g_tree_traverse (db->by_order,
		   (GTraverseFunc) gdam_song_info_add_xml, G_IN_ORDER, root);
  xmlSaveFile (fname, document);
}

void                 gdam_song_database_write_ext(GdamSongDatabase* db,
                                                  const char* fname,
						  GdamSongInfoDecideFunc func,
						  gpointer    ptr)
{
  FILE *fp;
  char *expanded_filename;
  char *extension;
  g_return_if_fail (db != NULL);
  g_return_if_fail (fname != NULL);
  expanded_filename = gdam_tilde_expand (fname);
  extension = strrchr (expanded_filename, '.');
  if (extension != NULL && strcasecmp (extension, ".xml") == 0)
    {
      gdam_song_database_write_xml (db, expanded_filename);
      return;
    }
  fp = fopen (expanded_filename, "w");
  g_free (expanded_filename);
  if (!fp)
    {
      fprintf (stderr, "couldn't open %s for writing.\n", fname);
      return;
    }
  gdam_song_database_write_fp_ext (db, fp, func, ptr);
  fclose (fp);
}

void                 gdam_song_database_write    (GdamSongDatabase* db,
                                                  const char* fname)
{
  gdam_song_database_write_ext (db, fname, NULL, NULL);
}

static GdamSongDatabase* parse_xml_song_database(const char* filename)
{
  xmlDoc *xml_database;
  xmlNode *node;
  GdamSongDatabase *song_database;
  xml_database = xmlParseFile (filename);
#if 0
  node = xmlDocGetRootElement (xml_database);
#else
  node = xml_database->root;
#endif
  song_database = gdam_song_database_new (filename);
  if (strcasecmp (node->name, "Gdam-Song-Database") != 0)
    {
      g_warning ("root node was %s; "
		 "should have been Gdam-Song-Database", node->name);
      return NULL;
    }
  for (node = node->childs; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "Song") == 0)
	{
	  GdamSongInfo *song_info;
	  song_info = gdam_song_info_from_xml (node->childs);
	  if (song_info == NULL)
	    {
	      g_warning ("error parsing song info");
	    }
	  else
	    {
	      gdam_song_database_add (song_database, song_info);
	    }
	}
      else
	{
	  g_warning ("xml database %s: "
		     "node with unexpected name `%s'", filename, node->name);
	}
    }
  return song_database;
}

GdamSongDatabase* 
gdam_song_database_add_from_file(GdamSongDatabase* _database,
				 const char* fname)
{
  char buf[2048];
  FILE *fp;
  int lineno = 0;
  GdamSongDatabase *database;
  GdamSongInfo *song_info = NULL;
  int order = 1;
  const char *extension;
  char *filename;
  g_return_val_if_fail (fname != NULL, NULL);
  filename = gdam_tilde_expand (fname);
  extension = strrchr (filename, '.');
  if (extension != NULL && strcasecmp (extension, ".xml") == 0)
    return parse_xml_song_database (filename);

  fp = fopen (filename, "r");
  if (!fp)
    {
      g_warning ("couldn't open `%s' for reading", filename);
      return NULL;
    }

  database = _database;
  if (!database)
    database = gdam_song_database_new (filename);

  while (fgets (buf, sizeof (buf), fp))
    {
      char *key = buf;
      char *colon;
      char *value;
      lineno++;
      while (*key && isspace (*key))
	key++;
      if (*key == '#' || *key == '\0')
	continue;
      colon = strchr (buf, ':');
      if (!colon)
	{
	  g_warning ("parsing %s, line %d: no colon", filename, lineno);
	  continue;
	}
      *colon = 0;
      value = colon + 1;
      while (*value && isspace (*value))
	value++;
      g_strchomp (value);
      g_strdown (key);
      /*
         gdam_debug("song_database_parse: key=%s; value=%s\n",
         key, value);
       */
      if (strcmp (key, "file") == 0)
	{
	  if (song_info)
	    gdam_song_database_add (database, song_info);
	  song_info = gdam_song_info_new (value, 0.0, 0.0);
	}
      else
	{
	  if (!song_info)
	    {
	      g_warning ("first key of %s was `%s', must be "
			 "`file' (line %d)", filename, key, lineno);
	      continue;
	    }
	  gdam_song_info_add_property (song_info, key, value);
	}
    }
  if (song_info)
    {
      gdam_song_info_set_order (song_info, order++);
      gdam_song_database_add (database, song_info);
    }
  else
    g_warning ("%s had no entries", filename);
  return database;
}

GdamSongDatabase* 
gdam_song_database_new_from_file(const char* filename){
	return gdam_song_database_add_from_file(NULL, filename);
}

static const char* fnames[] = {
	"/etc/gdam/song.database.xml",
	"/etc/gdam/song.database",
	"~/.gdam/song.database.xml",
	"~/.gdam/song.database",
	(const char*)NULL
};

const char* gdam_song_database_default_filename = NULL;

GdamSongDatabase*
gdam_song_database_system()
{
  const char *const *at;
  GdamSongDatabase *db;
  gdam_debug ("Song Database System called");
  if (default_song_database)
    {
      gdam_debug ("Song Database System..  returning default");
      return default_song_database;
    }
  if (gdam_song_database_default_filename != NULL)
    {
      db =
	gdam_song_database_new_from_file
	(gdam_song_database_default_filename);
      if (db == NULL)
	{
	  g_error ("song database %s does not exist",
		   gdam_song_database_default_filename);
	}
      return default_song_database = db;
    }
#if HAVE_GETENV
  {
    char *db_env;
    db_env = getenv ("GDAM_SONG_DATABASE");
    if (db_env != NULL)
      {
	db = gdam_song_database_new_from_file (db_env);
	if (db == NULL)
	  {
	    g_warning ("error loading song database %s", db_env);
	  }
	else
	  {
	    return default_song_database = db;
	  }
      }
  }
#endif
  for (at = fnames; *at; at++)
    {
      char *name = gdam_tilde_expand (*at);
      if (name)
	{
	  if (gdam_file_exists (name))
	    {
	      db = gdam_song_database_new_from_file (name);
	      gtk_object_ref (GTK_OBJECT (db));
	      gdam_debug ("Song Database from %s", name);
	      return default_song_database = db;
	    }
	  g_free (name);
	}
    }
  db = gdam_song_database_new ("system");
  gdam_debug ("Creating new database");
  return default_song_database = db;
}


static void gdam_song_database_init(GdamSongDatabase* song_database)
{
  g_return_if_fail (song_database != NULL);
  song_database->by_order
    = g_tree_new ((GCompareFunc) gdam_song_info_by_order);
  song_database->by_name = g_tree_new ((GCompareFunc) gdam_song_info_by_name);
  song_database->by_length
    = g_tree_new ((GCompareFunc) gdam_song_info_by_length);
  song_database->by_artist
    = g_tree_new ((GCompareFunc) gdam_song_info_by_artist);
  song_database->by_album
    = g_tree_new ((GCompareFunc) gdam_song_info_by_album);
  song_database->by_bpm = g_tree_new ((GCompareFunc) gdam_song_info_by_bpm);
  song_database->name_lookup = g_hash_table_new (g_str_hash, g_str_equal);
}
static void gdam_song_database_class_init(GdamSongDatabaseClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gdam_song_database_destroy;
  database_signals[CHANGED]
    = gtk_signal_new ("changed",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamSongDatabaseClass,
					 changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  database_signals[SONG_INFO_CHANGED]
    = gtk_signal_new ("song-info-changed",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamSongDatabaseClass,
					 song_info_changed),
		      gtk_marshal_NONE__POINTER,
		      GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  gtk_object_class_add_signals (object_class, database_signals, LAST_SIGNAL);
}

GtkType gdam_song_database_get_type()
{
  static GtkType song_database_type = 0;
  if (!song_database_type)
    {
      static const GtkTypeInfo song_database_info = {
	"GdamSongDatabase",
	sizeof (GdamSongDatabase),
	sizeof (GdamSongDatabaseClass),
	(GtkClassInitFunc) gdam_song_database_class_init,
	(GtkObjectInitFunc) gdam_song_database_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      parent_class = gtk_type_class (parent);
      song_database_type = gtk_type_unique (parent, &song_database_info);
    }
  return song_database_type;
}

static void gdam_song_database_destroy(GtkObject* object)
{
  GdamSongDatabase *database = GDAM_SONG_DATABASE (object);
  if (database == default_song_database) {
	  gdam_debug("destroying default database");
	  default_song_database = NULL;
  }
  g_tree_destroy (database->by_order);
  g_tree_destroy (database->by_name);
  g_tree_destroy (database->by_length);
  g_tree_destroy (database->by_bpm);
  g_hash_table_foreach (database->name_lookup, kill_songinfo, NULL);
  g_hash_table_destroy (database->name_lookup);
  database->by_order = NULL;
  database->by_name  = NULL;
  database->by_length = NULL;
  database->by_bpm = NULL;
  database->name_lookup = NULL;
  (*parent_class->destroy) (object);
}

void                 gdam_song_database_freeze(GdamSongDatabase* db)
{
  g_return_if_fail (db != NULL);
  gdam_debug ("songdb frozen");
  if (db->frozen_count == 0)
    {
      g_return_if_fail (db->changed_while_frozen == 0);
      db->changed_while_frozen = FALSE;
    }
  db->frozen_count++;
}

void                 gdam_song_database_thaw  (GdamSongDatabase* db)
{
  g_return_if_fail (db != NULL);
  gdam_debug ("songdb thawed");
  db->frozen_count--;
  if (db->frozen_count == 0 && db->changed_while_frozen != 0)
    {
      gtk_signal_emit (GTK_OBJECT (db), database_signals[CHANGED]);
      db->changed_while_frozen = FALSE;
    }
}
