/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsonginfo.h"
#include <common/gdamdebug.h>
#include <stdlib.h>

struct _GdamSongInfo {
    int            ref_count;
    float          length;
    float          gain;
    int            order;
    GdamBeatInfo*  beat_info;
    GHashTable*    song_properties;
    GList*	   loop_list;
    char*          filename;
};

/*=====================================================================*
 *                          gdam_song_info_new                         *
 *                                                                     *
 * Create a new song info.                                             *
 *=====================================================================*/
GdamSongInfo*        gdam_song_info_new         (const char*       filename,
                                                 float             bpm,
                                                 float             first_beat)
{
  GdamSongInfo *info = g_new0 (GdamSongInfo, 1);
  info->filename = g_strdup (filename);
  info->beat_info = gdam_beat_info_new ();
  gdam_beat_info_set_bpm (info->beat_info, bpm);
  gdam_beat_info_set_first_beat (info->beat_info, first_beat);
  info->length = -1;
  info->gain = -1;
  info->order = -1;
  info->song_properties = g_hash_table_new (g_str_hash, g_str_equal);
  info->ref_count = 1;
  return info;
}

/*=====================================================================*
 *                         gdam_song_info_copy                         *
 *                                                                     *
 * Make a copy of an existing songinfo.                                *
 *=====================================================================*/
static void do_one_hash_copy(gpointer key, gpointer val, gpointer user_data)
{
  g_hash_table_insert ((GHashTable *) user_data,
		       g_strdup ((char *) key), g_strdup ((char *) val));
}

GdamSongInfo* gdam_song_info_copy (GdamSongInfo* song_info)
{
  GdamSongInfo *new_info;
  new_info = g_new0 (GdamSongInfo, 1);
  *new_info = *song_info;
  new_info->ref_count = 1;
  new_info->song_properties = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_foreach (song_info->song_properties,
			do_one_hash_copy, new_info->song_properties);
  new_info->beat_info =
    gdam_beat_info_copy (gdam_song_info_get_beat_info (song_info));
  return new_info;
}

static gboolean parse_loop(const char* txt, GdamSongInfo* info)
{
  char *name;
  char *endp;
  gdouble length, start;
  char **data;

  data = g_strsplit (txt, ":", 4);
  if (*data == NULL || *(data + 1) == NULL || *(data + 2) == NULL)
    {
      g_warning ("missing fields in loop %s", txt);
      return FALSE;
    }
  name = data[0];
  start = strtod (data[1], &endp);
  if (endp == txt || *endp)
    return FALSE;
  length = strtod (data[2], &endp);
  if (endp == txt || *endp)
    return FALSE;
  gdam_song_info_add_loop (info, name, start, length);
  return TRUE;
}

/* TODO: make public so other files needn't copy?? */
static int sort_ranges(gconstpointer a, gconstpointer b)
{
  GdamBeatInfoRange *bira;
  GdamBeatInfoRange *birb;
  bira = (GdamBeatInfoRange *) a;
  birb = (GdamBeatInfoRange *) b;

  if (bira->first_beat < birb->first_beat)
    {
      return -1;
    }
  else if (bira->first_beat > birb->first_beat)
    {
      return 1;
    }
  return 0;
}

gboolean             gdam_song_info_add_range(GdamSongInfo*     info,
					      gdouble           bpm,
					      gdouble           firstbeat,
					      gdouble           start,
					      gdouble           end)
{
  GdamBeatInfoRange *bir;
  GdamBeatInfo *beat_info;
  beat_info = gdam_beat_info_copy (gdam_song_info_get_beat_info (info));
  bir = (GdamBeatInfoRange *) g_new0 (GdamBeatInfoRange, 1);
  bir->bpm = bpm;
  bir->first_beat = firstbeat;
  bir->range_start = start;
  bir->range_end = end;
  beat_info->range_list = g_list_insert_sorted (beat_info->range_list,
						bir, sort_ranges);
  gdam_song_info_set_beat_info (info, beat_info);
  gdam_beat_info_destroy (beat_info);
  return TRUE;
}

static gboolean parse_range(const char* txt, GdamSongInfo* info)
{
  char *name;
  char *endp;
  gdouble bpm, firstbeat;
  char **data;

  data = g_strsplit (txt, ":", 4);
  if (*data == NULL || *(data + 1) == NULL)
    {
      g_warning ("missing fields in range %s", txt);
      return FALSE;
    }
  bpm = strtod (data[0], &endp);
  if (endp == txt || *endp)
    return FALSE;
  firstbeat = strtod (data[1], &endp);
  if (endp == txt || *endp)
    return FALSE;
  gdam_song_info_add_range (info, bpm, firstbeat, 0, 0);
  return TRUE;
}


static gboolean parse_float(const char* txt, float* out)
{
  char *endp;
  *out = (float) strtod (txt, &endp);
  if (endp == txt || *endp)
    return FALSE;
  return TRUE;
}

/*=====================================================================*
 *                      gdam_song_info_set_length                      *
 *=====================================================================*/
void   	             gdam_song_info_set_length (GdamSongInfo*     song_info,
                                                gdouble           length)
{
  song_info->length = length;
}

/*=====================================================================*
 *                     gdam_song_info_add_property                     *
 *                                                                     *
 * Parameters:                                                         *
 *     info              the info to adjust.                           *
 *     key               name of the property.                         *
 *     value             value of the property.                        *
 *=====================================================================*/
gboolean             gdam_song_info_add_property(GdamSongInfo*     info,
                                                 const char*       key,
						 const char*       value)
{
  if (value)
    {
      gpointer old;
      if (strcmp (key, "bpm") == 0)
	{
	  float bpm;
	  if (!parse_float (value, &bpm))
	    {
	      g_warning ("error parsing bpm `%s'", value);
	      return FALSE;
	    }
	  gdam_beat_info_set_bpm (info->beat_info, bpm);
	  return TRUE;
	}
      else if (strcmp (key, "firstbeat") == 0)
	{
	  float first_beat;
	  if (!parse_float (value, &first_beat))
	    {
	      g_warning ("error parsing first beat `%s'", value);
	      return FALSE;
	    }
	  gdam_beat_info_set_first_beat (info->beat_info, first_beat);
	  return TRUE;
	}
      else if (strcmp (key, "length") == 0)
	{
	  if (!parse_float (value, &info->length))
	    {
	      g_warning ("error parsing length `%s'", value);
	      return FALSE;
	    }
	  return TRUE;
	}
      else if (strcmp (key, "gain") == 0)
	{
	  if (!parse_float (value, &info->gain))
	    {
	      g_warning ("error parsing gain `%s'", value);
	      return FALSE;
	    }
	  return TRUE;
	}
      else if (strcmp (key, "loop") == 0)
	{
	  if (!parse_loop (value, info))
	    {
	      g_warning ("error parsing loop `%s'", value);
	      return FALSE;
	    }
	  return TRUE;
	}
      else if (strcmp (key, "range") == 0)
	{
	  if (!parse_range (value, info))
	    {
	      g_warning ("error parsing range `%s'", value);
	      return FALSE;
	    }
	  return TRUE;
	}

      old = g_hash_table_lookup (info->song_properties, key);
      if (old)
	{
	  g_hash_table_insert (info->song_properties,
			       (gpointer) key, g_strdup (value));
	  g_free (old);
	}
      else
	{
	  g_hash_table_insert (info->song_properties,
			       g_strdup (key), g_strdup (value));
	}
    }
  /*  If 'value' is NULL, remove property from table.  */
  else
    {
      gpointer orig_key;
      gpointer orig_val;
      gboolean found;
      found = g_hash_table_lookup_extended (info->song_properties, key,
					    &orig_key, &orig_val);
      if (found)
	{
	  g_hash_table_remove (info->song_properties, key);
	  g_free (orig_key);
	  g_free (orig_val);
	}
    }
  return TRUE;
}

#define GDAM_IS_IN_DELTA(a,b,delta)		\
	( (a) >= (b) - (delta)			\
	  && (a) <= (b) + (delta) )

/*=====================================================================*
 *                     gdam_song_info_add_loop                         *
 *                                                                     *
 * Parameters:                                                         *
 *     info              the info to adjust.                           *
 *     name              name of the loop.                             *
 *     start             time within the song where the loop begins,   *
 *     			 in seconds.                                   *
 *     length		 length of the loop, in seconds.               *
 *=====================================================================*/
gboolean             gdam_song_info_add_loop(GdamSongInfo*     info,
                                             const char*       name,
					     gdouble           start,
					     gdouble           length)
{
  GList *at;
  GdamLoopInfo *name_match = NULL;
  GdamLoopInfo *info_match = NULL;

  /*  Check if we have a loop by that name or info...  */
  for (at = info->loop_list; at != NULL; at = at->next)
    {
      GdamLoopInfo *li;
      li = (GdamLoopInfo *) at->data;
      if (li->name && name && strcmp (li->name, name) == 0 && !name_match)
	{
	  name_match = li;

	}
      else if ((GDAM_IS_IN_DELTA (li->start, start, 0.01)
		&& GDAM_IS_IN_DELTA (li->length, length, 0.01))
	       && !info_match)
	{
	  info_match = li;
	}
    }
  if (name_match)
    {
      name_match->start = start;
      name_match->length = length;
    }
  else if (info_match)
    {
      if (info_match->name)
	g_free (info_match->name);
      info_match->name = g_strdup (name);
    }
  else
    {
      GdamLoopInfo *loop_info;
      loop_info = g_new0 (GdamLoopInfo, 1);
      loop_info->name = g_strdup (name);
      loop_info->start = start;
      loop_info->length = length;
      info->loop_list = g_list_append (info->loop_list, loop_info);
    }
  /* TODO: emit songinfo_changed?? */
  return TRUE;
}


/*=====================================================================*
 *                        gdam_song_info_set_bpm                       *
 *                                                                     *
 * Parameters:                                                         *
 *     song_info         the info to adjust.                           *
 *     bpm               the new beats-per-minute.                     *
 *=====================================================================*/
void gdam_song_info_set_bpm(GdamSongInfo* song_info, gdouble bpm)
{
  g_return_if_fail (song_info != NULL);
  g_return_if_fail (song_info->beat_info != NULL);
  gdam_beat_info_set_bpm (song_info->beat_info, bpm);
}
void gdam_song_info_set_first_beat(GdamSongInfo* song_info, gdouble first_beat)
{
  g_return_if_fail (song_info != NULL);
  g_return_if_fail (song_info->beat_info != NULL);
  gdam_beat_info_set_first_beat (song_info->beat_info, first_beat);
}


static void
kill_echo_song_property(gpointer key, gpointer value, gpointer unused)
{
  g_free (key);
  g_free (value);
}

const GHashTable* gdam_song_info_get_properties(GdamSongInfo* song_info)
{
  if (g_hash_table_size (song_info->song_properties))
    return song_info->song_properties;
  return NULL;
}

const GList* gdam_song_info_get_loops(GdamSongInfo* song_info)
{
  return song_info->loop_list;
}

static void gdam_song_info_destroy(GdamSongInfo* song_info)
{
  gdam_debug ("deleting songinfo for %s", song_info->filename);
  g_hash_table_foreach (song_info->song_properties,
			kill_echo_song_property, NULL);
  g_hash_table_destroy (song_info->song_properties);
  g_free (song_info->filename);
  g_free (song_info);
}
void gdam_song_info_unref(GdamSongInfo* song_info)
{
  g_return_if_fail (song_info->ref_count >= 1);
  if (--(song_info->ref_count) == 0)
    gdam_song_info_destroy (song_info);
}
void gdam_song_info_ref(GdamSongInfo* song_info)
{
  g_return_if_fail (song_info->ref_count >= 0);
  ++(song_info->ref_count);
}

const GdamBeatInfo* gdam_song_info_get_beat_info(GdamSongInfo* song_info)
{
  return song_info->beat_info;
}

void    gdam_song_info_set_beat_info(GdamSongInfo* song_info,
				     const GdamBeatInfo* beat_info)
{
  int num_ranges;
  num_ranges = g_list_length (beat_info->range_list);
  if (song_info->beat_info)
    gdam_beat_info_destroy (song_info->beat_info);
  song_info->beat_info = gdam_beat_info_copy (beat_info);
}

const char* gdam_song_info_get_filename(GdamSongInfo* song_info)
{
  g_return_val_if_fail (song_info != NULL, NULL);
  return song_info->filename;
}

double gdam_song_info_get_length(GdamSongInfo* song_info)
{
  return song_info->length;
}

double gdam_song_info_get_gain(GdamSongInfo* song_info)
{
  return song_info->gain;
}

int gdam_song_info_get_order(GdamSongInfo* song_info)
{
  return song_info->order;
}

void gdam_song_info_set_order(GdamSongInfo* song_info, int rank)
{
  song_info->order = rank;
}

static void add_prop_key_to_list(char* key, gpointer unused, GSList** list)
{
  *list = g_slist_prepend (*list, key);
}

static char* xml_escape(const char* word)
{
  char *buf;
  const char *tmp = word;
  int len = 0;
  while (*tmp)
    {
      if (*tmp == '<' || *tmp == '&' || *tmp == '>')
	len += 5;
      else
	len++;
      tmp++;
    }
  buf = g_new (char, len + 1);
  len = 0;
  tmp = word;
  while (*tmp)
    {
      if (*tmp == '<')
	{
	  memcpy (buf + len, "&lt;", 4);
	  len += 4;
	}
      else if (*tmp == '>')
	{
	  memcpy (buf + len, "&gt;", 4);
	  len += 4;
	}
      else if (*tmp == '&')
	{
	  memcpy (buf + len, "&amp;", 5);
	  len += 5;
	}
      else
	{
	  buf[len++] = *tmp;
	}
      tmp++;
    }
  buf[len] = 0;
  return buf;
}

void gdam_song_info_to_xml(GdamSongInfo* song_info, xmlNode* parent)
{
  GSList *keys = NULL;
  GSList *list;
  const GList *loop_list;
  const GList *range_list;
  xmlNode *node;
  char buf[256];
  char *tmp;
  GHashTable *properties = song_info->song_properties;
  int num_ranges;
  g_hash_table_foreach (properties, (GHFunc) add_prop_key_to_list, &keys);
  node = xmlNewChild (parent, NULL, "Song", NULL);

  /* XXX: how do we make this happen automatically??? */
  tmp = xml_escape (song_info->filename);
  xmlNewChild (node, NULL, "filename", tmp);
  g_free (tmp);

  sprintf (buf, "%.8g", song_info->length);
  xmlNewChild (node, NULL, "length", buf);

  sprintf (buf, "%.8g", gdam_beat_info_get_bpm (song_info->beat_info, 0));
  xmlNewChild (node, NULL, "bpm", buf);

  sprintf (buf, "%.8g",
	   gdam_beat_info_get_first_beat (song_info->beat_info, 0));
  xmlNewChild (node, NULL, "first_beat", buf);

  if (song_info->gain > 0)
    {
      sprintf (buf, "%5.5g", song_info->gain);
      xmlNewChild (node, NULL, "gain", buf);
    }

  for (list = keys; list != NULL; list = list->next)
    {
      char *value;
      char *key = list->data;
      value = g_hash_table_lookup (properties, key);
      xmlNewChild (node, NULL, key, value);
    }
  loop_list = gdam_song_info_get_loops (song_info);
  for (loop_list = loop_list; loop_list != NULL; loop_list = loop_list->next)
    {
      GdamLoopInfo *li = (GdamLoopInfo *) loop_list->data;
      sprintf (buf, "%s:%3.6f:%3.6f\n", li->name, li->start, li->length);
      xmlNewChild (node, NULL, "loop", buf);
    }

  num_ranges = g_list_length (song_info->beat_info->range_list);
  range_list = song_info->beat_info->range_list;
  for (range_list = range_list;
       range_list != NULL; range_list = range_list->next)
    {
      GdamBeatInfoRange *bir = (GdamBeatInfoRange *) range_list->data;
      sprintf (buf, "%3.8f:%3.8f:%3.8f:%3.8f\n", bir->bpm,
	       bir->first_beat, bir->range_start, bir->range_end);
      xmlNewChild (node, NULL, "range", buf);
    }
}

GdamSongInfo* gdam_song_info_from_xml(xmlNode* xml_node)
{
  GdamSongInfo *song_info;
  xmlNode *child;
  song_info = gdam_song_info_new (NULL, 90.0, 0.0);
  for (child = xml_node; child != NULL; child = child->next)
    {
      char *content;
      content = xmlNodeListGetString (NULL, child->childs, TRUE);
      if (strcmp (child->name, "filename") == 0)
	{
	  g_free (song_info->filename);
	  song_info->filename = g_strdup (content);
	}
      else
	{
	  gdam_song_info_add_property (song_info, child->name, content);
	}
    }
  if (song_info->filename == NULL)
    {
      gdam_song_info_unref (song_info);
      return NULL;
    }
  return song_info;
}
