/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambender.h"
#include <common/gdamsample.h>
#include <common/gdamplugin.h>
#include <gtk/gtksignal.h>
#include "../gdammodel.h"
#include "../gdamfilter.h"
#include "../gdamutil.h"

enum {
	SLOW_TO_STALL,
	LAST_SIGNAL
};

static guint bender_signals[LAST_SIGNAL];

/* Bender class function */
static void bender_real_slow_to_stall(GdamBender* bender, gdouble slow_time);
/* Filter class function */
static int gdam_bender_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);

static void gdam_bender_init(GdamBender* bender)
{
  bender->backbuffer = 8.0;
  bender->beat_factor = 0.5;
  bender->rate_factor = 1.0;
  bender->mode = GDAM_BENDER_KEEP_TIME;
}

static void gdam_bender_class_init(GdamBenderClass* klass);
GtkType gdam_bender_get_type()
{
  static GtkType bender_type = 0;
  if (!bender_type)
    {
      static const GtkTypeInfo bender_info = {
	"GdamBender",
	sizeof (GdamBender),
	sizeof (GdamBenderClass),
	(GtkClassInitFunc) gdam_bender_class_init,
	(GtkObjectInitFunc) gdam_bender_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      bender_type = gtk_type_unique (GDAM_TYPE_FILTER, &bender_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return bender_type;
}

static void made_request(gpointer data)
{
  GdamBender *bender = GDAM_BENDER (data);
  gdam_debug ("made request");
  if (bender->pending_request != NULL)
    {
      gdam_debug ("handling pending request");
      gdam_channel_append_request (GDAM_MODEL (bender)->channel,
				   bender->pending_request);
      bender->pending_request = NULL;
    }
  else
    {
      bender->making_request = FALSE;
    }
}

static void bender_real_slow_to_stall(GdamBender* bender, gdouble slow_time)
{
  g_warning ("deprecated method: slow_to_stall");
}

GdamBender* 	gdam_bender_new		()
{
  GdamBender *bender;
  bender = GDAM_BENDER (gtk_type_new (GDAM_TYPE_BENDER));
  return bender;
}

static int gdam_bender_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamBender *bender = GDAM_BENDER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  gtk_object_ref (GTK_OBJECT (bender));
  gdam_debug ("append_start_request: building bender from %d", new_source_id);
  model->source_id = gdam_channel_new_local_id (model->channel);
  /* TODO: set back buffer, mode */
  request = gdam_request_create_source ("GdasBender",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					bender,
					2,
					"filtered",
					GDAM_ARG_TYPE_INT,
					new_source_id,
					"buffer_length",
					GDAM_ARG_TYPE_INT,
					(int) (bender->backbuffer * 44100.0));
  return gdam_request_builder_append_create (builder, request);
}

void gdam_bender_slow_to_stall(GdamBender* bender, float slow_time)
{
  GtkObject *this_object = GTK_OBJECT (bender);

  if (gtk_signal_n_emissions (this_object, bender_signals[SLOW_TO_STALL]))
    return;

  gtk_signal_emit (this_object, bender_signals[SLOW_TO_STALL], slow_time);
}

void		gdam_bender_set_mode	(GdamBender*		bender,
					 GDAM_BENDER_MODE	mode)
{
  GdamArg args[1];

  gdam_debug ("Bender set_mode to %d", mode);

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  if (mode == GDAM_BENDER_KEEP_TIME)
    GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  else if (mode == GDAM_BENDER_SNAP_TIME)
    GDAM_ARG_STRING (args) = g_strdup ("snap_time");
  else if (mode == GDAM_BENDER_CONTINUOUS)
    GDAM_ARG_STRING (args) = g_strdup ("continuous");
  else if (mode == GDAM_BENDER_RETURN)
    GDAM_ARG_STRING (args) = g_strdup ("return");
  else
    return;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1, args, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
  bender->mode = mode;
}

void		gdam_bender_set_changes	(GdamBender*		bender,
					 int			num_points,
					 int*			changes)
{
  int *drag_data = changes;
  GdamArg args[2];

  /* TODO: fix this memory leak */
  /*memcpy(drag_data, changes, sizeof(int)*num_points*2); */

#if 0
  int n;
  gdam_debug ("Bender set_changes, %d points", num_points);
  fprintf (stderr, "Bender array:");
  for (n = 0; n < num_points * 2; n += 2)
    {
      fprintf (stderr, "  %f %f", (float) drag_data[n] / 44100,
	       (float) drag_data[n + 1] / GdamFixed_ONE);
    }
#endif

  /* TODO: copy this data? */
  args[0].name = "changes";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args).length = num_points ? num_points * 2 : 0;
  GDAM_ARG_INT_ARRAY (args).array = drag_data;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       made_request, made_request, bender);
      if (bender->making_request)
	{
	  gdam_debug ("delaying request...");
	  if (bender->pending_request != NULL)
	    gdam_request_destroy (bender->pending_request);
	  bender->pending_request = request;
	}
      else
	{
	  gdam_debug ("making request...");
	  bender->making_request = TRUE;
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }
}




void		gdam_bender_jump_to_mark(GdamBender*		bender)
{
  GdamArg args[2];

  gdam_debug ("Bender jump_to_mark");

  if (bender->mode != GDAM_BENDER_KEEP_TIME)
    {
      g_warning ("can't jump to mark in continuous mode...");
      return;
    }

  args[0].name = "jump_to_mark";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = 0;

  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = 0;
  GDAM_ARG_INT_ARRAY (args + 1).array = NULL;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       2,
				       args,
				       made_request, made_request, bender);
      /* TODO: should jump-to-mark ever delay it request?? */
      if (bender->making_request)
	{
	  gdam_debug ("delaying jump-to-mark request...");
	  if (bender->pending_request != NULL)
	    gdam_request_destroy (bender->pending_request);
	  bender->pending_request = request;
	}
      else
	{
	  bender->making_request = TRUE;
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }
}


void		gdam_bender_set_back_buffer	
					(GdamBender*		bender,
					 float			length)
{
  GdamArg args[1];

  gdam_debug ("Bender set_back_buffer");

  args[0].name = "back_buffer";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = length * 44100.0;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
  bender->backbuffer = length;
}

void		gdam_bender_rewind	(GdamBender*		bender,
					 float			rewind_length)
{
  float r_length = 0;
  int num_points = 1;
  int *drag_data = g_new (int, num_points * 2);
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  GdamArg args[2];
  gdam_debug ("Bender rewind, time: %f", rewind_length);
  if (!rewind_length)
    {
      float bpm;
      if (!bender->beat_factor)
	return;
      bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
      r_length = bender->beat_factor * 60.0 / bpm;
    }
  else if (rewind_length < 0)
    {
      r_length = 10.0;
    }
  else
    {
      r_length = rewind_length;
    }

  drag_data[0] = r_length * 44100.0 / r_f;
  drag_data[1] = -1 * r_f * GdamFixed_ONE;
  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       2,
				       args,
				       made_request, made_request, bender);
      if (bender->making_request)
	{
	  gdam_debug ("delaying rewind request");
	  if (bender->pending_request != NULL)
	    gdam_request_destroy (bender->pending_request);
	  bender->pending_request = request;
	}
      else
	{
	  bender->making_request = TRUE;
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }
}

static int delta_millis(GTimeVal* older, GTimeVal* younger)
{
  int millis = (older->tv_sec - younger->tv_sec) * 1000
    + (older->tv_usec - younger->tv_usec) / 1000;
  return millis;
}

static void gdam_bender_arg_jump_to_mark(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (GDAM_ARG_INT (arg) >= 1)
    gdam_bender_jump_to_mark (bender);
}

#define GDAM_BEND_DELTA_CEILING 200  
static void gdam_bender_arg_set_bend_amount(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GTimeVal real_time;
  int num_points = 10;
  float total_time = 0.2;
  int *change_array = g_new (int, num_points * 2);
  int n, delmil;
  float last_rate = bender->last_rate;
  float new_rate;
  float cur_rate = last_rate;
  float step_rate;
  float step_time;
  int pitch = (int) (GDAM_ARG_DOUBLE (arg) * 8192.0 * 2);
  gboolean jump_to_mark = FALSE;

  /* XXX: for now, out-of-range value means not bending. */
  if ((pitch < 0 || pitch > 16383))
    {
      if (!bender->is_bending)
	{
	  return;
	}
      gdam_debug ("HANDS OFF - no longer bending...");
      /*  TODO: re-enable this once it doesn't crash server. */
      jump_to_mark = TRUE;
      /*gdam_bender_jump_to_mark(bender); */
      new_rate = 1.0;
      bender->is_bending = FALSE;
    }
  else
    {
      bender->is_bending = TRUE;
      /* We fake (for now) raw pitch range: 0 <- 8192 -> 16383 */
      /* Pitch Wheel Mode:  0 <- 1 -> 2 */
      if (bender->bend_mode == 0)
	{
	  new_rate = (float) pitch / 8192;
	}
      /* Pitch Wheel Mode:  -2 <- 0 |1| 0-> 2 */
      else if (bender->bend_mode == 1)
	{
	  if (pitch == 8192)
	    new_rate = 1.0;
	  else if (pitch < 8192)
	    new_rate = (float) (8192 - pitch) * -2 / 8192.0;
	  else
	    new_rate = (float) (8192 - pitch) * -2 / 8192.0;
	}
      /* Pitch Wheel Mode:  -2 <- 1 -> 2 */
      else if (bender->bend_mode == 2)
	{
	  if (pitch >= 8192)
	    new_rate = (float) pitch / 8192.0;
	  else if (pitch > 5000)
	    new_rate = (float) (pitch - 5000) * -1 / 3192.0;
	  else
	    new_rate = (float) (pitch - 5000) * 2 / 5000.0;
	}
      else
	{
	  g_warning ("Confused in bend.");
	  new_rate = 1.0;
	}
    }

  step_rate = (new_rate - last_rate) / num_points;

  g_get_current_time (&real_time);
  delmil = delta_millis (&real_time, &bender->bend_adj_time);
  g_get_current_time (&bender->bend_adj_time);

  if (delmil < GDAM_BEND_DELTA_CEILING)
    {
      total_time = 0.001 * delmil;
    }
  if (delmil == 0)
    {
      total_time = 0.001;
    }

  step_time = total_time / num_points;
  bender->last_rate = new_rate;

  gdam_debug ("midi bend wheel changed!!!");
  gdam_debug ("delmil: %d", delmil);

  for (n = 0; n < num_points * 2; n += 2)
    {
      cur_rate += step_rate;
      change_array[n] = step_time * 44100;
      if (n == num_points * 2 - 2)
	change_array[n] = 10.0 * 44100;
      change_array[n + 1] = cur_rate * GdamFixed_ONE;
    }
  gdam_bender_set_changes (bender, num_points, change_array);
  if (jump_to_mark)
    gdam_bender_jump_to_mark (bender);
}



void		gdam_bender_rub		(GdamBender*		bender,
					 float			rub_width,
					 float			rub_length)
{
  float r_length = 0;
  float r_width = 0;
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  int num_points = 1;
  int *drag_data;
  int n;
  GdamArg args[2];

  if (!rub_width)
    {
      float bpm;
      if (bender->beat_factor <= 0)
	return;
      bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
      if (bpm > 1)
	r_width = bender->beat_factor * 60.0 / bpm;
      else
	r_width = bender->beat_factor;
    }
  else if (rub_width < 0)
    {
      /* TODO: what to do? */
      return;
    }
  else
    r_width = rub_width;
  if (!rub_length)
    {
      r_length = 10.0;
    }
  else if (rub_length < 0)
    {
      return;
    }
  else
    r_length = rub_length;

  g_return_if_fail (r_width > 0);
  g_return_if_fail (r_length > 0);


  gdam_debug ("Bender rub, width: %f length: %f", r_width, r_length);


  /*  Rub rates weighted +/- rate */
  if (bender->rub_mode == 1)
    {
      float rate0 = r_f;
      float rate1 = (1.0 + (1.0 - r_f));
      float time0 = rate1 * 2 * r_width / (rate0 + rate1);
      float time1 = 2 * r_width - time0;

      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = time0 * 44100;
	  drag_data[n + 1] = -1.0 * rate0 * GdamFixed_ONE;
	  drag_data[n + 2] = time1 * 44100;
	  drag_data[n + 3] = rate1 * GdamFixed_ONE;
	}
    }
  /*  Rub rates weighted +rate, normal */
  else if (bender->rub_mode == 2)
    {
      float rate0 = r_f;
      float rate1 = 1.0;
      float time0 = rate1 * 2 * r_width / (rate0 + rate1);
      float time1 = 2 * r_width - time0;

      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = time0 * 44100;
	  drag_data[n + 1] = -1.0 * rate0 * GdamFixed_ONE;
	  drag_data[n + 2] = time1 * 44100;
	  drag_data[n + 3] = rate1 * GdamFixed_ONE;
	}
    }
  /*  Rub rates even.  */
  else
    {
      /*if (bender->rub_mode == 0) */
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = r_width * 44100 / r_f;
	  drag_data[n + 3] = r_f * GdamFixed_ONE;
	}
    }

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       2,
				       args,
				       made_request, made_request, bender);
      if (bender->making_request)
	{
	  if (bender->pending_request != NULL)
	    gdam_request_destroy (bender->pending_request);
	  bender->pending_request = request;
	}
      else
	{
	  bender->making_request = TRUE;
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }
}

static void gdam_bender_arg_set_rate_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->rate_factor = GDAM_ARG_DOUBLE (arg);
}
static void gdam_bender_arg_get_rate_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->rate_factor;
}

static void gdam_bender_arg_set_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_set_mode (bender, GDAM_ARG_INT (arg));
}
static void gdam_bender_arg_get_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (arg->type == GDAM_ARG_TYPE_INT)
    {
      GDAM_ARG_INT (arg) = bender->mode;
    }
  else
    {
      GDAM_BENDER_MODE mode = bender->mode;
      if (mode == GDAM_BENDER_KEEP_TIME)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_time");
      else if (mode == GDAM_BENDER_SNAP_TIME)
	GDAM_ARG_STRING (arg) = g_strdup ("snap_time");
      else if (mode == GDAM_BENDER_CONTINUOUS)
	GDAM_ARG_STRING (arg) = g_strdup ("continuous");
      else if (mode == GDAM_BENDER_RETURN)
	GDAM_ARG_STRING (arg) = g_strdup ("return");
    }
}

static void gdam_bender_arg_set_bend_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->bend_mode = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_get_bend_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->bend_mode;
}

static void gdam_bender_arg_set_rub_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->rub_mode = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_get_rub_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->rub_mode;
}

static void gdam_bender_arg_set_eighth_beat_factor(GtkObject* object, 
						   GdamArg*   arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->beat_factor = GDAM_ARG_DOUBLE (arg) / 8;
}
static void gdam_bender_arg_get_eighth_beat_factor(GtkObject* object, 
						   GdamArg*   arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = 8 * bender->beat_factor;
}

static void gdam_bender_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->beat_factor = GDAM_ARG_DOUBLE (arg);
}

static void gdam_bender_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->beat_factor;
}


static void gdam_bender_class_init(GdamBenderClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  klass->slow_to_stall = bender_real_slow_to_stall;
  filter_class->append_start_request = gdam_bender_append_start_request;
  gdam_arg_register (object_class,
		     "rate_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_rate_factor,
		     gdam_bender_arg_get_rate_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_beat_factor,
		     gdam_bender_arg_get_beat_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "eighth_beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_eighth_beat_factor,
		     gdam_bender_arg_get_eighth_beat_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_mode,
		     gdam_bender_arg_get_mode, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bend_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_bend_mode,
		     gdam_bender_arg_get_bend_mode,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bend_amount",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_bend_amount, NULL, NULL, ",", 0);
  gdam_arg_register (object_class,
		     "jump_to_mark",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_mark, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "rub_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rub_mode,
		     gdam_bender_arg_get_rub_mode,
		     NULL, NULL, GDAM_ARG_VITAL);

  bender_signals[SLOW_TO_STALL] =
    gtk_signal_new ("slow_to_stall",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBenderClass, slow_to_stall),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, bender_signals, LAST_SIGNAL);
}

void gdam_bender_set_mark(GdamBender* bender){
	if (bender->mode != GDAM_BENDER_KEEP_TIME){
		g_warning("can't set mark in continuous mode...");
		return;
	}
	gdam_bender_set_mode(bender, GDAM_BENDER_KEEP_TIME); 
}

