/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamfiltervibrato.h"
#include <common/gdamplugin.h>
#include <gtk/gtksignal.h>

static void gdam_vibrato_real_rates_changed(GdamFilterVibrato* vibrato);
static int
gdam_filter_vibrato_append_start_request(GdamFilter*          filter,
				         int                  new_source_id,
				         GdamRequestBuilder*  builder);
static void gdam_filter_vibrato_handle_new_modifier(GdamFilterWavetable*,
	                                            int wavetable_local_id,
						    GdamRequestBuilder*);

enum {
	RATES_CHANGED,
	LAST_SIGNAL
};

static guint vibrato_signals[LAST_SIGNAL] = { 0 };

static void gdam_filter_vibrato_init(GdamFilterVibrato* filter_vibrato)
{
  filter_vibrato->min_rate = 1.0;
  filter_vibrato->max_rate = 1.0;
}

void gdam_vibrato_set_min_rate(GdamFilterVibrato* vibrato, double rate)
{
  GdamArg arg;
  arg.name = "min_rate";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = rate;
  gdam_arg_set_object_arg (GTK_OBJECT (vibrato), &arg);
}

void gdam_vibrato_set_max_rate(GdamFilterVibrato* vibrato, double rate)
{
  GdamArg arg;
  arg.name = "max_rate";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = rate;
  gdam_arg_set_object_arg (GTK_OBJECT (vibrato), &arg);
}

static void gdam_filter_vibrato_arg_set_min_rate(GtkObject* object, 
						  GdamArg* arg)
{
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (object);
  gdouble center_rate = (vibrato->min_rate + vibrato->max_rate) / 2.0;
  gdouble max_rate;
  vibrato->min_rate = GDAM_ARG_DOUBLE (arg);
  if (vibrato->min_rate > center_rate)
    vibrato->min_rate = center_rate;
  max_rate = center_rate + (center_rate - vibrato->min_rate);
  if (max_rate != vibrato->max_rate)
    gdam_vibrato_set_max_rate (vibrato, max_rate);
  /*vibrato->max_rate = max_rate; */
  gdam_filter_vibrato_rates_changed (vibrato);
}
static void gdam_filter_vibrato_arg_get_min_rate(GtkObject* object, 
						  GdamArg* arg)
{
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (object);
  GDAM_ARG_DOUBLE (arg) = vibrato->min_rate;
}

static void gdam_filter_vibrato_arg_set_mid_rate(GtkObject* object, 
						  GdamArg* arg)
{
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (object);
  gdouble old_center_rate = (vibrato->min_rate + vibrato->max_rate) / 2.0;
  gdouble delta = GDAM_ARG_DOUBLE (arg) - old_center_rate;
  gdouble min_rate, max_rate;
  min_rate = vibrato->min_rate + delta;
  max_rate = vibrato->max_rate + delta;
  vibrato->min_rate = min_rate;
  vibrato->max_rate = max_rate;
  gdam_vibrato_set_min_rate (vibrato, min_rate);
  gdam_vibrato_set_max_rate (vibrato, max_rate);
  gdam_filter_vibrato_rates_changed (vibrato);
}
static void gdam_filter_vibrato_arg_get_mid_rate(GtkObject* object, 
						  GdamArg* arg)
{
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (object);
  GDAM_ARG_DOUBLE (arg) = (vibrato->min_rate + vibrato->max_rate) / 2.0;
}

static void gdam_filter_vibrato_arg_set_max_rate(GtkObject* object, 
						  GdamArg* arg)
{
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (object);
  gdouble center_rate = (vibrato->min_rate + vibrato->max_rate) / 2.0;
  gdouble min_rate;
  vibrato->max_rate = GDAM_ARG_DOUBLE (arg);
  if (vibrato->max_rate < center_rate)
    vibrato->max_rate = center_rate;
  min_rate = center_rate - (vibrato->max_rate - center_rate);
  if (min_rate != vibrato->min_rate)
    gdam_vibrato_set_min_rate (vibrato, min_rate);
  /*vibrato->min_rate = min_rate; */
  gdam_filter_vibrato_rates_changed (vibrato);
}
static void gdam_filter_vibrato_arg_get_max_rate(GtkObject* object, 
						  GdamArg* arg)
{
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (object);
  GDAM_ARG_DOUBLE (arg) = vibrato->max_rate;
}

static void gdam_filter_vibrato_class_init(GdamFilterVibratoClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GdamFilterWavetableClass *wavetable_class;
  wavetable_class = GDAM_FILTER_WAVETABLE_CLASS (klass);
  GDAM_FILTER_CLASS (klass)->append_start_request
    = gdam_filter_vibrato_append_start_request;
  wavetable_class->handle_new_modifier
    = gdam_filter_vibrato_handle_new_modifier;
  klass->rates_changed = gdam_vibrato_real_rates_changed;
  vibrato_signals[RATES_CHANGED]
    = gtk_signal_new ("rates_changed",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamFilterVibratoClass,
					 rates_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, vibrato_signals, LAST_SIGNAL);
  gdam_arg_register (object_class,
		     "min_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_filter_vibrato_arg_set_min_rate,
		     gdam_filter_vibrato_arg_get_min_rate,
		     NULL, "0,30", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mid_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_filter_vibrato_arg_set_mid_rate,
		     gdam_filter_vibrato_arg_get_mid_rate,
		     NULL, "0,30", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "max_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_filter_vibrato_arg_set_max_rate,
		     gdam_filter_vibrato_arg_get_max_rate,
		     NULL, "0,30", GDAM_ARG_VITAL);
}

GtkType gdam_filter_vibrato_get_type()
{
  static GtkType filter_vibrato_type = 0;
  if (!filter_vibrato_type)
    {
      static const GtkTypeInfo filter_vibrato_info = {
	"GdamFilterVibrato",
	sizeof (GdamFilterVibrato),
	sizeof (GdamFilterVibratoClass),
	(GtkClassInitFunc) gdam_filter_vibrato_class_init,
	(GtkObjectInitFunc) gdam_filter_vibrato_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER_WAVETABLE;
      filter_vibrato_type = gtk_type_unique (parent, &filter_vibrato_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return filter_vibrato_type;
}
static int
gdam_filter_vibrato_append_start_request(GdamFilter*          filter,
				         int                  new_source_id,
				         GdamRequestBuilder*  builder)
{
  gint wavetable_local_id;
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  GdamFilterVibrato *vibrato = GDAM_FILTER_VIBRATO (filter);

  wavetable_local_id = gdam_filter_wavetable_make_wavetable_request
    (GDAM_FILTER_WAVETABLE (filter), builder);
  /*if (wavetable_local_id == 0) {
     g_warning("couldn't produce wavetable");
     return 0;
     } */
  gtk_object_ref (GTK_OBJECT (filter));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source
    ("GdasVibrato",
     model->source_id,
     gdam_model_got_subsource, gdam_model_didnt_get_subsource, filter, 4,
     //"mode",
     //GDAM_ARG_TYPE_STRING,
     //"one_ended",
     "vibrato",
     GDAM_ARG_TYPE_INT,
     wavetable_local_id,
     "filtered",
     GDAM_ARG_TYPE_INT,
     new_source_id,
     "min_rate",
     GDAM_ARG_TYPE_DOUBLE,
     vibrato->min_rate, "max_rate", GDAM_ARG_TYPE_DOUBLE, vibrato->max_rate);
  return gdam_request_builder_append_create (builder, request);
}
static void gdam_filter_vibrato_handle_new_modifier
				(GdamFilterWavetable* filter,
			         int wavetable_local_id,
	                         GdamRequestBuilder* builder)
{
  int source_id = GDAM_MODEL (filter)->source_id;
  GdamArg arg[1];
  g_return_if_fail (source_id != 0);
  arg[0].name = "vibrato";
  arg[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (arg + 0) = wavetable_local_id;
  gdam_request_builder_append_generic
    (builder, gdam_request_set_args (source_id, 1, arg, NULL, NULL, NULL));;
}

static void gdam_vibrato_real_rates_changed(GdamFilterVibrato* vibrato)
{
  GdamModel *model = GDAM_MODEL (vibrato);
  guint source_id = model->source_id;
  if (source_id != 0)
    {
      GdamArg args[2];
      args[0].name = "min_rate";
      args[0].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (args + 0) = vibrato->min_rate;
      args[1].name = "max_rate";
      args[1].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (args + 1) = vibrato->max_rate;
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args
				   (source_id, 2, args, NULL, NULL, NULL));
    }
}
void gdam_filter_vibrato_rates_changed(GdamFilterVibrato* vibrato)
{
  GtkObject *object = GTK_OBJECT (vibrato);
  if (vibrato->min_rate > vibrato->max_rate)
    {
      gdouble swap = vibrato->min_rate;
      vibrato->min_rate = vibrato->max_rate;
      vibrato->max_rate = swap;
    }
  if (gtk_signal_n_emissions (object, vibrato_signals[RATES_CHANGED]))
    return;
  gtk_signal_emit (object, vibrato_signals[RATES_CHANGED]);
}
