/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamfilterwavetable.h"
#include <common/gdamplugin.h>
#include "../gdambuffermanager.h"

static GtkObjectClass* parent_class = NULL;

static void gdam_filter_wavetable_destroy(GtkObject* object);
static void gdam_filter_wavetable_detach_from_buffer(GdamModel* model,
                                                     guint buffer_id);
						

static void gdam_filter_wavetable_init(GdamFilterWavetable* filter_wavetable)
{
  filter_wavetable->hertz = 60.0;
}

static void gdam_filter_wavetable_class_init(GdamFilterWavetableClass* klass);

GtkType gdam_filter_wavetable_get_type()
{
  static GtkType filter_wavetable_type = 0;
  if (!filter_wavetable_type)
    {
      static const GtkTypeInfo filter_wavetable_info = {
	"GdamFilterWavetable",
	sizeof (GdamFilterWavetable),
	sizeof (GdamFilterWavetableClass),
	(GtkClassInitFunc) gdam_filter_wavetable_class_init,
	(GtkObjectInitFunc) gdam_filter_wavetable_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      filter_wavetable_type = gtk_type_unique (parent,
					       &filter_wavetable_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return filter_wavetable_type;
}

static void on_wavetable_success(int id, gpointer data)
{
  GdamFilterWavetable *filter = GDAM_FILTER_WAVETABLE (data);
  filter->wavetable_source_id = id;
  gtk_object_unref (GTK_OBJECT (data));
}

static void on_wavetable_failure(gpointer data)
{
  GdamFilterWavetable *filter = GDAM_FILTER_WAVETABLE (data);
  g_warning ("failed to make wavetable for filter (buffer id = %d)",
	     filter->buffer_id);
  gtk_object_unref (GTK_OBJECT (data));
}

gint gdam_filter_wavetable_make_wavetable_request
                             (GdamFilterWavetable* wavetable,
                              GdamRequestBuilder*  builder)
{
  GdamModel *model = GDAM_MODEL (wavetable);
  int local_id;
  if (wavetable->buffer_id == 0)
    return 0;
  gtk_object_ref (GTK_OBJECT (wavetable));
  local_id = gdam_channel_new_local_id (model->channel);
  return gdam_request_builder_append_create
    (builder,
     gdam_request_create_source
     ("GdasWavetable",
      local_id,
      on_wavetable_success,
      on_wavetable_failure,
      wavetable,
      2,
      "buffer_id",
      GDAM_ARG_TYPE_INT,
      wavetable->buffer_id, "hertz", GDAM_ARG_TYPE_DOUBLE, wavetable->hertz));
}

void gdam_filter_wavetable_set_hertz(GdamFilterWavetable* filter, gdouble hertz)
{
  GdamRequest *request;
  filter->hertz = hertz;
  request = gdam_request_set_args
    (filter->wavetable_source_id,
     1,
     gdam_arg_new_from_valist
     (1, "hertz", GDAM_ARG_TYPE_DOUBLE, filter->hertz), NULL, NULL, NULL);
  gdam_channel_append_request (GDAM_MODEL (filter)->channel, request);
}

void gdam_filter_wavetable_set_buffer_id(GdamFilterWavetable* filter,
                                         int                  buffer_id,
					 gboolean             must_dealloc)
{
  GdamRequest *request;
  GdamModel *component = GDAM_MODEL (filter);
  GdamChannel *channel = component->channel;

  if (filter->buffer_id != 0 && filter->must_dealloc_buffer_id)
    {
      gdam_buffer_manager_unref (filter->buffer_id, component);
    }
  filter->buffer_id = buffer_id;
  filter->must_dealloc_buffer_id = must_dealloc;
  if (filter->wavetable_source_id == 0)
    {
      GdamRequestBuilder *builder;
      GdamFilterWavetableClass *klass;
      int wavetable_local_id;
      int source_id = component->source_id;

      /* Race condition: what if the request was pending? */
      if (source_id == 0)
	return;

      klass = GDAM_FILTER_WAVETABLE_CLASS (GTK_OBJECT (filter)->klass);
      g_return_if_fail (klass->handle_new_modifier != NULL);
      builder = gdam_request_builder_new ();
      wavetable_local_id =
	gdam_filter_wavetable_make_wavetable_request (filter, builder);
      (*klass->handle_new_modifier) (filter, wavetable_local_id, builder);
      gdam_channel_append_request
	(channel, gdam_request_builder_package (builder));
      gdam_request_builder_destroy (builder);
      return;
    }
  /* XXX: memory leak of args */
  request = gdam_request_set_args
    (filter->wavetable_source_id,
     2,
     gdam_arg_new_from_valist
     (2,
      "buffer_id",
      GDAM_ARG_TYPE_INT,
      filter->buffer_id,
      "hertz", GDAM_ARG_TYPE_DOUBLE, filter->hertz), NULL, NULL, NULL);
  gdam_channel_append_request (channel, request);
}

void gdam_filter_wavetable_set_default_buffer(GdamFilterWavetable* filter)
{
  guint buffer_id;
  buffer_id = gdam_buffer_manager_obtain_id (GDAM_MODEL (filter));
  if (buffer_id == 0)
    {
      g_warning ("no buffer has been selected");
      return;
    }
  gdam_filter_wavetable_set_buffer_id (filter, buffer_id, TRUE);
}

static void gdam_filter_wavetable_destroy(GtkObject* object)
{
  GdamFilterWavetable *filter = GDAM_FILTER_WAVETABLE (object);
  if (filter->buffer_id != 0 && filter->must_dealloc_buffer_id)
    {
      gdam_buffer_manager_unref (filter->buffer_id, GDAM_MODEL (filter));
    }
  (*parent_class->destroy) (object);
}
static void gdam_filter_wavetable_detach_from_buffer(GdamModel* model,
                                                     guint buffer_id)
{
  GdamFilterWavetable *wavetable;
  wavetable = GDAM_FILTER_WAVETABLE (model);
  if (buffer_id == wavetable->buffer_id)
    gdam_filter_wavetable_set_buffer_id (wavetable, 0, FALSE);
  else
    {
      /* Chain up to the parent class. */
      GdamModelClass *klass;
      klass = GDAM_MODEL_CLASS (parent_class);
      if (klass->detach_from_buffer != NULL)
	(*klass->detach_from_buffer) (model, buffer_id);
      else
	{
	  g_warning ("filter_wavetable: no way to detach "
		     "from buffer %d", buffer_id);
	}
    }
}

static void gdam_filter_wavetable_arg_set_buffer_id(GtkObject* object, 
						    GdamArg*   arg)
{
  GdamFilterWavetable *filter_wavetable = GDAM_FILTER_WAVETABLE (object);
  int id = GDAM_ARG_INT (arg);
  if (id == 0)
    {
      g_warning ("no buffer has been selected");
      return;
    }
  /*  TODO:  what value for third arg?? (must_dealloc) */
  gdam_filter_wavetable_set_buffer_id (filter_wavetable, id, FALSE);
}

static void gdam_filter_wavetable_arg_get_buffer_id(GtkObject* object, 
						    GdamArg* arg)
{
  GdamFilterWavetable *filter_wavetable = GDAM_FILTER_WAVETABLE (object);
  GDAM_ARG_INT (arg) = filter_wavetable->buffer_id;
}

static void gdam_filter_wavetable_arg_set_frequency(GtkObject* object, 
						    GdamArg* arg)
{
  GdamFilterWavetable *filter_wavetable = GDAM_FILTER_WAVETABLE (object);
  gdam_filter_wavetable_set_hertz (filter_wavetable, GDAM_ARG_DOUBLE (arg));
}

static void gdam_filter_wavetable_arg_get_frequency(GtkObject* object, 
						    GdamArg* arg)
{
  GdamFilterWavetable *filter_wavetable = GDAM_FILTER_WAVETABLE (object);
  GDAM_ARG_DOUBLE (arg) = filter_wavetable->hertz;
}

static void gdam_filter_wavetable_class_init(GdamFilterWavetableClass* klass)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gdam_filter_wavetable_destroy;
  GDAM_MODEL_CLASS (klass)->detach_from_buffer
    = gdam_filter_wavetable_detach_from_buffer;
  gdam_arg_register (object_class,
		     "frequency",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_filter_wavetable_arg_set_frequency,
		     gdam_filter_wavetable_arg_get_frequency,
		     NULL, "0,44100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdam_filter_wavetable_arg_set_buffer_id,
		     gdam_filter_wavetable_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
}
