/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamladspafilter.h"
#include <common/gdamsearchpath.h>
#include <common/gdam-protocol.h>
#include <stdlib.h>
#include <string.h>

/*
 *        _             _         _
 *  _ __ | |_   _  __ _(_)_ __   | |_ _   _ _ __   ___  ___
 * | '_ \| | | | |/ _` | | '_ \  | __| | | | '_ \ / _ \/ __|
 * | |_) | | |_| | (_| | | | | | | |_| |_| | |_) |  __/\__ \
 * | .__/|_|\__,_|\__, |_|_| |_|  \__|\__, | .__/ \___||___/
 * |_|            |___/               |___/|_|
 */

/* Here is a sample section from the xml file: */
/*
  <plugin>
    <name>...</name>
    <author>...</author>
    <unique_id>...</unique_id>
    <module>...</module>
    <module_index>...</module_index>
    <control>
      <name>...</name>
      <logarithmic>1</logarithmic>
      <min_value>0</min_value>
      <max_value>1</max_value>
    </control>
    <meter>
      <name>...</name>
      <logarithmic>1</logarithmic>
      <min_value>0</min_value>
      <max_value>1</max_value>
    </meter>
    <num_inputs>1</num_inputs>
    <num_outputs>1</num_outputs>
  </plugin>
*/

/* Random note:
 *
 * Although a LADSPA plugin may have any number of
 * inputs and outputs, here we only support 1 of them.
 * Furthermore we require that there be at least one of each.
 */

/* Hashtable mapping ``database names'' to PluginTypes. */
static GHashTable* plugin_hashtable = NULL;
static GPtrArray*  plugin_names = NULL;

static gint
xml_node_to_boolean(xmlNode* node)
{
  const char *contents = (const char *) node->content;
  if (contents == NULL && node->childs != NULL)
    contents = (const char *) node->childs->content;
  g_return_val_if_fail (contents != NULL, 0);
  return atoi (contents);
}

static gdouble
xml_node_to_double(xmlNode* node)
{
  const char *contents = (const char *) node->content;
  if (contents == NULL && node->childs != NULL)
    contents = (const char *) node->childs->content;
  g_return_val_if_fail (contents != NULL, 0);
  return strtod (contents, NULL);
}

char** gdam_ladspa_plugin_type_get_names  ()
{
  if (plugin_names == NULL)
    return NULL;

  return (char **) plugin_names->pdata;
}

static gboolean
parse_control_or_meter(xmlNode*                  children,
                       GdamLadspaParameterType*  param_out)
{
  xmlNode *node;
  memset (param_out, 0, sizeof (GdamLadspaParameterType));
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "name") == 0)
	{
	  if (param_out->name != NULL)
	    {
	      g_warning ("control had multiple <name> tags");
	      return FALSE;
	    }
	  param_out->name = (char *) xmlNodeGetContent (node);
	}
      else if (strcmp (node->name, "logarithmic") == 0)
	{
	  param_out->is_logarithmic = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "sample_rate_ratio") == 0)
	{
	  param_out->is_sample_rate_ratio = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "integer") == 0)
	{
	  param_out->is_integer = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "toggle") == 0)
	{
	  param_out->is_toggle = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "min_value") == 0)
	{
	  param_out->min_value = xml_node_to_double (node);
	  param_out->has_min_value = 1;
	}
      else if (strcmp (node->name, "max_value") == 0)
	{
	  param_out->max_value = xml_node_to_double (node);
	  param_out->has_max_value = 1;
	}
      else
	{
	  g_message ("unknown node type for ladspa control: %s", node->name);
	}
    }
  return TRUE;
}

static void gdam_ladspa_plugin_type_destroy(GdamLadspaPluginType* type)
{
  int i;
  g_free (type->database_name);
  g_free (type->module_name);
  g_free (type->author);
  for (i = 0; i < type->num_meters; i++)
    g_free (type->meters[i].name);
  g_free (type->meters);
  for (i = 0; i < type->num_controls; i++)
    g_free (type->controls[i].name);
  g_free (type->controls);
  g_free (type);
}


static GdamLadspaPluginType* gdam_ladspa_plugin_type_from_xml(xmlNode* node)
{
  GdamLadspaPluginType *rv;
  GArray *control_array;
  GArray *meter_array;
  xmlNode *child;

  rv = g_new0 (GdamLadspaPluginType, 1);
  control_array = g_array_new (FALSE, FALSE,
			       sizeof (GdamLadspaParameterType));
  meter_array = g_array_new (FALSE, FALSE, sizeof (GdamLadspaParameterType));

  /* XXX: still need to check that the string returned by
   *      xmlNodeGetContent should be g_free'd !!!!!!!!! */
  for (child = node->childs; child != NULL; child = child->next)
    {
      if (strcmp (child->name, "name") == 0)
	{
	  rv->database_name = (char *) xmlNodeGetContent (child);
	}
      else if (strcmp (child->name, "author") == 0)
	{
	  rv->author = (char *) xmlNodeGetContent (child);
	}
      else if (strcmp (child->name, "unique_id") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->unique_id = id;
	}
      else if (strcmp (child->name, "module") == 0)
	{
	  rv->module_name = (char *) xmlNodeGetContent (child);
	}
      else if (strcmp (child->name, "module_index") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->module_index = id;
	}
      else if (strcmp (child->name, "control") == 0)
	{
	  GdamLadspaParameterType param_type;
	  if (!parse_control_or_meter (child->childs, &param_type))
	    {
	      g_warning ("couldn't parse control: skipping");
	      continue;
	    }
	  g_array_append_val (control_array, param_type);
	}
      else if (strcmp (child->name, "meter") == 0)
	{
	  GdamLadspaParameterType param_type;
	  if (!parse_control_or_meter (child->childs, &param_type))
	    {
	      g_warning ("couldn't parse meter: skipping");
	      continue;
	    }
	  g_array_append_val (meter_array, param_type);
	}
      else if (strcmp (child->name, "num_inputs") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->num_inputs = id;
	}
      else if (strcmp (child->name, "num_outputs") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->num_outputs = id;
	}
      else
	{
	  g_message ("unknown node type for ladspa plugin: %s"
		     "... ignoring it", child->name);
	}
    }

  /* Shimmy the array we just constructed into the PluginType
   * structure. */
  rv->num_controls = control_array->len;
  rv->controls = (GdamLadspaParameterType *) control_array->data;
  rv->num_meters = meter_array->len;
  rv->meters = (GdamLadspaParameterType *) meter_array->data;
  g_array_free (control_array, FALSE);
  g_array_free (meter_array, FALSE);

  if (rv->database_name == NULL
      || rv->module_name == NULL || rv->module_index < 0)
    {
      g_warning ("imcomplete XML-LADSPA description");
      gdam_ladspa_plugin_type_destroy (rv);
      return NULL;
    }
  if (rv->num_inputs == 0)
    {
      g_message ("skipping LADSPA plugin with no inputs [%s]",
		 rv->database_name);
      gdam_ladspa_plugin_type_destroy (rv);
      return NULL;
    }
  if (rv->num_outputs == 0)
    {
      g_message ("skipping LADSPA plugin with no outputs [%s]",
		 rv->database_name);
      gdam_ladspa_plugin_type_destroy (rv);
      return NULL;
    }
  return rv;
}

static void initialize_ladspa_database()
{
  char *filename;
  xmlDoc *doc;
  xmlNode *root;
  xmlNode *child;

  if (plugin_hashtable == NULL)
    {
      plugin_hashtable = g_hash_table_new (g_str_hash, g_str_equal);
      plugin_names = g_ptr_array_new ();
    }

  filename = gdam_search_path ("ladspa.xml",
			       "/usr/share/ladspa:/usr/local/share/ladspa"
			       ":/usr/lib/ladspa:/usr/local/lib/ladspa"
			       ":/usr/share/gdam:/usr/local/share/gdam"
			       ":/usr/lib/gdam:/usr/local/lib/gdam"
			       ":~/.gdam", GDAM_SEARCH_PATH_FIND_FILE);
  if (filename == NULL)
    {
      g_warning ("couldn't find ladspa.xml -- "
		 "you need to run gdamladspapluginscanner");
      return;
    }


  doc = xmlParseFile (filename);
  if (doc == NULL)
    {
      g_warning ("xml in %s (ladspa database) malformed", filename);
      return;
    }

  root = doc->root;
  if (strcmp (root->name, "LADSPA-Plugin-Descriptions") != 0)
    {
      g_warning ("LADSPA plugin database didn't have appropriate "
		 "initial xml-tag <LADSPA-Plugin-Descriptions>");
      xmlFreeDoc (doc);
      return;
    }
  for (child = root->childs; child != NULL; child = child->next)
    {
      GdamLadspaPluginType *plugin_type;
      plugin_type = gdam_ladspa_plugin_type_from_xml (child);
      if (plugin_type == NULL)
	{
	  /*g_message("error parsing ladspa plugin- skipping"); */
	  continue;
	}
      if (g_hash_table_lookup (plugin_hashtable,
			       plugin_type->database_name) != NULL)
	{
	  g_warning ("multiple plugins named `%s' were found -- "
		     "ignoring duplicates", plugin_type->database_name);
	  gdam_ladspa_plugin_type_destroy (plugin_type);
	  continue;
	}
      g_hash_table_insert (plugin_hashtable,
			   plugin_type->database_name, plugin_type);
      g_ptr_array_add (plugin_names, plugin_type->database_name);
    }
  xmlFreeDoc (doc);
}

/*=====================================================================
 *          Server correspondence
 */
static int gdam_ladspa_filter_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamLadspaFilter *ladspa = GDAM_LADSPA_FILTER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamLadspaPluginType *plugin_type = ladspa->plugin_type;
  GdamRequest *request;
  gint *index_array;
  int i, local_id;
  int num_controls;
  GdamArg args[2];
  gtk_object_ref (GTK_OBJECT (ladspa));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasLadspaPlugin",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					ladspa, 0);
  local_id = gdam_request_builder_append_create (builder, request);

  args[0].name = "filename";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + 0) = plugin_type->module_name;
  args[1].name = "index";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 1) = plugin_type->module_index;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					local_id,
					"set_plugin",
					2, args, NULL, NULL, NULL);
  gdam_request_builder_append_generic (builder, request);

  num_controls = plugin_type->num_controls;
  index_array = alloca (sizeof (int) * num_controls);
  for (i = 0; i < num_controls; i++)
    index_array[i] = i;
  args[0].name = "control_indices";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 0).length = num_controls;
  GDAM_ARG_INT_ARRAY (args + 0).array = index_array;
  args[1].name = "controls";
  args[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).length = num_controls * 2;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).array = ladspa->controls;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					local_id,
					"set_controls",
					2, args, NULL, NULL, NULL);
  gdam_request_builder_append_generic (builder, request);

  args[0].name = "filtered";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 0) = new_source_id;
  request = gdam_request_set_args (local_id, 1, args, NULL, NULL, NULL);
  gdam_request_builder_append_generic (builder, request);

  return local_id;
}

static void gdam_ladspa_filter_reset(GtkObject*        object,
                                     GdamArg*          arg)
{
  GdamRequest *request;
  GdamModel *model = GDAM_MODEL (object);
  int source_id = model->source_id;
  request = gdam_request_set_args (source_id, 1, arg, NULL, NULL, NULL);
  gdam_channel_append_request (model->channel, request);
}

static void gdam_ladspa_filter_init(GdamLadspaFilter* ladspa_filter)
{
}
static void gdam_ladspa_filter_class_init(GdamFilterClass* filter_class)
{
  filter_class->append_start_request
    = gdam_ladspa_filter_append_start_request;
}

GtkType gdam_ladspa_filter_get_type()
{
  static GtkType ladspa_filter_type = 0;
  if (!ladspa_filter_type)
    {
      static const GtkTypeInfo ladspa_filter_info = {
	"GdamLadspaFilter",
	sizeof (GdamLadspaFilter),
	sizeof (GdamLadspaFilterClass),
	(GtkClassInitFunc) gdam_ladspa_filter_class_init,
	(GtkObjectInitFunc) gdam_ladspa_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      initialize_ladspa_database ();
      ladspa_filter_type = gtk_type_unique (parent, &ladspa_filter_info);
    }
  return ladspa_filter_type;
}

static void
gdam_ladspa_filter_set_plugin_type(GdamLadspaFilter*     filter,
                                   GdamLadspaPluginType* plugin_type)
{
  int num_controls = plugin_type->num_controls;
  int num_meters = plugin_type->num_meters;

  filter->controls = g_new0 (gdouble, 2 * num_controls);
  filter->meters = g_new0 (gdouble, 2 * num_meters);
  filter->which_meters_to_query = g_new0 (gboolean, num_meters);
  filter->plugin_type = plugin_type;
}

GdamLadspaFilter* gdam_ladspa_filter_new(const char* name)
{
  GdamLadspaPluginType *plugin_type;
  GdamLadspaFilter *filter;
  GtkType type;

  type = GDAM_TYPE_LADSPA_FILTER;

  plugin_type = g_hash_table_lookup (plugin_hashtable, name);
  if (plugin_type == NULL)
    {
      g_warning ("Sorry, no plugin in the database named %s", name);
      return NULL;
    }

  filter = GDAM_LADSPA_FILTER (gtk_type_new (type));
  gdam_ladspa_filter_set_plugin_type (filter, plugin_type);
  return filter;
}
void gdam_ladspa_filter_set_control     (GdamLadspaFilter* filter,
                                         int               control_index,
                                         gdouble           left_value,
                                         gdouble           right_value)
{
  GdamModel *model;
  GdamArg args[2];
  gdouble lr_values[2];
  GdamRequest *request;

  model = GDAM_MODEL (filter);
  filter->controls[control_index * 2 + 0] = left_value;
  filter->controls[control_index * 2 + 1] = right_value;

  if (model->source_id == 0)
    return;

  args[0].name = "control_indices";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 0).length = 1;
  GDAM_ARG_INT_ARRAY (args + 0).array = &control_index;
  lr_values[0] = left_value;
  lr_values[1] = right_value;
  args[1].name = "controls";
  args[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).length = 2;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).array = lr_values;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					model->source_id,
					"set_controls",
					2, args, NULL, NULL, NULL);
  gdam_channel_append_request (model->channel, request);
}
