/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdammididevice.h"
#include <common/gdamplugin.h>
#include <common/gdamsearchpath.h>
#include <common/gdamplugin.h>
#include "../../common/gdamdebug.h"
#include <math.h>
#include <gtk/gtksignal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

/* Maximum number of errors we may encounter in a midi file
 * before stopping processing. */
#define MAX_ERRORS	20

static const char* gdam_midi_note_names[] =
{ "c", "c#", "d", "d#", "e", "f", "f#", "g", "g#", "a", "a#", "b" };

/* A single connection to a trap. */
typedef struct _GdamMidiHook GdamMidiHook;

static gboolean gdam_midi_device_has_patchfile(GdamMidiDevice* device, 
                                               const char* patchfilename);
static void gdam_midi_device_add_pitchwheel_name(GdamMidiDevice* device,
                                                 int             channel,
						 const char*     name);
static void gdam_midi_device_add_parameter_name (GdamMidiDevice* device,
                                                 int             channel,
						 int             param,
						 const char*     name);
static void gdam_midi_device_add_sysex_name     (GdamMidiDevice* device,
						 int             make_id,
						 int             model_id,
                                                 int             length,
						 const guint8*   data,
						 const char*     name);
static gboolean process_midi_device_input(GIOChannel*   io_channel,
                                          GIOCondition  condition,
					  gpointer      user_data);


static void get_model_data(int       make_id,
		           int       model_id,
			   char**    make_name,
			   char**    model_name);

GdamInputEvent* gdam_midi_event_pitch_new (GdamMidiDevice*    device,
                                           int		      channel,
				           int		      pitch)
{
  GdamMidiEventPitch *mpitch = g_new (GdamMidiEventPitch, 1);
  GdamInputEventRange *range_event = (GdamInputEventRange *) mpitch;
  range_event->input_device = GDAM_INPUT_DEVICE (device);
  range_event->event_type = gdam_midi_event_pitch_wheel_get_type ();
  range_event->input_id = channel;
  range_event->min_value = 0;
  range_event->max_value = (1 << 14);
  range_event->value = pitch;
  return (GdamInputEvent *) mpitch;
}

GdamInputEvent* gdam_midi_event_param_new  (GdamMidiDevice*    device,
                                           int		      channel,
					   int		      parameter,
					   int		      value)
{
  GdamMidiEventParam *param = g_new (GdamMidiEventParam, 1);
  GdamInputEventRange *range_event = (GdamInputEventRange *) param;
  range_event->input_device = GDAM_INPUT_DEVICE (device);
  range_event->event_type = gdam_midi_event_parameter_get_type ();
  range_event->input_id = channel;
  range_event->min_value = 0;
  range_event->max_value = (1 << 7);
  range_event->value = value;
  param->parameter = parameter;
  return (GdamInputEvent *) param;
}

GdamInputEvent* gdam_midi_event_note_new(GdamMidiDevice*    device,
                                        int                instrument,
                                        int		   midi_id,
                                        int                velocity,
					gboolean           on)
{
  GdamMidiEventNote *note = g_new (GdamMidiEventNote, 1);
  GdamInputEvent *input_event = (GdamInputEvent *) note;
  input_event->input_device = GDAM_INPUT_DEVICE (device);
  if (on)
    input_event->event_type = gdam_midi_event_note_on_get_type ();
  else
    input_event->event_type = gdam_midi_event_note_off_get_type ();
  input_event->input_id = instrument;
  note->midi_index = midi_id;
  note->note_name = gdam_midi_note_names[midi_id % 12];
  note->octave = (midi_id - 12) / 12;
  note->hertz = 440.0 * pow (2.0, (gdouble) (midi_id - 57.0) / 12.0);
  note->velocity = velocity;
  return (GdamInputEvent *) note;
}

GdamInputEvent* gdam_midi_event_sysex_new(GdamMidiDevice*  device,
                                          const char*      maker,
					  const char*      model,
					  int              length,
					  const guint8*    bytes)
{
  GdamMidiEventSystemExclusive *sysex;
  GdamInputEvent *input_event;

  sysex = g_new (GdamMidiEventSystemExclusive, 1);
  input_event = (GdamInputEvent *) sysex;
  input_event->input_device = GDAM_INPUT_DEVICE (device);
  input_event->event_type = gdam_midi_event_sysex_get_type ();
  input_event->input_id = 0;
  sysex->message = g_memdup (bytes, length);
  sysex->message_length = length;
  return input_event;
}

static void gdam_midi_device_init(GdamMidiDevice* midi_device)
{
}
static void gdam_midi_device_class_init(GdamMidiDeviceClass* midi_device_class)
{
  /*GtkObjectClass* object_class = GTK_OBJECT_CLASS(midi_device_class); */
  GdamInputDeviceFileClass *device_file_class;
  device_file_class = GDAM_INPUT_DEVICE_FILE_CLASS (midi_device_class);

  device_file_class->parse_raw = process_midi_device_input;
}

GtkType gdam_midi_device_get_type()
{
  static GtkType midi_device_type = 0;
  if (!midi_device_type)
    {
      static const GtkTypeInfo midi_device_info = {
	"GdamMidiDevice",
	sizeof (GdamMidiDevice),
	sizeof (GdamMidiDeviceClass),
	(GtkClassInitFunc) gdam_midi_device_class_init,
	(GtkObjectInitFunc) gdam_midi_device_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_INPUT_DEVICE_FILE;
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
      midi_device_type = gtk_type_unique (parent, &midi_device_info);
      gdam_midi_event_note_on_get_type ();
      gdam_midi_event_note_off_get_type ();
      gdam_midi_event_pitch_wheel_get_type ();
      gdam_midi_event_parameter_get_type ();
      gdam_midi_event_sysex_get_type ();
    }
  return midi_device_type;
}

GdamMidiDevice* gdam_midi_device_new ()
{
  return GDAM_MIDI_DEVICE (gtk_type_new (GDAM_TYPE_MIDI_DEVICE));
}

/*
 *  ____                            ____  _             _
 * |  _ \ __ _ _ __ ___  ___ _ __  |  _ \| |_   _  __ _(_)_ __  ___
 * | |_) / _` | '__/ __|/ _ \ '__| | |_) | | | | |/ _` | | '_ \/ __|
 * |  __/ (_| | |  \__ \  __/ |    |  __/| | |_| | (_| | | | | \__ \
 * |_|   \__,_|_|  |___/\___|_|    |_|   |_|\__,_|\__, |_|_| |_|___/
 *                                                |___/
 */

static void gdam_midi_device_emit_program      (GdamMidiDevice* device,
                                                int             channel,
				                int             program)
{
  gdam_debug ("MIDI: Program: channel %d set to program %d [ignored]",
	      channel, program);
}

static void gdam_midi_device_emit_channel_pressure(GdamMidiDevice* device,
                                                   int             channel,
					           int             pressure)
{
	gdam_debug("MIDI: Channel Pressure: "
	          "channel %d set to program %d [ignored]",
				channel, pressure);
}

static void gdam_midi_device_emit_note_off      (GdamMidiDevice* device,
					         int             instrument,
                                                 int             pitch,
					         int             volume)
{
	GdamInputEvent* event;
	GdamMidiEventNote* note;
	event = gdam_midi_event_note_new(device,
	                                 instrument,
					 pitch,
					 volume,
					 FALSE);
	note = (GdamMidiEventNote*) event;
	gdam_debug("note_off: instrument %d, %s octave %d, volume %d",
		instrument,
		note->note_name,
		note->octave,
		note->velocity);
	gdam_input_trap_emit(NULL, event);
	gdam_input_event_destroy(event);
}
static void gdam_midi_device_emit_note_on(GdamMidiDevice* device,
					  int             instrument,
                                          int             pitch,
					  int             volume)
{
	GdamInputEvent* event;
	GdamMidiEventNote* note;
	if (volume == 0) {
		gdam_midi_device_emit_note_off(device,
		                               instrument, 
		                               pitch,
					       volume);
		return;
	}
	event = gdam_midi_event_note_new(device,
	                                 instrument,
					 pitch,
					 volume,
					 TRUE);
	note = (GdamMidiEventNote*) event;
	gdam_debug("note_on: instrument %d, %s octave %d, volume %d",
		instrument,
		note->note_name,
		note->octave,
		note->velocity);
	gdam_input_trap_emit(NULL, event);
	gdam_input_event_destroy(event);
}
static void gdam_midi_device_emit_key_pressure(GdamMidiDevice* device,
					       int             instrument,
                                               int             pitch,
					       int             volume)
{
#if 0
	GdamMidiNote* note;
	note = gdam_midi_note_new(instrument, pitch, volume);
	gtk_signal_emit(GTK_OBJECT(device),
	                midi_signals[KEY_PRESSURE],
			note);
	gdam_midi_note_destroy(note);
#endif
	g_warning("MIDI: Key Pressure: ignoring this, but really shouldn't");
}
	
static void gdam_midi_device_emit_parameter(GdamMidiDevice* device,
                                            int             channel,
					    int             which_param,
					    int             value)
{
	GdamInputEvent* event;
	event = gdam_midi_event_param_new(device, channel, which_param, value);
	gdam_input_trap_emit(NULL, event);
	gdam_input_event_destroy(event);

	gdam_debug("MIDI: Parameter: channel %d, param %d set to %d",
			channel, which_param, value);
}

static void gdam_midi_device_emit_pitch_wheel(GdamMidiDevice* device,
                                              int             channel,
				              int             pitch)
{
	GdamInputEvent* event;

	event = gdam_midi_event_pitch_new(device, channel, pitch);
	gdam_input_trap_emit(NULL, event);
	gdam_input_event_destroy(event);

	gdam_debug("MIDI: PitchWheel: channel %d, set to %d", channel, pitch);

}
static void gdam_midi_device_emit_sysex        (GdamMidiDevice* device,
				                const char*     maker,
				                const char*     model,
                                                int             length,
				                const guint8*   bytes)
{
	GdamInputEvent* event;

	event = gdam_midi_event_sysex_new(device, maker, model, length, bytes);
	gdam_input_trap_emit(NULL, event);
	gdam_input_event_destroy(event);

	gdam_debug("MIDI: SysEx: %s, %s, %d bytes", maker, model, length);
}

static void add_to_triplet(GdamMidiDevice* device, int c)
{
	if (device->triplet_length >= 3) {
		g_warning("midi protocol error -- command too long");
		return;
	}
	if (device->triplet_length == 0
	 && device->triplet[0] != '\0'
	 && c < 0x80) {
	 	device->triplet_length++;
	}

	device->triplet[device->triplet_length++] = (unsigned char) c;
}

static void maybe_process_triplet(GdamMidiDevice* device)
{
	int hi, lo, arg1, arg2;
	if (device->triplet_length == 0)
		return;
	if (device->triplet_length == 1) {
	 	if (device->triplet[0] < 0x80) {
			g_warning("protocol error: "
			          "command didn't begin with >0x80");
	 		device->triplet_length = 0;
		}
		else if (device->triplet[0] >= 0xF0 
					 && device->triplet[0] < 0xFF) {
			g_warning("not parsing: %d", device->triplet[0]);
	 		device->triplet_length = 0;
		}
		return;
	}

	hi = (device->triplet[0] / 16);
	lo = (device->triplet[0] % 16);

	arg1 = device->triplet[1];
	if (device->triplet_length == 2) {
		switch (hi) {
			case 12:
				gdam_midi_device_emit_program(device, lo, arg1);
				device->triplet_length = 0;
				break;
			case 13:
				gdam_midi_device_emit_channel_pressure
							(device, lo, arg1);
				device->triplet_length = 0;
				break;
			default:
				;
		}
		return;
	}

	arg2 = device->triplet[2];
	switch (hi) {
		case 8:
			gdam_midi_device_emit_note_off(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 9:
			gdam_midi_device_emit_note_on(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 10:
			gdam_midi_device_emit_key_pressure(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 11:
			gdam_midi_device_emit_parameter(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 14: {
			int pitch;
			pitch = (arg1 & 0x7f)
			      | ((arg2 & 0x7f) << 7);
			gdam_midi_device_emit_pitch_wheel(device, lo, pitch);
			device->triplet_length = 0;
			break;
		}
	}
	if (device->triplet_length > 0) {
		g_warning("error parsing midi commands");
		device->triplet_length = 0;
		return;
	}
	return;
}
static void append_to_system_exclusive(GdamMidiDevice*  device,
		                       int              c)
{
	if (device->system_exclusive == NULL) {
		char str[2];
		str[0] = c;
		str[1] = '\0';
		device->system_exclusive = g_string_new(str);
		return;
	}
	device->system_exclusive
			= g_string_append_c( device->system_exclusive,
					     (gchar) c);
}

static void process_system_exclusive  (GdamMidiDevice*  device)
{
	guint8* str;
	char* make_name, *model_name;
	int left;
	g_return_if_fail(device->system_exclusive != NULL);
	str = (guint8*) device->system_exclusive->str;
	g_return_if_fail(str != NULL);
	left = device->system_exclusive->len;

	if (left < 6) {
		g_message("system_exclusive too short, %d < 6", left);
		g_string_free(device->system_exclusive, TRUE);
		device->system_exclusive = NULL;
		return;
	}

	get_model_data(str[1], 
			  ((int)(unsigned char)str[3] << 8) 
			+ ((int)(unsigned char)str[4]),
			&make_name,
			&model_name);

	gdam_midi_device_emit_sysex(device, 
				    make_name,
				    model_name,
				    left - 6, 	/* 5 + checksum */
				    str + 5);
	g_string_free(device->system_exclusive, TRUE);
	device->system_exclusive = NULL;
}

static gboolean process_midi_device_input(GIOChannel*   io_channel,
                                          GIOCondition  condition,
					  gpointer      user_data)
{
	GdamMidiDevice* device = GDAM_MIDI_DEVICE(user_data);
	unsigned char midi_buf[1024];
	guint num_read;
	guint index;
	GIOError io_error;
	const char* filename = GDAM_INPUT_DEVICE(device)->device_name;
	io_error = g_io_channel_read(io_channel,
	                             (char*) midi_buf,
			             sizeof(midi_buf),
			             &num_read);
	if (io_error == G_IO_ERROR_AGAIN)
		return TRUE;
	if (io_error != G_IO_ERROR_NONE) {
		g_warning("error reading from midi device %s", filename);
		return FALSE;
	}
	if (num_read <= 0) {
		g_warning("MIDI device closed...");
		return FALSE;
	}

	for (index = 0; index < num_read; index++) {
		int c = midi_buf[index];
		/* This character is used by the midi device 
		 * to keep time. */
		/* TODO: obey these signals... */
		if (c >= 0xf8)
			continue;


		if (c == 0xf0 || device->system_exclusive != NULL) {
			int old_c = c;

			/* When we get 0xf0 in the middle,
			 * we want to transmit two sysex messages,
			 * as though they had not been joined:
			 *
			 *       0xf0  a b c  0xf0  d e f  0xf7
			 *
			 * should become 2 sysex events:
			 *       0xf0  a b c  0xf7
			 *       0xf0  d e f  0xf7
			 */
			if (c == 0xf0 && device->system_exclusive != NULL)
				c = 0xf7;
			append_to_system_exclusive(device, c);
			if (c == 0xf7) {
				process_system_exclusive(device);
			}
			if (old_c == 0xf0)
				device->system_exclusive = g_string_new("\360");
			continue;
		}

		add_to_triplet(device, c);
		maybe_process_triplet(device);
	}
	return TRUE;
}

static gboolean is_in_list_of_strings(GSList* list, const char* fname)
{
	for (   ; list != NULL; list = list->next)
		if (strcmp(fname, (char*)(list->data)) == 0)
			return TRUE;
	return FALSE;
}

static gboolean gdam_midi_device_has_patchfile(GdamMidiDevice* device, 
                                               const char* patchfilename)
{
	return is_in_list_of_strings(device->patch_list, patchfilename);
}

typedef struct _PendingPatches PendingPatches;
struct _PendingPatches {
	char* midi_device_name;
	GSList* patch_filenames;
};

#if 0
static GSList* pending_patches = NULL;
/*
 *                     add_pending_patch
 *
 * Add a patch file that should be parsed when the device is
 * opened.  We won't store duplicates.
 */
static void add_pending_patch(const char* device_name,
                              const char* mapfile)
{
	GSList* patches;
	PendingPatches* pending;
	for (patches = pending_patches;
	     patches != NULL;
	     patches = patches->next) {
	     	PendingPatches* p = patches->data;
		if (strcmp(device_name, p->midi_device_name) == 0) {
			if (!is_in_list_of_strings(p->patch_filenames, mapfile))
				p->patch_filenames = g_slist_prepend
					(p->patch_filenames, g_strdup(mapfile));
			return;
		}
	}

	pending = g_new0(PendingPatches, 1);
	pending->midi_device_name = g_strdup(device_name);
	pending->patch_filenames = g_slist_prepend(pending->patch_filenames,
	                                           g_strdup(mapfile));
	pending_patches = g_slist_prepend(pending_patches, pending);
}

static void load_map_file(const char* fname, GdamMidiDevice* device)
{
	gdam_midi_device_load_map(device, fname);
}

static void add_pending_patches(GdamMidiDevice* device)
{
	GSList* pp = pending_patches;
	const char* filename = GDAM_INPUT_DEVICE(device)->device_name;
	while (pp != NULL) {
		PendingPatches* pending = pp->data;
		if (strcmp(pending->midi_device_name, filename) == 0) {
			g_slist_foreach(pending->patch_filenames,
			                (GFunc) load_map_file,
					device);
			pending_patches = g_slist_remove(pending_patches,
			                                 pending);
			return;
		}
		pp = pp->next;
	}
}
#endif

struct _ModelInfo {
	int model_id;
	char* model_name;
};

typedef struct _MakeInfo MakeInfo;
typedef struct _ModelInfo ModelInfo;
struct _MakeInfo {
	int make_id;
	char* make_name;
	GHashTable* models_by_id;
};

GHashTable* all_makes = NULL;

void gdam_midi_device_record_make_name(int id, const char* name)
{
	MakeInfo* info;
	if (all_makes == NULL)
		all_makes = g_hash_table_new(NULL, NULL);
	info = g_hash_table_lookup(all_makes, GINT_TO_POINTER(id));
	if (info != NULL)
		return;
	info = g_new(MakeInfo, 1);
	info->make_id = id;
	info->make_name = g_strdup(name);
	info->models_by_id = NULL;
	g_hash_table_insert(all_makes, GINT_TO_POINTER(id), info);
}
void gdam_midi_device_record_model_name(int make_id, int model_id,
                                        const char* name)
{
	MakeInfo* make_info;
	ModelInfo* model_info;
	g_return_if_fail(all_makes != NULL);
	make_info = g_hash_table_lookup(all_makes, GINT_TO_POINTER(make_id));
	if (make_info == NULL) {
		g_warning("no make of type %d", make_id);
		return;
	}
	model_info = g_new(ModelInfo, 1);
	model_info->model_id = model_id;
	model_info->model_name = g_strdup(name);
	if (make_info->models_by_id == NULL)
		make_info->models_by_id = g_hash_table_new(NULL, NULL);
	g_hash_table_insert(make_info->models_by_id,
	                    GINT_TO_POINTER(model_id),
			    model_info);
	
}

static void get_model_data(int make_id,
		           int model_id,
			   char** make_name,
			   char** model_name)
{
	MakeInfo* make;
	ModelInfo* model;

	if (all_makes == NULL)
		goto unknown;

	make = g_hash_table_lookup(all_makes, GINT_TO_POINTER(make_id));
	if (make == NULL)
		goto unknown;

	if (make->models_by_id != NULL)
		model = g_hash_table_lookup(make->models_by_id, 
		                            GINT_TO_POINTER(model_id));
	else
		model = NULL;
	*make_name = make->make_name;
	*model_name = model ? model->model_name : NULL;
	return;
unknown:

	*make_name = "unknown make";
	*model_name = "unknown model";
	return;
}


static void gdam_midi_device_add_pitchwheel_name(GdamMidiDevice* device,
                                                 int             channel,
						 const char*     name)
{
	GdamInputTrap* trap;
	GdamInputEvent* event;
	event = gdam_midi_event_pitch_new(device, channel, 0);
	trap = gdam_input_trap_find_or_create(event);
	gdam_input_trap_set_name(trap, name);
	gdam_input_event_destroy(event);
}

static void gdam_midi_device_add_parameter_name (GdamMidiDevice* device,
                                                 int             channel,
						 int             param,
						 const char*     name)
{
	GdamInputTrap* trap;
	GdamInputEvent* event;
	event = gdam_midi_event_param_new(device, channel, param, 0);
	trap = gdam_input_trap_find_or_create(event);
	gdam_input_trap_set_name(trap, name);
	gdam_input_event_destroy(event);
}

static void gdam_midi_device_add_sysex_name     (GdamMidiDevice* device,
						 int             make_id,
						 int             model_id,
                                                 int             length,
						 const guint8*   data,
						 const char*     name)
{
	GdamInputTrap* trap;
	GdamInputEvent* event;
	const char* make_name;
	const char* model_name;
	
	get_model_data(make_id, model_id,
	               (char**) (&make_name), (char**) (&model_name));
	event = gdam_midi_event_sysex_new(device,
	                                  make_name, model_name,
					  length, data);
	trap = gdam_input_trap_find_or_create(event);
	gdam_input_trap_set_name(trap, name);
	gdam_input_event_destroy(event);
}

static const char* parse_pitch_wheel(GdamMidiDevice* device,
                                     const char* args)
{
	int channel;
	const char* endp;
	const char* description;

	/* Casts are for const portability... */
	channel = strtol( (char*) args, (char**) &endp, 10 );

	if (endp == args)
		return "error parsing pitchwheel channel";
	
	/* Find the description... */
	description = endp;
	while (isspace(*description) && *description != '\0')
		description++;
	
	if (*description == '\0')
		return "pitchwheel had no description";

	gdam_midi_device_add_pitchwheel_name(device, channel, description);
	return NULL;
}

static const char* parse_parameter(GdamMidiDevice* device, const char* args)
{
	int channel, param;
	const char* endp;
	const char* description;
	const char* start_param;
	channel = (int) strtol( (char*) args, (char**) &endp, 10 );
	if (endp == args)
		return "error parsing parameter channel";

	while (isspace(*endp) && *endp != '\0')
		endp++;

	start_param = endp;
	param = (int) strtol( (char*) start_param, (char**) &endp, 10 );
	if (endp == args)
		return "error parsing parameter id";
	
	description = endp;
	while (isspace(*description) && *description != '\0')
		description++;
	gdam_midi_device_add_parameter_name(device,
	                                    channel,
					    param,
					    description);
	return NULL;
}

static const char* parse_sysex(GdamMidiDevice* device, const char* args)
{
	int length, i;
	const char* endp;
	const char* at;
	const char* description;
	guint8* data;
	length = (int) strtol( (char*) args, (char**) &endp, 10 );
	if (endp == args)
		return "error parsing parameter channel";

	while (isspace(*endp) && *endp != '\0')
		endp++;
	at = endp;

	data = alloca(length);

	for (i = 0; i < length; i++) {
		int num;
		num = (int) strtol( (char*) at, (char**) &endp, 16 );
		if (at == endp)
			return "error parsing hex for sysex";
		data[i] = (guint8) num;
		at = endp;
		while (isspace(*at) && *at != '\0')
			at++;
	}
	description = at;
	gdam_midi_device_add_sysex_name(device,
					device->make_id,
					device->model_id,
	                                length,
					data,
					description);
	return NULL;
}

static const char* parse_make(GdamMidiDevice* device, const char* args)
{
	const char* endp;
	int make_id = strtol((char*) args, (char**) &endp, 0);
	if (endp == args) 
		return "error parsing make id";
	while (*endp && isspace(*endp))
		endp++;
	if (*endp != '\0')
		gdam_midi_device_record_make_name(make_id, endp);
	device->make_id = make_id;
	return NULL;
}

static const char* parse_model(GdamMidiDevice* device, const char* args)
{
	const char* endp = NULL;
	int model_id = strtol((char*) args, (char**) &endp, 0);
	if (endp == args) 
		return "error parsing model id";
	while (*endp && isspace(*endp))
		endp++;
	if (*endp != '\0')
		gdam_midi_device_record_model_name(device->make_id,
		                                   model_id,
						   endp);
	device->model_id = model_id;
	return NULL;
}
/*
 * Parse a single midi device patch line, in one of the following
 * formats:
 *
 *     make number name
 *     model number name
 *     pitchwheel 0 name
 *     parameter 0 0 name
 *     sysex length .. ... .. name
 *
 * This function returns an error message, or NULL on success. 
 */
static const char* parse_trap_line(GdamMidiDevice* device, const char* buf)
{
	int cmd_len;
	const char* end_word;
	const char* start_next;
	char* last_cmd;
	static char msg[256];
	while (*buf && isspace(*buf))
		buf++;
	if (*buf == '\0')
		return NULL;
	if (*buf == '#')
		return NULL;
	end_word = buf;
	while (*end_word && ! isspace(*end_word))
		end_word++;
	start_next = end_word;
	while (*start_next && isspace(*start_next))
		start_next++;
	if (*start_next == '\0')
		return "all commands require some parameters";
	cmd_len = (end_word - buf);
	last_cmd = alloca(cmd_len + 1);
	memcpy(last_cmd, buf, cmd_len);
	last_cmd[cmd_len] = '\0';
	if (g_strcasecmp(last_cmd, "pitchwheel") == 0)
		return parse_pitch_wheel(device, start_next);
	if (g_strcasecmp(last_cmd, "parameter") == 0)
		return parse_parameter(device, start_next);
	if (g_strcasecmp(last_cmd, "sysex") == 0)
		return parse_sysex(device, start_next);
	if (g_strcasecmp(last_cmd, "make") == 0)
		return parse_make(device, start_next);
	if (g_strcasecmp(last_cmd, "model") == 0)
		return parse_model(device, start_next);

	/* HACK : just so we get a little better error handling,
	          at the expense of totally nonthreadsafe code. */
	if (cmd_len > (int)sizeof(last_cmd) - 1)
		cmd_len = sizeof(last_cmd) - 1;
	memcpy(last_cmd, buf, cmd_len);
	last_cmd[cmd_len] = 0;
	g_snprintf(msg, sizeof(msg),
	           "no command %s found", last_cmd);
	return msg;
}

/* XXX: use gscanner instead. */
gboolean gdam_midi_device_load_map(GdamMidiDevice*    midi_dev,
                                   const char*        db_filename)
{
	FILE* fp;
	char buf[4096];
	int num_errors = 0;
	int lineno = 0;

	if (!gdam_midi_device_has_patchfile(midi_dev, db_filename))
		midi_dev->patch_list = g_slist_prepend(midi_dev->patch_list,
		                                     g_strdup(db_filename));

	gdam_debug("trying to load patch file %s", db_filename);
	fp = fopen(db_filename, "r");
	if (fp == NULL) {
		g_warning("couldn't open %s: %s", db_filename, strerror(errno));
		return FALSE;
	}

	while (fgets(buf, sizeof(buf), fp) != 0) {
		const char* error_message;
		char* nl;
		++lineno;
		nl = strchr(buf, '\n');
		if (nl == NULL) {
			g_warning("line too long or file truncated");
			fclose(fp);
			return FALSE;
		}

		*nl = '\0';	/* chomp */

		error_message = parse_trap_line(midi_dev, buf);
		if (error_message != NULL) {
			g_warning("error parsing trap file %s, line %d, "
				  "`%s': %s, continuing anyway", db_filename,
				  lineno, buf, error_message);
			if (++num_errors > MAX_ERRORS) {
				fclose(fp);
				return FALSE;
			}
		}
	}
	fclose(fp);
	gdam_debug("successfully parsed midi patch database %s", db_filename);
	return TRUE;
}

#if 0
static void assoc_patch_with_device(const char*   device_name,
                                    const char*   mapfile)
{
	GdamMidiDevice* device = NULL;
	if (midi_devices != NULL)
		device = g_hash_table_lookup(midi_devices, device_name);
	if (device != NULL) {
		if (!gdam_midi_device_has_patchfile(device, mapfile))
			gdam_midi_device_load_map(device, mapfile);
	} else {
		/* No device of that name yet... */
		add_pending_patch(device_name, mapfile);
	}
}
#endif

/*
 *  __  __ _     _ _   _____                 _
 * |  \/  (_) __| (_) | ____|_   _____ _ __ | |_ ___
 * | |\/| | |/ _` | | |  _| \ \ / / _ \ '_ \| __/ __|
 * | |  | | | (_| | | | |___ \ V /  __/ | | | |_\__ \
 * |_|  |_|_|\__,_|_| |_____| \_/ \___|_| |_|\__|___/
 * 
 */

/* XXX: note_on and note_off need to be better thought out... */
static char* gdam_midi_event_note_on_stringify(GdamInputEvent* event)
{
	GdamMidiEventNote* note = (GdamMidiEventNote*) event;
	return g_strdup_printf("note on, channel %d, %s octave %d, idx=%d",
				note->input_event.input_id,
				note->note_name,
				note->octave,
				note->midi_index);
}
/* CHECK this function.  */
static unsigned gdam_midi_event_note_on_hash(GdamInputEvent* event)
{
	GdamMidiEventNote* note;
	note = (GdamMidiEventNote*) event;
	return GPOINTER_TO_INT(event->input_device)
	     ^ (event->event_type << 6)
	     ^ (event->input_id << 12)
	     ^ (note->midi_index << 18);
}

static gboolean gdam_midi_event_note_on_equal(GdamInputEvent* event1,
                                               GdamInputEvent* event2)
{
	GdamMidiEventNote* note1;
	GdamMidiEventNote* note2;
	gboolean equal;
	note1 = (GdamMidiEventNote*) event1;
	note2 = (GdamMidiEventNote*) event2;
	equal = (event1->input_device == event2->input_device)
	    && (event1->event_type == event2->event_type)
	    && (event1->input_id == event2->input_id)
	    && (note1->midi_index == note2->midi_index);
	return equal;
}

static GHashTable* gdam_midi_event_note_on_to_keyvalues(GdamInputEvent* event)
{
	GHashTable* table;
	const char* device_name;
	GdamMidiEventNote* note = (GdamMidiEventNote*) event;
	table = g_hash_table_new(g_str_hash, g_str_equal);
	gdam_input_store_int(table, "input_id", event->input_id);
	gdam_input_store_int(table, "midi_index", note->midi_index);

	device_name = event->input_device->device_name;
	gdam_input_store_string(table, "device", device_name);
	return table;
}

static GdamInputEvent* gdam_midi_event_note_on_from_keyvalues(int type,
                                                            GHashTable* table)
{
	GdamInputEvent* event;
	GdamMidiEventNote* note;
	char* device_name;
	int midi_index;
	note = g_new(GdamMidiEventNote, 1);
	event = (GdamInputEvent*) note;
	event->event_type = type;
	event->input_id = gdam_input_fetch_int(table, "input_id");
	midi_index = gdam_input_fetch_int(table, "midi_index");
	note->midi_index = midi_index;
	note->note_name = gdam_midi_note_names[midi_index % 12];
	note->octave = (midi_index - 12) / 12;
	note->hertz = 440.0 * pow(2.0, (gdouble)(midi_index - 57.0) / 12.0);
	note->velocity = 0;

	device_name = g_hash_table_lookup(table, "device");
	gdam_input_event_parse_device(event, device_name);
	return event;
}


static GdamInputEvent*
gdam_midi_event_note_on_canonicalize(GdamInputEvent* event)
{
	GdamMidiEventNote* old_note = (GdamMidiEventNote*) event;
	return gdam_midi_event_note_new(GDAM_MIDI_DEVICE(event->input_device),
	                                event->input_id,
					old_note->midi_index,
					old_note->velocity,
					TRUE);
}

guint gdam_midi_event_note_on_get_type()
{
	static GdamInputEventTypeInfo type_info = {
		"GdamMidiEventNoteOn",
		gdam_midi_event_note_on_stringify,
		gdam_midi_event_note_on_hash,
		gdam_midi_event_note_on_equal,
		gdam_midi_event_note_on_to_keyvalues,
		gdam_midi_event_note_on_from_keyvalues,
		//NULL,
		//NULL,
		//NULL,
		//NULL,
		gdam_midi_event_note_on_canonicalize,
		NULL
	};
	static guint type = 0;
	if (type == 0)
		type = gdam_input_event_get_unique
				(gdam_input_event_int_get_type(),
				 GDAM_TYPE_MIDI_DEVICE,
				 &type_info);
	return type;
}
static char* gdam_midi_event_note_off_stringify(GdamInputEvent* event)
{
	GdamMidiEventNote* note = (GdamMidiEventNote*) event;
	return g_strdup_printf("note off, channel %d, %s octave %d, idx=%d",
				note->input_event.input_id,
				note->note_name,
				note->octave,
				note->midi_index);
}
/* CHECK this function.  */
static unsigned gdam_midi_event_note_off_hash(GdamInputEvent* event)
{
	GdamMidiEventNote* note;
	note = (GdamMidiEventNote*) event;
	return GPOINTER_TO_INT(event->input_device)
	     ^ (event->event_type << 6)
	     ^ (event->input_id << 12)
	     ^ (note->midi_index << 18);
}

static gboolean gdam_midi_event_note_off_equal(GdamInputEvent* event1,
                                               GdamInputEvent* event2)
{
	GdamMidiEventNote* note1;
	GdamMidiEventNote* note2;
	gboolean equal;
	note1 = (GdamMidiEventNote*) event1;
	note2 = (GdamMidiEventNote*) event2;

	equal = (event1->input_device == event2->input_device)
	    && (event1->event_type == event2->event_type)
	    && (event1->input_id == event2->input_id)
	    && (note1->midi_index == note2->midi_index);

	return equal;
}

static GHashTable* gdam_midi_event_note_off_to_keyvalues(GdamInputEvent* event)
{
	GHashTable* table;
	const char* device_name;
	GdamMidiEventNote* note = (GdamMidiEventNote*) event;
	table = g_hash_table_new(g_str_hash, g_str_equal);
	gdam_input_store_int(table, "input_id", event->input_id);
	gdam_input_store_int(table, "midi_index", note->midi_index);

	device_name = event->input_device->device_name;
	gdam_input_store_string(table, "device", device_name);
	return table;
}

static GdamInputEvent* gdam_midi_event_note_off_from_keyvalues(int type,
                                                            GHashTable* table)
{
	GdamInputEvent* event;
	GdamMidiEventNote* note;
	char* device_name;
	int midi_index;
	note = g_new(GdamMidiEventNote, 1);
	event = (GdamInputEvent*) note;
	event->event_type = type;
	event->input_id = gdam_input_fetch_int(table, "input_id");
	midi_index = gdam_input_fetch_int(table, "midi_index");
	note->midi_index = midi_index;
	note->note_name = gdam_midi_note_names[midi_index % 12];
	note->octave = (midi_index - 12) / 12;
	note->hertz = 440.0 * pow(2.0, (gdouble)(midi_index - 57.0) / 12.0);
	note->velocity = 0;

	device_name = g_hash_table_lookup(table, "device");
	gdam_input_event_parse_device(event, device_name);
	return event;
}


static GdamInputEvent*
gdam_midi_event_note_off_canonicalize(GdamInputEvent* event)
{
	GdamMidiEventNote* old_note = (GdamMidiEventNote*) event;
	return gdam_midi_event_note_new(GDAM_MIDI_DEVICE(event->input_device),
	                                event->input_id,
					old_note->midi_index,
					old_note->velocity,
					FALSE);
}

guint gdam_midi_event_note_off_get_type()
{
	static GdamInputEventTypeInfo type_info = {
		"GdamMidiEventNoteOff",
		gdam_midi_event_note_off_stringify,
		gdam_midi_event_note_off_hash,
		gdam_midi_event_note_off_equal,
		gdam_midi_event_note_off_to_keyvalues,
		gdam_midi_event_note_off_from_keyvalues,
		//NULL,
		//NULL,
		//NULL,
		//NULL,
		gdam_midi_event_note_off_canonicalize,
		NULL
	};
	static guint type = 0;
	if (type == 0)
		type = gdam_input_event_get_unique
				(gdam_input_event_int_get_type(),
				 GDAM_TYPE_MIDI_DEVICE,
				 &type_info);
	return type;
}
static char* gdam_midi_event_pitchwheel_stringify(GdamInputEvent* event)
{
	return g_strdup_printf("pitchwheel, channel %d", event->input_id);
}

guint gdam_midi_event_pitch_wheel_get_type()
{
	static GdamInputEventTypeInfo type_info = {
		"GdamMidiEventPitchWheel",
		gdam_midi_event_pitchwheel_stringify,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
	};
	static guint type = 0;
	if (type == 0)
		type = gdam_input_event_get_unique
				(gdam_input_event_range_get_type(),
				 GDAM_TYPE_MIDI_DEVICE,
				 &type_info);
	return type;
}
static char* gdam_midi_event_parameter_stringify(GdamInputEvent* event)
{
	GdamMidiEventParam* parameter;
	parameter = (GdamMidiEventParam*) event;
	return g_strdup_printf("parameter %d, channel %d",
				parameter->parameter,
				event->input_id);
}
static unsigned gdam_midi_event_parameter_hash(GdamInputEvent* event)
{
	GdamMidiEventParam* parameter;
	parameter = (GdamMidiEventParam*) event;
	return GPOINTER_TO_INT(event->input_device)
	     ^ (event->event_type << 6)
	     ^ (event->input_id << 12)
	     ^ (parameter->parameter << 18);
}
static gboolean gdam_midi_event_parameter_equal(GdamInputEvent* event1,
                                                GdamInputEvent* event2)
{
	GdamMidiEventParam* parameter1;
	GdamMidiEventParam* parameter2;
	parameter1 = (GdamMidiEventParam*) event1;
	parameter2 = (GdamMidiEventParam*) event2;
	return (event1->input_device == event2->input_device)
	    && (event1->event_type == event2->event_type)
	    && (event1->input_id == event2->input_id)
	    && (parameter1->parameter == parameter2->parameter);
}

static GHashTable* gdam_midi_event_parameter_to_keyvalues(GdamInputEvent* event)
{
	GHashTable* table;
	const char* device_name;
	GdamMidiEventParam* parameter = (GdamMidiEventParam*) event;
	table = g_hash_table_new(g_str_hash, g_str_equal);
	gdam_input_store_int(table, "input_id", event->input_id);
	gdam_input_store_int(table, "parameter", parameter->parameter);

	device_name = event->input_device->device_name;
	gdam_input_store_string(table, "device", device_name);
	return table;
}

static GdamInputEvent* gdam_midi_event_parameter_from_keyvalues(int type,
                                                            GHashTable* table)
{
	GdamInputEvent* event;
	GdamMidiEventParam* parameter;
	char* device_name;
	parameter = g_new(GdamMidiEventParam, 1);
	event = (GdamInputEvent*) parameter;
	event->event_type = type;
	event->input_id = gdam_input_fetch_int(table, "input_id");
	parameter->parameter = gdam_input_fetch_int(table, "parameter");

	device_name = g_hash_table_lookup(table, "device");
	gdam_input_event_parse_device(event, device_name);
	return event;
}
static GdamInputEvent*
gdam_midi_event_parameter_canonicalize(GdamInputEvent* old_event)
{
	GdamInputEvent* new_event;
	GdamMidiEventParam* new_parameter;
	GdamMidiEventParam* old_parameter = (GdamMidiEventParam*) old_event;
	new_parameter = g_new0(GdamMidiEventParam, 1);
	new_event = (GdamInputEvent*) new_parameter;
	*new_event = *old_event;
	new_parameter->parameter = old_parameter->parameter;
	return new_event;
}

guint gdam_midi_event_parameter_get_type()
{
	static GdamInputEventTypeInfo type_info = {
		"GdamMidiEventParameter",
		gdam_midi_event_parameter_stringify,
		gdam_midi_event_parameter_hash,
		gdam_midi_event_parameter_equal,
		gdam_midi_event_parameter_to_keyvalues,
		gdam_midi_event_parameter_from_keyvalues,
		gdam_midi_event_parameter_canonicalize,
		NULL
	};
	static guint type = 0;
	if (type == 0)
		type = gdam_input_event_get_unique
				(gdam_input_event_range_get_type(),
				 GDAM_TYPE_MIDI_DEVICE,
				 &type_info);
	return type;
}
static char* bin2hex(int length, const guint8* data)
{
	GString* string;
	char* rv;
	int i;
	string = g_string_new("");
	if (length > 0)
		g_string_sprintf(string, "0x%02x", data[0]);
	for (i = 1; i < length; i++)
		g_string_sprintfa(string, " 0x%02x", data[i]);
	rv = string->str;
	g_string_free(string, FALSE);
	return rv;
}
static char* skip_whitespace(char* at)
{
	while (*at && isspace(*at))
		at++;
	return at;
}
static gboolean hex2bin(const char* hex, int* length, guint8** data)
{
	char* endp;
	char* at = (char*) hex;
	GString* string;
	string = g_string_new("");
	at = skip_whitespace(at);
	while (*at) {
		int val = (int) strtol(at, &endp, 16);
		if (at == endp)
			return FALSE;
		at = skip_whitespace(endp);
		g_string_append_c(string, val);
	}
	*length = string->len;
	*data = (guint8*) (string->str);
	g_string_free(string, FALSE);
	return TRUE;
}
static char* gdam_midi_event_sysex_stringify(GdamInputEvent* event)
{
	GdamMidiEventSystemExclusive* sysex;
	char* hex;
	char* rv;
	sysex = (GdamMidiEventSystemExclusive*) event;
	hex = bin2hex(sysex->message_length, sysex->message);
	rv = g_strdup_printf("sysex, %s", hex);
	g_free(hex);
	return rv;
}
static unsigned gdam_midi_event_sysex_hash(GdamInputEvent* event)
{
	unsigned rv = event->event_type;
	GdamMidiEventSystemExclusive* sysex;
	int i;
	sysex = (GdamMidiEventSystemExclusive*) event;
	for (i = 0; i < sysex->message_length; i++) {
		rv ^= (rv << 5);
		rv += sysex->message[i];
	}
	return rv;
}
static gboolean gdam_midi_event_sysex_equal(GdamInputEvent* event1,
                                            GdamInputEvent* event2)
{
	GdamMidiEventSystemExclusive* sysex1, *sysex2;
	int length;
	sysex1 = (GdamMidiEventSystemExclusive*) event1;
	sysex2 = (GdamMidiEventSystemExclusive*) event2;
	if (event1->event_type != event2->event_type
	 || sysex1->message_length != sysex2->message_length)
	 	return FALSE;
	length = sysex1->message_length;
	if (memcmp(sysex1->message, sysex2->message, length) != 0)
		return FALSE;
	return TRUE;
}
static GHashTable* gdam_midi_event_sysex_to_keyvalues(GdamInputEvent* event)
{
	GHashTable* table;
	const char* device_name;
	char* hex;
	GdamMidiEventSystemExclusive* sysex;
	sysex = (GdamMidiEventSystemExclusive*) event;
	table = g_hash_table_new(g_str_hash, g_str_equal);
	hex = bin2hex(sysex->message_length, sysex->message);
	gdam_input_store_string(table, "sysex", hex);
	g_free(hex);
	device_name = event->input_device->device_name;
	gdam_input_store_string(table, "device", device_name);
	return table;
}
static GdamInputEvent* gdam_midi_event_sysex_from_keyvalues(int type, 
                                                            GHashTable* table)
{
	const char* hex;
	GdamMidiEventSystemExclusive* sysex;
	GdamInputEvent* event;
	char* device_name;
	sysex = g_new(GdamMidiEventSystemExclusive, 1);
	event = (GdamInputEvent*) sysex;
	event->event_type = type;

	hex = gdam_input_fetch_string(table, "sysex");
	hex2bin(hex, &sysex->message_length, &sysex->message);

	device_name = g_hash_table_lookup(table, "device");
	gdam_input_event_parse_device(event, device_name);
	return event;
}
static GdamInputEvent*
gdam_midi_event_sysex_canonicalize(GdamInputEvent* old_event)
{
	GdamMidiEventSystemExclusive* new_sysex, *old_sysex;
	GdamInputEvent* new_event;
	old_sysex = (GdamMidiEventSystemExclusive*) old_event;
	new_sysex = g_new(GdamMidiEventSystemExclusive, 1);
	new_event = (GdamInputEvent*) new_sysex;
	*new_event = *old_event;
	new_sysex->message_length = old_sysex->message_length;
	new_sysex->message = g_memdup(old_sysex->message,
	                              old_sysex->message_length);
	return new_event;
}
static void gdam_midi_event_sysex_destroy(GdamInputEvent* event)
{
	GdamMidiEventSystemExclusive* sysex;
	sysex = (GdamMidiEventSystemExclusive*) event;
	g_free(sysex->message);
	g_free(sysex);
}

guint gdam_midi_event_sysex_get_type()
{
	static GdamInputEventTypeInfo type_info = {
		"GdamMidiEventSysex",
		gdam_midi_event_sysex_stringify,
		gdam_midi_event_sysex_hash,
		gdam_midi_event_sysex_equal,
		gdam_midi_event_sysex_to_keyvalues,
		gdam_midi_event_sysex_from_keyvalues,
		gdam_midi_event_sysex_canonicalize,
		gdam_midi_event_sysex_destroy,
	};
	static guint type = 0;
	if (type == 0)
		type = gdam_input_event_get_unique
				(gdam_input_event_get_type(),
				 GDAM_TYPE_MIDI_DEVICE,
				 &type_info);
	return type;
}

