/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdammininetwork.h"
#include "../gdambuffermanager.h"
#include <common/gdam-protocol.h>
#include <common/gdamsearchpath.h>
#include <component/gdamxmlinfo.h>
#include <stdlib.h>
#include <string.h>

static int gdam_mini_network_append_start_request
                                         (GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);
static gboolean parse_param_value(GdamMiniNetworkParamValue* param_value,
                                  const char*                str,
				  GdamMiniNetworkParam*      param);
/*
 *            _       _                   _
 *  _ __ ___ (_)_ __ (_)  _ __   ___   __| | ___
 * | '_ ` _ \| | '_ \| | | '_ \ / _ \ / _` |/ _ \
 * | | | | | | | | | | | | | | | (_) | (_| |  __/
 * |_| |_| |_|_|_| |_|_| |_| |_|\___/ \__,_|\___|
 * 
 *      _                     _       _   _
 *   __| | ___  ___  ___ _ __(_)_ __ | |_(_) ___  _ __  ___
 *  / _` |/ _ \/ __|/ __| '__| | '_ \| __| |/ _ \| '_ \/ __|
 * | (_| |  __/\__ \ (__| |  | | |_) | |_| | (_) | | | \__ \
 *  \__,_|\___||___/\___|_|  |_| .__/ \__|_|\___/|_| |_|___/
 *                             |_|
 */
static int get_int(xmlNode* node)
{
  g_return_val_if_fail (node->content != NULL, -1);
  return atoi ((char *) node->content);
}
static gdouble get_double(xmlNode* node)
{
  g_return_val_if_fail (node->content != NULL, -1);
  return strtod ((char *) node->content, NULL);
}
GHashTable* gdam_mini_network_node_type_table = NULL;
#define node_type_table gdam_mini_network_node_type_table
static gboolean
process_input(xmlNode* nodes, GdamMiniNetworkInputType* input_type)
{
  memset (input_type, 0, sizeof (GdamMiniNetworkInputType));
  input_type->one_only = TRUE;
  for (; nodes != NULL; nodes = nodes->next)
    {
      if (strcmp (nodes->name, "single") == 0)
	{
	  char *content;
	  content = (char *) xmlNodeListGetString (NULL, nodes->childs, 1);
	  input_type->one_only = atoi (content);
	  g_free (content);
	  continue;
	}
      if (strcmp (nodes->name, "name") == 0)
	{
	  char *content;
	  content = (char *) xmlNodeListGetString (NULL, nodes->childs, 1);
	  input_type->name = content;
	  continue;
	}
      if (strcmp (nodes->name, "x") == 0)
	{
	  input_type->x = get_int (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "y") == 0)
	{
	  input_type->y = get_int (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "width") == 0)
	{
	  input_type->width = get_int (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "height") == 0)
	{
	  input_type->height = get_int (nodes->childs);
	  continue;
	}
      g_warning ("unknown node type under <node_type><input>");
    }
  return input_type->name != NULL;
}
static gboolean
process_parameter_mode(xmlNode* nodes, GdamMiniNetworkParamMode* param_mode)
{
  memset (param_mode, 0, sizeof (GdamMiniNetworkParamMode));
  param_mode->scale = 1.0;
  for (; nodes != NULL; nodes = nodes->next)
    {
      if (strcmp (nodes->name, "name") == 0)
	{
	  char *content;
	  content = (char *) xmlNodeListGetString (NULL, nodes->childs, 1);
	  param_mode->name = content;
	  continue;
	}
      if (strcmp (nodes->name, "default_value") == 0)
	{
	  param_mode->default_value = get_double (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "min_value") == 0)
	{
	  param_mode->min_value = get_double (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "max_value") == 0)
	{
	  param_mode->max_value = get_double (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "scale") == 0)
	{
	  param_mode->scale = get_double (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "logarithmic") == 0)
	{
	  param_mode->logarithmic = get_int (nodes->childs);
	  continue;
	}
    }
  return (param_mode->name != NULL);
}
static gboolean
process_parameter(xmlNode* nodes,
                  GdamMiniNetworkParam* input_param,
		  GdamMiniNetworkParamValue* default_value)
{
  GArray *array;
  GdamMiniNetworkParamType param_type = GDAM_MINI_NETWORK_RANGE;
  const char *def_value = NULL;
  memset (input_param, 0, sizeof (GdamMiniNetworkParam));
  array = g_array_new (FALSE, FALSE, sizeof (GdamMiniNetworkParamMode));
  for (; nodes != NULL; nodes = nodes->next)
    {
      if (strcmp (nodes->name, "name") == 0)
	{
	  char *content;
	  content = (char *) xmlNodeListGetString (NULL, nodes->childs, 1);
	  input_param->name = content;
	  continue;
	}
      if (strcmp (nodes->name, "type") == 0)
	{
	  char *type;
	  type = (char *) nodes->childs->content;
	  if (!type)
	    {
	      g_warning ("internal error parsing <type>");
	      continue;
	    }

	  if (strcmp (type, "range") == 0)
	    {
	      param_type = GDAM_MINI_NETWORK_RANGE;
	    }
	  else if (strcmp (type, "toggle") == 0)
	    {
	      param_type = GDAM_MINI_NETWORK_TOGGLE;
	    }
	  else if (strcmp (type, "buffer_id") == 0)
	    {
	      param_type = GDAM_MINI_NETWORK_BUFFER_ID;
	    }
	  else if (strcmp (type, "generic_range") == 0)
	    {
	      param_type = GDAM_MINI_NETWORK_GENERIC_RANGE;
	    }
	  else if (strcmp (type, "string") == 0)
	    {
	      param_type = GDAM_MINI_NETWORK_STRING;
	    }
	  else
	    {
	      g_warning ("parameter type `%s' unknown", type);
	    }
	  continue;
	}
      if (strcmp (nodes->name, "enter_not_required") == 0)
	{
	  input_param->enter_not_required = TRUE;
	  continue;
	}

      if (strcmp (nodes->name, "default_value") == 0)
	{
	  def_value = (const char *) nodes->childs->content;
	  continue;
	}
      if (strcmp (nodes->name, "min_value") == 0)
	{
	  input_param->min_value = get_double (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "max_value") == 0)
	{
	  input_param->max_value = get_double (nodes->childs);
	  continue;
	}
      if (strcmp (nodes->name, "mode") == 0)
	{
	  GdamMiniNetworkParamMode mode;
	  if (process_parameter_mode (nodes->childs, &mode))
	    {
	      g_array_append_val (array, mode);
	    }
	  continue;
	}
      g_warning ("unknown node type under <node_type><parameter>");
    }
  input_param->param_type = param_type;
  if (input_param->name != NULL)
    {
      input_param->modes = (GdamMiniNetworkParamMode *) (array->data);
      input_param->num_modes = array->len;
      g_array_free (array, 0);
    }
  else
    {
      g_array_free (array, 1);
    }
  if (def_value == NULL)
    {
      memset (default_value, 0, sizeof (GdamMiniNetworkParamValue));
    }
  else
    {
      if (!parse_param_value (default_value, def_value, input_param))
	{
	  g_warning ("error parsing default value for node %s "
		     "[string is `%s']", input_param->name, def_value);
	}
    }
  return input_param->name != NULL;
}


static GdamMiniNetworkNodeType* process_type_node_list(xmlNode* nodes)
{
  GdamMiniNetworkNodeType *type;
  GArray *params;
  GArray *default_values;
  GArray *inputs;
  int param_index = 0;
  if (node_type_table == NULL)
    node_type_table = g_hash_table_new (g_str_hash, g_str_equal);
  type = g_new0 (GdamMiniNetworkNodeType, 1);
  params = g_array_new (FALSE, FALSE, sizeof (GdamMiniNetworkParam));
  default_values
    = g_array_new (FALSE, FALSE, sizeof (GdamMiniNetworkParamValue));
  inputs = g_array_new (FALSE, FALSE, sizeof (GdamMiniNetworkInputType));
  for (; nodes != NULL; nodes = nodes->next)
    {
      if (strcmp (nodes->name, "name") == 0)
	{
	  type->type = (char *) xmlNodeListGetString (NULL,
						      nodes->childs, TRUE);
	}
      else if (strcmp (nodes->name, "pixmap") == 0)
	{
	  type->pixmap = (char *) xmlNodeListGetString (NULL,
							nodes->childs, TRUE);
	}
      else if (strcmp (nodes->name, "input") == 0)
	{
	  GdamMiniNetworkInputType input_type;
	  if (process_input (nodes->childs, &input_type))
	    {
	      g_array_append_val (inputs, input_type);
	    }
	  else
	    {
	      g_warning ("mini_network: "
			 "couldn't process input node [for %s]", type->type);
	    }
	}
      else if (strcmp (nodes->name, "parameter") == 0)
	{
	  GdamMiniNetworkParam param;
	  GdamMiniNetworkParamValue defval;
	  if (process_parameter (nodes->childs, &param, &defval))
	    {
	      param.param_index = param_index++;
	      g_array_append_val (params, param);
	      g_array_append_val (default_values, defval);
	    }
	  else
	    {
	      g_warning ("mini_network: "
			 "couldn't process parameter [for %s]", type->type);
	    }
	}
      else if (strcmp (nodes->name, "output_x") == 0)
	{
	  type->output_area_x = get_int (nodes->childs);
	}
      else if (strcmp (nodes->name, "output_y") == 0)
	{
	  type->output_area_y = get_int (nodes->childs);
	}
      else if (strcmp (nodes->name, "output_width") == 0)
	{
	  type->output_area_width = get_int (nodes->childs);
	}
      else if (strcmp (nodes->name, "output_height") == 0)
	{
	  type->output_area_height = get_int (nodes->childs);
	}
      else
	{
	  g_warning ("unknown tag %s: expected name,pixmap,input",
		     nodes->name);
	}
    }
  type->input_types = (GdamMiniNetworkInputType *) inputs->data;
  type->params = (GdamMiniNetworkParam *) params->data;
  type->param_defaults = (GdamMiniNetworkParamValue *) (default_values->data);
  type->num_inputs = inputs->len;
  type->num_params = params->len;
  g_array_free (inputs, FALSE);
  g_array_free (params, FALSE);
  g_array_free (default_values, FALSE);
  /*
     if (type->num_params > 0) {
     GString* string;
     int i;
     string = g_string_new("");
     for (i = 0; i < type->num_params; i++) {
     g_string_sprintfa(string, "%6.6g",
     type->params[i].default_value);
     if (i < type->num_params - 1)
     g_string_append_c(string, ':');
     }
     type->default_options = string->str;
     g_string_free(string, FALSE);
     }
   */
  if (type->type == NULL)
    {
      g_free (type->input_types);
      g_free (type);
      return NULL;
    }
  return type;
}

static gboolean parse_xml_node_types(const char* filename)
{
  xmlDoc *doc;
  xmlNode *root;
  xmlNode *types;
  doc = xmlParseFile (filename);
  if (doc == NULL)
    {
      g_warning ("no mininetwork node description file %s", filename);
      return FALSE;
    }
  root = doc->root;
  for (types = root->childs; types != NULL; types = types->next)
    {
      GdamMiniNetworkNodeType *type;
      if (strcmp (types->name, "node_type") != 0)
	{
	  g_warning ("expected node ``node_type'', got %s", types->name);
	  continue;
	}
      type = process_type_node_list (types->childs);
      if (type != NULL)
	{
	  g_hash_table_insert (node_type_table, (char *) type->type, type);
	}
    }
  xmlFreeDoc (doc);
  return TRUE;
}

GdamMiniNetworkNodeType* gdam_mini_network_get_node_type(const char* type)
{
  g_return_val_if_fail (node_type_table != NULL, NULL);
  return g_hash_table_lookup (node_type_table, (char *) type);
}

/*
 * Parsing/printing the generic model xml.
 *  <node id="%d">
 *    <type>...</type>
 *    <parameter name="%s">...</parameter>
 *    ...
 *  </node>
 *  ...
 *  <pipe id="%d">
 *    <input>%d</input>
 *    <output>%d</output>
 *    <input_name>%d</input_name>
 *  </pipe>
 *  ...
 */
static xmlNode* make_node_xml(GdamMiniNetworkNode* node, GdamXmlInfo* xml_info)
{
  xmlNode *xml;
  char buf[256];
  int xml_node_id;
  int i;
  GdamMiniNetworkNodeType *type = node->node_type;

  xml_node_id = gdam_xml_info_alloc_id (xml_info, node);
  xml = xmlNewNode (NULL, "node");
  snprintf (buf, sizeof (buf), "%d", xml_node_id);
  xmlSetProp (xml, "id", buf);
  xmlNewChild (xml, NULL, "type", node->node_type->type);
  for (i = 0; i < type->num_params; i++)
    {
      char *str;
      GString *string;
      xmlNode *child;
      string = g_string_new ("");
      gdam_mini_network_make_one_param_string
	(type->params + i, node->param_values + i, string,
	 GDAM_MINI_NETWORK_BUFFER_AS_STRING);
      str = string->str;
      g_string_free (string, FALSE);
      child = xmlNewChild (xml, NULL, "parameter", str);
      xmlSetProp (child, "name", type->params[i].name);
    }
  return xml;
}

static xmlNode* make_pipe_xml(GdamMiniNetworkPipe* pipe, GdamXmlInfo* xml_info)
{
  xmlNode *xml;
  char buf[256];
  int xml_pipe_id;
  int id;

  xml_pipe_id = gdam_xml_info_alloc_id (xml_info, pipe);
  xml = xmlNewNode (NULL, "pipe");
  snprintf (buf, sizeof (buf), "%d", xml_pipe_id);
  xmlSetProp (xml, "id", buf);

  id = gdam_xml_info_lookup_by_object (xml_info, pipe->source);
  g_return_val_if_fail (id != 0, NULL);
  snprintf (buf, sizeof (buf), "%d", id);
  xmlNewChild (xml, NULL, "input", buf);

  id = gdam_xml_info_lookup_by_object (xml_info, pipe->destination);
  g_return_val_if_fail (id != 0, NULL);
  snprintf (buf, sizeof (buf), "%d", id);
  xmlNewChild (xml, NULL, "output", buf);

  xmlNewChild (xml, NULL, "input_name", pipe->input_name);
  return xml;
}
static void gdam_mini_network_add_xml(GdamModel*     model,
	                              xmlNode*       parent,
				      GdamXmlInfo*   xml_info)
{
  GSList *list;
  GdamMiniNetwork *network;

  network = GDAM_MINI_NETWORK (model);

  for (list = network->nodes; list != NULL; list = list->next)
    {
      GdamMiniNetworkNode *node = list->data;
      xmlNode *node_node;
      node_node = make_node_xml (node, xml_info);
      if (node_node != NULL)
	{
	  xmlAddChild (parent, node_node);
	}
      else
	{
	  g_warning ("error producing xml for mininetwork node");
	  return;
	}
    }
  for (list = network->pipes; list != NULL; list = list->next)
    {
      GdamMiniNetworkPipe *pipe = list->data;
      xmlNode *pipe_node;
      pipe_node = make_pipe_xml (pipe, xml_info);
      if (pipe_node != NULL)
	{
	  xmlAddChild (parent, pipe_node);
	}
      else
	{
	  g_warning ("error producing xml for mininetwork pipe");
	  return;
	}
    }
}

/* Free ``*options_out'', ``*type_out''. */
static gboolean parse_node_xml(xmlNode* children,
                               char** type_out,
			       char** options_out)
{
  xmlNode *type_xml_node;
  xmlNode *param_node;
  GPtrArray *str_array;
  type_xml_node = children;

  if (type_xml_node == NULL)
    {
      g_warning ("<node> block was empty in mininetwork [invalid]");
      return FALSE;
    }
  if (strcmp (type_xml_node->name, "type") != 0)
    {
      g_warning ("sorry, but <type> must be first in a <node> block");
      return FALSE;
    }

  str_array = g_ptr_array_new ();

  /* XXX: must do error checking!!!  currently we just
   *      join the parameters together with
   *      colons, but if they are rearranged (by the user)
   *      they will be misordered!!! */
  for (param_node = type_xml_node->next;
       param_node != NULL; param_node = param_node->next)
    {
      if (strcmp (param_node->name, "parameter") == 0)
	{
	  char *content;
	  content = xmlNodeListGetString (NULL, param_node->childs, TRUE);
	  g_ptr_array_add (str_array, content);
	}
    }
  g_ptr_array_add (str_array, NULL);
  *options_out = g_strjoinv (":", (char **) str_array->pdata);
  *type_out = xmlNodeListGetString (NULL, type_xml_node->childs, TRUE);
  {
    guint i;
    for (i = 0; i < str_array->len; i++)
      g_free (str_array->pdata[i]);
  }
  g_ptr_array_free (str_array, TRUE);
  return TRUE;
}

static gboolean  parse_pipe_xml(xmlNode*              children,
			        GdamXmlInfo*          xml_info,
                                GdamMiniNetworkNode** source_out,
                                GdamMiniNetworkNode** dest_out,
                                char**                input_name_out)
{
  xmlNode *node;
  char *input = NULL;
  char *output = NULL;
  char *input_name = NULL;
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "input") == 0)
	{
	  input = xmlNodeListGetString (NULL, node->childs, 1);
	  continue;
	}
      if (strcmp (node->name, "output") == 0)
	{
	  output = xmlNodeListGetString (NULL, node->childs, 1);
	  continue;
	}
      if (strcmp (node->name, "input_name") == 0)
	{
	  input_name = xmlNodeListGetString (NULL, node->childs, 1);
	  continue;
	}
    }
  if (input == NULL || output == NULL || input_name == NULL)
    {
      g_warning ("pipe node wasn't complete need "
		 "<input>, <output>, <input_name> ... at least one was "
		 "missing");
      return FALSE;
    }
  *source_out = gdam_xml_info_lookup_by_id (xml_info, atoi (input));
  if (*source_out == NULL)
    {
      g_warning ("source node (%s) for <pipe> not found", input);
      return FALSE;
    }
  *dest_out = gdam_xml_info_lookup_by_id (xml_info, atoi (output));
  if (*dest_out == NULL)
    {
      g_warning ("dest node (%s) for <pipe> not found", output);
      return FALSE;
    }
  *input_name_out = g_strdup (input_name);
  return TRUE;
}

static int find_id(xmlNode* node)
{
  const char *property;
  property = xmlGetProp (node, "id");
  if (property == NULL)
    return 0;
  return atoi (property);
}

static void gdam_mini_network_parse_xml(GdamModel*     model,
					xmlNode*       xml,
					GdamXmlInfo*   xml_info)
{
  GdamMiniNetwork *mini_network;

  mini_network = GDAM_MINI_NETWORK (model);

  if (strcmp (xml->name, "node") == 0)
    {
      char *type;
      char *options;
      int id;
      GdamMiniNetworkNode *node;
      id = find_id (xml);
      if (id <= 0)
	{
	  g_warning ("no id=\"%%d\" tag to node <%s>", xml->name);
	  return;
	}
      if (!parse_node_xml (xml->childs, &type, &options))
	return;
      g_return_if_fail (type != NULL);
      if (strcmp (type, "input") == 0)
	{
	  node = mini_network->input;
	}
      else if (strcmp (type, "output") == 0)
	{
	  node = mini_network->output;
	}
      else
	{
	  node = gdam_mini_network_new_node (mini_network, type, options);
	}
      if (node == NULL)
	{
	  g_warning ("couldn't create network-node of type %s", type);
	}
      else
	{
	  gdam_xml_info_assoc (xml_info, id, node);
	}
    }
  else if (strcmp (xml->name, "pipe") == 0)
    {
      GdamMiniNetworkPipe *pipe;
      GdamMiniNetworkNode *source;
      GdamMiniNetworkNode *dest;
      char *input_name;
      int id;
      id = find_id (xml);
      if (id <= 0)
	{
	  g_warning ("no id=\"%%d\" tag to node <%s>", xml->name);
	  return;
	}
      if (!parse_pipe_xml (xml->childs, xml_info, &source, &dest,
			   &input_name))
	return;
      pipe = gdam_mini_network_new_pipe (mini_network,
					 source, dest, input_name);
      if (pipe == NULL)
	{
	  g_warning ("couldn't create network-pipe [%d]", id);
	}
      else
	{
	  gdam_debug ("creating pipe %d => %d",
		      source->node_id, dest->node_id);
	  gdam_xml_info_assoc (xml_info, id, pipe);
	}
    }
  else
    {
      g_warning ("node <%s> is not interpreted by mini_network", xml->name);
    }
}


/*
 *        _     _           _         _          __  __
 *   ___ | |__ (_) ___  ___| |_   ___| |_ _   _ / _|/ _|
 *  / _ \| '_ \| |/ _ \/ __| __| / __| __| | | | |_| |_
 * | (_) | |_) | |  __/ (__| |_  \__ \ |_| |_| |  _|  _|
 *  \___/|_.__// |\___|\___|\__| |___/\__|\__,_|_| |_|
 *           |__/
 */

static void gdam_mini_network_init(GdamMiniNetwork* mini_network)
{
  GdamMiniNetworkNode *node;
  GdamMiniNetworkPipe *pipe;
  GdamMiniNetworkNodeType *input_type;
  GdamMiniNetworkNodeType *output_type;
  mini_network->last_node_id = 2;
  input_type = g_hash_table_lookup (node_type_table, "input");
  g_return_if_fail (input_type != NULL);
  output_type = g_hash_table_lookup (node_type_table, "output");
  g_return_if_fail (output_type != NULL);

  node = g_new (GdamMiniNetworkNode, 1);
  node->node_type = input_type;
  node->node_id = 1;
  node->options = g_strdup ("");
  mini_network->input = node;
  mini_network->nodes = g_slist_prepend (mini_network->nodes, node);

  node = g_new (GdamMiniNetworkNode, 1);
  node->node_type = output_type;
  node->node_id = 2;
  node->options = g_strdup ("");
  mini_network->output = node;
  mini_network->nodes = g_slist_prepend (mini_network->nodes, node);

  pipe = g_new (GdamMiniNetworkPipe, 1);
  pipe->destination = mini_network->output;
  pipe->source = mini_network->input;
  pipe->input_name = g_strdup ("input");
  pipe->pipe_id = ++(mini_network->last_pipe_id);
  mini_network->pipes = g_slist_prepend (mini_network->pipes, pipe);
}
#define NETWORK_DESCRIPTION "mininetwork-description.xml"
#define SUBPREFIX "share/gdam"

/* Uh, this is ugly and nonportable!!!
 * 
 * XXX: we really need to resolve this kind of path
 *      problem in a civilized manner.
 */
#define SEARCH_PATH      PREFIX SUBPREFIX \
	      ":.:" SUBPREFIX ":/usr/local/" SUBPREFIX ":/usr/" SUBPREFIX 



static void gdam_mini_network_class_init(GdamMiniNetworkClass* klass)
{
  char *filename;
  GdamModelClass *model_class;
  filename = gdam_search_path (NETWORK_DESCRIPTION, SEARCH_PATH,
			       GDAM_SEARCH_PATH_FIND_FILE);
  if (filename == NULL)
    {
      g_warning ("couldn't find %s", NETWORK_DESCRIPTION);
      return;
    }
  if (!parse_xml_node_types (filename))
    {
      g_warning ("had trouble parsing mini-network node descriptions "
		 "from %s", filename);
    }

  model_class = GDAM_MODEL_CLASS (klass);
  model_class->parse_xml = gdam_mini_network_parse_xml;
  model_class->add_xml = gdam_mini_network_add_xml;

  GDAM_FILTER_CLASS (klass)->append_start_request
    = gdam_mini_network_append_start_request;
}

GtkType gdam_mini_network_get_type()
{
  static GtkType mini_network_type = 0;
  if (!mini_network_type)
    {
      static const GtkTypeInfo mini_network_info = {
	"GdamMiniNetwork",
	sizeof (GdamMiniNetwork),
	sizeof (GdamMiniNetworkClass),
	(GtkClassInitFunc) gdam_mini_network_class_init,
	(GtkObjectInitFunc) gdam_mini_network_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      mini_network_type = gtk_type_unique (parent, &mini_network_info);
    }
  return mini_network_type;
}

static void do_append_request(GdamMiniNetwork* network, GdamRequest* request)
{
  if (network->freeze_count > 0)
    {
      if (network->builder == NULL)
	network->builder = gdam_request_builder_new ();
      gdam_request_builder_append_generic (network->builder, request);
    }
  else
    {
      GdamModel *model;
      model = GDAM_MODEL (network);
      gdam_channel_append_request (model->channel, request);
    }
}

void gdam_mini_network_freeze(GdamMiniNetwork* network)
{
  network->freeze_count++;
}
void gdam_mini_network_thaw(GdamMiniNetwork* network)
{
  g_return_if_fail (network->freeze_count >= 1);
  if (--(network->freeze_count) == 0)
    {
      GdamRequestBuilder *builder = network->builder;
      if (builder != NULL)
	{
	  GdamRequest *package;
	  GdamChannel *channel;
	  network->builder = NULL;
	  package = gdam_request_builder_package (builder);
	  channel = GDAM_MODEL (network)->channel;
	  gdam_channel_append_request (channel, package);
	  gdam_request_builder_destroy (builder);
	}
    }
}

static GdamRequest* make_node_request(GdamMiniNetwork* network,
				      int source_id,
                                      GdamMiniNetworkNode* node)

{
  GdamRequest *request;
  GdamArg args[2];
  /* XXX: add error checking. */
  args[0].name = "type";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + 0) = (char *) node->node_type->type;
  args[1].name = "options";
  args[1].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + 1) = node->options;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					source_id,
					"add_node",
					2, args, NULL, NULL, NULL);
  return request;
}
void gdam_mini_network_param_value_copy(GdamMiniNetworkParamValue* dst,
                                        const GdamMiniNetworkParamValue* src,
					int type)
{
  if (type != GDAM_MINI_NETWORK_STRING)
    *dst = *src;
  else
    dst->string_value = g_strdup (src->string_value);
}

static gboolean parse_param_value(GdamMiniNetworkParamValue* param_value,
                                  const char*                str,
				  GdamMiniNetworkParam*      param)
{
  if (param->param_type == GDAM_MINI_NETWORK_RANGE)
    {
      param_value->double_value = strtod (str, NULL);
    }
  else if (param->param_type == GDAM_MINI_NETWORK_TOGGLE)
    {
      param_value->boolean_value = atoi (str);
    }
  else if (param->param_type == GDAM_MINI_NETWORK_BUFFER_ID)
    {
      param_value->buffer_id_value = atoi (str);
    }
  else if (param->param_type == GDAM_MINI_NETWORK_GENERIC_RANGE)
    {
      char **subs;
      subs = g_strsplit (str, ",", 5);
      if (subs[0] == NULL || subs[1] == NULL
	  || subs[2] == NULL || subs[3] == NULL || subs[4] == NULL)
	{
	  g_strfreev (subs);
	  return FALSE;
	}
      param_value->generic_range_value.value = strtod (subs[0], NULL);
      param_value->generic_range_value.min_value = strtod (subs[1], NULL);
      param_value->generic_range_value.max_value = strtod (subs[2], NULL);
      param_value->generic_range_value.is_logarithmic = atoi (subs[3]);
      param_value->generic_range_value.is_integer = atoi (subs[4]);
    }
  else if (param->param_type == GDAM_MINI_NETWORK_STRING)
    {
      param_value->string_value = g_strdup (str);
    }
  else
    {
      g_warning ("unknown param type: %d", param->param_type);
      return FALSE;
    }
  return TRUE;
}

GdamMiniNetworkNode* gdam_mini_network_new_node
                                    (GdamMiniNetwork* network,
                                     const char*      type,
				     const char*      options)
{
  GdamModel *model;
  GdamMiniNetworkNode *node;
  GdamMiniNetworkNodeType *node_type;
  GdamRequest *request;
  int i;

  model = GDAM_MODEL (network);

  node_type = gdam_mini_network_get_node_type (type);
  g_return_val_if_fail (node_type != NULL, NULL);

  node = g_new (GdamMiniNetworkNode, 1);
  node->node_type = node_type;
  node->node_id = ++(network->last_node_id);
  node->param_values = g_new (GdamMiniNetworkParamValue,
			      node_type->num_params);
  node->modes = g_new0 (int, node_type->num_params);
  gdam_debug ("new_node: type=%s, options=%s, node_id=%d",
	      type, options, node->node_id);
  if (options == NULL)
    {
      for (i = 0; i < node_type->num_params; i++)
	gdam_mini_network_param_value_copy
	  (&node->param_values[i],
	   &node_type->param_defaults[i], node_type->params[i].param_type);
    }
  else
    {
      char **split;
      split = g_strsplit (options, ":", node_type->num_params);
      for (i = 0; split[i] != NULL; i++)
	{
	  parse_param_value (&node->param_values[i],
			     split[i], &node_type->params[i]);
	}
      g_strfreev (split);
    }
  node->options = gdam_mini_network_make_param_string (node);
  network->nodes = g_slist_prepend (network->nodes, node);
  if (model->source_id != 0)
    {
      request = make_node_request (network, model->source_id, node);
      do_append_request (network, request);
    }
  network->seq_no++;
  return node;
}

void gdam_mini_network_configure_node (GdamMiniNetwork* network,
                                       GdamMiniNetworkNode* node,
				       const char*      options)
{
  GdamModel *model = GDAM_MODEL (network);
  g_free (node->options);
  node->options = g_strdup (options);
  if (model->source_id != 0)
    {
      GdamRequest *request;
      GdamArg args[2];
      args[0].name = "node_id";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args + 0) = node->node_id;
      args[1].name = "options";
      args[1].type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (args + 1) = (char *) options;

      request = gdam_request_call_function
	(GDAM_PROTOCOL_SOURCE_ARG,
	 model->source_id, "configure_node", 2, args, NULL, NULL, NULL);
      do_append_request (network, request);
    }
  network->seq_no++;
}

typedef struct _GdamPipeCreation GdamPipeCreation;
struct _GdamPipeCreation {
	GdamMiniNetwork* network;
	GdamMiniNetworkPipe* pipe;
};

/*XXX: totally need notification!!!! the skin is going to screw
 *     up and keep these entries in its hash table. */
static void on_pipe_failed(gpointer data)
{
  GdamPipeCreation *creation_data = data;
  creation_data->network->pipes =
    g_slist_remove (creation_data->network->pipes, creation_data->pipe);
  gtk_object_unref (GTK_OBJECT (creation_data->network));
  g_free (data);
}
static void on_pipe_created(int num_args, GdamArg* args, gpointer data)
{
  GdamPipeCreation *creation_data = data;
  if (num_args != 1
      || strcmp (args->name, "pipe_id") != 0
      || args->type != GDAM_ARG_TYPE_INT || GDAM_ARG_INT (args) == 0)
    {
      on_pipe_failed (creation_data);
    }
  else
    {
      gtk_object_unref (GTK_OBJECT (creation_data->network));
      g_free (creation_data);
    }
}

static GdamRequest* make_pipe_request(GdamMiniNetwork* network,
				      int source_id,
                                      GdamMiniNetworkPipe* pipe)

{
  GdamRequest *request;
  GdamArg args[3];
  GdamPipeCreation *creation_data;

  creation_data = g_new (GdamPipeCreation, 1);
  creation_data->network = network;
  creation_data->pipe = pipe;

  gtk_object_ref (GTK_OBJECT (network));
  args[0].name = "source";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 0) = pipe->source->node_id;
  args[1].name = "destination";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 1) = pipe->destination->node_id;
  args[2].name = "input_name";
  args[2].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + 2) = pipe->input_name;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					source_id,
					"add_pipe",
					3,
					args,
					on_pipe_created,
					on_pipe_failed, creation_data);
  return request;
}
gboolean gdam_mini_network_has_flow_internal
                                  (GdamMiniNetwork*       network,
                                   GdamMiniNetworkNode*   source,
                                   GdamMiniNetworkNode*   dest,
				   gboolean               must_be_0delay,
				   GHashTable*            visited)
{
  int i;
  GSList *list;
  if (source == dest)
    return TRUE;
  if (g_hash_table_lookup (visited, dest))
    return FALSE;

  g_hash_table_insert (visited, dest, dest);
  for (list = network->pipes; list != NULL; list = list->next)
    {
      GdamMiniNetworkPipe *pipe = list->data;
      if (pipe->destination != dest)
	continue;
      if (must_be_0delay)
	{
	  /*
	   * if we are looking for a zero-delay
	   * flow, then a flow through a delay-node
	   * that isn't the delay control line
	   * should be fine.
	   *
	   * HACK:  this should be virtual really badly!!!
	   */
	  if (strstr (dest->node_type->type, "delay") != NULL
	      && strstr (pipe->input_name, "delay") == NULL)
	    continue;
	}
      if (gdam_mini_network_has_flow_internal (network,
					       source,
					       pipe->source,
					       must_be_0delay, visited))
	{
	  return TRUE;
	}
    }
  return FALSE;
}

gboolean gdam_mini_network_has_flow(GdamMiniNetwork*       network,
                                    GdamMiniNetworkNode*   source,
                                    GdamMiniNetworkNode*   dest,
				    gboolean               must_be_0delay)
{
  GHashTable *table;
  gboolean rv;
  table = g_hash_table_new (NULL, NULL);
  rv = gdam_mini_network_has_flow_internal (network, source, dest,
					    must_be_0delay, table);
  g_hash_table_destroy (table);
  return rv;
}

GdamMiniNetworkPipe* gdam_mini_network_new_pipe 
                                    (GdamMiniNetwork* network,
                                     GdamMiniNetworkNode* source,
                                     GdamMiniNetworkNode* dest,
				     const char*      input_name)
{
  GdamModel *model;
  GdamMiniNetworkPipe *pipe;
  GdamRequest *request;

  model = GDAM_MODEL (network);

  if (gdam_mini_network_has_flow (network, dest, source, TRUE))
    {
      g_warning ("0-delay feedback loops not allowed");
      return NULL;
    }

  pipe = g_new (GdamMiniNetworkPipe, 1);
  pipe->destination = dest;
  pipe->source = source;
  pipe->input_name = g_strdup (input_name);
  pipe->pipe_id = ++(network->last_pipe_id);

  gdam_debug ("created pipe %d between %d => %d [%s]",
	      pipe->pipe_id, source->node_id, dest->node_id, input_name);

  network->pipes = g_slist_prepend (network->pipes, pipe);
  if (model->source_id != 0)
    {
      request = make_pipe_request (network, model->source_id, pipe);
      do_append_request (network, request);
    }
  network->seq_no++;

  return pipe;
}

void                 gdam_mini_network_remove_node
                                    (GdamMiniNetwork* network,
                                     GdamMiniNetworkNode* node)
{
  GdamModel *model = GDAM_MODEL (network);
  GSList **plist;
  gdam_debug ("removing node %d", node->node_id);
  network->nodes = g_slist_remove (network->nodes, node);
  if (model->source_id != 0)
    {
      GdamArg args[1];
      args[0].name = "remove_node";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args + 0) = node->node_id;
      do_append_request
	(network,
	 gdam_request_set_args (model->source_id, 1, args, NULL, NULL, NULL));
    }
  for (plist = &network->pipes; *plist;)
    {
      GdamMiniNetworkPipe *pipe = (*plist)->data;
      if (pipe->source == node || pipe->destination == node)
	{
	  GSList *next = (*plist)->next;
	  g_slist_free_1 (*plist);
	  g_free (pipe->input_name);
	  g_free (pipe);
	  *plist = next;
	}
      else
	{
	  plist = &(*plist)->next;
	}
    }
  network->seq_no++;

  g_free (node->modes);
  g_free (node->param_values);
  g_free (node->options);
  g_free (node);
}

void                 gdam_mini_network_remove_pipe 
                                    (GdamMiniNetwork* network,
                                     GdamMiniNetworkPipe* pipe)
{
  GdamModel *model = GDAM_MODEL (network);
  gdam_debug ("removing pipe %d", pipe->pipe_id);
  if (model->source_id != 0)
    {
      GdamArg args[1];
      args[0].name = "remove_pipe";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args + 0) = pipe->pipe_id;
      do_append_request
	(network,
	 gdam_request_set_args (model->source_id, 1, args, NULL, NULL, NULL));
    }
  network->seq_no++;
  network->pipes = g_slist_remove (network->pipes, pipe);
  g_free (pipe->input_name);
  g_free (pipe);
}


static int gdam_mini_network_append_start_request
                                         (GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamMiniNetwork *network = GDAM_MINI_NETWORK (filter);
  GdamModel *model;
  GdamRequest *request;
  int local_src_id;
  GSList *list;
  g_return_val_if_fail (network->freeze_count == 0, 0);
  model = GDAM_MODEL (network);
  gtk_object_ref (GTK_OBJECT (network));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasMiniNetwork",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					filter,
					1,
					"filtered",
					GDAM_ARG_TYPE_INT, new_source_id);
  local_src_id = gdam_request_builder_append_create (builder, request);
  network->last_node_id = 2;
  network->last_pipe_id = 0;
  for (list = network->nodes; list != NULL; list = list->next)
    {
      GdamMiniNetworkNode *node = list->data;
      GdamRequest *request;
      if (strcmp (node->node_type->type, "input") == 0
	  || strcmp (node->node_type->type, "output") == 0)
	continue;
      node->node_id = ++(network->last_node_id);
      request = make_node_request (network, local_src_id, node);
      gdam_request_builder_append_generic (builder, request);
    }
  for (list = network->pipes; list != NULL; list = list->next)
    {
      GdamMiniNetworkPipe *pipe = list->data;
      GdamRequest *request;
      pipe->pipe_id = ++(network->last_pipe_id);
      request = make_pipe_request (network, local_src_id, pipe);
      gdam_request_builder_append_generic (builder, request);
    }
  return local_src_id;
}

void gdam_mini_network_make_one_param_string
                            (GdamMiniNetworkParam* param,
                             GdamMiniNetworkParamValue* value,
			     GString* out,
			     guint flags)
{
  if (param->param_type == GDAM_MINI_NETWORK_RANGE)
    {
      g_string_sprintfa (out, "%6.6g", value->double_value);
    }
  else if (param->param_type == GDAM_MINI_NETWORK_TOGGLE)
    {
      gboolean val = value->boolean_value;
      g_string_append_c (out, val ? '1' : '0');
    }
  else if (param->param_type == GDAM_MINI_NETWORK_BUFFER_ID)
    {
      int buffer_id = value->buffer_id_value;
      if ((flags & GDAM_MINI_NETWORK_BUFFER_AS_STRING)
	  == GDAM_MINI_NETWORK_BUFFER_AS_STRING)
	{
	  GdamBufferManagerInfo *info;
	  info = gdam_buffer_manager_info_by_id (buffer_id);
	  if (info)
	    g_string_append (out, info->description);
	  else
	    g_string_sprintfa (out, "buffer %d", buffer_id);
	}
      else
	{
	  g_string_sprintfa (out, "%d", buffer_id);
	}
    }
  else if (param->param_type == GDAM_MINI_NETWORK_GENERIC_RANGE)
    {
      g_string_sprintfa (out, "%.8g,%.8g,%.8g,%d,%d",
			 value->generic_range_value.value,
			 value->generic_range_value.min_value,
			 value->generic_range_value.max_value,
			 value->generic_range_value.is_logarithmic,
			 value->generic_range_value.is_integer);
    }
  else if (param->param_type == GDAM_MINI_NETWORK_STRING)
    {
      if (value->string_value != NULL)
	g_string_append (out, value->string_value);
    }
  else
    {
      g_warning ("unknown param type %d", param->param_type);
    }
}

char* gdam_mini_network_make_param_string(GdamMiniNetworkNode* node)
{
  GString *string;
  int i;
  GdamMiniNetworkNodeType *node_type = node->node_type;
  char *rv;
  string = g_string_new ("");
  for (i = 0; i < node_type->num_params; i++)
    {
      gdam_mini_network_make_one_param_string (node_type->params + i,
					       node->param_values + i,
					       string, 0);
      if (i < node_type->num_params - 1)
	g_string_append_c (string, ':');
    }
  rv = string->str;
  g_string_free (string, FALSE);
  return rv;
}



#if 0

static void gdam_mini_network_new_source_id(GdamModel*     model,
	                                    int            source_id)
{
  GdamMiniNetwork *network = GDAM_MINI_NETWORK (model);
  GdamRequestBuilder *builder;
  model->source_id = source_id;
  if (network->pending_nodes || network->pending_pipes)
    {
      GSList *list;
      for (list = network->pending_nodes; list != NULL; list = list->next)
	{
	  GdamMiniNetworkNode *node = list->data;
	  GdamRequest *request;
	  request = make_node_request (network, node);
	  gdam_request_builder_append_generic (builder, request);
	}
      for (list = network->pending_pipes; list != NULL; list = list->next)
	{
	  GdamMiniNetworkPipe *pipe = list->data;
	  GdamRequest *request;
	  request = make_pipe_request (network, node);
	  gdam_request_builder_append_generic (builder, request);
	}
      gdam_channel_append_request (model->channel,
				   gdam_request_builder_package (builder));
      gdam_request_builder_destroy (builder);
      g_slist_free (network->pending_nodes);
      g_slist_free (network->pending_pipes);
      network->pending_nodes = NULL;
      network->pending_pipes = NULL;
    }

  (*GDAM_MODEL_CLASS (parent_class)->new_source_id) (model, source_id);
}

#endif
