/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamps2keyboard.h"
#include <common/gdamplugin.h>
#include <string.h>
#include <stdlib.h>

static void gdam_ps2_keyboard_init(GdamPs2Keyboard* ps2_keyboard)
{
}

static const char* scancode_table[256] = {
NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
NULL, NULL, "tab", NULL, NULL, NULL, "alt", "lshift", NULL, "control",
"q", "1", NULL, NULL, NULL, "z", "s", "a", "w", "2", NULL, NULL,
"c", "x", "d", "e", "4", "3", NULL, NULL, NULL, "v", "f", "t", "r",
"5", NULL, NULL, "n", "b", "h", "g", "y", "6", NULL, NULL, NULL, "m",
"j", "u", "7", "8", NULL, NULL, ",", "k", "i", "o", "0", "9", NULL,
NULL, ".", "/", "l", ";", "p", "-", NULL, NULL, NULL, "'", NULL,
"[", "=", NULL, NULL, "caps", "rshift", "enter", "]", NULL, "\"",
NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, "backspace", NULL,
NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
NULL, NULL, NULL, "esc", NULL, NULL, NULL, NULL, NULL,
};

static const char* translate_ps2_input_to_name(int c)
{
  if (c == -1)
    return "scancode -1";
  if (c < 0 || c > 256)
    return "scancode out of range";
  if (scancode_table[c] == NULL)
    scancode_table[c] = g_strdup_printf ("scancode 0x%02x", c);
  return scancode_table[c];
}

static int key_to_scancode(const char* key)
{
  int i;
  if (strncasecmp (key, "scancode ", 9) == 0)
    {
      return strtol (key + 9, NULL, 16);
    }
  for (i = 0; i < 256; i++)
    if (scancode_table[i] != NULL && strcmp (scancode_table[i], key) == 0)
      return i;
  return -1;
}

static void gdam_ps2_keyboard_parse_byte(GdamPs2Keyboard* keyboard, int c)
{
  GdamPs2KeyboardEvent *event;
  if (c == 0xf0)
    {
      keyboard->next_is_release = TRUE;
      return;
    }

  if (!keyboard->next_is_release)
    {
      if (keyboard->state_table[c])
	{
	  return;
	}
      keyboard->state_table[c] = TRUE;
    }
  else
    {
      keyboard->state_table[c] = FALSE;
    }

  event = g_new0 (GdamPs2KeyboardEvent, 1);
  event->event.input_device = (GdamInputDevice *) (keyboard);
  event->event.event_type = keyboard->next_is_release
    ? gdam_ps2_keyboard_release_event_get_type ()
    : gdam_ps2_keyboard_press_event_get_type ();
  event->scancode = c;
  event->name = translate_ps2_input_to_name (c);

  g_return_if_fail (0 <= c && c < 256);
  keyboard->next_is_release = FALSE;
  gdam_input_trap_emit (NULL, (GdamInputEvent *) event);
}


static gboolean gdam_ps2_keyboard_parse_raw(GIOChannel*   io_channel,
                                            GIOCondition  condition,
					    gpointer      user_data)
{
  GdamPs2Keyboard *keyboard = GDAM_PS2_KEYBOARD (user_data);
  unsigned char keyboard_buf[1024];
  int num_read;
  int index;
  GIOError io_error;
  const char *filename = GDAM_INPUT_DEVICE (keyboard)->device_name;
  io_error = g_io_channel_read (io_channel,
				keyboard_buf,
				sizeof (keyboard_buf), &num_read);
  if (io_error == G_IO_ERROR_AGAIN)
    return TRUE;
  if (io_error != G_IO_ERROR_NONE)
    {
      g_warning ("error reading from ps2 device %s", filename);
      return FALSE;
    }
  if (num_read <= 0)
    {
      g_warning ("ps2 keyboard device closed...");
      return FALSE;
    }
  for (index = 0; index < num_read; index++)
    gdam_ps2_keyboard_parse_byte (keyboard, keyboard_buf[index]);
  return TRUE;
}


static void gdam_ps2_keyboard_class_init(GdamInputDeviceFileClass* klass)
{
  klass->parse_raw = gdam_ps2_keyboard_parse_raw;
}

GtkType gdam_ps2_keyboard_get_type()
{
  static GtkType ps2_keyboard_type = 0;
  if (!ps2_keyboard_type)
    {
      static const GtkTypeInfo ps2_keyboard_info = {
	"GdamPs2Keyboard",
	sizeof (GdamPs2Keyboard),
	sizeof (GdamPs2KeyboardClass),
	(GtkClassInitFunc) gdam_ps2_keyboard_class_init,
	(GtkObjectInitFunc) gdam_ps2_keyboard_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_INPUT_DEVICE_FILE;
      ps2_keyboard_type = gtk_type_unique (parent, &ps2_keyboard_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
      (void) gdam_ps2_keyboard_press_event_get_type ();
      (void) gdam_ps2_keyboard_release_event_get_type ();
    }
  return ps2_keyboard_type;
}

static char* gdam_ps2_release_event_stringify(GdamInputEvent* event)
{
  GdamPs2KeyboardEvent *keyboard_event = (GdamPs2KeyboardEvent *) event;
  return g_strdup_printf ("ps2 key release %s", keyboard_event->name);
}
static char* gdam_ps2_press_event_stringify(GdamInputEvent* event)
{
  GdamPs2KeyboardEvent *keyboard_event = (GdamPs2KeyboardEvent *) event;
  return g_strdup_printf ("ps2 key press %s %d", keyboard_event->name,
			  keyboard_event->scancode);
}
static unsigned gdam_ps2_event_hash(GdamInputEvent* event)
{
  return GPOINTER_TO_UINT (event->input_device)
    ^ (((GdamPs2KeyboardEvent *) event)->scancode << 16)
    ^ (event->event_type << 8);
}
static GHashTable* gdam_ps2_event_to_keyvalues(GdamInputEvent* event)
{
  GHashTable *table = g_hash_table_new (g_str_hash, g_str_equal);
  const char *device_name;
  GdamPs2KeyboardEvent *ps2_event = (GdamPs2KeyboardEvent *) event;
  table = g_hash_table_new (g_str_hash, g_str_equal);

  device_name = event->input_device->device_name;
  gdam_input_store_string (table, "device", device_name);

  gdam_input_store_string (table, "key", ps2_event->name);
  return table;
}

static GdamInputEvent* gdam_ps2_event_from_keyvalues(int type,
                                                     GHashTable* table)
{

  GdamPs2KeyboardEvent *event;
  char *device_name;
  char *key;
  event = g_new (GdamPs2KeyboardEvent, 1);
  event->event.event_type = type;
  event->event.input_id = 0;
  device_name = g_hash_table_lookup (table, "device");
  gdam_input_event_parse_device ((GdamInputEvent *) event, device_name);
  key = gdam_input_fetch_string (table, "key");
  if (key == NULL)
    {
      g_warning ("no `key' in ps2 event");
      g_free (event);
      return NULL;
    }
  event->scancode = key_to_scancode (key);
  event->name = translate_ps2_input_to_name (event->scancode);
  return (GdamInputEvent *) event;
}
static gboolean gdam_ps2_event_equal(GdamInputEvent* event1,
                                     GdamInputEvent* event2)
{
  GdamPs2KeyboardEvent *kevent1 = (GdamPs2KeyboardEvent *) event1;
  GdamPs2KeyboardEvent *kevent2 = (GdamPs2KeyboardEvent *) event2;
  return event1->input_device == event2->input_device
    && event1->event_type == event2->event_type
    && kevent1->scancode == kevent2->scancode;
}

static GdamInputEvent* gdam_ps2_event_canonicalize(GdamInputEvent* event)
{
  return g_memdup (event, sizeof (GdamPs2KeyboardEvent));
}

int gdam_ps2_keyboard_release_event_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamPs2KeyboardRelease",
    gdam_ps2_release_event_stringify,
    gdam_ps2_event_hash,
    gdam_ps2_event_equal,
    gdam_ps2_event_to_keyvalues,
    gdam_ps2_event_from_keyvalues,
    gdam_ps2_event_canonicalize,
    NULL
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_get_type (), GDAM_TYPE_PS2_KEYBOARD, &type_info);
  return type;
}
int gdam_ps2_keyboard_press_event_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamPs2KeyboardPress",
    gdam_ps2_press_event_stringify,
    gdam_ps2_event_hash,
    gdam_ps2_event_equal,
    gdam_ps2_event_to_keyvalues,
    gdam_ps2_event_from_keyvalues,
    gdam_ps2_event_canonicalize,
    NULL
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_get_type (), GDAM_TYPE_PS2_KEYBOARD, &type_info);
  return type;
}
