/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamrecorderfilter.h"
#include <component/gdambuffermanager.h>
#include <common/gdamplugin.h>
#include <client/gdamchannel.h>
#include <gtk/gtksignal.h>

enum {
	SET_RECORDING,
	SET_SAVE_END_BUFFER,
	NEW_LENGTH,
	LAST_SIGNAL
};
static guint recorder_signals[LAST_SIGNAL] = {0};

static void gdam_recorder_filter_real_set_recording      (GdamRecorderFilter*,
                                                          gboolean recording);
static void gdam_recorder_filter_real_set_save_end_buffer(GdamRecorderFilter*,
                                                          gboolean should_save);
static void gdam_recorder_filter_destroy                 (GtkObject*);
static int update_length_thread(gpointer user_data);
                                                         
static GtkObjectClass* parent_class;

static int          gdam_recorder_filter_append_start_request
                                                    (GdamFilter* filter,
					             int new_source_id,
						     GdamRequestBuilder*);
static void         gdam_recorder_filter_append_stop_request
                                                    (GdamFilter* filter,
						     GdamRequestBuilder*);

static void gdam_recorder_filter_init(GdamRecorderFilter* recorder)
{
  recorder->eof_description = g_strdup ("recorder source ended");
}

static void gdam_recorder_filter_class_init(GdamRecorderFilterClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);

  klass->set_recording = gdam_recorder_filter_real_set_recording;
  klass->set_save_end_buffer = gdam_recorder_filter_real_set_save_end_buffer;
  object_class->destroy = gdam_recorder_filter_destroy;
  GDAM_FILTER_CLASS (klass)->append_start_request
    = gdam_recorder_filter_append_start_request;
  GDAM_FILTER_CLASS (klass)->append_stop_request
    = gdam_recorder_filter_append_stop_request;

  recorder_signals[SET_RECORDING] =
    gtk_signal_new ("set_recording",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamRecorderFilterClass,
				       set_recording),
		    gtk_marshal_NONE__BOOL, GTK_TYPE_NONE, 1, GTK_TYPE_BOOL);
  recorder_signals[SET_SAVE_END_BUFFER] =
    gtk_signal_new ("set_save_end_buffer",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamRecorderFilterClass,
				       set_save_end_buffer),
		    gtk_marshal_NONE__BOOL, GTK_TYPE_NONE, 1, GTK_TYPE_BOOL);
  recorder_signals[NEW_LENGTH] =
    gtk_signal_new ("new_length",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamRecorderFilterClass,
				       new_length),
		    gtk_marshal_NONE__INT, GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  gtk_object_class_add_signals (object_class, recorder_signals, LAST_SIGNAL);
}

GtkType gdam_recorder_filter_get_type()
{
  static GtkType recorder_filter_type = 0;
  if (!recorder_filter_type)
    {
      static const GtkTypeInfo recorder_filter_info = {
	"GdamRecorderFilter",
	sizeof (GdamRecorderFilter),
	sizeof (GdamRecorderFilterClass),
	(GtkClassInitFunc) gdam_recorder_filter_class_init,
	(GtkObjectInitFunc) gdam_recorder_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      recorder_filter_type = gtk_type_unique (parent, &recorder_filter_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return recorder_filter_type;
}

/*=====================================================================
 * Is recording.
 */

static void
gdam_recorder_filter_real_set_recording(GdamRecorderFilter* filter,
                                        gboolean            is_recording)
{

  GdamArg arg;
  GdamModel *component = GDAM_MODEL (filter);
  int is_rec = is_recording ? 1 : 0;

  if (is_rec == filter->is_recording)
    return;

  arg.name = "recording";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = is_rec;
  filter->is_recording = is_rec;
  gdam_channel_append_request (component->channel,
			       gdam_request_set_args (component->source_id,
						      1,
						      &arg,
						      NULL, NULL, NULL));
}

void gdam_recorder_filter_set_recording(GdamRecorderFilter* recorder,
                                        gboolean            is_recording)
{
  GtkObject *this_object = GTK_OBJECT (recorder);

  if (gtk_signal_n_emissions (this_object, recorder_signals[SET_RECORDING]))
    return;

  gtk_signal_emit (this_object,
		   recorder_signals[SET_RECORDING], is_recording ? 1 : 0);
}

/*=====================================================================
 * Save end buffer.
 */
typedef struct _EofEventData EofEventData;
struct _EofEventData {
	char* description;
	GdamChannel* channel;
};
static void made_eof_buffer(int id, gpointer user_data)
{
  GdamBufferManagerInfo *info;
  EofEventData *eof_data = user_data;
  if (id == 0)
    {
      g_free (eof_data->description);
      g_free (eof_data);
      g_warning ("error making end-buffer");
      return;
    }

  info = g_new0 (GdamBufferManagerInfo, 1);
  info->description = eof_data->description;
  info->buffer_id = id;
  info->channel = eof_data->channel;
  gdam_buffer_manager_add_buffer (info);
  g_free (eof_data);
}

static void got_eof_event(int num_args, GdamArg* args, gpointer user_data)
{
  GdamRecorderFilter *recorder = GDAM_RECORDER_FILTER (user_data);
  EofEventData *eof_data;
  g_return_if_fail (num_args == 1);

  eof_data = g_new (EofEventData, 1);
  eof_data->description = g_strdup (recorder->eof_description);
  eof_data->channel = GDAM_MODEL (recorder)->channel;

  /* XXX: if the recorder is already destroyed,
     then this wont be cleaned up. */
  recorder->eof_buffer_connect_id = GDAM_ARG_INT (args);
  gdam_channel_register_handler (GDAM_MODEL (recorder)->channel,
				 gdam_channel_int_handler_new
				 (recorder->eof_buffer_connect_id, TRUE,
				  made_eof_buffer, eof_data));
  gtk_object_unref (GTK_OBJECT (user_data));
}

static void
gdam_recorder_filter_real_set_save_end_buffer(GdamRecorderFilter* recorder,
                                              gboolean            should_save)
{
  GdamRequestBuilder *builder = gdam_request_builder_new ();
  GdamArg arg;
  int save = should_save ? 1 : 0;
  int source_id = GDAM_MODEL (recorder)->source_id;

  if (save == recorder->make_terminal)
    return;


  if (recorder->make_terminal)
    {
      int connect_id = recorder->eof_buffer_connect_id;
      if (connect_id != 0)
	gdam_request_builder_append_generic (builder,
					     /* XXX: leaves description as memory leak. */
					     gdam_request_unsubscribe
					     (connect_id, NULL, NULL, NULL));
      else
	{
	  /* TODO: make sure this can never happen... */
	  g_warning ("error disconnecting old eof event...");
	}
    }

  recorder->make_terminal = save;

  if (source_id == 0)
    return;

  gdam_request_builder_append_generic (builder,
				       gdam_request_set_args (source_id,
							      1,
							      &arg,
							      NULL,
							      NULL, NULL));

  if (save)
    {
      const char *eof_event = "eof_event";

      gtk_object_unref (GTK_OBJECT (recorder));

      gdam_request_builder_append_generic (builder,
					   gdam_request_get_args (source_id,
								  1,
								  &eof_event,
								  got_eof_event,
								  (GDestroyNotify)
								  gtk_object_unref,
								  builder));
    }
}



static int
gdam_recorder_filter_append_start_request(GdamFilter* filter,
					  int new_source_id,
					  GdamRequestBuilder* builder)
{
  GdamRecorderFilter *recorder = GDAM_RECORDER_FILTER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  g_return_val_if_fail (recorder->new_length_thread == 0, 0);
  model->source_id = gdam_channel_new_local_id (model->channel);
  recorder->new_length_thread = g_timeout_add (50,
					       update_length_thread, filter);
  return gdam_request_builder_append_create (builder,
					     gdam_request_create_source
					     ("GdasRecorder",
					      model->source_id,
					      gdam_model_got_subsource,
					      gdam_model_didnt_get_subsource,
					      recorder, 2, "recording",
					      GDAM_ARG_TYPE_INT,
					      recorder->is_recording,
					      "filtered", GDAM_ARG_TYPE_INT,
					      new_source_id));
}
static void gdam_recorder_filter_append_stop_request (GdamFilter* filter,
						     GdamRequestBuilder* b)
{
  GdamRecorderFilter *recorder = GDAM_RECORDER_FILTER (filter);
  if (recorder->new_length_thread)
    {
      g_source_remove (recorder->new_length_thread);
      recorder->new_length_thread = 0;
    }
}
	

/*=====================================================================
 * Length updating thread.
 */
static void got_new_length(int num_args, GdamArg* args, gpointer user_data)
{
  GdamRecorderFilter *recorder = GDAM_RECORDER_FILTER (user_data);
  g_return_if_fail (num_args == 1);
  g_return_if_fail (args->type == GDAM_ARG_TYPE_INT);
  g_return_if_fail (strcmp (args->name, "length") == 0);

  gdam_recorder_filter_new_length (recorder, GDAM_ARG_INT (args));
}

static int update_length_thread(gpointer user_data)
{
  GdamRecorderFilter *recorder = GDAM_RECORDER_FILTER (user_data);
  GdamModel *component = (GdamModel *) recorder;
  const char *length = "length";
  gtk_object_ref (GTK_OBJECT (user_data));
  gdam_channel_append_request (component->channel,
			       gdam_request_get_args (component->source_id,
						      1,
						      &length,
						      got_new_length,
						      (GDestroyNotify)
						      gtk_object_unref,
						      recorder));
  return TRUE;
}

/*=====================================================================
 * Buffer capture.
 */

static void got_buffer(int num_args,
                       GdamArg* args,
		       gpointer recorder)
{
  gboolean drain;
  GdamBufferManagerInfo *info;
  int buffer_id;
  g_return_if_fail (num_args == 1);
  drain = strcmp (args->name, "grab_buffer") == 0;
  buffer_id = GDAM_ARG_INT (args);

  info = g_new0 (GdamBufferManagerInfo, 1);
  info->channel = GDAM_MODEL (recorder)->channel;
  g_return_if_fail (info->channel != NULL);
  info->buffer_id = buffer_id;
  info->description = g_strdup_printf ("Recording %s",
				       drain ? "cut" : "copied");
  gdam_buffer_manager_add_buffer (info);
  gtk_object_unref (GTK_OBJECT (recorder));
}

static void got_buffer_failed(gpointer recorder)
{
  g_warning ("got buffer failed");
  gtk_object_unref (GTK_OBJECT (recorder));
}
                       

void gdam_recorder_filter_capture_buffer(GdamRecorderFilter* recorder,
                                         gboolean            drain)
{
  GdamModel *component = GDAM_MODEL (recorder);
  int source_id = component->source_id;
  const char *names;
  if (source_id == 0)
    {
      gdam_debug ("no recording source to capture from");
      return;
    }
  names = drain ? "grab_buffer" : "copy_buffer",
    gtk_object_ref (GTK_OBJECT (recorder));
  gdam_channel_append_request (component->channel,
			       gdam_request_get_args (source_id,
						      1,
						      &names,
						      got_buffer,
						      got_buffer_failed,
						      recorder));
}

/*=====================================================================
 * Cleanup.
 */
static void gdam_recorder_filter_destroy(GtkObject* object)
{
  GdamRecorderFilter *recorder = GDAM_RECORDER_FILTER (object);
  if (recorder->new_length_thread != 0)
    {
      g_source_remove (recorder->new_length_thread);
      recorder->new_length_thread = 0;
    }

  g_free (recorder->eof_description);
  recorder->eof_description = NULL;

  (*parent_class->destroy) (object);
}

void gdam_recorder_filter_new_length(GdamRecorderFilter* recorder,
                                     int                 new_length)
{
  GtkObject *this_object = GTK_OBJECT (recorder);

  if (gtk_signal_n_emissions (this_object, recorder_signals[NEW_LENGTH]))
    return;

  gtk_signal_emit (this_object, recorder_signals[NEW_LENGTH], new_length);
}

