/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamxmmsfilter.h"
#include <common/gdam-protocol.h>
#include <gtk/gtksignal.h>

enum {
	CONNECTED,
	DIDNT_CONNECT,
	LAST_SIGNAL
};
static guint signal_ids[LAST_SIGNAL] = { 0 };

static GtkObjectClass* parent_class = NULL;

static void on_connect_success(int       num_args,
                               GdamArg*  args,
			       gpointer  data)
{
  GdamXmmsFilter *filter;
  filter = GDAM_XMMS_FILTER (data);
  filter->is_connected = 1;
  filter->is_connecting = 0;
  gtk_signal_emit (GTK_OBJECT (filter), signal_ids[CONNECTED]);
  gtk_object_unref (GTK_OBJECT (filter));
}
static void on_connect_failure(gpointer  data)
{
  GdamXmmsFilter *filter;
  filter = GDAM_XMMS_FILTER (data);
  gtk_signal_emit (GTK_OBJECT (filter), signal_ids[DIDNT_CONNECT]);
  g_free (filter->path);
  filter->path = NULL;
  filter->is_connecting = 0;
  gtk_object_unref (GTK_OBJECT (filter));
}


static int  gdam_xmms_filter_append_start_request
                                        (GdamFilter*         filter,
                                         int                 local_src_id,
                                         GdamRequestBuilder* builder)
{
  GdamArg arg;
  GdamXmmsFilter *xmms;
  GdamModel *model = GDAM_MODEL (filter);
  int rv;
  xmms = GDAM_XMMS_FILTER (model);
  arg.name = "filtered";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = local_src_id;
  gtk_object_ref (GTK_OBJECT (model));
  model->source_id = gdam_channel_new_local_id (model->channel);
  rv = gdam_request_builder_append_create (builder,
					   gdam_request_create_source_v
					   ("GdasXmmsVisFilter",
					    model->source_id, 1, &arg,
					    gdam_model_got_subsource,
					    gdam_model_didnt_get_subsource,
					    model));
  if (xmms->path != NULL)
    {
      GdamRequest *request;
      gtk_object_ref (GTK_OBJECT (model));

      arg.name = "path";
      arg.type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (&arg) = xmms->path;
      request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					    rv,
					    "connect",
					    1,
					    &arg,
					    on_connect_success,
					    on_connect_failure, model);
      xmms->is_connecting = 1;
      gdam_request_builder_append_generic (builder, request);
    }
  return rv;
}


static void gdam_xmms_filter_init(GdamXmmsFilter* xmms_filter)
{
}
static void gdam_xmms_filter_finalize(GtkObject* object)
{
  GdamXmmsFilter *filter;
  filter = GDAM_XMMS_FILTER (object);
  g_free (filter->path);
  filter->path = NULL;
  if (parent_class->finalize != NULL)
    (*parent_class->finalize) (object);
}
static void gdam_xmms_filter_class_init(GdamXmmsFilterClass* xmms_filter_class)
{
  GtkObjectClass *object_class;
  GdamFilterClass *filter_class;
  object_class = GTK_OBJECT_CLASS (xmms_filter_class);
  filter_class = GDAM_FILTER_CLASS (xmms_filter_class);
  signal_ids[CONNECTED]
    = gtk_signal_new ("connected",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamXmmsFilterClass,
					 connected),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  signal_ids[DIDNT_CONNECT]
    = gtk_signal_new ("didnt_connect",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamXmmsFilterClass,
					 didnt_connect),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, signal_ids, LAST_SIGNAL);

  filter_class->append_start_request = gdam_xmms_filter_append_start_request;
}

GtkType gdam_xmms_filter_get_type()
{
  static GtkType xmms_filter_type = 0;
  if (!xmms_filter_type)
    {
      static const GtkTypeInfo xmms_filter_info = {
	"GdamXmmsFilter",
	sizeof (GdamXmmsFilter),
	sizeof (GdamXmmsFilterClass),
	(GtkClassInitFunc) gdam_xmms_filter_class_init,
	(GtkObjectInitFunc) gdam_xmms_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      xmms_filter_type = gtk_type_unique (parent, &xmms_filter_info);
    }
  return xmms_filter_type;
}
void gdam_xmms_filter_connect(GdamXmmsFilter*    xmms,
                              const char*        path)
{
  GdamModel *model;
  GdamArg arg;
  GdamRequest *request;
  model = GDAM_MODEL (xmms);
  if (xmms->is_connecting || xmms->is_connected)
    return;
  if (model->source_id == 0)
    return;

  arg.name = "path";
  arg.type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (&arg) = (char *) path;

  gtk_object_ref (GTK_OBJECT (model));
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					model->source_id,
					"connect",
					1,
					&arg,
					on_connect_success,
					on_connect_failure, xmms);
  xmms->is_connecting = 1;
  xmms->path = g_strdup (path);
  gdam_channel_append_request (model->channel, request);
}

