#include <common/gdamautoconf.h>

#if HAVE_LADSPA_H
#include <ladspa.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>
#include <gmodule.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>

static void duplicate_node(xmlNode* parent, xmlNode* orig)
{
  xmlNode *copy;
  copy = xmlCopyNode (orig, TRUE);
  xmlAddChild (parent, copy);
}

static gboolean ladspa_plugin_to_xml(const char*       plugin_module_name,
                                     xmlNode*          parent,
				     GHashTable*       skip_list)
{
  GModule *module;
  LADSPA_Descriptor_Function function;
  int i;
  char *final_dot;

  final_dot = strrchr ((char *) plugin_module_name, '.');
  if (final_dot == NULL || strcmp (final_dot, ".so") != 0)
    return TRUE;

  module = g_module_open (plugin_module_name, 0);
  if (module == NULL)
    {
      char *err_str;
      err_str = g_module_error ();
      g_warning ("couldn't open module: %s", err_str);
      g_free (err_str);
      return FALSE;
    }

  if (!g_module_symbol (module, "ladspa_descriptor", (gpointer *) & function))
    {
      char *err_str;
      err_str = g_module_error ();
      g_warning ("couldn't find `ladspa_descriptor': %s", err_str);
      g_free (err_str);
      return FALSE;
    }

  for (i = 0; TRUE; i++)
    {
      const LADSPA_Descriptor *descriptor;
      char tmp_buf[255];
      int j;
      int num_input_lines = 0;
      int num_output_lines = 0;
      xmlNode *top;

      descriptor = (*function) (i);
      if (descriptor == NULL)
	break;

      if (skip_list != NULL
	  && g_hash_table_lookup (skip_list, descriptor->Name) != NULL)
	{
	  xmlNode *old;
	  old = g_hash_table_lookup (skip_list, descriptor->Name);
	  duplicate_node (parent, old);
	  continue;
	}
      top = xmlNewNode (NULL, "plugin");
      xmlAddChild (parent, top);

      xmlNewChild (top, NULL, "name", descriptor->Name);
      xmlNewChild (top, NULL, "author", descriptor->Maker);

      snprintf (tmp_buf, sizeof (tmp_buf), "%ld", descriptor->UniqueID);
      xmlNewChild (top, NULL, "unique_id", tmp_buf);

      xmlNewChild (top, NULL, "module", plugin_module_name);
      snprintf (tmp_buf, sizeof (tmp_buf), "%d", i);
      xmlNewChild (top, NULL, "module_index", tmp_buf);

      for (j = 0; j < (int) descriptor->PortCount; j++)
	{
	  const LADSPA_PortRangeHint *range_hint;
	  const char *name = descriptor->PortNames[j];
	  int range_desc;
	  xmlNode *sub;
	  gboolean is_control, is_audio_data;
	  gboolean is_input, is_output;
	  LADSPA_PortDescriptor port_descriptor;

	  port_descriptor = descriptor->PortDescriptors[j];
	  is_input = LADSPA_IS_PORT_INPUT (port_descriptor);
	  is_output = LADSPA_IS_PORT_OUTPUT (port_descriptor);
	  is_control = LADSPA_IS_PORT_CONTROL (port_descriptor);
	  is_audio_data = LADSPA_IS_PORT_AUDIO (port_descriptor);
	  range_hint = descriptor->PortRangeHints + j;
	  range_desc = range_hint->HintDescriptor;

	  if (is_input && is_output)
	    {
	      g_warning ("%s: port %d was both input + output",
			 plugin_module_name, j);
	      break;
	    }
	  if (!is_input && !is_output)
	    {
	      g_warning ("%s: port %d was not input or output",
			 plugin_module_name, j);
	      break;
	    }
	  if (is_audio_data && is_control)
	    {
	      g_warning ("%s: port %d was both audio and"
			 " control", plugin_module_name, j);
	      break;
	    }
	  if (!is_audio_data && !is_control)
	    {
	      g_warning ("%s: port %d was neither audio nor"
			 " control", plugin_module_name, j);
	      break;
	    }
	  if (is_audio_data)
	    {
	      if (is_input)
		num_input_lines++;
	      else
		num_output_lines++;
	      continue;
	    }
	  sub = xmlNewNode (NULL, is_input ? "control" : "meter");
	  xmlAddChild (top, sub);

	  xmlNewChild (sub, NULL, "name", name);
	  if (LADSPA_IS_HINT_LOGARITHMIC (range_desc))
	    xmlNewChild (sub, NULL, "logarithmic", "1");
	  if (LADSPA_IS_HINT_BOUNDED_ABOVE (range_desc))
	    {
	      snprintf (tmp_buf, sizeof (tmp_buf),
			"%6.6g", range_hint->UpperBound);
	      xmlNewChild (sub, NULL, "max_value", tmp_buf);
	    }
	  if (LADSPA_IS_HINT_BOUNDED_BELOW (range_desc))
	    {
	      snprintf (tmp_buf, sizeof (tmp_buf),
			"%6.6g", range_hint->LowerBound);
	      xmlNewChild (sub, NULL, "min_value", tmp_buf);
	    }
	  if (LADSPA_IS_HINT_INTEGER (range_desc))
	    xmlNewChild (sub, NULL, "integer", "1");
	  if (LADSPA_IS_HINT_SAMPLE_RATE (range_desc))
	    {
	      xmlNewChild (sub, NULL, "sample_rate_ratio", "1");
	    }
	}
      snprintf (tmp_buf, sizeof (tmp_buf), "%d", num_input_lines);
      xmlNewChild (top, NULL, "num_inputs", tmp_buf);
      snprintf (tmp_buf, sizeof (tmp_buf), "%d", num_output_lines);
      xmlNewChild (top, NULL, "num_outputs", tmp_buf);
    }
  g_module_close (module);
  return TRUE;
}
static void usage(const char* prog_name)
{
  fprintf (stderr,
	   "usage: %s [--append] output_filename.xml\n\n"
	   "Scans your LADSPA plugins and produces a report file that\n"
	   "the GDAM client can use.\n", prog_name);
  exit (1);
}

static void scan_dir(const char* directory, xmlNode* parent, GHashTable* skip)
{
  DIR *dir;
  struct dirent *entry;
  dir = opendir (directory);
  if (dir == NULL)
    {
      /*g_warning("couldn't open directory %s", directory); */
      return;
    }
  while ((entry = readdir (dir)) != NULL)
    {
      char *extension;
      char *path;
      extension = strrchr (entry->d_name, '.');
      if (extension == NULL && strcmp (extension, ".so") != 0)
	continue;
      path = g_strconcat (directory, "/", entry->d_name, NULL);
      if (!ladspa_plugin_to_xml (path, parent, skip))
	g_warning ("error processing %s", path);
      g_free (path);
    }
  closedir (dir);

}
static const char* find_name(xmlNode* node)
{
  xmlNode *at;
  for (at = node->childs; at != NULL; at = at->next)
    {
      if (strcasecmp (at->name, "name") == 0)
	{
	  return xmlNodeGetContent (at);
	}
    }
  return NULL;
}

int main(int argc, char** argv)
{
  int i;
  const char *filename = NULL;
  xmlNode *top_node;
  xmlDoc *doc;
  gboolean append = FALSE;
  GHashTable *skip_list = NULL;
  for (i = 1; i < argc; i++)
    {
      char *arg = argv[i];
      if (strcmp (arg, "--help") == 0
	  || strcmp (arg, "-help") == 0 || strcmp (arg, "-h") == 0)
	usage (argv[0]);
      if (strcmp (arg, "--append") == 0)
	{
	  append = TRUE;
	  continue;
	}
      if (!filename)
	{
	  filename = arg;
	  continue;
	}
      g_error ("unrecognized option: %s", arg);
    }

  if (filename == NULL)
    usage (argv[0]);

  doc = xmlParseFile (filename);
  if (append && doc)
    {
      xmlNode *at;
      top_node = doc->root;
      skip_list = g_hash_table_new (g_str_hash, g_str_equal);
      for (at = top_node->childs; at != NULL; at = at->next)
	{
	  const char *name;
	  name = find_name (at);
	  if (name != NULL && g_hash_table_lookup (skip_list, name) == NULL)
	    {
	      g_hash_table_insert (skip_list, (char *) name, at);
	    }
	}
    }

  top_node = xmlNewNode (NULL, "LADSPA-Plugin-Descriptions");
  scan_dir ("/usr/local/lib/ladspa", top_node, skip_list);
  scan_dir ("/usr/lib/ladspa", top_node, skip_list);
  doc = xmlNewDoc (XML_DEFAULT_VERSION);
  doc->root = top_node;
  xmlSaveFile (filename, doc);
  return 0;
}
#else

#include <glib.h>

int main()
{
  g_error ("sorry -- w/o ladspa we shouldn't even compile this... "
	   "but we haven't fixed it yet");
  return 0;
}

#endif
