/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamcachebuffer.h"
#include "gdamconnection.h"
#include "gdamevent.h"
#include "gdamserverdebug.h"
#include "gdamdispatch.h"
#include <common/gdam-protocol.h>
#include <common/gdamsampleconverter.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <stdlib.h>

static GtkObjectClass* parent_class = NULL;

static void      gdam_cache_buffer_finished(GdamCacheBuffer*);
static void      gdam_cache_buffer_destroy(GtkObject*);
static gboolean  gdam_cache_buffer_prepare        (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data);
static gboolean  gdam_cache_buffer_check          (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);
static gboolean  gdam_cache_buffer_dispatch       (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);

static GSourceFuncs source_funcs = { 
	gdam_cache_buffer_prepare,
	gdam_cache_buffer_check,
	gdam_cache_buffer_dispatch,
	NULL
};

enum {
	FINISHED,
	LAST_SIGNAL
};

static guint cache_buffer_signals[LAST_SIGNAL];


static void gdam_cache_buffer_arg_get_length(GtkObject* object, GdamArg* arg)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  GDAM_ARG_INT64 (arg) = cache_buffer->buffer_length;
}

#if 0
static void gdam_cache_buffer_taper_region  (GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  int i;
  gint64 start = 0, length = -1;
  gboolean fade_in = TRUE;
  GdamFixed start_level, end_level;
  GdamSample *at;
  for (i = 0; i < num_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (strcmp (arg->name, "fade_in") == 0
	  && arg->type == GDAM_ARG_TYPE_INT)
	{
	  fade_in = (GDAM_ARG_INT (arg));
	  continue;
	}
      if (strcmp (arg->name, "start") == 0
	  && arg->type == GDAM_ARG_TYPE_INT64)
	{
	  start = GDAM_ARG_INT64 (arg);
	  continue;
	}
      if (strcmp (arg->name, "length") == 0
	  && arg->type == GDAM_ARG_TYPE_INT64)
	{
	  length = GDAM_ARG_INT64 (arg);
	  continue;
	}
      g_warning ("unknown type/arg %s", arg->name);
      return;
    }

  if (fade_in)
    {
      start_level = 0;
      delta_level = GdamFixed_ONE;
    }
  else
    {
      start_level = GdamFixed_ONE;
      delta_level = -GdamFixed_ONE;
    }

  at = cache_buffer->buffer + start;
  for (i = 0; i < length; i++)
    {
    volume = (GdamSampleProduct) start_level *i +}
}
#endif

static void gdam_cache_buffer_process_region(GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  int length = -1, start = 0, i;
  gboolean cut = FALSE, make_buffer = FALSE;
  GdamCacheBuffer *orig = GDAM_CACHE_BUFFER (object);
  if (orig->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "cut") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  cut = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "make_buffer") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  make_buffer = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_cut: %s",
		 args_in[i].name);
    }
  if (start < 0)
    {
      length += start;
      start = 0;
    }
  if (start + length > orig->buffer_length)
    length = orig->buffer_length - start;
  if (length <= 0)
    return;

  if (make_buffer)
    {
      GdamCacheBuffer *new_buffer;
      GdamSample *new_samples;
      *num_out = 1;
      *args_out = g_new (GdamArg, 1);
      new_samples = g_new (GdamSample, length);
      gdam_sample_copy (new_samples, orig->buffer + start, length);
      new_buffer = gdam_cache_buffer_memory (length, new_samples, g_free);
      (*args_out)[0].name = g_strdup ("buffer_id");
      (*args_out)[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT ((*args_out) + 0) = gdam_cache_buffer_get_id (new_buffer);
    }
  if (cut)
    {
      int new_length = orig->buffer_length - length;
      memmove (orig->buffer + start,
	       orig->buffer + start + length,
	       (new_length - start) * sizeof (GdamSample));
      gdam_debug ("cut: removed %d: %d => %d", length,
		  orig->buffer_length, new_length);
      orig->buffer_length = new_length;
      orig->buffer = g_renew (GdamSample, orig->buffer, new_length);
    }
}
static void gdam_cache_buffer_replace_region(GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  int length = 0, start = 0, i, new_length;
  int insert_length = -1, insert_start = 0;
  int insert_buffer_id = 0;
  GdamCacheBuffer *insert_buffer;
  GdamCacheBuffer *orig = GDAM_CACHE_BUFFER (object);
  int post_length;
  if (orig->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_buffer_id") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_buffer_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_cut: %s",
		 args_in[i].name);
    }
  if (start < 0)
    {
      length += start;
      start = 0;
    }
  if (start + length > orig->buffer_length)
    length = orig->buffer_length - start;
  if (length <= 0)
    return;
  if (insert_buffer_id == 0)
    return;
  insert_buffer = gdam_cache_buffer_get_by_id (insert_buffer_id);
  if (insert_buffer == NULL)
    {
      gdam_debug ("buffer to insert %d not found", insert_buffer_id);
      return;
    }
  if (insert_length == -1)
    insert_length = insert_buffer->buffer_length;
  if (insert_start < 0)
    {
      insert_length += insert_start;
      insert_start = 0;
    }
  if (insert_start + insert_length > insert_buffer->buffer_length)
    insert_length = insert_buffer->buffer_length - insert_start;
  if (insert_length < 0)
    return;

  post_length = orig->buffer_length - start - length;
  new_length = orig->buffer_length - length + insert_length;
  orig->buffer = g_renew (GdamSample, orig->buffer, new_length);
  memmove (orig->buffer + start + insert_length,
	   orig->buffer + start + length, post_length * sizeof (GdamSample));
  gdam_sample_copy (orig->buffer + start,
		    insert_buffer->buffer + insert_start, insert_length);
  orig->buffer_length = new_length;
}

static void gdam_cache_buffer_insert        (GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  GdamCacheBuffer *insert;
  GdamSample *at;
  const GdamSample *start_copy;
  int insertion_point = 0, multiplicity = 1, buffer_id = 0;
  int insert_offset = 0, insert_length = -1;
  int i;
  int new_length;
  int second_half_length;	/* length of original buffer after
				 * the insertion point. */

  if (cache_buffer->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "multiplicity") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  multiplicity = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_buffer_id") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  buffer_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insertion_point = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_offset") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_offset = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_length = GDAM_ARG_INT (args_in + i);
	  continue;
	}



      gdam_debug ("buffer_insert: unknown argument: %s", args_in[i].name);
    }

  if (buffer_id == 0
      || (insert = gdam_cache_buffer_get_by_id (buffer_id)) == NULL)
    {
      gdam_debug ("insert: no buffer id %d found", buffer_id);
      return;
    }

  if (insert_length == -1)
    insert_length = insert->buffer_length;
  if (insert_offset < 0)
    {
      insert_length += insert_offset;
      insert_offset = 0;
    }
  if (insert_offset + insert_length > insert->buffer_length)
    insert_length = insert->buffer_length - insert_offset;
  if (insert_length <= 0)
    return;

  if (insertion_point > cache_buffer->buffer_length)
    insertion_point = cache_buffer->buffer_length;
  if (insertion_point < 0)
    insertion_point = 0;

  new_length = cache_buffer->buffer_length + insert_length * multiplicity;
  second_half_length = cache_buffer->buffer_length - insertion_point;
  cache_buffer->buffer = g_renew (GdamSample,
				  cache_buffer->buffer, new_length);
  memmove (cache_buffer->buffer + new_length - second_half_length,
	   cache_buffer->buffer + insertion_point,
	   second_half_length * sizeof (GdamSample));
  at = cache_buffer->buffer + insertion_point;
  start_copy = insert->buffer + insert_offset;
  for (i = 0; i < multiplicity; i++)
    {
      gdam_sample_copy (at, start_copy, insert_length);
      at += insert_length;
    }
  cache_buffer->buffer_length = new_length;

}
static void gdam_cache_buffer_reverse_region(GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  int length = -1, start = 0, i;
  GdamCacheBuffer *cache = GDAM_CACHE_BUFFER (object);
  GdamSample *start_ptr, *end_ptr;
  if (cache->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_reverse: %s",
		 args_in[i].name);
    }
  if (start < 0)
    {
      length += start;
      start = 0;
    }
  if (start + length > cache->buffer_length)
    length = cache->buffer_length - start;
  if (length <= 0)
    return;
  start_ptr = cache->buffer + start;
  end_ptr = cache->buffer + start + length - 1;
  while (start_ptr < end_ptr)
    {
      GdamSample tmp = *start_ptr;
      *start_ptr = *end_ptr;
      *end_ptr = tmp;
      start_ptr++;
      end_ptr--;
    }
}

typedef struct _SavingThread SavingThread;

struct _SavingThread {
	char* filename;
	FILE* fp;
	GdamCacheBuffer* buffer;
	int at;
	int remaining;
};

static void free_saving_thread(SavingThread* saving_thread)
{
  if (saving_thread->fp)
    {
      FILE *fp = saving_thread->fp;
      saving_thread->fp = NULL;
      fclose (fp);
    }
  gtk_object_unref (GTK_OBJECT (saving_thread->buffer));
  g_free (saving_thread);
}
#define MAX_ATOMIC_SAVE		8192

static GdamSampleConverter wav_converter = NULL;

static gboolean save_more(gpointer user_data)
{
  SavingThread *saving_thread = user_data;
  int to_write = MAX_ATOMIC_SAVE;
  int amt;
  char *data;
  GdamCacheBuffer *buffer = saving_thread->buffer;
  if (saving_thread->at >= buffer->buffer_length)
    {
      free_saving_thread (saving_thread);
      return FALSE;
    }
  if (saving_thread->remaining >= 0 && to_write > saving_thread->remaining)
    to_write = saving_thread->remaining;
  if (to_write + saving_thread->at > buffer->buffer_length)
    to_write = buffer->buffer_length - saving_thread->at;
  if (to_write <= 0)
    {
      free_saving_thread (saving_thread);
      return FALSE;
    }
  data = alloca (4 * to_write);
  if (wav_converter == NULL)
    wav_converter = gdam_sample_converter_get
      (GDAM_SAMPLE_BITS,
       TRUE,
       (G_BYTE_ORDER == G_BIG_ENDIAN),
       16, TRUE, (G_BYTE_ORDER == G_BIG_ENDIAN));

  (*wav_converter) (data, buffer->buffer + saving_thread->at, to_write * 2);
  amt = fwrite (data, 4, to_write, saving_thread->fp);
  if (amt <= 0)
    {
      g_warning ("fwrite failed writing %s", saving_thread->filename);
      free_saving_thread (saving_thread);
      return FALSE;
    }
  saving_thread->at += amt;
  if (saving_thread->at == buffer->buffer_length)
    {
      free_saving_thread (saving_thread);
      return FALSE;
    }
  if (saving_thread->remaining >= 0)
    {
      if ((saving_thread->remaining -= amt) <= 0)
	{
	  free_saving_thread (saving_thread);
	  return FALSE;
	}
    }
  return TRUE;
}

#define WAV_RIFF   0x46464952	/* "RIFF" */
#define WAV_WAVE   0x45564157	/* "WAVE" */
#define WAV_LIST   0x5453494c	/* "LIST" */
#define WAV_INFO   0x4f464e49	/* "INFO" */
#define WAV_ISFT   0x54465349	/* "ISFT" */
#define WAV_ICRD   0x44524349	/* "ICRD" */
#define WAV_FMT    0x20746d66	/* "fmt " */
#define WAV_DATA   0x61746164	/* "data" */
#define WAV_MONO   0x00010001	/* 0x0001 0x0001 */
#define WAV_STEREO 0x00020001	/* 0x0001 0x0002 */

static gboolean write_wave_file_header(FILE* fp, int length)
{
  guint32 byte_length;
  guint32 h[11];
  guint32 fmt = (16 << 16) | 4;

  byte_length = length * 4;
  memcpy (h + 0, "RIFF", 4);
  h[1] = byte_length + 36;
  memcpy (h + 2, "WAVE", 4);
  memcpy (h + 3, "fmt ", 4);
  h[4] = GINT32_TO_LE (16);
  h[5] = GINT32_TO_LE (WAV_STEREO);
  h[6] = 44100;
  h[7] = 44100 * 4;
  h[8] = fmt;			/* TODO: see what this field should be */
  memcpy (h + 9, "data", 4);
  h[10] = byte_length;
  if (fwrite (h, 4, 11, fp) != 11 || ferror (fp))
    return FALSE;
  return TRUE;
}

static void gdam_cache_buffer_save_region   (GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  int length = -1, start = 0, i;
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  char *filename = NULL;
  SavingThread *thread_data;
  char *extension;
  FILE *fp;
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "filename") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_STRING)
	{
	  filename = g_strdup (GDAM_ARG_STRING (args_in + i));
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_save: %s",
		 args_in[i].name);
    }
  if (filename == NULL)
    {
      g_warning ("no filename was given to save_region");
      return;
    }
  extension = strrchr (filename, '.');
  if (extension == NULL || g_strcasecmp (extension, ".wav") != 0)
    {
      g_warning ("files must end in .wav (for security)");
      return;
    }
  fp = fopen (filename, "w");
  if (fp == NULL)
    {
      g_warning ("opening create %s", filename);
      return;
    }
  thread_data = g_new (SavingThread, 1);
  thread_data->filename = filename;
  thread_data->fp = fp;
  thread_data->buffer = cache_buffer;
  thread_data->at = start;
  thread_data->remaining = length;
  gtk_object_ref (GTK_OBJECT (cache_buffer));
  if (length < 0)
    length = cache_buffer->buffer_length - start;
  if (!write_wave_file_header (fp, length))
    {
      g_warning ("error writing wave file header");
      free_saving_thread (thread_data);
      return;
    }
  g_idle_add (save_more, thread_data);
}

static void gdam_cache_buffer_class_init(GdamCacheBufferClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;

  klass->finished = gdam_cache_buffer_finished;
  object_class->destroy = gdam_cache_buffer_destroy;

  cache_buffer_signals[FINISHED] =
    gtk_signal_new ("finished",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamCacheBufferClass,
				       finished),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, cache_buffer_signals,
				LAST_SIGNAL);

  gdam_arg_register (object_class,
		     "length",
		     GDAM_ARG_TYPE_INT64,
		     NULL, gdam_cache_buffer_arg_get_length, NULL, NULL, 0);
  gdam_arg_function_register (object_class,
			      "process_region",
			      gdam_cache_buffer_process_region);
  gdam_arg_function_register (object_class,
			      "replace_region",
			      gdam_cache_buffer_replace_region);
  gdam_arg_function_register (object_class,
			      "insert_region", gdam_cache_buffer_insert);
  gdam_arg_function_register (object_class,
			      "reverse_region",
			      gdam_cache_buffer_reverse_region);
  gdam_arg_function_register (object_class,
			      "save", gdam_cache_buffer_save_region);
}

static void gdam_cache_buffer_init(GdamCacheBuffer* cache)
{
  gdam_buffer_construct (&cache->loading_buffer);
  cache->buffer_length = -1;
}

GtkType gdam_cache_buffer_get_type ()
{
  static GtkType cache_buffer_type = 0;
  if (!cache_buffer_type)
    {
      static const GtkTypeInfo cache_buffer_info = {
	"GdamCacheBuffer",
	sizeof (GdamCacheBuffer),
	sizeof (GdamCacheBufferClass),
	(GtkClassInitFunc) gdam_cache_buffer_class_init,
	(GtkObjectInitFunc) gdam_cache_buffer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAM_TYPE_SERVER_OBJECT);
      cache_buffer_type = gtk_type_unique (GDAM_TYPE_SERVER_OBJECT,
					   &cache_buffer_info);
    }
  return cache_buffer_type;
}

GdamCacheBuffer* gdam_cache_buffer_new(GdasSource* to_drain)
{
  GdamCacheBuffer *buffer;
  GtkObject *to_drain_object = GTK_OBJECT (to_drain);
  g_return_val_if_fail (to_drain_object != NULL, NULL);

  buffer = (GdamCacheBuffer *) (gtk_type_new (GDAM_TYPE_CACHE_BUFFER));
  buffer->to_drain = to_drain;
  gtk_object_ref (to_drain_object);
  gtk_object_sink (to_drain_object);
  buffer->source_id = g_source_add (G_PRIORITY_LOW, FALSE, &source_funcs, NULL,	/* no source_data */
				    buffer, NULL);
  g_return_val_if_fail (buffer->source_id != 0, NULL);
  gdam_debug ("new cache buffer (source_id=%d) ref_count %d",
	      buffer->source_id, GTK_OBJECT (buffer)->ref_count);
  return buffer;
}

GdamCacheBuffer*  gdam_cache_buffer_memory    (int              buffer_length,
                                               GdamSample*      buffer,
					       GDestroyNotify   free_func)
{
  GdamCacheBuffer *cache_buffer;
  cache_buffer = (GdamCacheBuffer *) (gtk_type_new (GDAM_TYPE_CACHE_BUFFER));
  cache_buffer->buffer_length = buffer_length;
  cache_buffer->buffer = buffer;
  cache_buffer->free_func = free_func;
  return cache_buffer;
}

static void my_free_wrapper(gpointer ptr)
{
  if (ptr)
    free (ptr);
}


/* We use malloc and free here rather than g_malloc/g_free
 * b/c this is a rather likely place to run out of memory,
 * and b/c handling the out-of-memory condition in this case
 * is quite easy...
 */
static void gdam_cache_buffer_finished(GdamCacheBuffer* cache)
{
  int n;
  g_return_if_fail (cache->buffer_length == -1);
  cache->buffer_length = cache->loading_buffer.size / sizeof (GdamSample);
  cache->buffer = malloc (sizeof (GdamSample) * cache->buffer_length);
  cache->free_func = my_free_wrapper;
  if (cache->buffer == NULL)
    {
      cache->buffer_length = 0;
      g_message ("couldn't create buffer: out of memory");
      gdam_buffer_destruct (&cache->loading_buffer);
      return;
    }
  n = gdam_buffer_read (&cache->loading_buffer,
			cache->buffer,
			cache->buffer_length * sizeof (GdamSample));
  gdam_debug ("done caching: buffer id = %d, length = %d, ref_count = %d",
	      gdam_cache_buffer_get_id (cache),
	      cache->buffer_length, GTK_OBJECT (cache)->ref_count);
  g_return_if_fail (n == cache->buffer_length * (int) sizeof (GdamSample));
  gdam_buffer_destruct (&cache->loading_buffer);
}

static gboolean  gdam_cache_buffer_prepare        (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data)
{
  GdamCacheBuffer *cache_buffer = (GdamCacheBuffer *) user_data;
  gboolean rv = gdas_source_is_ready_to_read (cache_buffer->to_drain);
  return rv;
}
static gboolean  gdam_cache_buffer_check          (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  GdamCacheBuffer *cache_buffer = (GdamCacheBuffer *) user_data;
  gboolean rv = gdas_source_is_ready_to_read (cache_buffer->to_drain);
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdam_cache_buffer_check: is_ready=%d", rv);
  return rv;
}

static gboolean  gdam_cache_buffer_dispatch       (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  GdamSample tmp_buf[4096];
  int n;
  GdamCacheBuffer *cache = GDAM_CACHE_BUFFER (user_data);
  g_return_val_if_fail (cache->to_drain != NULL, TRUE);
  g_return_val_if_fail (cache->buffer_length == -1, TRUE);

  n = gdas_source_read (cache->to_drain, tmp_buf,
			sizeof (tmp_buf) / sizeof (GdamSample));
  if (n <= 0)
    {
      GtkObject *sub;

      gtk_signal_emit (GTK_OBJECT (cache), cache_buffer_signals[FINISHED]);
      cache->source_id = 0;
      sub = GTK_OBJECT (cache->to_drain);
      cache->to_drain = NULL;
      gtk_object_unref (sub);
      return FALSE;
    }
  gdam_buffer_append (&cache->loading_buffer,
		      tmp_buf, n * sizeof (GdamSample));
  return TRUE;
}

static void  gdam_cache_buffer_destroy(GtkObject* cache)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (cache);
  if (cache_buffer->source_id)
    {
      int source_id = cache_buffer->source_id;
      cache_buffer->source_id = 0;
      g_source_remove (source_id);
    }
  if (cache_buffer->to_drain != NULL)
    {
      GtkObject *sub;
      sub = GTK_OBJECT (cache_buffer->to_drain);
      cache_buffer->to_drain = NULL;
      gtk_object_unref (sub);
    }
  gdam_debug ("destroying buffer %d, size %d",
	      gdam_cache_buffer_get_id (cache_buffer),
	      cache_buffer->buffer_length);
  gdam_buffer_destruct (&cache_buffer->loading_buffer);
  if (cache_buffer->free_func)
    (*cache_buffer->free_func) (cache_buffer->buffer);
  (*parent_class->destroy) (cache);
}

GdamCacheBuffer*  gdam_cache_buffer_get_by_id (guint            buffer_id)
{
  GdamServerObject *object;
  object = gdam_dispatch_get_server_object (buffer_id);
  if (object == NULL)
    return NULL;
  gtk_object_unref ((GtkObject *) object);
  if (!GDAM_IS_CACHE_BUFFER (object))
    return NULL;
  return (GdamCacheBuffer *) object;
}
int               gdam_cache_buffer_get_id    (GdamCacheBuffer* buffer)
{
  GdamServerObject *object = GDAM_SERVER_OBJECT (buffer);
  return object->global_id;
}


static void marshal_cache_buffer_finish(GdamConnection*    registered,
                                        GtkObject*         actor,
			                int                connection_id,
			                gpointer           event_type_data,
			                gpointer           connection_data)
{
  GdamCacheBuffer *external = GDAM_CACHE_BUFFER (actor);
  GdamBuffer event;
  gdam_buffer_construct (&event);
  gdam_buffer_append_int (&event, GDAM_EVENT_CACHE_BUFFER_FINISH);
  gdam_buffer_append_int (&event, connection_id);
  gdam_buffer_append_int (&event, gdam_cache_buffer_get_id (external));
  gdam_connection_emit_event (registered, &event);
}

guint gdam_cache_buffer_get_done_event()
{
  static guint id = 0;
  if (id == 0)
    {
      id = gdam_event_register (GDAM_TYPE_CACHE_BUFFER,
				"finished",
				marshal_cache_buffer_finish, NULL, NULL);
    }
  return id;
}
