/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamconfigfile.h"
#include <stdio.h>
#include <ctype.h>
#include <common/gdammodule.h>
#include "gdamdispatch.h"
#include <errno.h>
#include <netdb.h>
#include <pwd.h>
#include <unistd.h>
#include <stdlib.h>
#include "gdamdevice.h"

const char*           gdam_username       = NULL;
const char*           gdam_groupname      = NULL;
const char*           gdam_pid_file       = NULL;
FILE*                 gdam_log_fp         = NULL;
gboolean              gdam_daemonize      = FALSE;
gboolean              gdam_bind_localhost = TRUE;
GHashTable*           gdas_microphones    = NULL;
int                   gdam_external_nice  = 0;

extern FILE*          gdam_debugging_fp;

/* The notable property of this config file is the it is always of
 * the form:
 *    Word "param1" "param2" ...
 * So this structure suffices: */
typedef struct _ConfigFileLine {
	const char* filename;
	int lineno;
	char* word;
	int num_params;
	char** params;
} ConfigFileLine;

static ConfigFileLine*      config_file_line_parse  (FILE* fp,
						     const char* filename,
                                                     int* lineno);
static void                 config_file_line_destroy(ConfigFileLine* line);
static void                 config_file_line_shift  (ConfigFileLine* line);
static char*                config_file_line_handle (ConfigFileLine* line);

/* Map from device_type => GdamDeviceOpenFunc. */
static GHashTable*          device_types = NULL;


static GHashTable* port_to_host_list = NULL;
static GHashTable* unix_to_uid_list = NULL;
GSList* gdam_config_lookup_tcp_auth(int port)
{
  if (port_to_host_list == NULL)
    return NULL;
  return g_hash_table_lookup (port_to_host_list, GINT_TO_POINTER (port));
}
GSList* gdam_config_lookup_unix_auth(const char* path)
{
  if (unix_to_uid_list == NULL)
    return NULL;
  return g_hash_table_lookup (unix_to_uid_list, path);
}

static GSList* make_ip_list(const char* host_list_str)
{
  char **strs = g_strsplit (host_list_str, ":", 0);
  char **at = strs;
  GSList *list = NULL;
  if (at == NULL)
    return NULL;
  while (*at)
    {
      const char *host = *at;
      int uid;
      struct hostent *entry;
      entry = gethostbyname (host);
      if (entry != NULL)
	{
	  char *ip = g_new (char, 4);
	  memcpy (ip, entry->h_addr_list[0], 4);
	  list = g_slist_prepend (list, ip);
	}
      else
	{
	  g_warning ("couldn't get ip number for %s", host);
	}
      at++;
    }
  g_strfreev (strs);
  return g_slist_reverse (list);
}

static GSList* make_uid_list(const char* uid_list_str)
{
  char **strs = g_strsplit (uid_list_str, ":", 0);
  char **at = strs;
  GSList *list = NULL;
  if (at == NULL)
    return NULL;
  while (*at)
    {
      const char *username = *at;
      int uid;
      if (isdigit (username[0]))
	{
	  uid = atoi (username);
	}
      else
	{
	  struct passwd *entry;
	  entry = getpwnam (username);
	  uid = entry->pw_uid;
	}
      list = g_slist_prepend (list, GINT_TO_POINTER (uid));
      at++;
    }
  g_strfreev (strs);
  return g_slist_reverse (list);
}


/* Handlers.  One for each Word allowed in the config file. */
typedef char*  (*ConfigFileLineHandler)(ConfigFileLine*);

static char* handle_DeviceType(ConfigFileLine* file_line)
{
  gpointer symbol;
  char *type;
  char *symbol_name;
  gboolean is_builtin, is_plugin;
  GModule *module;
  if (file_line->num_params != 2)
    {
      return g_strdup_printf ("%s:%d: expected two params, "
			      "\"module_type\" \"module_name\"",
			      file_line->filename, file_line->lineno);
    }
  gdam_debug ("handle_devicetype");
  if (!device_types)
    device_types = g_hash_table_new (g_str_hash, g_str_equal);

  type = file_line->params[1];
  if (g_hash_table_lookup (device_types, type))
    {
      return g_strdup_printf ("%s:%d: device type %s already "
			      "registered",
			      file_line->filename, file_line->lineno,
			      file_line->params[1]);
    }
  is_plugin = (strcmp (file_line->params[0], "plugin") == 0);
  is_builtin = (strcmp (file_line->params[0], "builtin") == 0);

  /* hack for user friendly upgrading */
  if (is_builtin && strcmp (type, "oss") == 0)
    {
      is_builtin = FALSE;
      is_plugin = TRUE;
      fprintf (stderr,
	       "OSS support has now been moved into a plugin.\n"
	       "This means your server config file (%s) requires\n"
	       "a minor update to line %d, which currently looks\n"
	       "like:\n"
	       "      DeviceType \"builtin\" \"oss\"\n"
	       "It should be changed to be:\n"
	       "      DeviceType \"plugin\" \"oss\"\n"
	       "Thanks.\n", file_line->filename, file_line->lineno);
    }
  if (is_builtin)
    {
      module = gdam_module_self ();
    }
  else if (is_plugin)
    {
      char *module_name;
      module_name = g_strconcat ("gdamdevice", type, (char *) NULL);
      module = gdam_module_by_name ("devices", module_name);
      if (module == NULL)
	{
	  return g_strdup_printf ("%s:%d: module %s not found",
				  file_line->filename, file_line->lineno,
				  module_name);
	}
    }
  else
    {
      return g_strdup_printf ("%s:%d: only the `builtin' are `plugin'"
			      "module types are allowed for now...",
			      file_line->filename, file_line->lineno);
    }

  symbol_name = g_strdup_printf ("gdam_device_%s_new", file_line->params[1]);
  if (!g_module_symbol (module, symbol_name, &symbol))
    {
      char *rv = g_strdup_printf ("%s:%d: devicetype %s (fct %s) "
				  "was not found",
				  file_line->filename, file_line->lineno,
				  file_line->params[1], symbol_name);
      g_free (symbol_name);
      return rv;
    }

  /* Although we never cast it from a gpointer here,
     `symbol' is actually a function pointer of type
     GdamDeviceOpenFunc. */

  g_hash_table_insert (device_types, g_strdup (type), symbol);
  return NULL;
}

static char* handle_Device(ConfigFileLine* file_line)
{
  gpointer symbol;
  char **params = file_line->params;
  GdamDeviceOpenFunc creator;
  if (file_line->num_params < 2 || file_line->num_params > 3)
    {
      return g_strdup_printf ("%s:%d: expected two or three params, "
			      "\"module_name\" \"device_name\""
			      " [\"device_options\"]",
			      file_line->filename, file_line->lineno);
    }
  gdam_debug ("handle_device");

  if (!device_types)
    {
      return g_strdup_printf ("%s:%d: you should have at least "
			      "one DeviceType line before a "
			      "Device line",
			      file_line->filename, file_line->lineno);
    }

  if (device_types == NULL
      || (symbol = g_hash_table_lookup (device_types, params[0])) == NULL)
    return g_strdup_printf ("%s:%d: Device: no module of name `%s'",
			    file_line->filename, file_line->lineno,
			    params[0]);
  creator = (GdamDeviceOpenFunc) symbol;
  gdam_debug ("registering device %s", params[1]);
  gdam_device_register (params[1],
			creator,
			NULL,
			file_line->num_params >= 3 ? params[2] : "",
			gdam_acl_new (TRUE));
  return NULL;
}

static char* handle_DeviceAlias(ConfigFileLine* file_line)
{
  g_warning("DeviceAlias is deprecated, remove from config file.");
  return NULL;
}


static char* handle_User(ConfigFileLine* file_line)
{
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: User: must have exactly "
			    "one parameter, the username",
			    file_line->filename, file_line->lineno);
  gdam_username = g_strdup (file_line->params[0]);
  return NULL;
}
static char* handle_Group(ConfigFileLine* file_line)
{
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: Group: must have exactly "
			    "one parameter, the groupname",
			    file_line->filename, file_line->lineno);
  gdam_groupname = g_strdup (file_line->params[0]);
  return NULL;
}
static char* handle_PidFile(ConfigFileLine* file_line)
{
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: PidFile: must have exactly "
			    "one parameter, the pid-file's filename",
			    file_line->filename, file_line->lineno);
  gdam_pid_file = g_strdup (file_line->params[0]);
  return NULL;
}
static char* handle_LogFile(ConfigFileLine* file_line)
{
  const char *logfilename;
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: LogFile: must have exactly "
			    "one parameter, the log's filename",
			    file_line->filename, file_line->lineno);
  logfilename = file_line->params[0];
  if (gdam_log_fp != NULL)
    return g_strdup_printf ("%s:%d: LogFile: tried to open two "
			    "log files",
			    file_line->filename, file_line->lineno);
  gdam_log_fp = fopen (logfilename, "a");
  if (gdam_log_fp == NULL)
    return g_strdup_printf ("%s:%d: LogFile: couldn't open %s: %s",
			    file_line->filename, file_line->lineno,
			    logfilename, strerror (errno));
  gdam_debugging_fp = gdam_log_fp;
  return NULL;
}
static char* handle_DispatchLog(ConfigFileLine* file_line)
{
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: DispatchLog: must have exactly "
			    "one parameter, the log's filename",
			    file_line->filename, file_line->lineno);
  gdam_dispatch_set_log (file_line->params[0]);
  return NULL;
}
static char* handle_Microphone(ConfigFileLine* file_line)
{
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: Microphone: must have exactly "
			    "one parameter, the device's filename",
			    file_line->filename, file_line->lineno);
  if (gdas_microphones == NULL)
    gdas_microphones = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (gdas_microphones,
		       g_strdup (file_line->params[0]), GINT_TO_POINTER (1));
  return NULL;
}
static char* handle_ExternalPriorityDrop(ConfigFileLine* file_line)
{
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: ExternalPriorityDrop: "
			    "must have exactly "
			    "one parameter, the device's filename",
			    file_line->filename, file_line->lineno);
  gdam_external_nice = atoi (file_line->params[0]);
  return NULL;
}

static gboolean parse_bool(const char* string, gboolean* rv)
{
  if (strcasecmp (string, "yes") == 0
      || strcasecmp (string, "y") == 0
      || strcasecmp (string, "true") == 0
      || strcasecmp (string, "t") == 0 || strcasecmp (string, "1") == 0)
    *rv = TRUE;
  else
    if (strcasecmp (string, "no") == 0
	|| strcasecmp (string, "n") == 0
	|| strcasecmp (string, "false") == 0
	|| strcasecmp (string, "f") == 0 || strcasecmp (string, "0") == 0)
    *rv = FALSE;
  else
    return FALSE;
  return TRUE;
}
static char* handle_Daemon(ConfigFileLine* file_line)
{
  const char *daemonize;
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: Daemon: must have exactly "
			    "one parameter, whether to daemonize",
			    file_line->filename, file_line->lineno);
  daemonize = file_line->params[0];
  if (!parse_bool (daemonize, &gdam_daemonize))
    {
      return g_strdup_printf ("%s:%d: Daemon: unrecognized boolean: %s",
			      file_line->filename, file_line->lineno,
			      daemonize);
    }
  return NULL;
}
static char* handle_Debug(ConfigFileLine* file_line)
{
  const char *debug;
  if (file_line->num_params != 1)
    return g_strdup_printf ("%s:%d: Debug: must have exactly "
			    "one boolean parameter",
			    file_line->filename, file_line->lineno);
  debug = file_line->params[0];
  if (!parse_bool (debug, &gdam_debugging_on))
    {
      return g_strdup_printf ("%s:%d: Debug: unrecognized boolean: %s",
			      file_line->filename, file_line->lineno, debug);
    }
  return NULL;
}
static char* handle_BindLocalhost(ConfigFileLine* file_line)
{
  gboolean bind_localhost = TRUE;
  if (file_line->num_params >= 1)
    {
      if (!parse_bool (file_line->params[0], &bind_localhost))
	{
	  return g_strdup_printf ("%s:%d: BindLocalhost: "
				  "unrecognized boolean: %s",
				  file_line->filename, file_line->lineno,
				  file_line->params[0]);
	}
    }
  gdam_bind_localhost = bind_localhost;
  return NULL;
}

/*
 * UnixSocket "path_name" "allowed_users"
 */
static char* handle_UnixSocket(ConfigFileLine* file_line)
{
  char *path;
  GSList *list;
  GSList *old_list;
  if (file_line->num_params != 2)
    {
      return g_strdup_printf ("%s:%d: UnixSocket: "
			      "should have two parameters, path and user_list",
			      file_line->filename, file_line->lineno);
    }
  path = g_strdup (file_line->params[0]);
  list = make_uid_list (file_line->params[1]);
  if (unix_to_uid_list == NULL)
    unix_to_uid_list = g_hash_table_new (g_str_hash, g_str_equal);
  old_list = g_hash_table_lookup (unix_to_uid_list, path);
  list = g_slist_concat (old_list, list);
  g_hash_table_insert (unix_to_uid_list, path, list);
  return NULL;
}

/*
 * TcpSocket "tcp_port" "allowed_hosts"
 */
static char* handle_TcpSocket(ConfigFileLine* file_line)
{
  int port;
  gpointer port_as_ptr;
  GSList *list;
  GSList *old_list;
  if (file_line->num_params != 2)
    {
      return g_strdup_printf ("%s:%d: TcpSocket: "
			      "should have two parameters, path and host_list",
			      file_line->filename, file_line->lineno);
    }
  port = atoi (file_line->params[0]);
  if (port <= 0)
    {
      return g_strdup_printf ("%s:%d: TcpSocket: "
			      "port must be > 0",
			      file_line->filename, file_line->lineno);
    }
  list = make_ip_list (file_line->params[1]);
  port_as_ptr = GINT_TO_POINTER (port);
  if (port_to_host_list == NULL)
    port_to_host_list = g_hash_table_new (NULL, NULL);
  old_list = g_hash_table_lookup (port_to_host_list, port_as_ptr);
  list = g_slist_concat (old_list, list);
  g_hash_table_insert (port_to_host_list, port_as_ptr, list);
  return NULL;
}


static char* handle_IfRoot(ConfigFileLine* file_line)
{
  if (getuid () == 0)
    {
      config_file_line_shift (file_line);
      return config_file_line_handle (file_line);
    }
  return NULL;
}

/*
 * Table of all config file lines, and a function to scan them.
 */
static struct ConfigFileLineType {
	const char* word;
	ConfigFileLineHandler handler;
} config_file_line_types[] = {
	{ "DeviceType", handle_DeviceType },
	{ "Device", handle_Device },
	{ "DeviceAlias", handle_DeviceAlias },
	{ "User", handle_User },
	{ "Group", handle_Group },
	{ "IfRoot", handle_IfRoot },
	{ "PidFile", handle_PidFile },
	{ "LogFile", handle_LogFile },
	{ "Daemon", handle_Daemon },
	{ "Debug", handle_Debug },
	{ "DispatchLog", handle_DispatchLog },
	{ "ExternalPriorityDrop", handle_ExternalPriorityDrop },
	{ "BindLocalhost", handle_BindLocalhost },
	{ "Microphone", handle_Microphone },
	{ "TcpSocket", handle_TcpSocket },
	{ "UnixSocket", handle_UnixSocket },
};
#define num_config_file_line_types  (sizeof(config_file_line_types)	\
                                   / sizeof(config_file_line_types[0]))

/* TODO:  rewrite this to use bsearch (after one qsort...) */
static ConfigFileLineHandler get_handler(const char* word)
{
  guint i;
  for (i = 0; i < num_config_file_line_types; i++)
    if (strcasecmp (config_file_line_types[i].word, word) == 0)
      return config_file_line_types[i].handler;
  return NULL;
}

static void config_file_line_shift(ConfigFileLine* file_line)
{
  g_free (file_line->word);

  if (file_line->num_params == 0)
    file_line->word = NULL;
  else
    {
      file_line->word = file_line->params[0];
      g_memmove (file_line->params, file_line->params + 1,
		 sizeof (char *) * file_line->num_params);
      file_line->num_params--;
    }
}

static char* config_file_line_handle(ConfigFileLine* file_line)
{
  ConfigFileLineHandler handler;
  handler = get_handler (file_line->word);
  if (handler == NULL)
    {
      return g_strdup_printf ("%s:%d: no handler for `%s'",
			      file_line->filename,
			      file_line->lineno, file_line->word);
    }
  else
    {
      char *msg = (*handler) (file_line);
      if (msg != NULL)
	return msg;
    }
  return NULL;
}

gboolean  gdam_config_file_parse(const char* filename)
{
  FILE *fp = fopen (filename, "r");
  int lineno = 0;
  ConfigFileLine *line;
  if (!fp)
    {
      g_warning ("configuration file %s could not be found", filename);
      return FALSE;
    }
  while ((line = config_file_line_parse (fp, filename, &lineno)) != NULL)
    {
      char *message = config_file_line_handle (line);
      if (message != NULL)
	g_warning ("%s", message);
      config_file_line_destroy (line);
    }
  fclose (fp);
  return TRUE;
}

/*=====================================================================
 * Bleh, actual parsing.
 */

static ConfigFileLine* parse_from_word(char* str,
                                       const char* filename,
                                       int lineno)
{
  GPtrArray *exprs = NULL;
  char *at;
  char *word_start = str;
  char *word_end;
  ConfigFileLine *line;
  word_end = word_start;
  while (*word_end && !isspace (*word_end))
    word_end++;
  exprs = g_ptr_array_new ();

  at = word_end;
  while (*at)
    {
      char *dq_end;
      while (*at && isspace (*at))
	at++;
      if (!*at)
	break;
      if (*at != '"')
	{
	  g_warning ("%s:%d: expected double-quote (char %d)",
		     filename, lineno, at - str);
	  goto fail;
	}
      dq_end = strchr (at + 1, '"');
      if (!dq_end)
	{
	  g_warning ("%s:%d: mismatched double-quotes...", filename, lineno);
	  goto fail;
	}
      {
	*dq_end = 0;
	g_ptr_array_add (exprs, g_strdup (at + 1));
	*dq_end = '"';
      }
      at = dq_end + 1;
    }

  line = g_new (ConfigFileLine, 1);
  g_ptr_array_set_size (exprs, exprs->len);
  line->params = (char **) (exprs->pdata);
  line->num_params = exprs->len;
  g_ptr_array_free (exprs, FALSE);
  {
    char backup = *word_end;
    *word_end = 0;
    line->word = g_strdup (word_start);
    *word_end = backup;
  }
  line->filename = filename;
  line->lineno = lineno;
  return line;

fail:
  {
    guint i;
    for (i = 0; i < exprs->len; i++)
      g_free (g_ptr_array_index (exprs, i));
  }

  g_ptr_array_free (exprs, TRUE);
  return NULL;
}

static ConfigFileLine* config_file_line_parse  (FILE* fp,
						const char* filename,
                                                int* lineno)
{
  char buf[4096];
  while (fgets (buf, sizeof (buf), fp))
    {
      char *first_nonws = buf;
      (*lineno)++;
      while (*first_nonws && isspace (*first_nonws))
	first_nonws++;
      if (*first_nonws == 0 || *first_nonws == '#')
	continue;
      return parse_from_word (first_nonws, filename, *lineno);
    }
  return NULL;
}

static void config_file_line_destroy(ConfigFileLine* line)
{
  int i;
  for (i = 0; i < line->num_params; i++)
    g_free (line->params[i]);
  g_free (line->params);
  g_free (line->word);
  g_free (line);
}
