/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamdispatch.h"
#include "gdamdevice.h"
#include <common/gdam-protocol.h>
/*
 * Code which handles the global communication.
 *
 * In particular the source_id/device_id interface
 * is accessible here.
 */
static GHashTable* server_object_map     = NULL;
static int         last_global_id        = 0;
extern GHashTable* gdam_dispatch_methods;

GdasSource* gdam_dispatch_get_source(int gdas_source_id)
{
  gpointer rv;
  if (server_object_map == NULL)
    return NULL;
  rv = g_hash_table_lookup (server_object_map,
			    GINT_TO_POINTER (gdas_source_id));
  if (rv == NULL)
    return NULL;
  if (GDAS_IS_SOURCE (rv))
    {
      gtk_object_ref ((GtkObject *) rv);
      return (GdasSource *) rv;
    }
  return NULL;
}

GdamSink* gdam_dispatch_get_sink(int sink_id)
{
  gpointer rv;
  if (server_object_map == NULL)
    return NULL;
  rv = g_hash_table_lookup (server_object_map, GINT_TO_POINTER (sink_id));
  if (rv == NULL)
    return NULL;
  if (GDAM_IS_SINK (rv))
    {
      gtk_object_ref ((GtkObject *) rv);
      return (GdamSink *) rv;
    }
  return NULL;
}

GdamServerObject* gdam_dispatch_get_server_object(int sink_id)
{
  gpointer rv;
  if (server_object_map == NULL)
    return NULL;
  rv = g_hash_table_lookup (server_object_map, GINT_TO_POINTER (sink_id));
  if (rv == NULL)
    return NULL;
  g_return_val_if_fail (GDAM_IS_SERVER_OBJECT (rv), NULL);
  gtk_object_ref ((GtkObject *) rv);
  return (GdamServerObject *) rv;
}

GdamBuffer* gdam_dispatch_buffer(GdamConnection* connection,
                                 GdamPermissions* perms,
				 GdamBuffer* command)
{
  int type;
  GdamBuffer *rv = gdam_buffer_new ();
  GdamDispatchMessageFunc dispatcher;

  if (!gdam_buffer_parse_int (command, &type))
    {
      g_warning ("error parsing command type from header");
      return NULL;
    }
#if DEBUG_DISPATCHING
  gdam_debug ("gdam_dispatch: message of type %d", type);
#endif
  dispatcher = gdam_dispatch_message_find_by_type (type);
  if (!dispatcher)
    {
      if (GDAM_MESSAGE_FIRST_MESSAGE < type
	  && type < GDAM_MESSAGE_FIRST_RESPONSE)
	{
	  g_warning ("message was in valid range..."
		     "presumably a unimplemented command");
	  gdam_buffer_append_int (rv, GDAM_RESPONSE_FAILED);
	  gdam_buffer_append_int (rv, type);
	  {
	    int tmp;
	    g_assert (gdam_buffer_peek_int (rv, &tmp));
	    g_assert (tmp == GDAM_RESPONSE_FAILED);
	    g_assert (rv->size == 8);
	  }
	  return rv;
	}
      else if (GDAM_MESSAGE_FIRST_RESPONSE <= type
	       && type < GDAM_MESSAGE_LAST_MESSAGE)
	{
	  g_warning ("got a response from the client... " "protocol error");
	}
      else if (type == GDAM_MESSAGE_FIRST_MESSAGE)
	{
	  g_warning ("GDAM_MESSAGE_FIRST_MESSAGE "
		     "should never be a message type");
	}
      else if (type == GDAM_MESSAGE_LAST_MESSAGE)
	{
	  g_warning ("GDAM_MESSAGE_LAST_MESSAGE "
		     "should never be a message type");
	}
      else
	{
	  g_warning ("client message was completely "
		     "invalid-- type id out of range");
	}
      gdam_buffer_destroy (rv);
      return NULL;
    }

  /* Strange as it may seem, we don't pass whether the
   * command succeeded or failed to the caller.
   * Maybe we should... */
  switch ((*dispatcher) (connection, perms, command, rv))
    {
    case GDAM_DISPATCH_INDICATE_SUCCESS:
      gdam_buffer_append_int (rv, GDAM_RESPONSE_SUCCESS);
      gdam_buffer_append_int (rv, type);
    case GDAM_DISPATCH_SUCCESS:
      break;

    case GDAM_DISPATCH_TOO_SHORT:
      g_warning ("message of type `%s' was too short",
		 gdam_protocol_get_enum_name ("message", type));
    case GDAM_DISPATCH_INDICATE_FAIL:
      gdam_buffer_append_int (rv, GDAM_RESPONSE_FAILED);
      gdam_buffer_append_int (rv, type);
    case GDAM_DISPATCH_FAIL:
      break;

    default:
      g_assert_not_reached ();
    }
  return rv;
}

int gdam_dispatch_obtain_global_id(GdamServerObject* object)
{
  if (last_global_id++ == 0)
    {
      server_object_map = g_hash_table_new (NULL, NULL);
    }
  g_hash_table_insert (server_object_map,
		       GINT_TO_POINTER (last_global_id), object);
  return last_global_id;
}

void gdam_dispatch_release_global_id(int source_id)
{
  g_hash_table_remove (server_object_map, GINT_TO_POINTER (source_id));
}
