/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamevent.h"
#include <gtk/gtksignal.h>

typedef struct _GdamEventInfo GdamEventInfo;

struct _GdamEventInfo {
	gboolean		is_signal;
	GtkType          	object_type;
	char*			signal_name;
	GdamEventMarshal	marshal;
	GDestroyNotify		destroyer;
	gpointer		event_type_data;
};

static GHashTable* event_registry = NULL;
static GHashTable* connection_table = NULL;

static guint gdam_event_alloc_type_id()
{
  static guint id = 0;
  return ++id;
}
static GdamEventInfo* lookup_event_info(guint type_id)
{
  if (!event_registry)
    return NULL;
  return (GdamEventInfo *) g_hash_table_lookup (event_registry,
						GUINT_TO_POINTER (type_id));
}
static void gdam_event_assoc_type_id(guint type_id, GdamEventInfo* info)
{
  if (!event_registry)
    event_registry = g_hash_table_new (NULL, NULL);
  g_hash_table_insert (event_registry, GUINT_TO_POINTER (type_id), info);
}

/* Alloc and free connection id's. */
static guint gdam_event_alloc_connect_id()
{
  static guint id = 0;
  return ++id;
}
static void gdam_event_free_connect_id(guint type_id)
{
  g_hash_table_remove (connection_table, GUINT_TO_POINTER (type_id));
}

/*=====================================================================*
 *                         gdam_event_register                         *
 *                                                                     *
 * Register a new type of event that you can connect a                 *
 * to a channel to have it issue a packet on the stream.               *
 *                                                                     *
 * Parameters:                                                         *
 *     signal_id         the signal's identifying number.              *
 *     marshal           the function which fills the buffer when the  *
 *                       event arrives.                                *
 *     destroyer         function to call when the signal is detached. *
 *     event_type_data   data specific to this registration.           *
 *                                                                     *
 * Returns the id for the event type.                                  *
 *=====================================================================*/
guint        gdam_event_register         (GtkType          object_type,
                                          const char*      signal_name,
                                          GdamEventMarshal marshal,
					  GDestroyNotify   destroyer,
					  gpointer         event_type_data)
{
  GdamEventInfo *info;
  guint event_type_id;


  info = g_new (GdamEventInfo, 1);
  info->is_signal = TRUE;
  info->object_type = object_type;
  info->signal_name = g_strdup (signal_name);
  info->marshal = marshal;
  info->destroyer = destroyer;
  info->event_type_data = event_type_data;

  event_type_id = gdam_event_alloc_type_id ();
  gdam_event_assoc_type_id (event_type_id, info);
  return event_type_id;
}

/*=====================================================================*
 *                         gdam_timer_register                         *
 *                                                                     *
 * Register a new type of event that you can connect a                 *
 * to a channel to have it issue a packet on the stream.               *
 *                                                                     *
 * Parameters:                                                         *
 *     object_type       the type of object that may be tied to the    *
 *                       timer.                                        *
 *     marshal           the function which fills the buffer when the  *
 *                       event arrives.                                *
 *     destroyer         function to call when the signal is detached. *
 *     event_type_data   data specific to this registration.           *
 *=====================================================================*/
guint        gdam_timer_register         (GtkType          object_type,
                                          GdamEventMarshal marshal,
					  GDestroyNotify   destroyer,
					  gpointer         event_type_data)
{
  GdamEventInfo *info;
  guint event_type_id;

  info = g_new (GdamEventInfo, 1);
  info->is_signal = FALSE;
  info->object_type = object_type;
  info->signal_name = NULL;
  info->marshal = marshal;
  info->destroyer = destroyer;
  info->event_type_data = event_type_data;

  event_type_id = gdam_event_alloc_type_id ();
  gdam_event_assoc_type_id (event_type_id, info);
  return event_type_id;
}

typedef struct _GdamEventConnection GdamEventConnection;
struct _GdamEventConnection {
	GdamEventInfo*			info;

	/* Our key in the connection hash table. */
	guint				connection_id;

	/* Our handler_id for the gtk-signal, or source_id for a timer. */
	guint				handler_id;

	/* If we need to trap destroy on `object'... */
	guint				destroy_id;

	/* The object responsible for the event. */
	GtkObject*			object;

	/* Who to notify of the event. */
	GdamConnection*			connection;

	/* Data specific to this particular connection to this
	 * kind of event. */
	gpointer			connect_data;
};

/* To make it easy (since this must be a signal handler for
 * any kind of signal), we perform the marshalling using
 * one pointer `info', which we reliably get as the 
 * first argument by making this an "object signal". */
static void gdam_event_marshal(GdamEventConnection* info)
{
  (*info->info->marshal) (info->connection,
			  info->object,
			  info->connection_id,
			  info->info->event_type_data, info->connect_data);
}

static void destroy_connection(GdamEventConnection* info)
{
  if (info->info->destroyer)
    info->info->destroyer (info->connect_data);
  if (info->connection_id)
    gdam_event_free_connect_id (info->connection_id);
  g_free (info);
}

guint        gdam_event_connect          (GtkObject*       object,
					  GdamConnection*  to_connect,
                                          guint            signal_type_id,
					  gpointer         connect_data)
{
  GdamEventConnection *event_connection;
  GdamEventInfo *info;
  info = lookup_event_info (signal_type_id);
  g_return_val_if_fail (info != NULL, 0);

  /* Connect the signal up. */
  event_connection = g_new (GdamEventConnection, 1);
  event_connection->info = info;
  event_connection->connection = to_connect;
  event_connection->object = object;
  event_connection->connect_data = connect_data;
  event_connection->handler_id = gtk_signal_connect_full (object,
							  info->signal_name,
							  GTK_SIGNAL_FUNC
							  (gdam_event_marshal),
							  NULL,
							  event_connection,
							  (GDestroyNotify)
							  destroy_connection,
							  TRUE, FALSE);

  /* Add it to the connection hash. */
  event_connection->connection_id = gdam_event_alloc_connect_id ();
  if (!connection_table)
    connection_table = g_hash_table_new (NULL, NULL);
  g_hash_table_insert (connection_table,
		       GUINT_TO_POINTER (event_connection->connection_id),
		       event_connection);
  return event_connection->connection_id;
}

static void destroy_timer_connection(GdamEventConnection* info)
{
  info->handler_id = 0;
  if (info->info->destroyer)
    info->info->destroyer (info->connect_data);
  if (info->destroy_id)
    {
      guint id = info->destroy_id;
      info->destroy_id = 0;
      gtk_signal_disconnect (info->object, id);
    }
  if (info->connection_id)
    g_hash_table_remove (connection_table,
			 GUINT_TO_POINTER (info->connection_id));
  g_free (info);
}

static void destroy_timer_object(GdamEventConnection* info)
{
  info->destroy_id = 0;

  /* This will invoke the destroy_timer_connection. */
  if (info->handler_id)
    g_source_remove (info->handler_id);
}

/* Connect a timer event. */
guint        gdam_timer_connect          (GtkObject*       object,
                                          guint            event_type_id,
					  GdamConnection*  to_connect,
					  int              period,
					  gpointer         connect_data)
{
  GdamEventConnection *event_connection;
  GdamEventInfo *event_info;
  event_info = lookup_event_info (event_type_id);
  g_return_val_if_fail (event_info != NULL, 0);

  /* Connect the signal up. */
  event_connection = g_new (GdamEventConnection, 1);
  event_connection->info = event_info;
  event_connection->connection = to_connect;
  event_connection->object = object;
  event_connection->handler_id = g_timeout_add_full (G_PRIORITY_DEFAULT,
						     period,
						     (GSourceFunc)
						     gdam_event_marshal,
						     event_connection,
						     (GDestroyNotify)
						     destroy_timer_connection);

  event_connection->destroy_id = gtk_signal_connect_object (object,
							    "destroy",
							    GTK_SIGNAL_FUNC
							    (destroy_timer_object),
							    (GtkObject *)
							    event_connection);

  /* Add it to the connection hash. */
  event_connection->connection_id = gdam_event_alloc_connect_id ();
  if (!connection_table)
    connection_table = g_hash_table_new (NULL, NULL);
  g_hash_table_insert (connection_table,
		       GUINT_TO_POINTER (event_connection->connection_id),
		       event_connection);
  return event_connection->connection_id;
}

gboolean     gdam_event_disconnect(guint            connection_id)
{
  GdamEventConnection *con_info;
  g_return_val_if_fail (connection_table != NULL, FALSE);
  con_info = (GdamEventConnection *)
    g_hash_table_lookup (connection_table, GUINT_TO_POINTER (connection_id));
  if (!con_info)
    return FALSE;
  if (con_info->info->is_signal)
    gtk_signal_disconnect (con_info->object, con_info->handler_id);
  else
    g_source_remove (con_info->handler_id);
  return TRUE;
}

typedef struct _DisconnectorData DisconnectorData;
struct _DisconnectorData {
	GArray*	connections_to_kill;
	GdamConnection* connection;
};

static void append_disconnect(gpointer key, gpointer val, gpointer dd)
{
  DisconnectorData *data = (DisconnectorData *) dd;
  GdamEventConnection *ev_con = (GdamEventConnection *) val;
  int id;
  if (data->connection != ev_con->connection)
    return;
  id = GPOINTER_TO_INT (key);
  data->connections_to_kill =
    g_array_append_val (data->connections_to_kill, id);
}
void    gdam_event_disconnect_all_from_connection(GdamConnection* connection)
{
  DisconnectorData data;
  guint i;
  if (connection_table == NULL)
    return;

  data.connections_to_kill = g_array_new (FALSE, FALSE, sizeof (int));
  data.connection = connection;

  g_hash_table_foreach (connection_table, append_disconnect, &data);

  for (i = 0; i < data.connections_to_kill->len; i++)
    {
      int id = g_array_index (data.connections_to_kill, int, i);
      gdam_event_disconnect (id);
    }
  g_array_free (data.connections_to_kill, TRUE);
}
