/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if NOT, WRITe to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamoutputdevice.h"
#include "gdamserverdebug.h"
#include "gdamarctan.h"
#include "gdammain.h"
#include <common/gdamsampleconverter.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#define MAX_ATOMIC_SOURCE_READ		2048

static GtkObjectClass* parent_class = NULL;

#define GDAM_DEVICE_PRIORITY G_PRIORITY_HIGH

/* GdamOutputDevice virtuals. */
static void     gdam_output_device_on_write_block(GdamOutputDevice* device);
static void     gdam_output_device_close_fd      (GdamOutputDevice* device);
/* GdamSink virtuals. */
static gboolean gdam_output_device_real_open     (GdamSink*         sink,
                                                  const char*       name,
				                  const char*       opts);
static void     gdam_output_device_real_close    (GdamSink*         sink);
static gboolean gdam_output_device_set_source    (GdamSink*         sink,
                                                  GdasSource*       source);
static void     gdam_output_device_remove_source (GdamSink*         sink);
static void gdam_output_device_source_destroy_notify(gpointer user_data);
#if 0
static int      gdam_output_device_write         (GdamSink*         output,
	                                          const GdamSample* samples,
					          int               max_read);
#endif
/* GtkObject virtual. */
static void     gdam_output_device_destroy       (GtkObject*        to_kill);


/* Add/remove the sink from the main loop. */
static void     gdam_output_device_stop          (GdamOutputDevice* device);
static void     gdam_output_device_play          (GdamOutputDevice* device);


/* Functions passed to g_source_add. */
static gboolean gdam_output_device_prepare       (gpointer          src_data,
                                                  GTimeVal*         curtime,
                                                  gint*             timeout,
		                                  gpointer          user_data);
static gboolean gdam_output_device_check         (gpointer          src_data,
                                                  GTimeVal*         curtime,
		                                  gpointer          user_data);
static gboolean gdam_output_device_dispatch      (gpointer          src_data,
                                                  GTimeVal*         curtime,
		                                  gpointer          user_data);
static gboolean maybe_fill_prebuffer             (GdamOutputDevice* device);
static GSourceFuncs gdam_output_device_source_funcs = {
	gdam_output_device_prepare,
	gdam_output_device_check,
	gdam_output_device_dispatch,
	NULL
};

/* And this function does a single fragment-sized write out to
 * the file descriptor. */
static gboolean gdam_output_device_write_one_frag(GdamOutputDevice* device);

static void gdam_output_device_options_clear(GdamOutputDeviceOptions* options)
{
  /* options which control the output mechanism. */
  options->sampling_rate = 44100;
  options->num_channels = 2;
  options->bits_per_sample = 16;
  options->is_signed = TRUE;
  options->is_wrong_endian = FALSE;
  options->is_wrong_stereo_order = FALSE;
  options->block_size = 4096;
  options->do_forking = FALSE;
  options->log_frag_size = 12;
  options->num_fragments = 4;
}

static void gdam_output_device_init(GdamOutputDevice* output_device)
{
  output_device->device_fd = -1;
  output_device->poll_fd.fd = -1;
  gdam_output_device_options_clear (&output_device->options);
  output_device->to_native_samples = NULL;
  output_device->to_native_channels = NULL;
  output_device->sample_size = 4;
  output_device->limit_mode = 2;
}

static void gdam_output_device_class_init(GdamOutputDeviceClass* klass)
{
  GdamSinkClass *sink_class;
  sink_class = GDAM_SINK_CLASS (klass);
  klass->on_write_block = gdam_output_device_on_write_block;
  klass->close_fd = gdam_output_device_close_fd;

  //sink_class->write = gdam_output_device_write;
  sink_class->set_source = gdam_output_device_set_source;
  sink_class->remove_source = gdam_output_device_remove_source;
  sink_class->open = gdam_output_device_real_open;
  sink_class->close = gdam_output_device_real_close;
}

GtkType gdam_output_device_get_type()
{
  static GtkType output_device_type = 0;
  if (!output_device_type)
    {
      static const GtkTypeInfo output_device_info = {
	"GdamOutputDevice",
	sizeof (GdamOutputDevice),
	sizeof (GdamOutputDeviceClass),
	(GtkClassInitFunc) gdam_output_device_class_init,
	(GtkObjectInitFunc) gdam_output_device_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SINK;
      parent_class = gtk_type_class (parent);
      output_device_type = gtk_type_unique (parent, &output_device_info);
    }
  return output_device_type;
}

/*
 * GdamOutputDevice virtuals
 */
static void gdam_output_device_on_write_block(GdamOutputDevice* device)
{
  const char *device_name;
  device_name = GDAM_SINK (device)->device_name;
  fprintf (stderr,
	   "Hello...\n\n"
	   "You seem to be using a soundcard (%s), \n"
	   "which doesn't support ``polling'', a feature\n"
	   "GDAM uses extensively.  In order to use GDAM,\n"
	   "you must specify the ``fork'' option to GDAM.\n"
	   "\n"
	   "Modify the server.config file, which is probably\n"
	   "in %s, but you might also look in\n"
	   "\t~/.gdam/server.config\n"
	   "\t/usr/local/gdam/lib/server.config\n"
	   "\t/usr/lib/server.config\n"
	   "Are also places we look...\n\n"
	   "The exact location should have been printed when we"
	   "started.\n"
	   "\nPlease modify the line that looks like\n"
	   "\tDevice          \"oss\"           \"%s\"     \"\"\n"
	   "to look like:\n"
	   "\tDevice          \"oss\"           \"%s\"     "
	   "\"fork\"\n"
	   "(if there is already an option, put a colon in between,\n"
	   "\t\"fork:numfragments=16\"\n"
	   "or whatever...)\n", device_name,
	   G_STRINGIFY (GDAM_CONFIG_LOCATION_TOKEN), device_name,
	   device_name);
}

static void gdam_output_device_close_fd      (GdamOutputDevice* device)
{
  if (device->poll_fd.fd >= 0)
    {
      close (device->poll_fd.fd);
      device->poll_fd.fd = -1;
    }
}

static gboolean parse_options(const char*               options_string,
                              GdamOutputDeviceOptions*  options)
{
  char **option_strings;
  char **at;
  option_strings = g_strsplit (options_string, ":", 0);
  gdam_output_device_options_clear (options);
  for (at = option_strings; *at != NULL; at++)
    {
      char *opt = *at;
      if (strncmp (opt, "logfragsize=", 12) == 0)
	{
	  int log_frag_size = atoi (opt + 12);
	  if (log_frag_size < 6 || log_frag_size > 100)
	    {
	      g_warning ("log_frag_size was out of bounds...");
	    }
	  else
	    {
	      options->log_frag_size = log_frag_size;
	    }
	}
      else if (strncmp (opt, "bits=", 5) == 0)
	{
	  int bits = atoi (opt + 5);
	  options->bits_per_sample = bits;
	  if (bits != 8 && bits != 16)
	    {
	      g_warning ("unknown bits/sample %d", bits);
	    }
	}
      else if (strcmp (opt, "mono") == 0)
	{
	  options->num_channels = 1;
	}
      else if (strcmp (opt, "stereo") == 0)
	{
	  options->num_channels = 2;
	}
      else if (strncmp (opt, "rate=", 5) == 0)
	{
	  options->sampling_rate = atoi (opt + 5);
	}
      else if (strcmp (opt, "fork") == 0)
	{
	  options->do_forking = TRUE;
	}
      else if (strcmp (opt, "fork_rt") == 0)
	{
	  options->do_forking = TRUE;
	  options->do_forking_with_helper = TRUE;
	}
      else if (strcmp (opt, "wrong_endian") == 0)
	{
	  options->is_wrong_endian = TRUE;
	}
      else if (strcmp (opt, "no_ioctls") == 0)
	{
	  options->no_ioctls = TRUE;
	}
      else if (strcmp (opt, "duplex") == 0)
	{
	  g_warning ("duplex device option no longer supported");
	}
      else if (strncmp (opt, "numfragments=", 13) == 0)
	{
	  int num_fragments = atoi (opt + 13);
	  if (num_fragments < 3 || num_fragments > 300)
	    {
	      g_warning ("numfragments was out of bounds...");
	    }
	  else
	    {
	      options->num_fragments = num_fragments;
	    }
	}
      else
	{
	  g_warning ("unknown option to output device: %s", opt);
	}
      g_free (opt);
    }
  g_free (option_strings);
  return TRUE;
}
static gboolean writen(int fd, const char* buf, int len)
{
  while (len > 0)
    {
      int n = write (fd, buf, len);
      if (n < 0)
	{
	  if (errno == EINTR)
	    continue;
	  return FALSE;
	}
      buf += n;
      len -= n;
    }
  return TRUE;
}

static gboolean readn(int fd, char* buf, int len)
{
  while (len > 0)
    {
      int n = read (fd, buf, len);
      if (n < 0)
	{
	  if (errno == EINTR)
	    continue;
	  return FALSE;
	}
      if (n == 0)
	return FALSE;
      buf += n;
      len -= n;
    }
  return TRUE;
}

static void exec_fork_helper(int           in_fd,
                             int           out_fd,
			     int           log_frag_size,
			     int           num_frags)
{
  char in_str[256], out_str[256];
  char block_size_str[256];
  char *args[128];
  fcntl (in_fd, F_SETFD, 0);
  fcntl (out_fd, F_SETFD, 0);
  g_snprintf (in_str, sizeof (in_str), "%d", in_fd);
  g_snprintf (out_str, sizeof (out_str), "%d", out_fd);
  g_snprintf (block_size_str, sizeof (block_size_str),
	      "%d", 1 << log_frag_size);
  args[0] = "gdamrtcat";
  args[1] = "--block-size";
  args[2] = block_size_str;
  args[3] = "--num-blocks";
  args[4] = "3";
  args[5] = in_str;
  args[6] = out_str;
  args[7] = NULL;
  execvp ("gdamrtcat", args);
  execv ("./gdamrtcat", args);

  /* normal situation if you cd server/, run ./gdam-server. */
  execv ("./helper/gdamrtcat", args);

  /* normal situation if the server was autolaunched. */
  execv ("../server/helper/gdamrtcat", args);

  g_warning ("fork_rt: couldn't find gdamrtcat");
}

/*
 * GdamSink virtuals
 */
static gboolean gdam_output_device_fork_open     (GdamOutputDevice* device)
{
  int pid;
  GdamOutputDeviceClass *klass;
  int pipe_fds[2];

  /* XXX: should use fragment size! */
  char buffer[4096];

  klass = GDAM_OUTPUT_DEVICE_CLASS (GTK_OBJECT (device)->klass);

  /* Child thread. */
  if (!(*klass->open_fd) (device))
    {
      _exit (5);
    }

  if (pipe (pipe_fds) < 0)
    {
      g_warning ("pipe failed: %s", strerror (errno));
      return FALSE;
    }

retry_fork:
  pid = fork ();
  if (pid == -1)
    {
      if (errno == EINTR || errno == EAGAIN)
	{
	  sleep (1);
	  goto retry_fork;
	}
    }

  if (pid == 0)
    {
      int soundcard_fd;
      GdamOutputDeviceOptions *options = &device->options;
      close (pipe_fds[1]);
      soundcard_fd = device->device_fd;
      if (options->do_forking_with_helper)
	exec_fork_helper (pipe_fds[0], soundcard_fd,
			  options->log_frag_size, options->num_fragments);
      while (readn (pipe_fds[0], buffer, sizeof (buffer))
	     && writen (soundcard_fd, buffer, sizeof (buffer)))
	;
      _exit (0);
    }

  /* XXX: not sure what to do about the problem that
   *      many devices may not like this forking
   *      arrangement... */
  close (device->device_fd);

  device->device_fd = pipe_fds[1];
  close (pipe_fds[0]);
  return TRUE;
}

#define PREBUFFER_MULTIPLE	2

void gdam_output_device_options_changed(GdamOutputDevice* device)
{
  GdamOutputDeviceOptions *options = &device->options;
  gboolean is_big_endian;
  is_big_endian = (G_BYTE_ORDER == G_BIG_ENDIAN);
  if (options->is_wrong_endian)
    is_big_endian = !is_big_endian;
  device->to_native_samples =
    gdam_sample_converter_get (GDAM_SAMPLE_BITS, TRUE,
			       (G_BYTE_ORDER == G_BIG_ENDIAN),
			       options->bits_per_sample,
			       options->is_signed, is_big_endian);
  device->to_native_channels =
    gdam_sample_converter_get_channel_changer (2, options->num_channels);
  if (options->bytes_per_sample == 0)
    {
      if (options->bits_per_sample == 24)
	options->bytes_per_sample = 4;
      else
	options->bytes_per_sample = options->bits_per_sample / 8;
    }
  {
    int input_size = sizeof (GdamSample);
    int num_channels = options->num_channels;
    int mid_size = sizeof (GdamSampleValue) * num_channels;
    int output_size = options->bytes_per_sample * num_channels;
    int i;
    i = MAX (input_size, mid_size);
    i = MAX (i, output_size);
    device->conversion_sample_size = i;
  }
  device->total_sample_size = options->bytes_per_sample
    * options->num_channels;
  device->pre_buffer_alloced = device->total_sample_size
    * options->block_size * PREBUFFER_MULTIPLE;
  device->pre_buffer_amount = device->pre_buffer_start = 0;
  device->pre_buffer = g_realloc (device->pre_buffer,
				  device->pre_buffer_alloced
				  * device->total_sample_size);
}

void       gdam_output_device_set_fd         (GdamOutputDevice* output_device,
                                              int               fd)
{
  GdamOutputDeviceClass *klass;

  klass = GDAM_OUTPUT_DEVICE_CLASS (GTK_OBJECT (output_device)->klass);

  g_return_if_fail (fd >= 0);
  g_return_if_fail (output_device->device_fd < 0);

  if (output_device->options.do_forking)
    {
#if 0
      if (!gdam_output_device_fork_set_fd (output_device))
	return FALSE;
      gdam_output_device_options_changed (output_device);
      return TRUE;
#else
      g_warning ("do_forking not supported with set_fd yet");
#endif
    }

  g_main_add_poll (&output_device->poll_fd, GDAM_DEVICE_PRIORITY);
  output_device->device_fd = fd;
  output_device->poll_fd.fd = fd;
  output_device->poll_fd.events = G_IO_OUT;
  if (output_device->source_id == 0)
    {
      output_device->source_id = g_source_add (GDAM_DEVICE_PRIORITY,
					       TRUE,
					       &gdam_output_device_source_funcs,
					       NULL,
					       output_device,
					       gdam_output_device_source_destroy_notify);
    }
  if (output_device->pre_buffer_busy_func == 0)
    {
      output_device->pre_buffer_busy_func
	= gdam_main_add_busy_func
	((GHookCheckFunc) maybe_fill_prebuffer, output_device);
    }

  if (output_device->options.make_non_blocking)
    {
      int fd;
      fd = output_device->device_fd;
      fcntl (fd, F_SETFL, fcntl (fd, F_GETFL) | O_NONBLOCK);
      g_warning ("using experimental, nonblocking options");
    }

  gdam_debug ("Device set at %d bits/sample, %d channels, rate %d",
	      output_device->options.bits_per_sample,
	      output_device->options.num_channels,
	      output_device->options.sampling_rate);
  gdam_output_device_options_changed (output_device);
}


static gboolean gdam_output_device_real_open     (GdamSink*         sink,
                                                  const char*       name,
				                  const char*       opts)
{
  GdamOutputDevice *output_device;
  GdamOutputDeviceClass *klass;

  output_device = GDAM_OUTPUT_DEVICE (sink);
  klass = GDAM_OUTPUT_DEVICE_CLASS (GTK_OBJECT (sink)->klass);

  if (name != sink->device_name)
    {
      g_free (sink->device_name);
      sink->device_name = g_strdup (name);
    }

  if (!parse_options (opts, &output_device->options))
    {
      return FALSE;
    }

  if (output_device->options.do_forking)
    {
      if (!gdam_output_device_fork_open (output_device))
	return FALSE;
      gdam_output_device_options_changed (output_device);
      return TRUE;
    }

  if (!(*klass->open_fd) (output_device))
    return FALSE;

  if (output_device->options.make_non_blocking)
    {
      int fd;
      fd = output_device->device_fd;
      fcntl (fd, F_SETFL, fcntl (fd, F_GETFL) | O_NONBLOCK);
      g_warning ("using experimental, nonblocking options");
    }

  gdam_debug ("Device set at %d bits/sample, %d channels, rate %d",
	      output_device->options.bits_per_sample,
	      output_device->options.num_channels,
	      output_device->options.sampling_rate);
  gdam_output_device_options_changed (output_device);
  return TRUE;
}

static void     gdam_output_device_real_close     (GdamSink*         sink)
{
  GdamOutputDevice *output_device;
  output_device = GDAM_OUTPUT_DEVICE (sink);
  gdam_output_device_stop (output_device);
  if (output_device->device_fd >= 0)
    {
      close (output_device->device_fd);
      output_device->device_fd = -1;
    }
}
static gboolean gdam_output_device_set_source   (GdamSink*         sink,
                                                 GdasSource*       source)
{
  sink->feed = source;
  source->device = sink;
  gtk_object_ref (GTK_OBJECT (source));
  gdam_output_device_play (GDAM_OUTPUT_DEVICE (sink));
  return TRUE;
}
static void     gdam_output_device_remove_source(GdamSink*         sink)
{
  gdam_output_device_stop (GDAM_OUTPUT_DEVICE (sink));
  gtk_object_unref (GTK_OBJECT (sink->feed));
}
#if 0
static int      gdam_output_device_write        (GdamSink*         output,
	                                         const GdamSample* samples,
					         int               max_read)
{
  g_assert_not_reached ();
  return -1;
}
#endif

/*
 * GtkObject virtual
 */
static void     gdam_output_device_destroy      (GtkObject*        to_kill)
{
  GdamOutputDevice *output_device;
  GdamSink *sink;
  output_device = GDAM_OUTPUT_DEVICE (to_kill);
  sink = GDAM_SINK (to_kill);
  gdam_output_device_stop (output_device);
  gdam_sink_close (sink);
  (*parent_class->destroy) (to_kill);
}

static void do_limit(int           limit_mode,
                     GdamSample*   buffer,
		     int           num_read)
{
#if (GDAM_SAMPLE_BITS == 24)
  if (limit_mode == 1)
    {
      while (num_read > 0)
	{
	  int t;
	  t = buffer->s[0];
	  buffer->s[0] = GDAM_SAMPLE_CLAMP (t);
	  t = buffer->s[1];
	  buffer->s[1] = GDAM_SAMPLE_CLAMP (t);
	  num_read--;
	  buffer++;
	}
    }
  else if (limit_mode == 2)
    {
      while (num_read > 0)
	{
	  int t;
	  t = buffer->s[0];
	  buffer->s[0] = gdam_arctan_b24 (t);
	  t = buffer->s[1];
	  buffer->s[1] = gdam_arctan_b24 (t);

	  num_read--;
	  buffer++;
	}
    }
#endif
}

/*
 * Helper functions.
 */
/* Remove the sink from the main loop. */
static void     gdam_output_device_stop         (GdamOutputDevice* device)
{
  if (device->poll_fd.fd >= 0)
    {
      g_main_remove_poll (&device->poll_fd);
      device->poll_fd.fd = -1;
    }
  if (device->source_id != 0)
    {
      int source_id = device->source_id;
      device->source_id = 0;
      g_source_remove (source_id);
    }
  if (device->pre_buffer_busy_func != 0)
    {
      gdam_main_remove_busy_func (device->pre_buffer_busy_func);
      device->pre_buffer_busy_func = 0;
    }
}

static gboolean too_long_to_write_trigger(gpointer user_data)
{
  GdamOutputDevice *device;
  GtkObject *object;
  GdamOutputDeviceClass *device_class;
  device = GDAM_OUTPUT_DEVICE (user_data);
  object = GTK_OBJECT (device);
  device_class = GDAM_OUTPUT_DEVICE_CLASS (object->klass);
  if (!device->has_written)
    (*device_class->on_write_block) (device);
  gtk_object_unref (object);

  return FALSE;
}

/* XXX: this is not a very good way to do things.
 *      better would be to do this whenever we return FALSE
 *      from dispatch or g_source_remove.
 */
static void gdam_output_device_source_destroy_notify(gpointer user_data)
{
  GdamOutputDevice *output_device = GDAM_OUTPUT_DEVICE (user_data);
  output_device->source_id = 0;
  gdam_output_device_stop (output_device);
}


/* Add the sink to the main loop. */
static void     gdam_output_device_play         (GdamOutputDevice* device)
{
  g_return_if_fail (device->source_id == 0);
  g_return_if_fail (device->poll_fd.fd == -1);

  if (device->device_fd < 0)
    {
      GdamOutputDeviceClass *klass;
      GtkObject *object;
      object = GTK_OBJECT (device);
      klass = GDAM_OUTPUT_DEVICE_CLASS (object->klass);
      if (!(*klass->open_fd) (device))
	return;
    }
  g_return_if_fail (device->device_fd >= 0);

  device->poll_fd.fd = device->device_fd;
  device->poll_fd.events = G_IO_OUT;
  if (device->device_fd >= 0)
    g_main_add_poll (&device->poll_fd, GDAM_DEVICE_PRIORITY);

  device->source_id = g_source_add (GDAM_DEVICE_PRIORITY,
				    TRUE,
				    &gdam_output_device_source_funcs,
				    NULL,
				    device,
				    gdam_output_device_source_destroy_notify);
  g_assert (device->source_id != 0);

  if (device->has_written == FALSE)
    {
      gtk_object_ref (GTK_OBJECT (device));
      g_timeout_add (3000, too_long_to_write_trigger, device);
    }
  if (device->pre_buffer_busy_func == 0)
    {
      device->pre_buffer_busy_func
	= gdam_main_add_busy_func
	((GHookCheckFunc) maybe_fill_prebuffer, device);
    }
}

static gboolean maybe_fill_prebuffer(GdamOutputDevice* device)
{
  GdasSource *feed = GDAM_SINK (device)->feed;
  char *pre_buffer = device->pre_buffer;
  char *tmp_buffer;
  int fragment_size = device->options.block_size;
  int num_read, to_read;
  int sample_size = device->total_sample_size;
  int end_index;
  if (feed == NULL)
    return FALSE;
  if (!gdas_source_is_ready_to_read (feed))
    return FALSE;

  to_read = device->pre_buffer_alloced - device->pre_buffer_amount;
  if (to_read > MAX_ATOMIC_SOURCE_READ)
    to_read = MAX_ATOMIC_SOURCE_READ;
  if (to_read + device->pre_buffer_amount < device->options.block_size)
    to_read = device->options.block_size - device->pre_buffer_amount;

  tmp_buffer = alloca (device->conversion_sample_size * to_read);

  num_read = gdas_source_read (feed, (GdamSample *) tmp_buffer, to_read);
  g_return_val_if_fail (num_read <= to_read, FALSE);
  if (num_read <= 0)
    return FALSE;
  do_limit (device->limit_mode, (GdamSample *) tmp_buffer, num_read);
  if (device->to_native_channels != NULL)
    (*device->to_native_channels) (tmp_buffer, tmp_buffer, num_read);
  if (device->to_native_samples != NULL)
    {
      int channel_samps;
      channel_samps = device->options.num_channels * num_read;
      (*device->to_native_samples) (tmp_buffer, tmp_buffer, channel_samps);
    }

  /* Copy the samples into pre_buffer */
  end_index = device->pre_buffer_start + device->pre_buffer_amount;
  /*g_message("num_read=%d, end_index=%d, num_alloced=%d",num_read,end_index,device->pre_buffer_alloced); */
  if (end_index >= device->pre_buffer_alloced)
    end_index -= device->pre_buffer_alloced;
  if (num_read + end_index >= device->pre_buffer_alloced)
    {
      int p1 = device->pre_buffer_alloced - end_index;
      int p2 = num_read - p1;
      /*g_message("p1/p2=%d/%d",p1,p2); */
      memcpy (pre_buffer + end_index * sample_size,
	      tmp_buffer, p1 * sample_size);
      memcpy (pre_buffer, tmp_buffer + p1 * sample_size, p2 * sample_size);
    }
  else
    {
      /*g_message("num_read=%d",num_read); */
      memcpy (pre_buffer + end_index * sample_size,
	      tmp_buffer, num_read * sample_size);
    }
  device->pre_buffer_amount += num_read;

  return TRUE;
}


/* Write a single fragment-sized out to the file descriptor. */
static gboolean
gdam_output_device_write_one_frag(GdamOutputDevice* output_device)
{
  GdamSample *fragment_buffer;
  GdamOutputDeviceClass *klass;
  int num_samples;
  int num_read;
  GdamSink *sink;
  GdasSource *feed;
  int fragment_size;
  int block_size;
  int bits_per_sample;
  int frag_buffer_size;
  int sample_size;
  int old_pre_buffer_start, old_pre_buffer_amount;
  gboolean was_prebuffer = FALSE;

  num_read = output_device->options.block_size;
  sink = (GdamSink *) output_device;
  feed = sink->feed;

  block_size = output_device->options.block_size;
  bits_per_sample = output_device->options.bits_per_sample;
  bits_per_sample *= output_device->options.num_channels;
  sample_size = bits_per_sample / 8;
  num_samples = block_size;

  g_return_val_if_fail (feed != NULL, FALSE);

  frag_buffer_size = output_device->total_sample_size * block_size;
  fragment_buffer = alloca (frag_buffer_size);
  old_pre_buffer_start = output_device->pre_buffer_start;
  old_pre_buffer_amount = output_device->pre_buffer_amount;
  if (output_device->pre_buffer_amount < block_size)
    {
      bzero (fragment_buffer, frag_buffer_size);
    }
  else
    {
      if (output_device->pre_buffer_start + block_size
	  >= output_device->pre_buffer_alloced)
	{
	  int p1 = output_device->pre_buffer_alloced
	    - output_device->pre_buffer_start;
	  int p2 = block_size - p1;
	  int p1_bytes = p1 * sample_size;
	  memcpy (fragment_buffer,
		  output_device->pre_buffer
		  + output_device->pre_buffer_start * sample_size, p1_bytes);
	  memcpy (fragment_buffer + p1_bytes,
		  output_device->pre_buffer, p2 * sample_size);
	  output_device->pre_buffer_start = p2;
	}
      else
	{
	  fragment_buffer = output_device->pre_buffer
	    + sample_size * output_device->pre_buffer_start;
	  output_device->pre_buffer_start += block_size;
	}
      output_device->pre_buffer_amount -= block_size;
    }

  output_device->has_written = TRUE;

  /*klass=GDAM_OUTPUT_DEVICE_CLASS(GTK_OBJECT(output_device)->klass); */
  klass = (GdamOutputDeviceClass *) (((GtkObject *) output_device)->klass);

  for (;;)
    {

      int num_written;
      if (klass->write_pcm)
	{
	  num_written = klass->write_pcm (output_device,
					  fragment_buffer, frag_buffer_size);
	}
      else
	{
	  num_written = write (output_device->device_fd,
			       fragment_buffer, frag_buffer_size);
	}
      if (num_written < 0)
	{
	  if (errno == EINTR)
	    continue;
	  if (errno == EWOULDBLOCK || errno == EAGAIN)
	    {
	      goto revert_back;
	    }
	  g_warning ("error writing to sound device: %s", strerror (errno));

	  /* Will set device_fd to -1. */
	  gdam_sink_close (sink);

	  return FALSE;
	}

      if (num_written < frag_buffer_size)
	{
	  static gboolean one_shot = FALSE;
	  if (!one_shot)
	    {
	      g_warning ("hmmm... your sound device didn't "
			 "accept a whole fragment...");
	      g_warning ("something may be wrong, but " "we'll proceed");
	      one_shot = TRUE;
	    }
	  g_assert (num_written != 0);
	  //fragment_size -= num_written;
	  //fragment_buffer += num_written;
	  //continue;
	}
      break;
    }

  return TRUE;

revert_back:
  output_device->pre_buffer_start = old_pre_buffer_start;
  output_device->pre_buffer_amount = old_pre_buffer_amount;
  return TRUE;
}


/*
 * GSourceFunc functions.
 */
static gboolean gdam_output_device_prepare      (gpointer          src_data,
                                                 GTimeVal*         curtime,
                                                 gint*             timeout,
		                                 gpointer          user_data)
{
  *timeout = -1;
  return FALSE;
}
static gboolean gdam_output_device_check        (gpointer          src_data,
                                                 GTimeVal*         curtime,
		                                 gpointer          user_data)
{
  GdamOutputDevice *output_device = (GdamOutputDevice *) user_data;
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdam_output_device_check: revents=%d",
		output_device->poll_fd.revents);
  if (output_device->poll_fd.fd < 0)
    return FALSE;
  return output_device->poll_fd.revents == 0 ? FALSE : TRUE;
}
static gboolean gdam_output_device_dispatch     (gpointer          src_data,
                                                 GTimeVal*         curtime,
		                                 gpointer          user_data)
{
  GdamOutputDevice *output_device;
  output_device = GDAM_OUTPUT_DEVICE (user_data);
  if ((output_device->poll_fd.revents & (G_IO_ERR | G_IO_HUP)) != 0)
    {
      g_message ("output_device had problem.  terminating");
      return FALSE;
    }
  if ((output_device->poll_fd.revents & G_IO_OUT) == G_IO_OUT)
    {
      if (GDAM_SERVER_DEBUG_POLL)
	gdam_debug ("gdam_output_device_dispatch: " "writing one fragment");
      if (!gdam_output_device_write_one_frag (output_device))
	return FALSE;
    }
  if ((output_device->poll_fd.revents & G_IO_IN) == G_IO_IN)
    {
      g_warning ("no support for duplex mode available yet");
      /*if (!gdam_device_oss_read_one_fragment(dev_oss)) {
         g_warning("error reading from microphone");
         } */
    }
  return TRUE;
}
