/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <glib.h>
#include <errno.h>
#include <string.h>
#include <sys/poll.h>
#include <unistd.h>


#define MIN_READ 4096


/*
 * Returns:
 *        1 if ready to write on fd_write
 *        0 if ready to read on fd_read
 *       -1 if an error occurred polling.
 *
 * Gives preference to writers.
 */
static int bipoll(int fd_write, int fd_read)
{
  struct pollfd pollfds[2];
  pollfds[0].fd = fd_write;
  pollfds[0].events = POLLOUT;
  pollfds[1].fd = fd_read;
  pollfds[1].events = POLLIN | POLLHUP;
  while (poll (pollfds, 2, -1) < 0)
    {
      if (errno == EINTR)
	continue;
      g_error ("poll failed: %s", strerror (errno));
    }
  if (pollfds[0].revents & POLLOUT)
    return 1;
  if (pollfds[1].revents & (POLLIN | POLLHUP))
    return 0;
  g_warning ("polling on sound dev + pipe: timed out?");
  return -1;
}

/*
 * Write `len' bytes of `buf' to `fd'.
 */
gboolean writen(int fd, const char* buf, int len)
{
  while (len > 0)
    {
      int n = write (fd, buf, len);
      if (n < 0)
	{
	  if (errno == EINTR)
	    continue;
	  return FALSE;
	}
      buf += n;
      len -= n;
    }
  return TRUE;
}

static gboolean readn(int fd, char* buf, int len)
{
  while (len > 0)
    {
      int n = read (fd, buf, len);
      if (n < 0)
	{
	  if (errno == EINTR)
	    continue;
	  return FALSE;
	}
      if (n == 0)
	return FALSE;
      buf += n;
      len -= n;
    }
  return TRUE;
}

/*
 * Write out, always topping off to a multiple
 * of block_size, and returning the number of blocks.
 */
static int write_by_blocks(int fd, char* buf, int num_blocks, int block_size)
{
  int wrote, leftover;

  while ((wrote = write (fd, buf, num_blocks * block_size)) < 0)
    {
      if (errno == EINTR)
	continue;
      g_error ("error writing: %s", strerror (errno));
    }
  if (wrote % block_size == 0)
    leftover = 0;
  else
    leftover = block_size - (wrote % block_size);
  if (leftover && !writen (fd, buf + wrote, leftover))
    g_error ("error writing: %s", strerror (errno));
  return (wrote + leftover) / block_size;
}

void  gdam_send_file(int fd_write,
                     int fd_read,
		     int block_size,
		     int max_buffer)
{
  char *circular_buffer;
  char *tmpbuf;
  int buffer_start = 0, buffer_length = 0;
  gboolean done_reading = FALSE;

  if (block_size <= 0)
    block_size = 1;

  tmpbuf = g_new (char, MIN_READ);

  max_buffer += block_size - 1;
  max_buffer /= block_size;
  max_buffer *= block_size;

  circular_buffer = g_new (char, max_buffer);
  for (;;)
    {
      gboolean do_reading = TRUE;


      /* 
       * Figure out if we should read or write.
       */
      if (done_reading)
	{
	  do_reading = FALSE;
	}
      else if (buffer_length < block_size)
	{
	  do_reading = TRUE;
	}
      else if (buffer_length == max_buffer)
	{
	  do_reading = FALSE;
	}
      else
	{
	  switch (bipoll (fd_write, fd_read))
	    {
	    case 0:
	      do_reading = TRUE;
	      break;
	    case 1:
	      do_reading = FALSE;
	      break;
	    default:
	      {
		g_error ("error polling");
	      }
	    }
	}
      if (do_reading)
	{
	  /*
	   * Read from read_fd.
	   */

	  /* Figure out how many bytes we can sneak
	   * contiguously at the end of the buffer. */
	  int end_max_length = max_buffer - buffer_length;
	  if (buffer_start + buffer_length <= max_buffer)
	    end_max_length -= buffer_start;

	  if (end_max_length >= MIN_READ)
	    {
	      int amt;
	    retry_read1:
	      if ((amt = read (fd_read,
			       circular_buffer + buffer_start
			       + buffer_length, end_max_length) < 0))
		{
		  if (errno == EINTR)
		    goto retry_read1;
		  g_error ("error reading from "
			   "pipe or socket: %s", strerror (errno));
		}
	      if (amt == 0)
		done_reading = TRUE;
	      buffer_length += amt;
	    }
	  else
	    {
	      int amt;
	      int min_read = MIN_READ;
	      if (min_read > max_buffer - buffer_length)
		min_read = max_buffer - buffer_length;
	    retry_read2:
	      if ((amt = read (fd_read, tmpbuf, min_read)) < 0)
		{
		  if (errno == EINTR)
		    goto retry_read2;
		  g_error ("error reading from "
			   "pipe or socket: %s", strerror (errno));
		}
	      if (amt == 0)
		done_reading = TRUE;
	      if (amt > end_max_length)
		{
		  memcpy (circular_buffer
			  + buffer_length - end_max_length,
			  tmpbuf, end_max_length);
		  memcpy (circular_buffer,
			  tmpbuf + end_max_length, amt - end_max_length);
		}
	      else
		{
		  memcpy (circular_buffer
			  + buffer_length - end_max_length, tmpbuf, amt);
		}
	    }
	}
      else
	{
	  /*
	   * Read from write_fd.
	   */
	  int to_write;
	  int wrote;
	  /* Do writing (to sound device presumably...) */
	  g_assert (buffer_length >= block_size);
	  if (buffer_start == max_buffer)
	    buffer_start = 0;
	  to_write = buffer_length;
	  if (max_buffer - buffer_start < to_write)
	    to_write = max_buffer - buffer_start;
	  wrote = write_by_blocks (fd_write,
				   circular_buffer + buffer_start,
				   to_write / block_size, block_size);
	  if (wrote <= 0)
	    g_error ("error writing (to sound device)");
	}
    }
  g_free (circular_buffer);
  g_free (tmpbuf);
}

void  gdam_send_file_cannot_select_for_write(int fd_write,
                                             int fd_read,
					     int block_size)
{
  char *buffer;
  buffer = g_new (char, block_size);
  while (readn (fd_read, buffer, block_size)
	 && writen (fd_write, buffer, block_size))
    ;
  g_free (buffer);
}
