/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsink.h"
#include "gdamdevice.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"
#include "gdasfilter.h"
#include <common/gdamplugin.h>

static void gdam_sink_class_init       (GdamSinkClass*    klass);
static void gdam_sink_init             (GdamSink*         sink);
static void gdam_sink_destroy          (GdamSink*         sink);

#if 0
static void gdam_sink_set_arg          (GtkObject*        sink,
                                        GtkArg*           arg,
					guint             arg_id);
static void gdam_sink_get_arg          (GtkObject*        sink,
                                        GtkArg*           arg,
					guint             arg_id);
#endif

static GtkObjectClass* parent_class = NULL;

GtkType gdam_sink_get_type()
{
  static GtkType sink_type = 0;
  if (!sink_type)
    {
      static const GtkTypeInfo sink_info = {
	"GdamSink",
	sizeof (GdamSink),
	sizeof (GdamSinkClass),
	(GtkClassInitFunc) gdam_sink_class_init,
	(GtkObjectInitFunc) gdam_sink_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SERVER_OBJECT;
      parent_class = gtk_type_class (parent);
      sink_type = gtk_type_unique (parent, &sink_info);
    }
  return sink_type;
}

static void gdam_sink_class_init(GdamSinkClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  object_class->destroy = (void (*)(GtkObject *)) gdam_sink_destroy;
}

static void gdam_sink_init(GdamSink* sink)
{
  sink->feed = NULL;
}

static void gdam_sink_destroy(GdamSink* sink)
{
  if (sink->feed)
    gdam_sink_remove_source (sink);
  parent_class->destroy ((GtkObject *) sink);
  if (sink->device_name)
    {
      g_free (sink->device_name);
      sink->device_name = NULL;
    }
}

#if 0
int gdam_sink_write(GdamSink* input, const GdamSample* samples, int max_write)
{
  GtkObjectClass *klass = ((GtkObject *) input)->klass;
  GdamSinkClass *sink_class = (GdamSinkClass *) klass;
  return sink_class->write (input, samples, max_write);
}
#endif

gboolean gdam_sink_set_source(GdamSink* sink, GdasSource* source)
{
  GtkObjectClass *klass = ((GtkObject *) sink)->klass;
  GdamSinkClass *sink_class = (GdamSinkClass *) klass;
  GtkObject *sub;
  if (sink->feed)
    gdam_sink_remove_source (sink);
  if (!source)
    return TRUE;

  if (sink->device_name)
    {
      GdasSource *feed = source;
      for (;;)
	{
	  GdasFilter *filter;
	  if (!GDAM_IS_FILTER (feed))
	    break;
	  filter = GDAS_FILTER (feed);
	  if (filter->filtered == NULL)
	    break;
	  feed = filter->filtered;
	}
      gdam_namespace_int_set ("device-feeds",
			      sink->device_name,
			      GDAM_SERVER_OBJECT (feed)->global_id);
    }
  else
    {
      g_warning ("error, NULL device_name in gdam_sink_set_source");
    }

  sub = GTK_OBJECT (source);
  gtk_object_ref (sub);
  if (sink_class->set_source (sink, source))
    {
      gtk_object_sink (GTK_OBJECT (source));
      gtk_object_unref (sub);
      return TRUE;
    }
  gtk_object_unref (sub);
  return FALSE;
}

void gdam_sink_remove_source(GdamSink* sink)
{
  GtkObjectClass *klass = ((GtkObject *) sink)->klass;
  GdamSinkClass *sink_class = (GdamSinkClass *) klass;
  if (sink_class->remove_source)
    sink_class->remove_source (sink);
  else
    gtk_object_unref (GTK_OBJECT (sink->feed));
  sink->feed = NULL;
}

gboolean gdam_sink_open(GdamSink*   sink,
                        const char* device_name,
			const char* device_options)
{
  GtkObjectClass *klass = ((GtkObject *) sink)->klass;
  GdamSinkClass *sink_class = (GdamSinkClass *) klass;
  g_return_val_if_fail (sink_class->open != NULL, FALSE);
  gdam_debug ("gdam_sink_open device %s", device_name);
  return sink_class->open (sink, device_name, device_options);
}
void gdam_sink_close(GdamSink* sink)
{
  GtkObjectClass *klass = ((GtkObject *) sink)->klass;
  GdamSinkClass *sink_class = (GdamSinkClass *) klass;
  g_return_if_fail (sink_class->close != NULL);
  sink_class->close (sink);
}

GdamSink* gdam_sink_new_generic_v(const char* type_name,
                                  int         num_args,
			          GdamArg*    args)
{
  GtkType gtk_type;
  GdamSink *sink;
  if (!type_name)
    {
      gdam_debug ("to gtktype for `%s'", type_name);
      return NULL;
    }
  gtk_type = gdam_plugin_get_type_by_name ("server", type_name);
  if (gtk_type == 0)
    {
      g_warning ("no plugin of type %s found", type_name);
      return NULL;
    }
  sink = GDAM_SINK (gtk_type_new (gtk_type));
  if (!sink)
    {
      gdam_debug ("error instantiating sink of type %s", type_name);
      return NULL;
    }
  if (num_args > 0)
    {
      int i;
      g_return_val_if_fail (args != NULL, NULL);
      for (i = 0; i < num_args; i++)
	gdam_arg_set_object_arg ((GtkObject *) sink, args + i);
    }
  return sink;
}

