/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasbufferingfilter.h"
#include "gdamdispatch.h"
#include <common/gdam-protocol.h>
static int gdas_buffering_filter_read(GdasSource* source,
                                      GdamSample* sample,
				      int         num_read);

static void        gdas_buffering_filter_class_init         (GdasBufferingFilterClass*   klass);
static void        gdas_buffering_filter_init               (GdasBufferingFilter*        this);

static GtkObjectClass* parent_class;

#ifndef G_CAN_INLINE
G_INLINE_FUNC void gdas_buffering_filter_append_sample
						(GdasBufferingFilter *filter,
						 GdamSample           sample)
{
  if (filter->buffer_length == 0)
    return;
  ++(filter->buffer_inbound_pos);
  if (filter->buffer_inbound_pos == filter->buffer_length)
    filter->buffer_inbound_pos = 0;
  filter->buffer[filter->buffer_inbound_pos] = sample;
  if (filter->buffer_current_amount < filter->buffer_length)
    filter->buffer_current_amount++;
}
	

G_INLINE_FUNC GdamSample gdas_buffering_filter_sample_at
						(GdasBufferingFilter *filter,
						 int                  delay)
{
  int pos = filter->buffer_inbound_pos - delay;
  if (pos < 0)
    pos += filter->buffer_length;
  return filter->buffer[pos];
}

G_INLINE_FUNC GdamSample gdas_buffering_filter_sample_from_start
						(GdasBufferingFilter *filter,
						 int                  offset)
{
  int pos = filter->buffer_inbound_pos
    - filter->buffer_current_amount + offset;
  if (pos < 0)
    pos += filter->buffer_length;
  return filter->buffer[pos];
}
#endif /* ! G_CAN_INLINE */

GtkType gdas_buffering_filter_get_type()
{
  static GtkType buffering_filter_type = 0;
  if (!buffering_filter_type)
    {
      static const GtkTypeInfo buffering_filter_info = {
	"GdasBufferingFilter",
	sizeof (GdasBufferingFilter),
	sizeof (GdasBufferingFilterClass),
	(GtkClassInitFunc) gdas_buffering_filter_class_init,
	(GtkObjectInitFunc) gdas_buffering_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_FILTER);
      buffering_filter_type = gtk_type_unique (GDAS_TYPE_FILTER,
					       &buffering_filter_info);
    }
  return buffering_filter_type;
}

static void gdas_buffering_filter_class_init(GdasBufferingFilterClass* klass)
{
  GDAS_SOURCE_CLASS (klass)->read = gdas_buffering_filter_read;
}

static void gdas_buffering_filter_init(GdasBufferingFilter* this)
{
}

void gdas_buffering_filter_buffer(GdasBufferingFilter* this,
				  const GdamSample* src, int n)
{
  int max1 = this->buffer_length - this->buffer_inbound_pos;
  GdamSample *inbound = this->buffer + this->buffer_inbound_pos;
  g_return_if_fail (n <= this->buffer_length);
  if (n < max1)
    {
      gdam_sample_copy (inbound, src, n);
      this->buffer_inbound_pos += n;
    }
  else
    {
      int remainder = n - max1;
      gdam_sample_copy (inbound, src, max1);
      gdam_sample_copy (this->buffer, src + max1, remainder);
      this->buffer_inbound_pos = remainder;
    }
  this->buffer_current_amount += n;
}

int gdas_buffering_filter_buffer_from_source(GdasBufferingFilter* this,
				             GdasSource* src, int n)
{
  int max1 = this->buffer_length - this->buffer_inbound_pos;
  GdamSample *inbound = this->buffer + this->buffer_inbound_pos;
  if (src == NULL)
    return 0;
  g_return_val_if_fail (n <= this->buffer_length, -1);
  if (n < max1)
    {
      int amt_read;
      if ((amt_read = gdas_source_read (src, inbound, n)) == -1)
	return amt_read;
      this->buffer_inbound_pos += amt_read;
      this->buffer_current_amount += amt_read;
      /* PARANOIA: can this happen? */
      if (this->buffer_inbound_pos == this->buffer_length)
	this->buffer_inbound_pos = 0;
      return amt_read;
    }
  else
    {
      int remainder = n - max1;
      int amt_read;
      if ((amt_read = gdas_source_read (src, inbound, max1)) == max1)
	{
	  if (remainder > 0 && gdas_source_is_ready_to_read (src))
	    {
	      amt_read = gdas_source_read (src,
					   this->buffer + max1, remainder);
	    }
	  else
	    {
	      amt_read = 0;
	    }
	  this->buffer_inbound_pos = amt_read;
	  this->buffer_current_amount += amt_read;
	  if (this->buffer_current_amount > this->buffer_length)
	    this->buffer_current_amount = this->buffer_length;
	  if (this->buffer_inbound_pos == this->buffer_length)
	    this->buffer_inbound_pos = 0;
	  return max1 + amt_read;
	}
      else
	{
	  this->buffer_inbound_pos += amt_read;
	  this->buffer_current_amount += amt_read;
	  if (this->buffer_current_amount > this->buffer_length)
	    this->buffer_current_amount = this->buffer_length;
	  /* PARANOIA: can this happen? */
	  if (this->buffer_inbound_pos == this->buffer_length)
	    this->buffer_inbound_pos = 0;
	  return amt_read;
	}
    }
}

int gdas_buffering_filter_peek_from_buffer(GdasBufferingFilter* this,
					   GdamSample* dest, int n)
{
  int outbound_pos;
  int max1;
  GdamSample *outbound;

  if (this->buffer_length <= 0)
    return 0;

  outbound_pos = (this->buffer_inbound_pos +
		  this->buffer_current_amount) % this->buffer_length;
  max1 = this->buffer_length - outbound_pos;
  outbound = this->buffer + outbound_pos;

  if (n > this->buffer_current_amount)
    n = this->buffer_current_amount;
  if (n < max1)
    {
      gdam_sample_copy (dest, outbound, n);
    }
  else
    {
      int remainder = n - max1;
      gdam_sample_copy (dest, outbound, max1);
      gdam_sample_copy (dest + max1, this->buffer, remainder);
    }
  return n;
}
void gdas_buffering_filter_discard(GdasBufferingFilter* filter,
				   int                  n)
{
  if (filter->buffer_current_amount < n)
    filter->buffer_current_amount = 0;
  else
    filter->buffer_current_amount -= n;
}

int gdas_buffering_filter_append(GdasBufferingFilter* filter,
				  const GdamSample*    buf,
				  int                  n)
{
  int free_space = filter->buffer_length - filter->buffer_current_amount;
  int discard = 0;
  if (free_space < n)
    {
      discard = n - free_space;
      gdas_buffering_filter_discard (filter, discard);
    }
  gdas_buffering_filter_ensure_space (filter, n);
  if (filter->buffer_inbound_pos + n <= filter->buffer_length)
    {
      gdam_sample_copy (filter->buffer + filter->buffer_inbound_pos, buf, n);
      filter->buffer_inbound_pos += n;
    }
  else
    {
      int part_2 = filter->buffer_inbound_pos + n - filter->buffer_length;
      int part_1 = n - part_2;
      gdam_sample_copy (filter->buffer + filter->buffer_inbound_pos,
			buf, part_1);
      gdam_sample_copy (filter->buffer, buf + part_1, part_2);
      filter->buffer_inbound_pos = part_2;
    }
  filter->buffer_current_amount += n;
  if (filter->buffer_inbound_pos >= filter->buffer_length)
    {
      g_return_val_if_fail (filter->buffer_inbound_pos
			    == filter->buffer_length, discard);
      filter->buffer_inbound_pos = 0;
    }
  return discard;
}

void gdas_buffering_filter_ensure_space(GdasBufferingFilter* this,
		                        int       	     amount)
{
  GdamSample *new_buffer;
  int new_size;
  if (amount < this->buffer_length)
    return;
  if (this->buffer_length == 0)
    new_size = 32;
  else
    new_size = this->buffer_length * 2;
  while (new_size < amount)
    new_size *= 2;
  new_buffer = g_new (GdamSample, new_size);
  if (this->buffer)
    {
      gdas_buffering_filter_peek_from_buffer (this, new_buffer,
					      this->buffer_current_amount);
      g_free (this->buffer);
    }
  this->buffer = new_buffer;
  this->buffer_length = new_size;
  this->buffer_inbound_pos = this->buffer_current_amount;
  g_assert (this->buffer_inbound_pos < this->buffer_length);
}

static int gdas_buffering_filter_read(GdasSource* source,
                                      GdamSample* sample,
				      int         num_read)
{
  GdasBufferingFilter *buffering = GDAS_BUFFERING_FILTER (source);
  GdasSource *filtered;
  int amt_read;
  g_return_val_if_fail (buffering != NULL, -1);
  filtered = GDAS_FILTER (source)->filtered;
  if (filtered == NULL)
    return 0;
  amt_read = gdas_source_read (filtered, sample, num_read);
  (void) gdas_buffering_filter_append (buffering, sample, amt_read);
  return amt_read;
}
