/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdascache.h"
#include <common/gdamarg.h>

static int       gdas_cache_read               (GdasSource*      input,
	                                        GdamSample*      samples,
						int              max_read);
static GdamTime  gdas_cache_get_position       (GdasSource*      input);
static int       gdas_cache_seek               (GdasSource*      input,
                                                GdamTime         offset,
						int              whence);
static void      gdas_cache_miss_data          (GdasSource*      input,
	                                        int              samples_lost);
static void      gdas_cache_destroy            (GtkObject*       object);
static gboolean  gdas_cache_setup              (GdasCache*       cache,
                                                guint            cache_id);

static GtkObjectClass* parent_class = NULL;

static void gdas_cache_arg_set_buffer_id(GtkObject* object, GdamArg* arg)
{
  gdas_cache_setup (GDAS_CACHE (object), GDAM_ARG_INT (arg));
}
static void gdas_cache_arg_get_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdamCacheBuffer *buffer = GDAS_CACHE (object)->cache;
  GDAM_ARG_INT (arg) = buffer ? gdam_cache_buffer_get_id (buffer) : 0;
}
static void gdas_cache_arg_set_loop(GtkObject* object, GdamArg* arg)
{
  GDAS_CACHE (object)->should_loop = GDAM_ARG_INT (arg);
}
static void gdas_cache_arg_get_loop(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_CACHE (object)->should_loop;
}
static void gdas_cache_arg_set_length(GtkObject* object, GdamArg* arg)
{
  GDAS_CACHE (object)->remaining = GDAM_ARG_INT (arg);
}
static void gdas_cache_arg_get_length(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_CACHE (object)->remaining;
}


static void gdas_cache_class_init(GdasSourceClass* to_init)
{
  GtkObjectClass *object_class = (GtkObjectClass *) to_init;
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (to_init);
  to_init->read = gdas_cache_read;
  to_init->miss_data = gdas_cache_miss_data;
  object_class->destroy = gdas_cache_destroy;
  source_class->get_position = gdas_cache_get_position;
  source_class->seek = gdas_cache_seek;

  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdas_cache_arg_set_buffer_id,
		     gdas_cache_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "length",
		     GDAM_ARG_TYPE_INT,
		     gdas_cache_arg_set_length,
		     gdas_cache_arg_get_length, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "loop",
		     GDAM_ARG_TYPE_INT,
		     gdas_cache_arg_set_loop,
		     gdas_cache_arg_get_loop,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

static void gdas_cache_init(GdasCache* to_init)
{
  to_init->remaining = -1;
}

GtkType gdas_cache_get_type ()
{
  static GtkType cache_type = 0;
  if (!cache_type)
    {
      static const GtkTypeInfo cache_info = {
	"GdasCache",
	sizeof (GdasCache),
	sizeof (GdasSourceClass),
	(GtkClassInitFunc) gdas_cache_class_init,
	(GtkObjectInitFunc) gdas_cache_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      cache_type = gtk_type_unique (GDAS_TYPE_SOURCE, &cache_info);
    }
  return cache_type;
}

static gboolean  	gdas_cache_setup        (GdasCache*        cache,
                                                 guint             cache_id)
{
  GdamCacheBuffer *cache_buf;
  cache_buf = gdam_cache_buffer_get_by_id (cache_id);
  if (cache_buf == NULL)
    return FALSE;
  if (cache_buf->buffer_length <= 0)
    {
      gdam_debug ("cache %u: still loading...", cache_id);
      return FALSE;
    }
  gtk_object_ref (GTK_OBJECT (cache_buf));
  if (cache->cache)
    gtk_object_unref (GTK_OBJECT (cache->cache));
  cache->cache = cache_buf;
  cache->cache_index %= cache_buf->buffer_length;
  return TRUE;
}

static int       gdas_cache_read               (GdasSource*      input,
	                                        GdamSample*      samples,
						int              max_read)
{
  GdasCache *cache = GDAS_CACHE (input);
  int buffer_len;
  GdamSample *buffer;
  int rv = 0;

  if (cache->cache == NULL)
    return 0;

  buffer = cache->cache->buffer;
  buffer_len = cache->cache->buffer_length;

  if (cache->remaining >= 0 && max_read > cache->remaining)
    max_read = cache->remaining;
  /* TODO: see if this code is notably more efficient pre-splitting
   *       into looped/non-looped cases... (probably not) */
  while (max_read > 0)
    {
      int amt = buffer_len - cache->cache_index;
      if (amt > max_read)
	amt = max_read;
      if (amt <= 0)
	break;

      if (cache->cache_index < 0)
	{
	  int to_zero = -cache->cache_index;
	  if (to_zero > amt)
	    to_zero = amt;
	  gdam_sample_zero (samples, to_zero);
	  gdam_sample_copy (samples + to_zero,
			    buffer + cache->cache_index, amt - to_zero);
	}
      else
	{
	  gdam_sample_copy (samples, buffer + cache->cache_index, amt);
	}

      cache->cache_index += amt;
      rv += amt;
      samples += amt;
      max_read -= amt;

      if (cache->cache_index >= buffer_len)
	{
	  g_assert (cache->cache_index == buffer_len);
	  if (cache->should_loop)
	    cache->cache_index = 0;
	  else
	    break;
	}
    }
  if (cache->remaining > 0)
    cache->remaining -= rv;
  return rv;
}

static GdamTime  gdas_cache_get_position       (GdasSource*      input)
{
  GdasCache *cache = GDAS_CACHE (input);
  return cache->cache_index;
}
static int       gdas_cache_seek               (GdasSource*      input,
                                                GdamTime         offset,
						int              whence)
{
  GdasCache *cache = GDAS_CACHE (input);
  switch (whence)
    {
    case 0:
      cache->cache_index = offset;
      return TRUE;
    case 1:
      cache->cache_index += offset;
      return TRUE;
    case 2:
      {
	if (cache->cache)
	  {
	    cache->cache_index = cache->cache->buffer_length + offset;
	  }
	return TRUE;
      }
    }
  return FALSE;
}
static void      gdas_cache_miss_data          (GdasSource*      input,
	                                        int              samples_lost)
{
  GdasCache *cache = GDAS_CACHE (input);
  int new_index = cache->cache_index;
  new_index += samples_lost;
  new_index %= cache->cache->buffer_length;
  cache->cache_index = new_index;
}

static void      gdas_cache_destroy            (GtkObject*       object)
{
  GdasCache *cache = GDAS_CACHE (object);
  if (cache->cache)
    {
      GtkObject *buf = GTK_OBJECT (cache->cache);
      cache->cache = NULL;
      gtk_object_unref (buf);
    }
  parent_class->destroy (object);
}
