/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdassource.h"
#include "gdasexternal.h"
#include "gdamcleanup.h"
#include "gdamconfigfile.h"
#include <common/gdambuffer.h>
#include <common/gdam-protocol.h>
#include <common/gdamsampleconverter.h>
#include "gdamconnection.h"
#include "gdamevent.h"
#include "gdamserverdebug.h"
#include "gdamrealtime.h"
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

/* Wait until 1/4 the maximum buffer is full,
 * or MAX_TO_WAIT_FOR samples are ready
 * before saying the source is ready to read. */
#define FRACTION_TO_WAIT_FOR	4
#define MAX_TO_WAIT_FOR		8192
#define GDAS_EXTERNAL_PRIORITY       G_PRIORITY_HIGH

/*
 * Notes:
 *   - there is no way to run an external source generally from
 *     the server.  That would be a big security hole.
 */

static void        gdas_external_class_init       (GdasSourceClass* klass);
static void        gdas_external_init             (GdasExternal*    external);
static void        gdas_external_destroy          (GdasExternal*    source);

static GdamTime    gdas_external_get_position     (GdasSource*      source);
static gboolean    gdas_external_is_ready_to_read (GdasSource*      source);
static int	   gdas_external_read             (GdasSource*      input,
					           GdamSample*      samples,
					           int              max_read);
static void        gdas_external_miss_data        (GdasSource*      input,
					           int              lost);
static gboolean    gdas_external_prepare          (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data);
static gboolean    gdas_external_check            (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);
static gboolean    gdas_external_dispatch         (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);

static inline void add_poll_fd(GdasExternal* external)
{
  g_main_add_poll (&external->poll_fd, G_PRIORITY_DEFAULT);
  external->has_poll_fd = TRUE;
}
static inline void remove_poll_fd(GdasExternal* external)
{
  g_main_remove_poll (&external->poll_fd);
  external->has_poll_fd = FALSE;
}


	/** derivers might usefully define these,
	 ** but we didn't want to give the client end
	 ** the general controls until they stabilize....
	 ** nonetheless, a generic `pause' and `resume'
	 ** and so on might be later implemented here. */
#if 0
static void        gdas_external_set_arg   (GtkObject*        external,
                                            GtkArg*           arg,
					    guint             arg_id);
static void        gdas_external_get_arg   (GtkObject*        external,
                                            GtkArg*           arg,
					    guint             arg_id);
#endif

static GtkObjectClass*  parent_class = NULL;

GtkType gdas_external_get_type()
{
  static GtkType external_type = 0;
  if (!external_type)
    {
      static const GtkTypeInfo external_info = {
	"GdasExternal",
	sizeof (GdasExternal),
	sizeof (GdasExternalClass),
	(GtkClassInitFunc) gdas_external_class_init,
	(GtkObjectInitFunc) gdas_external_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      external_type = gtk_type_unique (GDAS_TYPE_SOURCE, &external_info);
    }
  return external_type;
}
static void gdas_external_set_length(GtkObject* object, GdamArg* arg)
{
  GdasExternal *external = GDAS_EXTERNAL (object);
  external->samples_remaining = GDAM_ARG_INT64 (arg);
}
static void gdas_external_get_length(GtkObject* object, GdamArg* arg)
{
  GdasExternal *external = GDAS_EXTERNAL (object);
  GDAM_ARG_INT64 (arg) = external->samples_remaining;
}

static void gdas_external_set_buffer_size(GtkObject* object, GdamArg* arg)
{
  GdasExternal *external = GDAS_EXTERNAL (object);
  external->max_buffer_size = GDAM_ARG_INT (arg);
}
static void gdas_external_get_buffer_size(GtkObject* object, GdamArg* arg)
{
  GdasExternal *external = GDAS_EXTERNAL (object);
  GDAM_ARG_INT (arg) = external->max_buffer_size;
}

#if 0
static void debug_command(char** command)
{
  char *word;
  char **cmd = command;
  g_return_if_fail (command != NULL);
  g_message ("DEBUG COMMAND: ");
  for (; *cmd != NULL; cmd++)
    {
      word = *cmd;
      fprintf (stderr, "%s ", word);
    }
  fprintf (stderr, "\n\n");
}
#endif

static void gdas_external_real_do_exec(GdasExternal* external,
                                       char**        command)
{
  gdam_realtime_unset ();
  //debug_command(command);
  if (external->stdin_redirect_fd != STDIN_FILENO)
    dup2 (external->stdin_redirect_fd, STDIN_FILENO);
  execvp (command[0], command);
}

static void gdas_external_class_init(GdasSourceClass* source_class)
{
  GtkObjectClass *object_class = (GtkObjectClass *) source_class;
  GdasExternalClass *external_class = GDAS_EXTERNAL_CLASS (source_class);
  if (!parent_class)
    parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
  source_class->read = gdas_external_read;
  source_class->miss_data = gdas_external_miss_data;
  source_class->is_ready_to_read = gdas_external_is_ready_to_read;
  source_class->get_position = gdas_external_get_position;
  external_class->do_exec = gdas_external_real_do_exec;
  object_class->destroy = (void (*)(GtkObject *)) gdas_external_destroy;

  gdam_arg_register (object_class,
		     "length",
		     GDAM_ARG_TYPE_INT64,
		     gdas_external_set_length,
		     gdas_external_get_length, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_size",
		     GDAM_ARG_TYPE_INT,
		     gdas_external_set_buffer_size,
		     gdas_external_get_buffer_size,
		     NULL, "0,", GDAM_ARG_VITAL);
}

static int default_buffer_size = 64 * 1024;

int gdas_external_get_default_buffer_size()
{
  return default_buffer_size;
}
void gdas_external_set_default_buffer_size(int buffer_size)
{
  default_buffer_size = buffer_size;
}

static void gdas_external_init(GdasExternal* external)
{
  int src_bits_per_sample;
  gboolean is_big_endian = (G_BYTE_ORDER == G_BIG_ENDIAN);
  external->process_id = -1;
  external->poll_fd.fd = -1;
  external->poll_fd.events = G_IO_IN | G_IO_ERR | G_IO_HUP;
  external->reader_source = 0;
  external->skipped = 0;
  external->max_buffer_size = gdas_external_get_default_buffer_size ();
  external->samples_remaining = -1;
  external->num_channels = 2;
  external->bytes_per_sample = 2;
  external->stdin_redirect_fd = STDIN_FILENO;
  src_bits_per_sample = external->bytes_per_sample * 8;
  external->sample_converter = gdam_sample_converter_get
    (src_bits_per_sample, TRUE, is_big_endian,
     GDAM_SAMPLE_BITS, TRUE, is_big_endian);
  external->channel_converter = NULL;
  gdam_buffer_construct (&external->buffer);
}

/*=====================================================================*
 *                      gdas_external_kill_command                     *
 *                                                                     *
 * external                                                            *
 * the source which should stop running the command.                   *
 *=====================================================================*/
gboolean gdas_external_kill_command(GdasExternal* external,
                                    int*          exit_status)
{
  int tmp, pid;
  if (!exit_status)
    exit_status = &tmp;
  if (external->poll_fd.fd >= 0)
    {
      if (external->has_poll_fd)
	remove_poll_fd (external);
      if (!external->dont_close)
	close (external->poll_fd.fd);
      external->poll_fd.fd = -1;
    }
  if (external->reader_source)
    {
      guint source = external->reader_source;
      external->reader_source = 0;
      gdam_debug ("external_kill_command: gsource=%d, external=%p,%d",
		  source, external, GDAM_SERVER_OBJECT (external)->global_id);
      g_source_remove (source);
    }
  if (external->process_id < 0)
    return FALSE;
  external->skipped = 0;
  pid = external->process_id;
  gdam_cleanup_unregister_child (pid);
  external->process_id = -1;
  if (waitpid (pid, exit_status, WNOHANG) > 0)
    {
      return TRUE;
    }
  kill (pid, SIGTERM);
  kill (pid, SIGKILL);
  while (waitpid (pid, exit_status, 0) < 0)
    {
      if (errno == EINTR)
	continue;
      perror ("waitpid");
      return FALSE;
    }
  return TRUE;
}

static GSourceFuncs external_source_funcs = {
	gdas_external_prepare,
	gdas_external_check,
	gdas_external_dispatch,
	NULL
};

/*=====================================================================*
 *                       gdas_external_has_opened                      *
 *                                                                     *
 * Tell the external to start reading off a given                      *
 * file descriptor.                                                    *
 *                                                                     *
 * Parameters:                                                         *
 *     external          the source to affect.                         *
 *     fd                the file descriptor to read from              *
 *=====================================================================*/
 void gdas_external_has_opened(GdasExternal* external, int fd)
 {
	g_return_if_fail(external->poll_fd.fd == -1);
	g_return_if_fail(external->poll_fd.events==(G_IO_IN|G_IO_ERR|G_IO_HUP));
	g_return_if_fail(external->skipped == 0);
	g_return_if_fail(external->reader_source == 0);

	external->poll_fd.fd = fd;

	external->reader_source = g_source_add(GDAS_EXTERNAL_PRIORITY,
	                                       TRUE,
		                               &external_source_funcs,
		                               NULL,
		                               external,
		                               NULL);
}

/*=====================================================================*
 *                      gdas_external_run_command                      *
 *                                                                     *
 * Start running a command in the gdas_external class.                 *
 *                                                                     *
 * TODO:  simulate `/bin/sh -c command' like perl does.                *
 *                                                                     *
 * Parameters:                                                         *
 *     external          the source which should run the command.      *
 *     sh_command        the command to run.                           *
 *=====================================================================*/
gboolean gdas_external_run_command(GdasExternal* external, 
                                   char**        command)
{
  int fds[2];

  if (external->process_id >= 0)
    {
      (void) gdas_external_kill_command (external, NULL);
    }

  g_return_val_if_fail (external->poll_fd.fd == -1, FALSE);
  g_return_val_if_fail (external->poll_fd.events ==
			(G_IO_IN | G_IO_ERR | G_IO_HUP), FALSE);
  g_return_val_if_fail (external->skipped == 0, FALSE);
  g_return_val_if_fail (external->process_id == -1, FALSE);
  g_return_val_if_fail (external->reader_source == 0, FALSE);
  g_return_val_if_fail (!GTK_OBJECT_DESTROYED (external), FALSE);

  if (pipe (fds) < 0)
    {
      perror ("creating pipe");
      return FALSE;
    }
  /*  Set close-on-exec flag  */
  fcntl (fds[0], F_SETFD, 1);
  gdam_buffer_construct (&external->buffer);
  if (external->command)
    g_strfreev (external->command);
  external->skipped = 0;
retry_fork:
  external->process_id = fork ();
  if (external->process_id < 0)
    {
      if (errno == EAGAIN)
	{
	  sleep (3);
	  goto retry_fork;
	}
      if (errno == EINTR)
	goto retry_fork;
      perror ("error forking");
      g_free (external);
      close (fds[0]);
      close (fds[1]);
      return FALSE;
    }
  if (external->process_id == 0)
    {
      GdasExternalClass *external_class;
      GtkObject *object;
      object = GTK_OBJECT (external);
      external_class = GDAS_EXTERNAL_CLASS (object->klass);

      if (gdam_external_nice != 0)
	nice (gdam_external_nice);

      dup2 (fds[1], STDOUT_FILENO);
      (*external_class->do_exec) (external, command);

      _exit (127);
    }
  close (fds[1]);


  gdas_external_has_opened (external, fds[0]);
  gdam_cleanup_register_child (external->process_id);

  return TRUE;
}
static gboolean do_one_read(GdasExternal* external)
{
  int n;
  char buf[32 * 1024];
redo_read:
  n = read (external->poll_fd.fd, buf, sizeof (buf));
  //fprintf(stderr, "dispatch: read %d into buffer\n",n);
  if (n < 0)
    {
      if (errno == EINTR)
	goto redo_read;
      if (errno == EWOULDBLOCK)
	return TRUE;
      perror ("reading from external source");
      external->reader_source = 0;
      remove_poll_fd (external);
      return FALSE;
    }
  if (n == 0)
    {
      gdam_debug ("end-of-file from external source "
		  "[gsource %d, %p,%d]",
		  external->reader_source, external,
		  GDAM_SERVER_OBJECT (external)->global_id);
      external->reader_source = 0;
      remove_poll_fd (external);
      return FALSE;
    }
  if (n <= external->skipped)
    {
      external->skipped -= n;
      return TRUE;
    }
  else if (external->skipped)
    {
      int s = external->skipped;
      gdam_buffer_append (&external->buffer, buf + s, n - s);
      external->skipped = 0;
      return TRUE;
    }

  gdam_buffer_append (&external->buffer, buf, n);
  return TRUE;
}

/*=====================================================================*
 *                          The GSourceFuncs.                          *
 *                                                                     *
 * These set up the PollFD (and indicate generally how                 *
 * to process the events involved.)                                    *
 *                                                                     *
 * AFAICT there is no convenient way to do this otherwise--            *
 * we'd have to add/remove watches all the time.                       *
 *=====================================================================*/

/* This function tells whether we know we need servicing 
 * before doing the poll.  It is also a good time to 
 * set up the PollFD.
 *
 * We always only need servicing if the fd is readable,
 * so we always return FALSE.
 *
 * But we do determine whether to poll at all via setting
 * external->poll_fd.events appropriately. */
static gboolean
gdas_external_prepare(gpointer  source_data,
                      GTimeVal* curtime,
                      gint*     timeout,
		      gpointer  user_data)
{
  GdasExternal *external = GDAS_EXTERNAL (user_data);

  g_return_val_if_fail (!GTK_OBJECT_DESTROYED (external), FALSE);

  /* We never have anything to do unless the source is ready. */
  *timeout = -1;

  /* Returning events == 0 means we have nothing to process. */
  external->poll_fd.events = external->poll_fd.revents = 0;

  /* We shouldn't be added as a source unless we have a valid
   * file descriptor... */
  g_return_val_if_fail (external->poll_fd.fd >= 0, FALSE);

  /* Set up the read fd only if our buffer isn't
   * too full already. */
  if (external->buffer.size >= external->max_buffer_size)
    {
      g_return_val_if_fail (external->skipped == 0, FALSE);
      external->poll_fd.events = 0;
      return FALSE;
    }

  if (!external->has_poll_fd)
    add_poll_fd (external);

  external->poll_fd.events = G_IO_IN | G_IO_ERR | G_IO_HUP;
  return FALSE;
}


/* Return whether we need servicing.
 * This is simple-- we just test revents. */
static gboolean
gdas_external_check(gpointer  source_data,
                    GTimeVal* curtime,
		    gpointer  user_data)
{
  GdasExternal *external = GDAS_EXTERNAL (user_data);
  g_return_val_if_fail (external->poll_fd.fd >= 0, FALSE);
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdas_external_check: revents=%d", external->poll_fd.revents);

  return external->poll_fd.revents ? TRUE : FALSE;
}
static gboolean
gdas_external_dispatch(gpointer  source_data,
                       GTimeVal* curtime,
		       gpointer  user_data)
{
  GdasExternal *external = GDAS_EXTERNAL (user_data);
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdas_external_dispatch");
  return do_one_read (external);
}


static gboolean
gdas_external_is_ready_to_read(GdasSource* source)
{
  GdasExternal *external = GDAS_EXTERNAL (source);
  int min_size = 1;
  int sample_size = external->num_channels * external->bytes_per_sample;

  /* Sources with no readers are either buffered,
   * or at eof.  Never blocking on anything at least... */
  if (external->reader_source == 0)
    return TRUE;

  if (external->buffer.size < sample_size)
    return FALSE;

  /* TEST: don't say we're ready to read unless we have 
   * a 1/FRACTION_TO_WAIT_FOR our buffer
   * or MAX_TO_WAIT_FOR samples, whichever is
   * lower. */
  if (!external->has_been_ready)
    {
      min_size = external->max_buffer_size / FRACTION_TO_WAIT_FOR;
      min_size = MIN (min_size, MAX_TO_WAIT_FOR);
      if (external->buffer.size >= sample_size * min_size)
	{
	  external->has_been_ready = 1;
	}
    }

  return (external->buffer.size >= sample_size * min_size);
}

static int
gdas_external_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasExternal *external = GDAS_EXTERNAL (source);
  int num_channels;
  int rv = 0;
  int sample_size;
  if (external->zero_time > 0)
    {
      int i = MIN (n, external->zero_time);
      external->zero_time -= i;
      n -= i;
      gdam_sample_zero (buf, i);
      rv = i;
      buf += i;
    }

  if (external->buffer.size < (int) sizeof (GdamSample))
    {
      if (external->reader_source == 0)
	return rv;

      if (rv > 0)
	return rv;

      bzero (buf, sizeof (GdamSample));
      gdas_external_miss_data (source, 1);
      return rv + 1;
    }
  num_channels = external->num_channels;
  sample_size = external->num_channels * external->bytes_per_sample;
  if (external->buffer.size / sample_size < n)
    n = external->buffer.size / sample_size;
  if (external->samples_remaining >= 0)
    {
      if (n > external->samples_remaining)
	n = (int) external->samples_remaining;
      external->samples_remaining -= n;
    }
  gdam_buffer_read (&external->buffer, (char *) buf, n * sample_size);
  external->samples_output += n;

  if (external->sample_converter != NULL)
    (*external->sample_converter) (buf, buf, n * num_channels);
  if (external->channel_converter != NULL)
    (*external->channel_converter) (buf, buf, n);

  return rv + n;
}

static void gdas_external_miss_data(GdasSource* source, int n)
{
  int discarded;
  GdasExternal *external = GDAS_EXTERNAL (source);

  if (external->ignore_miss_data)
    return;

  if (external->samples_remaining > 0)
    {
      if (n > external->samples_remaining)
	external->samples_remaining = 0;
      else
	external->samples_remaining -= n;
    }
  if (external->zero_time > 0)
    {
      int i = MIN (n, external->zero_time);
      external->zero_time -= i;
      n -= i;
    }

  external->samples_output += n;

  n *= external->num_channels * external->bytes_per_sample;
  discarded = gdam_buffer_discard (&external->buffer, n);
  n -= discarded;		/* remaining bytes we cannot yet discard... */
  external->skipped += n;
}

static void gdas_external_destroy(GdasExternal* external)
{
  g_return_if_fail (external != NULL);
  gdas_external_kill_command (external, NULL);

  if (external->command)
    {
      g_strfreev (external->command);
      external->command = NULL;
    }

  gdam_buffer_destruct (&external->buffer);
  parent_class->destroy ((GtkObject *) external);
}

void gdas_external_set_num_channels(GdasExternal* external, int num_channels)
{
  g_return_if_fail (num_channels == 1 || num_channels == 2);
  external->channel_converter
    = gdam_sample_converter_get_channel_changer (num_channels, 2);
  external->num_channels = num_channels;
}

static GdamTime gdas_external_get_position(GdasSource* source)
{
  GdasExternal *external = GDAS_EXTERNAL (source);
  return (gint64) external->samples_output - (gint64) external->zero_time;
}
