/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasgain.h"
#include "gdamdispatch.h"
#include <common/gdam-protocol.h>


static int         gdas_gain_read               (GdasSource*	  gain,
                                       	         GdamSample*	  buf,
				       	         int		  n);

static void        gdas_gain_class_init         (GdasFilterClass* klass);
static void        gdas_gain_init               (GdasGain*      gain);

static GtkObjectClass* parent_class;

GtkType gdas_gain_get_type()
{
  static GtkType gain_type = 0;
  if (!gain_type)
    {
      static const GtkTypeInfo gain_info = {
	"GdasGain",
	sizeof (GdasGain),
	sizeof (GdasFilterClass),
	(GtkClassInitFunc) gdas_gain_class_init,
	(GtkObjectInitFunc) gdas_gain_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_FILTER);
      gain_type = gtk_type_unique (GDAS_TYPE_FILTER, &gain_info);
    }
  return gain_type;
}
static void gdas_gain_arg_set_gain(GtkObject* object, GdamArg* arg)
{
  gdouble gain = GDAM_ARG_DOUBLE (arg);
  gdas_gain_set_level (GDAS_GAIN (object),
		       (GdamFixed) (gain * GdamFixed_ONE));
}
static void gdas_gain_arg_get_gain(GtkObject* object, GdamArg* arg)
{
  gdouble gain = GDAS_GAIN (object)->current_gain;
  gain /= GdamFixed_ONE;
  GDAM_ARG_DOUBLE (arg) = gain;
}

static void gdas_gain_class_init(GdasFilterClass* klass)
{
  GdasSourceClass *source_class = (GdasSourceClass *) klass;
  source_class->read = gdas_gain_read;

  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_gain_arg_set_gain,
		     gdas_gain_arg_get_gain,
		     NULL, "-100,100", GDAM_ARG_VITAL);
}

static void gdam_sample_regain_null(GdamSample* buf,
                                    int buf_len,
                                    GdamFixed unused)
{
}

static void gdas_gain_init(GdasGain* gain)
{
  gain->regainer = gdam_sample_regain_null;
  gain->current_gain = GdamFixed_ONE;
  gain->desired_gain = GdamFixed_ONE;

  /* by default, adjust the volume from 0 => 1 over 64 samples */
  gain->max_delta_per_sample = GdamFixed_ONE / 4096;
}

static int gdas_gain_read(GdasSource*	source, GdamSample* buf, int n)
{
  GdasGain *gain = GDAS_GAIN (source);
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  int num_read;
  int at;
  g_return_val_if_fail (gain != NULL, 0);
  filtered = filter->filtered;
  if (!filtered)
    return 0;
  num_read = gdas_source_read (filtered, buf, n);
  at = 0;
  if (gain->current_gain != gain->desired_gain)
    {
      int max_delta = gain->max_delta_per_sample;
      GdamFixed cur_gain = gain->current_gain;
      GdamFixed des_gain = gain->desired_gain;
      while (cur_gain < des_gain && at < num_read)
	{
	  buf[at].s[0] = GDAM_FIXED_MUL (cur_gain, buf[at].s[0]);
	  buf[at].s[1] = GDAM_FIXED_MUL (cur_gain, buf[at].s[1]);
	  cur_gain += max_delta;
	  if (cur_gain > des_gain)
	    cur_gain = des_gain;
	  at++;
	}
      while (cur_gain > des_gain && at < num_read)
	{
	  buf[at].s[0] = GDAM_FIXED_MUL (cur_gain, buf[at].s[0]);
	  buf[at].s[1] = GDAM_FIXED_MUL (cur_gain, buf[at].s[1]);
	  cur_gain -= max_delta;
	  if (cur_gain < des_gain)
	    cur_gain = des_gain;
	  at++;
	}
      gain->current_gain = cur_gain;
    }
  if (at < num_read)
    (*gain->regainer) (buf + at, num_read - at, gain->current_gain);
  if (num_read > 0)
    gain->has_output = TRUE;
  return num_read;
}

void gdas_gain_set_level(GdasGain* gain, GdamFixed fixed)
{
  if (fixed == GdamFixed_ONE)
    gain->regainer = gdam_sample_regain_null;
  else
    gain->regainer = gdam_sample_scale;
  gain->desired_gain = fixed;
  if (gain->max_delta_per_sample <= 0 || !gain->has_output)
    gain->current_gain = fixed;
}

