/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasresample.h"
#include "gdamdispatch.h"
#include <common/gdam-protocol.h>
#include <common/gdamarg.h>
#include <stdio.h>
#include <math.h>



static int         gdas_resample_read             (GdasSource*	    resample,
                                       	           GdamSample*	    buf,
				       	           int		    n);
static void        gdas_resample_miss_data        (GdasSource*	    source,
                                       	           int              amount);
static gboolean    gdas_resample_seek             (GdasSource*      source,
                                                   GdamTime         offset,
		                                   int              whence);
static GdamTime    gdas_resample_get_position     (GdasSource*      source);

static void        gdas_resample_get_latency_info (GdasSource*      source,
					           GdasSource*      subsource,
					           GdamLatencyInfo* latency);

static void        gdas_resample_class_init       (GdasFilterClass* klass);
static void        gdas_resample_init             (GdasResample*    resample);


static void        compute_filter_coeffs();

static GtkObjectClass* parent_class;

GtkType gdas_resample_get_type()
{
  static GtkType resample_type = 0;
  if (!resample_type)
    {
      static const GtkTypeInfo resample_info = {
	"GdasResample",
	sizeof (GdasResample),
	sizeof (GdasFilterClass),
	(GtkClassInitFunc) gdas_resample_class_init,
	(GtkObjectInitFunc) gdas_resample_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_FILTER);
      resample_type = gtk_type_unique (GDAS_TYPE_FILTER, &resample_info);
    }
  return resample_type;
}
static void gdas_resample_arg_set_rate(GtkObject* object, GdamArg* arg)
{
  GdasResample *resample = GDAS_RESAMPLE (object);
  guint64 rate = (guint64) (GDAM_ARG_DOUBLE (arg) * 65536.0 * 65536.0);
  resample->rate = rate;
  resample->last_skipped = 0;
}
static void gdas_resample_arg_get_rate(GtkObject* object, GdamArg* arg)
{
  GdasResample *resample = GDAS_RESAMPLE (object);
  GDAM_ARG_DOUBLE (arg) = (double) resample->rate / (65536.0 * 65536.0);
}
static void gdas_resample_arg_set_hi_quality(GtkObject* object, GdamArg* arg)
{
  GdasResample *resample = GDAS_RESAMPLE (object);
  resample->hi_quality = GDAM_ARG_INT (arg);
}
static void gdas_resample_arg_get_hi_quality(GtkObject* object, GdamArg* arg)
{
  GdasResample *resample = GDAS_RESAMPLE (object);
  GDAM_ARG_INT (arg) = resample->hi_quality;
}

static void gdas_resample_get_latency_info(GdasSource*      source,
					   GdasSource*      subsource,
					   GdamLatencyInfo* latency)
{
  latency->latency = 0;
  g_return_if_fail (GDAS_RESAMPLE (source) != NULL);
  latency->resample = (GdamFixed) (((GdasResample *) source)->rate >> 16);
}


static void gdas_resample_class_init(GdasFilterClass* klass)
{
  GdasSourceClass *source_class = (GdasSourceClass *) klass;
  source_class->read = gdas_resample_read;
  source_class->miss_data = gdas_resample_miss_data;
  source_class->seek = gdas_resample_seek;
  source_class->get_position = gdas_resample_get_position;
  source_class->get_latency_info = gdas_resample_get_latency_info;
  compute_filter_coeffs ();
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_resample_arg_set_rate,
		     gdas_resample_arg_get_rate,
		     NULL, "0,100", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "hi_quality",
		     GDAM_ARG_TYPE_INT,
		     gdas_resample_arg_set_hi_quality,
		     gdas_resample_arg_get_hi_quality,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

static void gdas_resample_init(GdasResample* resample)
{
  resample->accum = 0;
  resample->last_sample.s[0] = resample->last_sample.s[1] = 0;
  resample->rate = (1LL << 32);
}

static GdamFixed** filter_coeffs = NULL;

/* the interpolation ratio. */
#define RATIO 500

/* passband specified, larger makes longer filters */
#define PERCENT_PASS (85.0)

/* minimum attenuation in stopbands (dB), larger make long filters */
/*float min_attenuation = 60.0;*/
#define MIN_ATTENUATION (60.0)

/* Use min_attenuation to get beta (for Kaiser window function)
 * and nfilt (always odd valued and = 2*npair +1)
 * using Kaiser's empirical formulas.           */
static int get_filter_length(gfloat    min_attenuation, 
                             gfloat    delta_f,
			     gfloat*   beta)
{
  int num_pair;
  *beta = 0.0;			/* value of beta if min_attenuation < 21 */
  if (min_attenuation >= 50.0)
    *beta = 0.1102 * (min_attenuation - 8.71);
  if (min_attenuation < 50.0 && min_attenuation >= 21.0)
    {
      *beta = 0.5842 * pow ((min_attenuation - 21.0), 0.4)
	+ 0.07886 * (min_attenuation - 21.0);
    }
  num_pair = ((min_attenuation - 8.0) / (28.72 * delta_f));
  return 2 * num_pair + 1;
}

static gfloat bessel_i0(gfloat y)
{
  float s = 1.0, ds = 1.0, d = 0.0;
  do
    {
      d = d + 2;
      ds = ds * (y * y) / (d * d);
      s = s + ds;
    }
  while (ds > 1E-7 * s);
  return s;
}

static void compute_filter_coeffs()
{
  int i, j;
  int filter_length, npair, n, k, lsize;
  gfloat delta_f;
  gfloat beta, valizb, alpha;
  gfloat w, ck, y, npair_inv, pi_ratio;
  gfloat phase, dec;

  delta_f = (200 - 2 * PERCENT_PASS) / (200.0 * RATIO);

  filter_length = get_filter_length (MIN_ATTENUATION, delta_f, &beta);
  lsize = filter_length / RATIO;
  filter_length = (long) lsize *RATIO + 1;
  npair = (filter_length - 1) / 2;

  filter_coeffs = g_new0 (GdamFixed *, RATIO);
  for (i = 0; i < RATIO; i++)
    filter_coeffs[i] = g_new0 (GdamFixed, lsize);

  /* Compute Kaiser window sample values */
  i = 0;
  j = 0;
  valizb = 1.0 / bessel_i0 (beta);
  npair_inv = 1.0 / npair;
  pi_ratio = M_PI / RATIO;
  filter_coeffs[i++][j] = 0.0;	/* n = 0 case */

  for (n = 1; n < npair; n++)
    {
      k = npair - n;
      alpha = k * npair_inv;
      y = beta * sqrt (1.0 - (alpha * alpha));
      w = valizb * bessel_i0 (y);
      ck = RATIO * sin (k * pi_ratio) / (k * M_PI);
      filter_coeffs[i++][j] = GDAM_FIXED_FROM_DOUBLE (w * ck);
      if (i == RATIO)
	{
	  i = 0;
	  j++;
	}
    }

  /* force the pass through point */
  filter_coeffs[i][lsize / 2] = 1.0;

  /* second half of response */
  for (n = 1; n < npair; n++)
    {
      i = npair - n;		/* "from" location */
      k = npair + n;		/* "to" location */
      filter_coeffs[k % RATIO][k / RATIO]
	= filter_coeffs[i % RATIO][i / RATIO];
    }
}

/* add to the hi-quality filter;  and remove from the hi-quality filter. */
static inline void add_sample_to_filter   (GdasResample*  resample,
                                           GdamSample*    to_add)
{
  g_error ("XXX: unimplemented");
}
static inline GdamSample get_filter_output(GdasResample*  resample)
{
  GdamSample rv;
  g_error ("XXX: unimplemented");
  return rv;
}

static int gdas_resample_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasResample *resample = GDAS_RESAMPLE (source);
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  int num_read, rv;
  guint64 rate;
  guint accum;
  GdamSample last_sample;
  GdamSample *at_sample;	/* will point inside the buffer `tmp' */
  /* How many samples can we get without overfilling buf 
     and last_sample? */
  int max_read;
  int num_output;
  int leftover_input;
  int last_skipped;

  GdamSample *tmp;
  GdamSample *out;

  g_return_val_if_fail (resample != NULL, 0);
  filtered = filter->filtered;
  if (!filtered)
    return 0;
  rate = resample->rate;
  accum = resample->accum;

  if (rate == (1LL << 32))
    return gdas_source_read (filtered, buf, n);

  /* Determine the constraints on how much data to read,
   * we subtract 1 because the leftover counts as a sample. */
  if (resample->last_skipped > 0)
    {
      last_skipped = resample->last_skipped;
      if (last_skipped > (int) (rate >> 32))
	{
	  g_warning ("cropping last_skipped [was %d]", last_skipped);
	  last_skipped = (rate >> 32) - 1;
	}
    }
  else
    last_skipped = 0;
  max_read = (int) ((resample->rate * n + accum) >> 32) - last_skipped;

  if (max_read <= 0)
    {
      tmp = NULL;
      num_read = 0;
    }
  else
    {
      /* Read the data. */
      tmp = (GdamSample *) alloca (sizeof (GdamSample) * max_read);
      num_read = gdas_source_read (filtered, tmp, max_read);

      if (num_read <= 0)
	{
	  /* Just discard fractional ending data... who cares? */
	  return num_read;
	}
    }

  /* Prevent division by zero and overflows... */
  if (num_read >= resample->rate)
    {
      num_output = n;
    }
  else
    {
      num_output = ((((gint64) (num_read + 1 + last_skipped)
		      << 32)) - accum - 1) / resample->rate;
      if (num_output > n)
	{
	  num_output = n;
	}
    }

  if (num_read == max_read && num_output != n)
    {
      g_warning ("resample: n=%d, num_output=%d, num_read=max_read=%d",
		 n, num_output, max_read);
    }

  /* The use of `tmp - 1' is to denote that we are
   * outputting `last_sample' right now. */
  at_sample = tmp - 1 - last_skipped;
  out = buf;
  rv = num_output;
  last_sample = resample->last_sample;

  /* Rescale the data. */
  if (!resample->hi_quality)
    {
      guint32 fraction = (guint32) rate;
      guint32 whole = (guint32) (rate >> 32);
      while (num_output-- > 0)
	{
	  guint32 new_accum = accum + fraction;
	  at_sample += whole;
	  if (new_accum < accum)
	    {
	      last_sample = *(++at_sample);
	    }
	  else if (whole)
	    last_sample = *at_sample;
	  accum = new_accum;

	  *out++ = last_sample;
	}
    }
  else
    {
      guint32 fraction = (guint32) rate;
      guint32 whole = (guint32) (rate >> 32);
      while (num_output-- > 0)
	{
	  guint32 new_accum = accum + fraction;
	  int i;
	  for (i = 0; i < whole; i++)
	    {
	      add_sample_to_filter (resample, at_sample);
	      at_sample++;
	    }
	  if (new_accum < accum)
	    {
	      add_sample_to_filter (resample, at_sample);
	      at_sample++;
	    }
	  accum = new_accum;
	  *out++ = get_filter_output (resample);
	}
    }

  resample->last_sample = last_sample;
  resample->accum = accum;

  /* Not maintaining the following condition will lead
   * to pops and so forth, as the source above us
   * will think we are blocking... */
  if (num_read == max_read)
    g_return_val_if_fail (rv == n, rv);

  if (tmp != NULL)
    {
      leftover_input = tmp + num_read - 1 - at_sample;
      resample->last_skipped = leftover_input;
      g_return_val_if_fail (leftover_input >= 0, rv);
    }
  else
    {
      resample->last_skipped = 0;
    }
  g_return_val_if_fail (out - buf == rv, rv);

  return rv;
}
static void        gdas_resample_miss_data        (GdasSource*	    source,
                                       	           int              n)
{
  GdasResample *resample = GDAS_RESAMPLE (source);
  GdasSource *filtered = GDAS_FILTER (source)->filtered;
  if (filtered != NULL)
    {
      int submiss;
      submiss = (int) ((resample->accum + resample->rate * n) >> 32);
      gdas_source_miss_data (filtered, submiss);
    }
  resample->accum += (guint32) (resample->rate * n);
  return;
}

static gboolean
gdas_resample_seek(GdasSource* source,
                   GdamTime    offset,
		   int         whence)
{
  GdasResample *resample = GDAS_RESAMPLE (source);
  GdasFilter *filter = GDAS_FILTER (source);
  g_return_val_if_fail (resample != NULL, FALSE);
  g_return_val_if_fail (filter != NULL, FALSE);

  if (filter->filtered == NULL)
    {
      gdam_debug ("seek attempted on resample with no subsource");
      return FALSE;
    }

  /* XXX: not safe to 64-bits at all!!! */
  offset *= resample->rate;
  offset >>= 32;

  return gdas_source_seek (filter->filtered, offset, whence);
}

static GdamTime gdas_resample_get_position(GdasSource* source)
{
  GdasResample *resample = GDAS_RESAMPLE (source);
  GdasFilter *filter = GDAS_FILTER (source);
  gdouble position;
  if (filter->filtered == NULL)
    {
      gdam_debug ("tried to get position of an empty resample");
      return -1;
    }
  position = gdas_source_get_position (filter->filtered);
  if (position == -1)
    return -1;

  position *= (1LL << 32);
  if (resample->rate == 0)
    {
      g_warning ("resample rate was 0");
      return -1;
    }
  position /= resample->rate;
  return (gint64) position;
}


