/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaswalker.h"
#include <common/gdam-protocol.h>
#include <stdlib.h>
#include <math.h>

static int         gdas_walker_read             (GdasSource*   input,
	                                         GdamSample*   samples,
						 int           max_read);
static void        gdas_walker_miss_data        (GdasSource*   input,
	                                         int           samples_lost);
static void        gdas_walker_destroy          (GtkObject*    object);
static gboolean    gdas_walker_setup            (GdasWalker*   walker,
                                                 guint         buffer_id);
static GtkObjectClass* parent_class = NULL;
static GdasSourceClass* parent_source_class = NULL;

static inline double gdas_walker_drand48(GdasWalker *this)
{
  return erand48 (this->random_state);
}
static inline void gdas_walker_srand48(GdasWalker *this, int seed)
{
  this->random_state[0] = 0x330e;
  *((int *) &this->random_state[1]) = seed;
}
static inline int gdas_walker_get_seed(GdasWalker *this)
{
  return *((int *) &this->random_state[1]);
}

static void gdas_walker_set_mean_length(GtkObject* object, GdamArg* arg)
{
  GDAS_WALKER (object)->mean_length = GDAM_ARG_DOUBLE (arg);
}
static void gdas_walker_get_mean_length(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_WALKER (object)->mean_length;
}
static void gdas_walker_set_random_seed(GtkObject* object, GdamArg* arg)
{
  gdas_walker_srand48 (GDAS_WALKER (object), GDAM_ARG_INT (arg));
}
static void gdas_walker_get_random_seed(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = gdas_walker_get_seed (GDAS_WALKER (object));
}
static void gdas_walker_set_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdasWalker *walker = GDAS_WALKER (object);
  if (!gdas_walker_setup (walker, GDAM_ARG_INT (arg)))
    {
      gdam_debug ("walker_setup failed (buffer_id=%d)", GDAM_ARG_INT (arg));
      return;
    }
  if (walker->position >= walker->cache->buffer_length)
    walker->position = walker->cache->buffer_length - 1;
}
static void gdas_walker_get_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdamCacheBuffer *buffer = GDAS_WALKER (object)->cache;
  GDAM_ARG_INT (arg) = buffer ? gdam_cache_buffer_get_id (buffer) : 0;
}
static gboolean	gdas_walker_seek(GdasSource*      input,
	                         GdamTime         offset,
				 int              whence)
{
  GdasWalker *walker = GDAS_WALKER (input);
  int buffer_len = 0;
  int position = walker->position;
  int new_position;
  if (walker->cache != NULL)
    buffer_len = walker->cache->buffer_length;
  if (whence == 2)
    new_position = buffer_len + offset;
  else if (whence == 1)
    new_position = position + offset;
  else
    new_position = offset;
  if (buffer_len > 0)
    walker->position = new_position % buffer_len;
  return TRUE;
}
static GdamTime gdas_walker_get_position(GdasSource* source)
{
  return GDAS_WALKER (source)->position;
}


static void gdas_walker_class_init(GdasSourceClass* to_init)
{
  GtkObjectClass *object_class = (GtkObjectClass *) to_init;
  to_init->read = gdas_walker_read;
  to_init->miss_data = gdas_walker_miss_data;
  to_init->seek = gdas_walker_seek;
  to_init->get_position = gdas_walker_get_position;
  object_class->destroy = gdas_walker_destroy;

  gdam_arg_register (object_class,
		     "mean_length",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_walker_set_mean_length,
		     gdas_walker_get_mean_length, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "random_seed",
		     GDAM_ARG_TYPE_INT,
		     gdas_walker_set_random_seed,
		     gdas_walker_get_random_seed, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdas_walker_set_buffer_id,
		     gdas_walker_get_buffer_id, NULL, NULL, GDAM_ARG_VITAL);
}

static void gdas_walker_init(GdasWalker* to_init)
{
  to_init->mean_length = 441;
  gdas_walker_srand48 (to_init, 0xdeadbeef);
}

GtkType gdas_walker_get_type ()
{
  static GtkType walker_type = 0;
  if (!walker_type)
    {
      static const GtkTypeInfo walker_info = {
	"GdasWalker",
	sizeof (GdasWalker),
	sizeof (GdasSourceClass),
	(GtkClassInitFunc) gdas_walker_class_init,
	(GtkObjectInitFunc) gdas_walker_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      parent_source_class = (GdasSourceClass *) parent_class;
      walker_type = gtk_type_unique (GDAS_TYPE_SOURCE, &walker_info);
    }
  return walker_type;
}

static gboolean    gdas_walker_setup(GdasWalker*    walker,
                                        guint             buffer_id)
{
  GdamCacheBuffer *cache_buf;
  cache_buf = gdam_cache_buffer_get_by_id (buffer_id);
  if (cache_buf == NULL)
    return FALSE;
  if (cache_buf->buffer_length <= 0)
    {
      gdam_debug ("walker %u: still loading...", buffer_id);
      return FALSE;
    }
  gtk_object_ref (GTK_OBJECT (cache_buf));
  if (walker->cache)
    gtk_object_unref (GTK_OBJECT (walker->cache));
  walker->cache = cache_buf;
  return TRUE;
}

GdasWalker* gdas_walker_new(guint buffer_id)
{
  GdasWalker *walker;
  walker = GDAS_WALKER (gtk_type_new (GDAS_TYPE_WALKER));
  if (!gdas_walker_setup (walker, buffer_id))
    {
      gtk_object_destroy (GTK_OBJECT (walker));
      return NULL;
    }
  return walker;
}

static int       gdas_walker_read           (GdasSource*      input,
	                                        GdamSample*      samples,
						int              max_read)
{
  GdasWalker *walker = GDAS_WALKER (input);
  int buffer_len;
  GdamSample *buffer;
  int to_read = max_read;
  int position = walker->position;
  int direction = walker->direction;
  int length = walker->length;

  //gdam_debug("walker_read: %d bytes [start offset=%d]", max_read, position);

  if (walker->cache == NULL)
    return 0;

  buffer = walker->cache->buffer;
  buffer_len = walker->cache->buffer_length;

  while (to_read > 0)
    {
      double x;
      while (to_read > 0 && length > 0)
	{
	  if (direction == -1)
	    {
	      if (--position < 0)
		{
		  position = 0;
		  length++;
		  direction = +1;
		}
	    }
	  else
	    {
	      if (++position >= buffer_len)
		{
		  position = buffer_len - 1;
		  length++;
		  direction = -1;
		}
	    }
	  *samples++ = buffer[position];
	  to_read--;
	  length--;
	}
      if (length == 0)
	{
	  do
	    x = gdas_walker_drand48 (walker);
	  while (x == 0.0);
	  length = (int) (-walker->mean_length * log (x));
	  direction = (gdas_walker_drand48 (walker) < 0.5 ? -1 : 1);
	}
    }
  walker->position = position;
  walker->direction = direction;
  walker->length = length;

  return max_read;
}

static void      gdas_walker_miss_data      (GdasSource*      input,
	                                        int              samples_lost)
{
  //GdasWalker* walker = GDAS_WALKER(input);
}

static void      gdas_walker_destroy            (GtkObject*       object)
{
  GdasWalker *walker = GDAS_WALKER (object);
  if (walker->cache)
    {
      GtkObject *buf = GTK_OBJECT (walker->cache);
      walker->cache = NULL;
      gtk_object_unref (buf);
    }
  parent_class->destroy (object);
}
