/* Gdamdsp -- a feed from a normal dsp device into a gdam-server.
 *
 * Based on:
 *
 * Vsound - a virtual audio loopback cable used for recording /dev/dsp streams
 * Copyright (C) 1999 James Henstridge <james@daa.com.au>
 *
 *   Based on the esddsp utility that is part of esound.  Esddsp's copyright:
 *   Copyright (C) 1998, 1999 Manish Singh <yosh@gimp.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gdamdsppreload.h"
#include <sys/un.h>
#include <sys/socket.h>
#include <errno.h>

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif

/* #define DSP_DEBUG */

#ifdef DSP_DEBUG
#define DPRINTF(format, args...)	printf(format, ## args)
#else
#define DPRINTF(format, args...)
#endif

#define guint32 unsigned

#include <dlfcn.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#ifdef HAVE_MACHINE_SOUNDCARD_H
#  include <machine/soundcard.h>
#else
#  ifdef HAVE_SOUNDCARD_H
#    include <soundcard.h>
#  else
#    include <sys/soundcard.h>
#  endif
#endif

#define REAL_LIBC RTLD_NEXT

#ifdef __FreeBSD__
typedef unsigned long request_t;
#else
typedef int request_t;
#endif

static int sndfd = -1;
static FILE *dest = NULL;
static const char* gdamdsp_device = NULL;



static void
transmit_message (int message_type,
                  int message_length,
		  const char* message)
{
  char header[8];
  int tmp;

  /* the transmitted message length includes the message_type. */
  tmp = message_length + 4;
  header[0] = (char) (tmp >> 24);
  header[1] = (char) (tmp >> 16);
  header[2] = (char) (tmp >> 8);
  header[3] = (char) (tmp >> 0);
  header[4] = (char) (message_type >> 24);
  header[5] = (char) (message_type >> 16);
  header[6] = (char) (message_type >> 8);
  header[7] = (char) (message_type >> 0);

  if (fwrite (header, sizeof(guint32), 2, dest) != 2
   ||   (message_length > 0
      && fwrite (message, 1, message_length, dest) != message_length) )
    {
      fprintf (stderr, "warning:  error writing to socket: %s\n",
              strerror (errno));
      /* Give up:  the dsp stream will be hopelessly hosed. */
      exit(1);
    }
  fflush(dest);
}

static void transmit_message_str (int         message_type,
                                  const char* string)
{
  transmit_message (message_type, strlen(string), string);
}


static void
dsp_init (void)
{
  static int inited = 0;
  struct sockaddr_un unix_addr;
  const char* dest_filename;
  int dest_fd;

  if (inited) return;
  inited = 1;

  dest_filename = getenv ("GDAMDSP_SOCKET");
  if (dest_filename == NULL)
    dest_filename = "/tmp/gdamdsp.socket";

  /* connect to the unix domain socket */
  unix_addr.sun_family = AF_UNIX;
  strncpy (unix_addr.sun_path, dest_filename, UNIX_PATH_MAX);
  dest_fd = socket (PF_UNIX, SOCK_STREAM, 0);
  if (dest_fd < 0)
    {
      perror ("socket(PF_UNIX, SOCK_STREAM)");
      exit(1);
    }
  if (connect (dest_fd, (struct sockaddr*) &unix_addr, sizeof(unix_addr)) < 0)
    {
      fprintf (stderr, "couldn't connect to unix socket %s: %s\n",
               dest_filename, strerror(errno));
      exit (1);
    }
  dest = fdopen (dest_fd, "r+");

  gdamdsp_device = getenv ("GDAMDSP_DEVICE");
  if (gdamdsp_device == NULL)
    {
      gdamdsp_device = "/dev/dsp";
    }
}

int
open (const char *pathname, int flags, ...)
{
  static int (*func) (const char *, int, mode_t) = NULL;
  va_list args;
  mode_t mode;

  if (!func)
    func = (int (*) (const char *, int, mode_t)) dlsym (REAL_LIBC, "open");

  dsp_init ();

  va_start (args, flags);
  mode = va_arg (args, mode_t);
  va_end (args);

  if (strcmp (pathname, gdamdsp_device) == 0)
    {
      transmit_message_str (GDAMDSP_OPEN, pathname);
      return (sndfd = (*func) ("/dev/null", O_WRONLY, 0));
    }
  else
    return (*func) (pathname, flags, mode);
}

static int
dspctl (int fd, request_t request, void *argp)
{
  int *arg = (int *) argp;

  DPRINTF ("dspctl: mogrifying ioctl request");
  switch (request)
    {
    case SNDCTL_DSP_RESET:
      transmit_message_str (GDAMDSP_IOCTL, "reset");
      break;
    case SNDCTL_DSP_POST:
      transmit_message_str (GDAMDSP_IOCTL, "post");
      break;

    case SNDCTL_DSP_SETFMT:
      
      switch (*arg) {
      case AFMT_QUERY:
#ifdef WORDS_BIGENDIAN
	*arg = AFMT_S16_BE;  /* this is arbitrary */
#else
	*arg = AFMT_S16_LE;  /* this is arbitrary */
#endif
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:s16");
	break;
      case AFMT_MU_LAW:
	fprintf (stderr, "dsp_setfmt: mulaw not supported by gdamdsp yet.\n");
	/*exit (1);*/
	return -1;

      case AFMT_A_LAW:
	/*fprintf (stderr, "dsp_setfmt: alaw not supported by gdamdsp yet.\n");*/
	return -1;
      case AFMT_S8:
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:s8");
	break;
      case AFMT_U8:
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:u8");
	break;
      case AFMT_S16_LE:
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:s16le");
	break;
      case AFMT_S16_BE:
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:s16be");
	break;
      case AFMT_U16_LE:
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:u16le");
	break;
      case AFMT_U16_BE:
        transmit_message_str (GDAMDSP_IOCTL, "setfmt:u16be");
	break;
      }
      break;

    case SNDCTL_DSP_SPEED:
      {
        char buf[255];
        sprintf (buf, "speed %d", *arg);
        transmit_message_str (GDAMDSP_IOCTL, buf);
        break;
      }

    case SNDCTL_DSP_STEREO:
      {
        char buf[255];
	sprintf (buf, "channels %d", *arg ? 2 : 1);
        transmit_message_str (GDAMDSP_IOCTL, buf);
        break;
      }

    case SNDCTL_DSP_CHANNELS:
      {
        char buf[255];
	sprintf (buf, "channels %d", *arg);
        transmit_message_str (GDAMDSP_IOCTL, buf);
        break;
      }

    case SNDCTL_DSP_GETBLKSIZE:
      DPRINTF(ioctls, "dsp_getblksize (returning 4k)\n");
      *arg = 4 * 1024;
      break;

    case SNDCTL_DSP_GETFMTS:
      {
        int rv = AFMT_U8 | AFMT_S16_LE | AFMT_S16_BE |
	         AFMT_S8 | AFMT_U16_LE | AFMT_U16_BE;
        DPRINTF ("dsp_getfmts (returning 0x%02x)\n", rv);
        *arg = rv;
	break;
      }

    case SNDCTL_DSP_GETCAPS:
      DPRINTF (ioctls, "dsp_getcaps\n");
      *arg = 0;
      break;

    case SNDCTL_DSP_GETOSPACE:
      DPRINTF ("dsp_getospace\n");
      {
	audio_buf_info *bufinfo = (audio_buf_info *) argp;
	bufinfo->bytes = 4 * 1024;
      }
      break;


    default:
      DPRINTF ("unhandled /dev/dsp ioctl (%x - %p)\n", request, argp);
      break;
    }

  return 0;
}

int
ioctl (int fd, request_t request, ...)
{
  static int (*func) (int, request_t, void *) = NULL;
  va_list args;
  void *argp;

  if (!func)
    func = (int (*) (int, request_t, void *)) dlsym (REAL_LIBC, "ioctl");
  va_start (args, request);
  argp = va_arg (args, void *);
  va_end (args);

  if (fd != sndfd)
    return (*func) (fd, request, argp);
  else
    return dspctl (fd, request, argp);
}

ssize_t
dspwrite (int fd, const void* buffer, size_t n)
{
  transmit_message (GDAMDSP_WRITE, n, buffer);
  return n;
}

ssize_t
write (int fd, const void* buffer, size_t n)
{
  static ssize_t (*func) (int, const void*, size_t) = NULL;

  if (!func)
    func = (ssize_t (*) (int, const void*, size_t)) dlsym (REAL_LIBC, "write");

  if (fd != sndfd)
    return (*func) (fd, buffer, n);
  else
    return dspwrite (fd, buffer, n);
}

int
close (int fd)
{
  static int (*func) (int) = NULL;

  if (!func)
    func = (int (*) (int)) dlsym (REAL_LIBC, "close");

  if (fd == sndfd)
    {
      sndfd = -1;
      transmit_message (GDAMDSP_CLOSE, 0, NULL);
    }
 
  return (*func) (fd);
}

