/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasexpander.h"
#include <common/gdamplugin.h>
#include <math.h>

/*
 * Maps the position values by raising to various powers,
 * usually 0<p<1.
 */

#define EXP_TABLE_SIZE	32768

/* Table from sample 0.5 to sample 1.0   =>   (log2(x) + 1) * EXP_TABLE_SIZE)
 *      			      (in the range 0 <= x <= EXP_TABLE_SIZE)
 *
 * So, if we have a channel's sample x, -32768<x<32768,
 * First make it positive and store the sign...
 *
 *
 * if (sample == 0) return 0;
 * 
 * while (sample < GDAM_SAMPLE_MAX / 2) {
 *	sample += sample;
 *      total += EXP_TABLE_SIZE;
 * }
 * 
 */
static int* log_table = NULL;

/* Table:
 *              2^(x / EXP_TABLE_SIZE - 1.0) * GDAM_SAMPLE_ONE
 */
static int* exp_table = NULL;



#if 0
static inline int map_sample(int in, gdouble power)
{
  gboolean was_negative;
  int rv;
  int exp_index = 0;
  int log_index;
  int shift;

  if (in == 0)
    return 0;

  if (in < 0)
    {
      was_negative = TRUE;
      in = (-in);
    }
  else
    {
      was_negative = FALSE;
    }

  while (in < GDAM_SAMPLE_MAX / 2)
    {
      exp_index -= EXP_TABLE_SIZE;
      in += in;
    }

  log_index = (in - (GDAM_SAMPLE_MAX / 2))
    * EXP_TABLE_SIZE / (GDAM_SAMPLE_MAX / 2);
  g_return_val_if_fail (0 <= log_index && log_index < EXP_TABLE_SIZE, 0);
  exp_index += log_table[log_index];

  exp_index = (exp_index - EXP_TABLE_SIZE) * power + EXP_TABLE_SIZE;

  if (exp_index >= EXP_TABLE_SIZE)
    return (was_negative ? GDAM_SAMPLE_MIN : GDAM_SAMPLE_MAX);

  if (exp_index < 0)
    {
      shift = (-exp_index + EXP_TABLE_SIZE - 1) / EXP_TABLE_SIZE;
      exp_index += shift * EXP_TABLE_SIZE;
    }
  else
    {
      shift = 0;
    }

  g_return_val_if_fail (0 <= exp_index && exp_index < EXP_TABLE_SIZE, 0);
  rv = (exp_table[exp_index] >> shift);
  if (was_negative)
    rv = (-rv);
  return CLAMP (rv, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
}

#else

static inline int map_sample(int in, gdouble power)
{
  gdouble v;
  gboolean was_negative;
  if (in == 0)
    return 0;
  if (in < 0)
    {
      was_negative = TRUE;
      in = (-in);
    }
  else
    {
      was_negative = FALSE;
    }
  v = (gdouble) in / GDAM_SAMPLE_ONE;
  v = pow (v, power) * GDAM_SAMPLE_ONE;
  return was_negative ? (int) (-v) : (int) (v);
}

#endif

static void gdas_expander_init(GdasExpander* expander)
{
  expander->power = 1.0;
}

static int gdas_expander_read(GdasSource*        source,
                              GdamSample*        buf,
			      int                n)
{
  int num_read;
  int i;
  gdouble power = GDAS_EXPANDER (source)->power;

  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buf, n);

  for (i = 0; i < num_read; i++)
    {
      buf[i].s[0] = map_sample (buf[i].s[0], power);
      buf[i].s[1] = map_sample (buf[i].s[1], power);
    }
  return num_read;
}

static void gdas_expander_arg_set_power(GtkObject* object, GdamArg* arg)
{
  GDAS_EXPANDER (object)->power = GDAM_ARG_DOUBLE (arg);
}
static void gdas_expander_arg_get_power(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_EXPANDER (object)->power;
}

static inline gdouble log2(gdouble x)
{
  return log (x) * M_LOG2E;
}

		
static void gdas_expander_class_init(GdasExpanderClass* expander_class)
{
  int i;
  log_table = g_new (int, EXP_TABLE_SIZE + 1);
  exp_table = g_new (int, EXP_TABLE_SIZE + 1);

  for (i = 0; i <= EXP_TABLE_SIZE; i++)
    {
      gdouble sample = (gdouble) i / EXP_TABLE_SIZE;
      sample /= 2.0;
      sample += 0.5;
      sample = (log2 (sample) + 1) * EXP_TABLE_SIZE;
      log_table[i] = (int) sample;
    }
  for (i = 0; i <= EXP_TABLE_SIZE; i++)
    {
      gdouble sample = (gdouble) i / EXP_TABLE_SIZE;
      sample -= 1.0;
      sample = pow (2.0, sample);
      sample *= GDAM_SAMPLE_ONE;
      exp_table[i] = (int) sample;
    }

  GDAS_SOURCE_CLASS (expander_class)->read = gdas_expander_read;

  gdam_arg_register (GTK_OBJECT_CLASS (expander_class),
		     "power",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_expander_arg_set_power,
		     gdas_expander_arg_get_power,
		     NULL, "0,40", GDAM_ARG_VITAL);
}

GtkType gdas_expander_get_type()
{
  static GtkType expander_type = 0;
  if (!expander_type)
    {
      static const GtkTypeInfo expander_info = {
	"GdasExpander",
	sizeof (GdasExpander),
	sizeof (GdasExpanderClass),
	(GtkClassInitFunc) gdas_expander_class_init,
	(GtkObjectInitFunc) gdas_expander_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      expander_type = gtk_type_unique (parent, &expander_info);
      gdam_plugin_check_version (__FILE__, VERSION);
    }
  return expander_type;
}
