/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasfeedbackfilter.h"
#include "../gdamconnection.h"
#include <common/gdamsample.h>
#include <gtk/gtksignal.h>
#include <common/gdamplugin.h>

#define MIN_FEEDBACK_TIME	256
#define MAX_READ_EXPECTED	(1<<16)

static void gdas_feedback_filter_class_init(GdasSourceClass* source_class);
static void gdas_feedback_filter_init(GdasFeedbackFilter* feedback_filter)
{
}

GtkType gdas_feedback_filter_get_type()
{
  static GtkType feedback_filter_type = 0;
  if (!feedback_filter_type)
    {
      static const GtkTypeInfo feedback_filter_info = {
	"GdasFeedbackFilter",
	sizeof (GdasFeedbackFilter),
	sizeof (GdasFeedbackFilterClass),
	(GtkClassInitFunc) gdas_feedback_filter_class_init,
	(GtkObjectInitFunc) gdas_feedback_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      feedback_filter_type = gtk_type_unique (parent, &feedback_filter_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return feedback_filter_type;
}

static int gdas_feedback_filter_read(GdasSource*     source,
                                     GdamSample*     buffer,
				     int             max_samples)
{
  GdasFeedbackFilter *feedback = GDAS_FEEDBACK_FILTER (source);
  GdasFeedbackFilter *feed = feedback->feed_into;
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered = filter->filtered;
  int num_read;

  if (filtered == NULL)
    return 0;

  /* Reduce max_samples if we have enough pent-up data
   * to output some now, but not `max_samples' worth. */
  {
    int max_from_feed;
    max_from_feed = feedback->buffer_amount - feedback->delay;

    /* Ignore max_from_feed if we are really running out
     * of data... */
    if (max_from_feed >= MIN_FEEDBACK_TIME && max_from_feed < max_samples)
      max_samples = max_from_feed;
  }

  num_read = gdas_source_read (filtered, buffer, max_samples);
  if (num_read <= 0)
    return num_read;

  if (feedback->feedback_volume != 0)
    {
      int p1 = feedback->buffer_alloc - feedback->buffer_start;
      int num_feedback = num_read;
      GdamFixed volume = feedback->feedback_volume;
      if (num_feedback > feedback->buffer_amount)
	num_feedback = feedback->buffer_amount;
      if (p1 > num_feedback && num_feedback > 0)
	{
	  gdam_sample_add_scaled
	    (buffer,
	     feedback->buffer + feedback->buffer_start, num_feedback, volume);
	  feedback->buffer_start += num_feedback;
	  feedback->buffer_amount -= num_feedback;
	}
      else if (num_feedback > 0)
	{
	  int p2 = num_feedback - p1;
	  gdam_sample_add_scaled
	    (buffer, feedback->buffer + feedback->buffer_start, p1, volume);
	  gdam_sample_add_scaled (buffer + p1, feedback->buffer, p2, volume);
	  feedback->buffer_start = p2;
	  feedback->buffer_amount -= num_feedback;
	}
    }
  else
    {
      int num_feedback = num_read;
      if (num_feedback > feedback->buffer_amount)
	num_feedback = feedback->buffer_amount;
      if (num_feedback > 0)
	{
	  feedback->buffer_start += num_feedback;
	  feedback->buffer_start %= feedback->buffer_alloc;
	  feedback->buffer_amount -= num_feedback;
	}
    }

  if (feed != NULL && feed->buffer != NULL && feed->buffer_alloc > 0)
    {
      int index, empty;
      int to_append = num_read;
      GdamSample *copy_from;
      if (feed->buffer_amount < feed->delay)
	{
	  int zeroes = feed->delay - feed->buffer_amount;
	  int start = feed->buffer_start + feed->buffer_amount;
	  int p1;
	  start %= feed->buffer_alloc;
	  p1 = feed->buffer_alloc - start;
	  if (p1 <= zeroes)
	    {
	      gdam_sample_zero (feed->buffer + start, p1);
	      gdam_sample_zero (feed->buffer, zeroes - p1);
	    }
	  else
	    {
	      gdam_sample_zero (feed->buffer + start, zeroes);
	    }
	  feed->buffer_amount = feed->delay;
	}
      else if (feed->buffer_amount > feed->delay)
	{
	  int delta = feed->buffer_amount - feed->delay;
	  feed->buffer_start += delta;
	  feed->buffer_start %= feed->buffer_alloc;
	  feed->buffer_amount = feed->delay;
	}

      /* append num_read to buffer, if it will fit. */
      index = feed->buffer_amount + feed->buffer_start;
      if (index >= feed->buffer_alloc)
	index -= feed->buffer_alloc;
      empty = feed->buffer_alloc - feed->buffer_amount;
      if (empty < to_append)
	to_append = empty;

      g_assert (to_append >= 0);

      feed->buffer_amount += to_append;
      copy_from = buffer;
      while (to_append > 0)
	{
	  int o = to_append;
	  int contig = feed->buffer_alloc - index;
	  if (o > contig)
	    o = contig;
	  gdam_sample_copy (feed->buffer + index, copy_from, o);
	  copy_from += o;
	  index += o;
	  if (index == feed->buffer_alloc)
	    index = 0;
	  to_append -= o;
	  g_assert (o > 0);
	}
    }

  return num_read;
}

static void gdam_feedback_arg_set_feedback_dest(GtkObject* object, GdamArg* arg)
{
  GdasSource *source;
  source = gdam_connection_get_source (NULL, GDAM_ARG_INT (arg));
  if (source == NULL)
    {
      gdam_debug ("feedback_filter: set_dest: %d not found",
		  GDAM_ARG_INT (arg));
      return;
    }
  if (!GDAM_IS_FEEDBACK_FILTER (source))
    {
      gdam_debug ("feedback_filter: set_dest: %d wasn't another "
		  "feedback filter");
      gtk_object_unref (GTK_OBJECT (source));
      return;
    }
  gdas_feedback_filter_set_dest (GDAS_FEEDBACK_FILTER (object),
				 (GdasFeedbackFilter *) source);
  gtk_object_unref (GTK_OBJECT (source));
}
static void gdam_feedback_arg_get_feedback_dest(GtkObject* object, GdamArg* arg)
{
  GdasFeedbackFilter *filter = GDAS_FEEDBACK_FILTER (object)->feed_into;
  GDAM_ARG_INT (arg) = filter ? GDAM_SERVER_OBJECT (filter)->global_id : 0;
}
static void gdam_feedback_arg_set_feedback_source(GtkObject* object,
                                                  GdamArg* arg)
{
  GdasSource *source;
  source = gdam_connection_get_source (NULL, GDAM_ARG_INT (arg));
  if (source == NULL)
    {
      gdam_debug ("feedback_filter: set_source: %d not found",
		  GDAM_ARG_INT (arg));
      return;
    }
  if (!GDAM_IS_FEEDBACK_FILTER (source))
    {
      gdam_debug ("feedback_filter: set_source: %d wasn't another "
		  "feedback filter");
      gtk_object_unref (GTK_OBJECT (source));
      return;
    }
  gdas_feedback_filter_set_dest ((GdasFeedbackFilter *) source,
				 GDAS_FEEDBACK_FILTER (object));
  gtk_object_unref (GTK_OBJECT (source));
}
static void gdam_feedback_arg_set_gain(GtkObject* object, GdamArg* arg)
{
  GdasFeedbackFilter *filter = GDAS_FEEDBACK_FILTER (object);
  filter->feedback_volume = (int) (GDAM_ARG_DOUBLE (arg) * GdamFixed_ONE);
}
static void gdam_feedback_arg_get_gain(GtkObject* object, GdamArg* arg)
{
  GdasFeedbackFilter *filter = GDAS_FEEDBACK_FILTER (object);
  gdouble gain = (gdouble) filter->feedback_volume / GdamFixed_ONE;
  GDAM_ARG_DOUBLE (arg) = gain;
}


/*
 *                     ``set_delay''
 *
 * Basically the buffer starts out looking like this:
 *
 *         ======************===========================
 *	   0     ^           ^            
 *               |           +-----------+
 *           buffer_start                data_end
 *          (where the next              (where the last sample
 *	    sample would be read from)    was written)
 *
 *   There are `buffer_amount' stars, and the line is `buffer_alloc'
 *   characters long.
 *
 *  (The data may be circularly wrapped:
 *         *******===============================*******
 *                ^data_end                      ^buffer_start)
 *
 * If a resize is necessary, new_alloc becomes the new line
 * length:
 *
 *                                    buffer_start will
 *                                    be at new_alloc - delay.
 *                                           v
 *         00000000000000000000===============================************
 *                             ^                              ^
 *                       new_alloc-old_alloc         new_alloc-buffer_amount
 *                                                            
 */
static void
gdas_feedback_filter_set_delay(GdasFeedbackFilter* filter, int delay)
{
  int desired_size = delay + MAX_READ_EXPECTED;
  int old_alloc = filter->buffer_alloc;
  int data_end;
  data_end = filter->buffer_start + filter->buffer_amount;
  if (data_end > old_alloc)
    data_end -= old_alloc;

  filter->delay = delay;

  if (desired_size > filter->buffer_alloc)
    {
      int new_alloc = filter->buffer_alloc;
      GdamSample *sample;
      if (new_alloc == 0)
	new_alloc = 2048;
      while (new_alloc < desired_size)
	new_alloc += new_alloc;
      sample = g_new (GdamSample, new_alloc);
      {
	gdam_sample_copy (sample + new_alloc - old_alloc,
			  filter->buffer + data_end, old_alloc - data_end);
	gdam_sample_copy (sample + new_alloc - data_end,
			  filter->buffer, data_end);
	gdam_sample_zero (sample, new_alloc - old_alloc);
      }
      filter->buffer_start = filter->buffer_alloc - delay;
      if (filter->buffer_start < 0)
	filter->buffer_start += new_alloc;
      g_free (filter->buffer);
      filter->buffer = sample;
      filter->buffer_alloc = new_alloc;
    }
  else
    {
      filter->buffer_start = data_end - delay;
      if (filter->buffer_start < 0)
	filter->buffer_start += old_alloc;
    }
  g_assert (0 <= filter->buffer_start
	    && filter->buffer_start < filter->buffer_alloc);
}

static void gdam_feedback_arg_set_delay(GtkObject* object, GdamArg* arg)
{
  GdasFeedbackFilter *filter = GDAS_FEEDBACK_FILTER (object);
  int delay = GDAM_ARG_INT (arg);
  gdas_feedback_filter_set_delay (filter, delay);
}


static void gdam_feedback_arg_get_delay(GtkObject* object, GdamArg* arg)
{
  GdasFeedbackFilter *filter = GDAS_FEEDBACK_FILTER (object);
  GDAM_ARG_INT (arg) = filter->delay;
}


static void gdas_feedback_filter_class_init(GdasSourceClass* source_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (source_class);
  source_class->read = gdas_feedback_filter_read;
  gdam_arg_register (object_class,
		     "feedback_dest",
		     GDAM_ARG_TYPE_INT,
		     gdam_feedback_arg_set_feedback_dest,
		     gdam_feedback_arg_get_feedback_dest,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "feedback_source",
		     GDAM_ARG_TYPE_INT,
		     gdam_feedback_arg_set_feedback_source,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_feedback_arg_set_gain,
		     gdam_feedback_arg_get_gain, NULL, "0,3", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "delay",
		     GDAM_ARG_TYPE_INT,
		     gdam_feedback_arg_set_delay,
		     gdam_feedback_arg_get_delay, NULL, "0,", GDAM_ARG_VITAL);
}

static void dest_was_destroyed(GdasFeedbackFilter* dest,
                               GdasFeedbackFilter* source)
{
  source->feed_into = NULL;
  source->feed_into_destroy_id = 0;
}

void   gdas_feedback_filter_set_dest(GdasFeedbackFilter*  source,
                                     GdasFeedbackFilter*  dest)
{
  g_return_if_fail (!GTK_OBJECT_DESTROYED (source));
  if (source->feed_into != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (source->feed_into),
			     source->feed_into_destroy_id);
      source->feed_into = NULL;
      source->feed_into_destroy_id = 0;
    }
  source->feed_into_destroy_id = gtk_signal_connect (GTK_OBJECT (dest),
						     "destroy",
						     GTK_SIGNAL_FUNC
						     (dest_was_destroyed),
						     source);
  source->feed_into = dest;
}
