/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasiirfilter.h"
#include <common/gdamplugin.h>
#include <stdlib.h>

static void gdas_iir_filter_init(GdasIirFilter* iir_filter)
{
}

static GdamSample*
resize_buffer(GdamSample* src, int offset, int length, int new_length)
{
  GdamSample *new_buf = g_new (GdamSample, new_length);
  gdam_sample_copy (new_buf, src + offset, length - offset);
  gdam_sample_copy (new_buf + length - offset, src, offset);
  gdam_sample_zero (new_buf + length, new_length - length);
  g_free (src);
  return new_buf;
}

static void gdas_iir_filter_ensure_space(GdasIirFilter* filter, int length)
{
  int index = filter->next_index, old_len = filter->length;
  if (length <= old_len)
    return;
  filter->input = resize_buffer (filter->input, index, old_len, length);
  filter->output = resize_buffer (filter->output, index, old_len, length);
  filter->next_index = filter->length;
  filter->length = length;
}

static int compare_iir_pole(gconstpointer p1, gconstpointer p2)
{
  const GdamIirPole *pole_1 = p1;
  const GdamIirPole *pole_2 = p2;
  if (pole_1->delay < pole_2->delay)
    return -1;
  if (pole_1->delay > pole_2->delay)
    return +1;
  return 0;
}

static void gdas_iir_filter_set_poles(GdasIirFilter*     filter,
				      int                num_poles,
                                      GdamIirPole*       poles)
{
  g_free (filter->poles);
  filter->num_poles = num_poles;
  filter->poles = poles;
  qsort (filter->poles, num_poles, sizeof (GdamIirPole), compare_iir_pole);
  if (num_poles > 0)
    {
      int needed_samples = poles[num_poles - 1].delay + 1;
      gdas_iir_filter_ensure_space (filter, needed_samples);
    }
}

static int gdas_iir_filter_read(GdasSource*    source,
                                GdamSample*    buf,
				int            max_read)
{
  int num_read;
  int i;
  GdasIirFilter *iir_filter = GDAS_IIR_FILTER (source);
  GdamSample *input = iir_filter->input;
  GdamSample *output = iir_filter->output;
  int next_index = iir_filter->next_index;
  int length = iir_filter->length;
  int num_poles = iir_filter->num_poles;

  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buf, max_read);

  for (i = 0; i < num_read; i++)
    {
      gint64 vals[2] = { 0 };
      GdamIirPole *pole = iir_filter->poles;
      int pole_index;
      input[next_index] = buf[i];
      for (pole_index = 0; pole_index < num_poles; pole_index++)
	{
	  int offset;
	  offset = next_index - pole->delay;
	  if (offset < 0)
	    offset += length;

	  vals[0] += (gint64) pole->iir * output[offset].s[0];
	  vals[1] += (gint64) pole->iir * output[offset].s[1];

	  vals[0] += (gint64) pole->fir * input[offset].s[0];
	  vals[1] += (gint64) pole->fir * input[offset].s[1];

	  pole++;
	}
      vals[0] /= GdamFixed_ONE;
      vals[1] /= GdamFixed_ONE;
      buf[i].s[0] = GDAM_SAMPLE_CLAMP (vals[0]);
      buf[i].s[1] = GDAM_SAMPLE_CLAMP (vals[1]);
      output[next_index] = buf[i];
      if (++next_index == length)
	next_index = 0;
    }
  iir_filter->next_index = next_index;
  return num_read;
}

static void gdas_iir_filter_arg_set_poles(GtkObject* object, GdamArg* arg)
{
  int *array = GDAM_ARG_INT_ARRAY (arg).array;
  int length = GDAM_ARG_INT_ARRAY (arg).length;
  GdamIirPole *poles;
  int num_poles;
  int i;

  if (length % 3 != 0)
    {
      g_warning ("set_poles wasn't have an array of triplets");
      return;
    }

  num_poles = length / 3;
  poles = g_new (GdamIirPole, num_poles);
  for (i = 0; i < num_poles; i++)
    {
      poles[i].delay = array[3 * i + 0];
      poles[i].fir = array[3 * i + 1];
      poles[i].iir = array[3 * i + 2];
    }
  gdas_iir_filter_set_poles (GDAS_IIR_FILTER (object), num_poles, poles);
}

static void gdas_iir_filter_arg_get_poles(GtkObject* object, GdamArg* arg)
{
  GdasIirFilter *iir_filter = GDAS_IIR_FILTER (object);
  int num_poles = iir_filter->num_poles;
  int *array;
  int i;
  GDAM_ARG_INT_ARRAY (arg).length = 3 * num_poles;
  array = g_new (int, 3 * num_poles);
  GDAM_ARG_INT_ARRAY (arg).array = array;
  for (i = 0; i < num_poles; i++)
    {
      array[3 * i + 0] = iir_filter->poles[i].delay;
      array[3 * i + 1] = iir_filter->poles[i].fir;
      array[3 * i + 2] = iir_filter->poles[i].iir;
    }
}

static void gdas_iir_filter_class_init(GdasIirFilterClass* iir_filter_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (iir_filter_class);
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (iir_filter_class);
  source_class->read = gdas_iir_filter_read;
  gdam_arg_register (object_class, "poles", GDAM_ARG_TYPE_INT_ARRAY,
		     gdas_iir_filter_arg_set_poles,
		     gdas_iir_filter_arg_get_poles, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
}

GtkType gdas_iir_filter_get_type()
{
  static GtkType iir_filter_type = 0;
  if (!iir_filter_type)
    {
      static const GtkTypeInfo iir_filter_info = {
	"GdasIirFilter",
	sizeof (GdasIirFilter),
	sizeof (GdasIirFilterClass),
	(GtkClassInitFunc) gdas_iir_filter_class_init,
	(GtkObjectInitFunc) gdas_iir_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      iir_filter_type = gtk_type_unique (parent, &iir_filter_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return iir_filter_type;
}
