/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasmicrophone.h"
#include "../gdamdevice.h"
#include <sys/soundcard.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>

/* XXX: TODO:
 *      
 *      Doesn't properly do multiple-open protection.
 *      Some sound cards hang when that is done, so fix this!!!
 */

/* XXX: From gdamconfigfile.c.
 *
 *      It's in that header, but since it doesn't follow
 *      naming convention, i copy it here for reference. */
extern GHashTable*           gdas_microphones;

static int open_oss_microphone(const char* device_name, gboolean* must_close)
{
  int fd;
  int t;
  int num_channels;
  int rate = 44100;
  int sample_size;
  int log_frag_size = 11, num_fragments = 3;
  gboolean stereo = TRUE;

  *must_close = TRUE;

#if 0				/* old duplex hack that didn't work [maybe it worked] */
  sink = gdam_device_is_open (device_name);
  if (sink != NULL)
    {
      sink_class = GDAM_SINK_CLASS (GTK_OBJECT (sink)->klass);
      if (sink_class->get_write_fd != NULL)
	{
	  int fd;
	  fd = (*sink_class->get_write_fd) (sink);
	  if (fd >= 0)
	    {
	      g_message ("using duplex capabilities of %s", device_name);
	      *must_close = FALSE;
	      return fd;
	    }
	  else
	    {
	      g_warning ("%s was not opened duplex; "
			 "this might be bad", device_name);
	    }
	}
    }
#endif

  if (gdas_microphones == NULL
      || g_hash_table_lookup (gdas_microphones, device_name) == NULL)
    {
      g_warning ("I cannot open %s as a Microphone until it is "
		 "in your server config file, like this:", device_name);
      fprintf (stderr, "\n\tMicrophone \"%s\"\n\n", device_name);
      return -1;
    }

  fd = open (device_name, O_RDONLY);
  if (fd < 0)
    {
      g_warning ("opening %s: %s", device_name, strerror (errno));
      return fd;
    }

  /* Close the device on exec. */
  fcntl (fd, F_SETFD, FD_CLOEXEC);

  t = (log_frag_size & 0xffff) | ((num_fragments & 0xffff) << 16);
  if (ioctl (fd, SNDCTL_DSP_SETFRAGMENT, &t) < 0)
    {
      g_warning ("error going to correct amount of frag (%s)",
		 strerror (errno));
    }
  t = 16;
  if (ioctl (fd, SNDCTL_DSP_SAMPLESIZE, &t) < 0)
    {
      g_warning ("error going to %d bits per sample", t);
      close (fd);
      return -1;
    }
  t = stereo ? 1 : 0;
  if (ioctl (fd, SNDCTL_DSP_STEREO, &t) < 0)
    {
      g_warning ("error setting device to stereo: %s", strerror (errno));
      return -1;
    }
  if (ioctl (fd, SNDCTL_DSP_SPEED, &rate) < 0)
    {
      g_warning ("error setting dev to %d samples per sec", t);
      return -1;
    }

  rate = num_channels = sample_size = -1;
  if (ioctl (fd, SOUND_PCM_READ_RATE, &rate) < 0
      || ioctl (fd, SOUND_PCM_READ_BITS, &sample_size) < 0
      || ioctl (fd, SOUND_PCM_READ_CHANNELS, &num_channels) < 0
      || rate == -1 || num_channels == -1 || sample_size == -1)
    {
      g_warning ("error getting device parameters");
      g_warning ("you may need to manually specify the soundcard "
		 "rate if it does not support the defaults");
    }
  if (num_channels == 1)
    stereo = 0;
  else if (num_channels == 2)
    stereo = 1;
  else
    {
      g_warning ("unsupported number of channels %d", num_channels);
      close (fd);
      return -1;
    }
  fprintf (stderr, "Device set at %d bits/sample, stereo %s, rate %d.\n",
	   sample_size, stereo ? "on" : "off", rate);
  return fd;
}

static void gdas_microphone_init(GdasMicrophone* microphone)
{
  GDAS_EXTERNAL (microphone)->ignore_miss_data = 1;
  GDAS_EXTERNAL (microphone)->max_buffer_size = 2048;
}
static void gdas_microphone_maybe_open(GdasSource* source)
{
  GdasMicrophone *microphone = GDAS_MICROPHONE (source);
  char *command[3];
  if (microphone->device_name == NULL)
    return;

  if (microphone->use_helper)
    {
      command[0] = "gdamrecordoss";
      command[1] = microphone->device_name;
      command[2] = NULL;
      gdas_external_run_command (GDAS_EXTERNAL (source), command);
    }
  else
    {
      int fd;
      gboolean must_close;
      fd = open_oss_microphone (microphone->device_name, &must_close);
      if (fd < 0)
	{
	  g_warning ("couldn't open microphone %s", microphone->device_name);
	  return;
	}
      gdas_external_has_opened (GDAS_EXTERNAL (source), fd);
      if (!must_close)
	GDAS_EXTERNAL (source)->dont_close = TRUE;
    }
}
static void gdas_microphone_get_device_name(GtkObject* object, GdamArg* arg)
{
  const char *device_name = GDAS_MICROPHONE (object)->device_name;
  if (device_name == NULL)
    return;
  GDAM_ARG_STRING (arg) = g_strdup (device_name);
}

static void gdas_microphone_set_device_name(GtkObject* object, GdamArg* arg)
{
  GdasMicrophone *microphone = GDAS_MICROPHONE (object);
  const char *device_name = microphone->device_name;
  if (device_name != NULL)
    {
      gdam_debug ("you cannot set a microphone's device twice");
      return;
    }
  microphone->device_name = g_strdup (GDAM_ARG_STRING (arg));
  gdas_microphone_maybe_open (GDAS_SOURCE (microphone));
}
static void gdas_microphone_set_use_helper(GtkObject* object, GdamArg* arg)
{
  GDAS_MICROPHONE (object)->use_helper = GDAM_ARG_INT (arg);
}
static void gdas_microphone_get_use_helper(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_MICROPHONE (object)->use_helper;
}

static void gdas_microphone_class_init(GdasMicrophoneClass* microphone_class)
{
  GdasSourceClass *source_class;
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (microphone_class);
  source_class = GDAS_SOURCE_CLASS (microphone_class);
  gdam_arg_register (object_class,
		     "device_name",
		     GDAM_ARG_TYPE_STRING,
		     gdas_microphone_set_device_name,
		     gdas_microphone_get_device_name,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "use_helper",
		     GDAM_ARG_TYPE_INT,
		     gdas_microphone_set_use_helper,
		     gdas_microphone_get_use_helper,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

GtkType gdas_microphone_get_type()
{
  static GtkType microphone_type = 0;
  if (!microphone_type)
    {
      static const GtkTypeInfo microphone_info = {
	"GdasMicrophone",
	sizeof (GdasMicrophone),
	sizeof (GdasMicrophoneClass),
	(GtkClassInitFunc) gdas_microphone_class_init,
	(GtkObjectInitFunc) gdas_microphone_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_EXTERNAL;
      microphone_type = gtk_type_unique (parent, &microphone_info);
    }
  return microphone_type;
}
