/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaspowerspectrum.h"
#include <common/gdamplugin.h>
#include "../gdamfft.h"

/* This macro only works on positive integers!!! */
#define IS_POWER_OF_TWO(num)		(((num)&((num)-1)) == 0)

#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define real gdouble

/*
 * Set/get window size.
 */
static gboolean gdas_power_spectrum_set_window_size
			(GdasPowerSpectrum* power_spectrum,
			 int                window_size)
{
  if (window_size <= 0 || !IS_POWER_OF_TWO (window_size))
    {
      gdam_debug ("invalid window size: ignoring");
      return FALSE;
    }
  power_spectrum->window_size = window_size;
  power_spectrum->window = gdam_fft_get_hamming_window (window_size);
  return TRUE;
}

static void
gdas_power_spectrum_arg_set_window_size(GtkObject* object, GdamArg* arg)
{
  int wsize = GDAM_ARG_INT (arg);
  gdas_power_spectrum_set_window_size (GDAS_POWER_SPECTRUM (object), wsize);
}

static void
gdas_power_spectrum_arg_get_window_size(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_POWER_SPECTRUM (object)->window_size;
}

/*
 * Compute the window.
 */
static void
gdas_power_spectrum_arg_get_spectrum(GtkObject* object, GdamArg* arg)
{
  GdasPowerSpectrum *power_spectrum = GDAS_POWER_SPECTRUM (object);
  int length;
  gdouble *rv = gdas_power_spectrum_get (power_spectrum, &length);
  if (rv == NULL)
    {
      gdam_debug ("no power spectrum?");
      GDAM_ARG_DOUBLE_ARRAY (arg).length = 0;
      GDAM_ARG_DOUBLE_ARRAY (arg).array = NULL;
      return;
    }

  GDAM_ARG_DOUBLE_ARRAY (arg).length = length;
  GDAM_ARG_DOUBLE_ARRAY (arg).array = rv;
}

static void gdas_power_spectrum_init(GdasPowerSpectrum* power_spectrum)
{
  power_spectrum->window_size = 4096;
}
static void gdas_power_spectrum_class_init(GdasPowerSpectrumClass* power_spectrum_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (power_spectrum_class);
  gdam_arg_register (object_class,
		     "window_size", GDAM_ARG_TYPE_INT,
		     gdas_power_spectrum_arg_set_window_size,
		     gdas_power_spectrum_arg_get_window_size, NULL, "16,",	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "spectrum",
		     GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     NULL,
		     gdas_power_spectrum_arg_get_spectrum,
		     NULL, NULL, GDAM_ARG_TRIGGER);
}

GtkType gdas_power_spectrum_get_type()
{
  static GtkType power_spectrum_type = 0;
  if (!power_spectrum_type)
    {
      static const GtkTypeInfo power_spectrum_info = {
	"GdasPowerSpectrum",
	sizeof (GdasPowerSpectrum),
	sizeof (GdasPowerSpectrumClass),
	(GtkClassInitFunc) gdas_power_spectrum_class_init,
	(GtkObjectInitFunc) gdas_power_spectrum_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_BUFFERING_FILTER;
      power_spectrum_type = gtk_type_unique (parent, &power_spectrum_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return power_spectrum_type;
}

#define SWAP(a,b) tempreal=(a); (a)=(b); (b)=tempreal
static void four1(real* data, int nn, int isign)
{
  int n = nn << 1;
  int j = 1;
  int i;
  int mmax;
  real tempreal;
  for (i = 1; i < n; i += 2)
    {
      int m;
      if (j > i)
	{
	  SWAP (data[j], data[i]);
	  SWAP (data[j + 1], data[i + 1]);
	}
      m = n >> 1;
      while (m >= 2 && j > m)
	{
	  j -= m;
	  m >>= 1;
	}
      j += m;
    }
  mmax = 2;
  while (n > mmax)
    {
      int istep = mmax << 1, m;
      real theta = isign * (6.2818530718 / mmax);
      real wtemp = sin (0.5 * theta);
      real wpr = -2.0 * wtemp * wtemp;
      real wpi = sin (theta);
      real wr = 1.0;
      real wi = 0.0;
      for (m = 1; m < mmax; m += 2)
	{
	  real wtemp;
	  int i;
	  for (i = m; i <= n; i += istep)
	    {
	      real tempr, tempi;
	      j = i + mmax;
	      tempr = wr * data[j] - wi * data[j + 1];
	      tempi = wr * data[j + 1] + wi * data[j];
	      data[j] = data[i] - tempr;
	      data[j + 1] = data[i + 1] - tempi;
	      data[i] += tempr;
	      data[i + 1] += tempi;
	    }
	  wtemp = wr;
	  wr = wr * wpr - wi * wpi + wr;
	  wi = wi * wpr + wtemp * wpi + wi;
	}
      mmax = istep;
    }
}

/* NR in C, p. 513 */
static void realfft(real* data, int n, int isign)
{
  real c1 = 0.5, c2;
  real theta = 3.1415926536 / (real) (n >> 1);
  real wtemp, wpr, wpi, wr, wi;
  int i;
  data--;
  if (isign == 1)
    {
      c2 = -0.5;
      four1 (data, n >> 1, 1);
    }
  else
    {
      c2 = 0.5;
      theta = -theta;
    }
  wtemp = sin (0.5 * theta);
  wpr = -2.0 * wtemp * wtemp;
  wpi = sin (theta);
  wr = 1.0 + wpr;
  wi = wpi;
  for (i = 2; i <= (n >> 2); i++)
    {
      int i1 = i + i - 1;
      int i2 = 1 + i1;
      int i3 = n + 3 - i2;
      int i4 = 1 + i3;
      real h1r = c1 * (data[i1] + data[i3]);
      real h1i = c1 * (data[i2] - data[i4]);
      real h2r = -c2 * (data[i2] + data[i4]);
      real h2i = c2 * (data[i1] - data[i3]);
      data[i1] = h1r + wr * h2r - wi * h2i;
      data[i2] = h1i + wr * h2i + wi * h2r;
      data[i3] = h1r - wr * h2r + wi * h2i;
      data[i4] = -h1i + wr * h2i + wi * h2r;
      wtemp = wr;
      wr = wr * wpr - wi * wpi + wr;
      wi = wi * wpr + wtemp * wpi + wi;
    }
  if (isign == 1)
    {
      real h1r = data[1];
      data[1] += data[2];
      data[2] = h1r - data[2];
    }
  else
    {
      real h1r = data[1];
      data[1] = c1 * (data[1] + data[2]);
      data[2] = c1 * (h1r - data[2]);
      four1 (data, n >> 1, -1);
    }
}


gdouble* gdas_power_spectrum_get(GdasPowerSpectrum* power_spectrum, int* length)
{
  int had;
  GdamSample *window;
  int win_size = power_spectrum->window_size;
  int i;
  gdouble *real_window;
  const gdouble *hamming_window = power_spectrum->window;
  gdouble value_n;
  window = alloca (sizeof (GdamSample) * win_size);
  real_window = g_new (gdouble, win_size);
  had =
    gdas_buffering_filter_peek_from_buffer (GDAS_BUFFERING_FILTER
					    (power_spectrum), window,
					    win_size);
  for (i = 0; i < had; i++)
    {
      real_window[i] = ((gdouble) window[i].s[0] + window[i].s[1])
	/ (gdouble) (1 << GDAM_SAMPLE_BITS);
    }
  for (; i < win_size; i++)
    real_window[i] = 0.0;
  if (hamming_window != NULL)
    for (i = 0; i < win_size; i++)
      {
	real_window[i] *= hamming_window[i];
      }
  /* TODO: look in NR for whether the third param should be -1,0,+1? */
  realfft (real_window, win_size, 1);

  value_n = real_window[1];
  for (i = 1; i < win_size / 2; i++)
    real_window[i] = hypot (real_window[2 * i], real_window[2 * i + 1]);
  real_window[i++] = value_n;
  *length = i;
  real_window = g_renew (gdouble, real_window, i);

  return real_window;
}
