/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamdial.h"
#include "pixmapcache.h"
#include "gdamskin.h"		/* needed for path stuff. */
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

enum {
	MOTION,
	LAST_SIGNAL
};
static guint gdam_dial_signals[LAST_SIGNAL] = { 0 };

static int  gdam_dial_expose_event           (GtkWidget*        widget,
                                              GdkEventExpose*   event);
static int  gdam_dial_configure_event        (GtkWidget*        widget,
                                              GdkEventConfigure*event);
static void gdam_dial_draw                   (GtkWidget*        widget,
                                              GdkRectangle*     area);
static void gdam_dial_draw_default           (GtkWidget*        widget);
static void gdam_dial_draw_focus             (GtkWidget*        widget);
static void gdam_dial_size_request           (GtkWidget*        widget,
                                              GtkRequisition*   event);
static int  gdam_dial_press_event            (GtkWidget*        widget,
                                              GdkEventButton*   event);
static int  gdam_dial_release_event          (GtkWidget*        widget,
                                              GdkEventButton*   event);
static int  gdam_dial_motion_event           (GtkWidget*        widget,
                                              GdkEventMotion*   event);
static GtkObjectClass* parent_class;
static void try_load_images                  (GdamDial*         dial);
static void gdam_dial_init(GdamDial* pix_dial)
{

  GTK_WIDGET_SET_FLAGS (pix_dial, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);
  gtk_widget_add_events (GTK_WIDGET (pix_dial),
			 GDK_EXPOSURE_MASK
			 | GDK_STRUCTURE_MASK
			 | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
  pix_dial->wrapcount = 1;
}
static void gdam_dial_class_init(GdamDialClass* klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  widget_class->expose_event = gdam_dial_expose_event;
  widget_class->configure_event = gdam_dial_configure_event;
  widget_class->button_press_event = gdam_dial_press_event;
  widget_class->button_release_event = gdam_dial_release_event;
  widget_class->motion_notify_event = gdam_dial_motion_event;
  widget_class->size_request = gdam_dial_size_request;
  widget_class->draw = gdam_dial_draw;
  widget_class->draw_focus = gdam_dial_draw_focus;
  widget_class->draw_default = gdam_dial_draw_default;
  gdam_dial_signals[MOTION] =
    gtk_signal_new ("motion",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamDialClass, motion),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, gdam_dial_signals, LAST_SIGNAL);
}

GtkType gdam_dial_get_type()
{
  static GtkType pix_dial_type = 0;
  if (!pix_dial_type)
    {
      static const GtkTypeInfo pix_dial_info = {
	"GdamDial",
	sizeof (GdamDial),
	sizeof (GdamDialClass),
	(GtkClassInitFunc) gdam_dial_class_init,
	(GtkObjectInitFunc) gdam_dial_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_EVENT_BOX;
      parent_class = gtk_type_class (parent);
      pix_dial_type = gtk_type_unique (parent, &pix_dial_info);
    }
  return pix_dial_type;
}

static void try_load_images(GdamDial*         dial)
{
  GdkWindow *window = GTK_WIDGET (dial)->window;
  PixmapInfo *pixmap_info;
  g_return_if_fail (dial->num_tiles > 0);
  if (dial->pixmap != NULL)
    return;
  if (dial->image_filename == NULL)
    return;
  g_return_if_fail (window != NULL);
  pixmap_info = pixmap_cache_get_pixmap (window, dial->image_filename);
  if (pixmap_info == NULL)
    {
      g_warning ("dial: error loading image %s", dial->image_filename);
      return;
    }
  dial->pixmap = pixmap_info->pixmap;
  dial->mask = pixmap_info->mask;
  dial->tile_width = pixmap_info->width;
  dial->tile_height = pixmap_info->height / dial->num_tiles;
}

static void
gdam_dial_draw_focus (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void
gdam_dial_draw_default (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void maybe_redraw_image(GdamDial* dial)
{
  int new_index;
  gfloat fraction;
  GtkAdjustment *adj = dial->adjustment;
  fraction = (adj->value - adj->lower) / (adj->upper - adj->lower);
  new_index = floor (fraction * dial->num_tiles * dial->wrapcount);
  new_index = new_index % dial->num_tiles;
  new_index = CLAMP (new_index, 0, dial->num_tiles - 1);
  if (dial->images_reversed)
    new_index = dial->num_tiles - 1 - new_index;
  if (new_index != dial->last_image_drawn)
    {
      dial->last_image_drawn = new_index;
      gtk_widget_draw_default (GTK_WIDGET (dial));
    }
}

static void
reset_delta_x(GdamDial* dial)
{
  GtkAdjustment *adjustment = dial->adjustment;
  g_return_if_fail (adjustment != NULL);
  gtk_adjustment_set_value (dial->adjustment, dial->orig_value);
}


static void
update_delta_x(GdamDial* dial, int delta_x)
{
  GtkAdjustment *adjustment = dial->adjustment;
  gfloat amount = delta_x;
  if (dial->screen_range == 0)
    {
      return;
    }
  g_return_if_fail (adjustment != NULL);
  amount /= dial->screen_range;
  amount *= dial->adjustment->upper - dial->adjustment->lower;
  amount += adjustment->value;
  amount = CLAMP (amount, adjustment->lower, adjustment->upper);
  gtk_adjustment_set_value (dial->adjustment, amount);
}

static void gdam_dial_draw(GtkWidget* widget, GdkRectangle* area)
{
  GdamDial *dial = GDAM_DIAL (widget);
  GdkWindow *window = widget->window;
  int window_width = widget->allocation.width;
  int window_height = widget->allocation.height;
  int tile_top, x_margin, y_margin;

  if (!GTK_WIDGET_DRAWABLE (widget))
    return;

  if (dial->pixmap == NULL)
    try_load_images (dial);
  x_margin = (window_width - dial->tile_width) / 2;
  y_margin = (window_height - dial->tile_height) / 2;
  tile_top = dial->tile_height * dial->last_image_drawn;
#if 1
  gdk_draw_pixmap (widget->window, widget->style->fg_gc[0], dial->pixmap, 0, tile_top,	/* src x,y */
		   x_margin, y_margin,	/* dest x,y */
		   dial->tile_width, dial->tile_height);	/* extent */
#else
  gdk_window_copy_area (widget->window, widget->style->fg_gc[0], x_margin, y_margin,	/* dest x,y */
			dial->pixmap, 0, tile_top,	/* src x,y */
			dial->tile_width, dial->tile_height);	/* extent */
#endif
}

static int gdam_dial_press_event(GtkWidget*       widget,
                                 GdkEventButton*  event)
{
  GdamDial *dial = GDAM_DIAL (widget);
  gtk_grab_add (widget);
  dial->last_drag_x = event->x;
  dial->is_dragging = TRUE;
  return FALSE;
}

static int gdam_dial_release_event(GtkWidget*       widget,
                                   GdkEventButton*  event)
{
  gboolean clicked = FALSE;
  int x, y;
  guint state;
  GdamDial *dial = GDAM_DIAL (widget);
  gtk_grab_remove (widget);

  if (event->window != widget->window)
    {
      gdk_window_get_pointer (widget->window, &x, &y, &state);
    }
  else
    {
      x = event->x;
      y = event->y;
      state = event->state;
    }


  if (dial->reset_on_release)
    {
      reset_delta_x (dial);
    }
  else
    {
      update_delta_x (dial, x - dial->last_drag_x);
    }

  dial->is_dragging = FALSE;

  return TRUE;
}

static int gdam_dial_expose_event(GtkWidget*       widget,
                                  GdkEventExpose*  event)
{
  GdamDial *dial = GDAM_DIAL (widget);
  gtk_widget_draw_default (GTK_WIDGET (dial));
  return FALSE;
}
static int gdam_dial_configure_event(GtkWidget*         widget,
                                     GdkEventConfigure* event)
{
  GdamDial *dial = GDAM_DIAL (widget);
  if (dial->pixmap == NULL)
    try_load_images (dial);
  gtk_widget_draw_default (GTK_WIDGET (dial));
  return FALSE;
}

static void gdam_dial_size_request    (GtkWidget*       widget,
                                       GtkRequisition*  requisition)
{
  GdamDial *dial = GDAM_DIAL (widget);
  if (widget->window != NULL)
    try_load_images (dial);
  requisition->width = dial->tile_width;
  requisition->height = dial->tile_height;
}

static int  gdam_dial_motion_event          (GtkWidget*       widget,
                                             GdkEventMotion*  event)
{
  int x, y;
  guint state;
  GdamDial *dial;
  dial = GDAM_DIAL (widget);
  if (!dial->is_dragging)
    return 0;
  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      if (event->window != widget->window)
	{
	  gdk_window_get_pointer (widget->window, &x, &y, &state);
	}
      else
	{
	  x = event->x;
	  y = event->y;
	  state = event->state;
	}
    }

  update_delta_x (dial, x - dial->last_drag_x);
  dial->last_drag_x = x;
  return 0;
}

void gdam_dial_set_pixmap_by_filename(GdamDial*       dial,
                                      const char*     image_filename)
{
  dial->image_filename = g_strdup (image_filename);
  if (GTK_WIDGET (dial)->window)
    try_load_images (dial);
}

/*
 * To make a GdamDial:
 *
 *   string1:    image_filename
 *   string2:    cur_value:min_value:max_value[:reversed][:reset][:pix_wrapcount]
 *   int1:       screen_range
 *   int2:       num_tiles
 *
 * Here is a description of the fields.
 *
 *     image_filename:   a vertically stacked array of pixmaps.
 *     cur_value:        initial value for the adjustment.
 *     min_value:        minimum value for the adjustment.
 *     max_value:        maximum value for the adjustment.
 *     screen_range:     number of pixels to move to go from
 *                       min_value to max_value.
 *     num_tiles:        number of vertically stacked images in the image file
 */
GtkWidget* gdam_dial_create_func           (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char **coordinates;
  gfloat cur_value, min_value, max_value, step_incr;
  GtkAdjustment *adjustment;
  GdamDial *dial;
  int id;
  int i;
  dial = GDAM_DIAL (gtk_type_new (GDAM_TYPE_DIAL));
  gdam_dial_set_pixmap_by_filename (dial, string1);

  coordinates = g_strsplit (string2, ":", 100);
  if (coordinates[0] == NULL
      || coordinates[1] == NULL || coordinates[2] == NULL)
    {
      g_warning ("gdam_dial_create_func: too few arguments to "
		 "init adjustment");
      g_strfreev (coordinates);
      gtk_object_sink (GTK_OBJECT (dial));
      return NULL;
    }

  cur_value = atof (coordinates[0]);
  min_value = atof (coordinates[1]);
  max_value = atof (coordinates[2]);
  for (i = 3; coordinates[i] != NULL; i++)
    {
      const char *arg = coordinates[i];
      if (strcmp (arg, "reversed") == 0)
	{
	  dial->images_reversed = 1;
	  continue;
	}
      else if (strcmp (arg, "reset") == 0)
	{
	  dial->reset_on_release = 1;
	  continue;
	}
      else
	{
	  int wcount;
	  wcount = atoi (arg);
	  if (wcount > 0)
	    {
	      dial->wrapcount = wcount;
	    }
	  continue;
	}
    }
  step_incr = 0.01;
  g_strfreev (coordinates);

  adjustment = (GtkAdjustment *) gtk_adjustment_new (cur_value,
						     min_value,
						     max_value,
						     step_incr,
						     step_incr * 5,
						     step_incr);
  gdam_dial_set_adjustment (dial, adjustment);
  dial->screen_range = int1;
  dial->num_tiles = int2;
  dial->orig_value = cur_value;

  return GTK_WIDGET (dial);
}


void gdam_dial_set_adjustment(GdamDial*  dial,
                              GtkAdjustment*    adjustment)
{
  GtkAdjustment *old_adj = dial->adjustment;
  if (old_adj == adjustment)
    return;
  if (old_adj != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (old_adj), dial->connect_id);
      gtk_object_unref (GTK_OBJECT (old_adj));
    }
  dial->adjustment = adjustment;
  if (adjustment == NULL)
    dial->connect_id = 0;
  else
    {
      GtkObject *object = GTK_OBJECT (adjustment);
      gtk_object_ref (object);
      gtk_object_sink (object);
      dial->connect_id
	= gtk_signal_connect_object
	(object,
	 "value_changed",
	 GTK_SIGNAL_FUNC (maybe_redraw_image), GTK_OBJECT (dial));
      update_delta_x (dial, 0);
    }
}
