/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include "gdamhistory.h"

static char* preferences_directory = NULL;
gchar* gdam_dir = NULL;
static GHashTable* histories = NULL;

void initialize_root_gdam_directory()
{
  struct stat stat_buf;
  static gboolean has_initialized = FALSE;

  if (has_initialized)
    return;
  has_initialized = TRUE;

  gdam_dir = g_strdup_printf ("%s/.gdam", g_get_home_dir ());

  if (stat (gdam_dir, &stat_buf) < 0)
    {
      g_warning ("no preferences directory -- I will create one");
      if (mkdir (gdam_dir, 0755) < 0)
	{
	  g_warning ("I cannot make the directory `%s'...", gdam_dir);
	  return;
	}
    }
  if (!S_ISDIR (stat_buf.st_mode))
    {
      g_warning ("currently ~/.gdam isn't a directory, "
		 "guess I won't save your preferences");
      return;
    }
}


static void initialize_preferences_directory()
{
  static gboolean inited = FALSE;
  struct stat stat_buf;

  if (inited)
    return;

  inited = TRUE;
  initialize_root_gdam_directory ();
  if (!gdam_dir)
    return;
  preferences_directory = g_strdup_printf ("%s/history", gdam_dir);
  if (stat (preferences_directory, &stat_buf) < 0)
    {
      g_warning ("no preferences directory -- I will create one");
      if (mkdir (preferences_directory, 0755) < 0)
	{
	  g_warning ("I cannot make the directory `%s'...",
		     preferences_directory);
	  g_free (preferences_directory);
	  preferences_directory = NULL;
	  return;
	}
      return;
    }
  if (!S_ISDIR (stat_buf.st_mode))
    {
      g_warning ("currently %s isn't a directory, "
		 "guess I won't save your preferences",
		 preferences_directory);
      g_free (preferences_directory);
      preferences_directory = NULL;
      return;
    }
  return;
}
static GHashTable* load_hash_table(const char* category)
{
  char buf[4096];
  char *filename;
  FILE *fp;
  GHashTable *rv = g_hash_table_new (g_str_hash, g_str_equal);
  if (preferences_directory == NULL)
    return rv;
  filename = g_strdup_printf ("%s/%s", preferences_directory, category);
  fp = fopen (filename, "r");
  g_free (filename);
  if (fp == NULL)
    return rv;
  while (fgets (buf, sizeof (buf), fp) != NULL)
    {
      char *colon = strchr (buf, ':');
      if (colon == NULL)
	continue;
      *colon++ = '\0';
      while (*colon && isspace (*colon))
	colon++;
      if (g_hash_table_lookup (rv, buf))
	{
	  g_warning ("internal preferences file format error: "
		     "the category %s / key %s "
		     "appeared twice", category, buf);
	  continue;
	}
      g_strchomp (colon);
      g_hash_table_insert (rv, g_strdup (buf), g_strdup (colon));
    }
  fclose (fp);
  return rv;
}

static void write_one_keyvalue(const char* key, const char* val, FILE* fp)
{
  fprintf (fp, "%s: %s\n", key, val);
}
static void rewrite_hash_table(const char* category, GHashTable* table)
{
  char *filename;
  FILE *fp;
  if (preferences_directory == NULL)
    return;
  filename = g_strdup_printf ("%s/%s", preferences_directory, category);
  fp = fopen (filename, "w");
  g_free (filename);
  if (fp == NULL)
    return;
  g_hash_table_foreach (table, (GHFunc) write_one_keyvalue, fp);
  fclose (fp);
}

static GHashTable* get_cat_table(const char* category)
{
  GHashTable *cat_table;
  initialize_preferences_directory ();
  if (histories == NULL)
    histories = g_hash_table_new (g_str_hash, g_str_equal);
  cat_table = g_hash_table_lookup (histories, category);
  if (cat_table == NULL)
    {
      cat_table = load_hash_table (category);
      g_hash_table_insert (histories, g_strdup (category), cat_table);
    }
  return cat_table;
}

const char*   gdam_history_get_entry    (const char*     category,
                                          const char*     entry)
{
  GHashTable *cat_table;
  cat_table = get_cat_table (category);
  return (const char *) g_hash_table_lookup (cat_table, entry);
}

void          gdam_history_set_entry     (const char*     category,
                                          const char*     entry,
					  const char*     value)
{
  GHashTable *cat_table;
  char *old_value;
  char *val;
  cat_table = get_cat_table (category);
  old_value = g_hash_table_lookup (cat_table, entry);
  val = g_strdup (value);
  if (old_value)
    {
      g_free (old_value);
      g_hash_table_insert (cat_table, (char *) entry, val);
    }
  else
    {
      g_hash_table_insert (cat_table, g_strdup (entry), val);
    }
  rewrite_hash_table (category, cat_table);
}
