/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include <gtk/gtksignal.h>
#include <gtk/gtktree.h>
#include <gtk/gtktreeitem.h>
#include <stdlib.h>
#include "gdaminputselector.h"
#include "gdamskinutil.h"
#include "gdampacker.h"

static guint gdam_new_trap_signal_id = 0;
static guint gdam_input_event_signal_id = 0;

static void gdam_input_selector_class_init (GdamSkinClass* skin_class);

static void gdam_input_selector_destroy(GtkObject* object)
{
  GdamInputSelector *selector = GDAM_INPUT_SELECTOR (object);
  if (selector->emission_id != 0)
    gtk_signal_remove_emission_hook (gdam_new_trap_signal_id,
				     selector->emission_id);
#if 0
  if (selector->grope_emission_id != 0)
    gtk_signal_remove_emission_hook (gdam_new_trap_signal_id,
				     selector->grope_emission_id);
#endif
}

static void gdam_input_selector_init(GdamInputSelector* input_selector)
{
}
GtkType gdam_input_selector_get_type()
{
  static GtkType input_selector_type = 0;
  if (!input_selector_type)
    {
      static const GtkTypeInfo input_selector_info = {
	"GdamInputSelector",
	sizeof (GdamInputSelector),
	sizeof (GdamInputSelectorClass),
	(GtkClassInitFunc) gdam_input_selector_class_init,
	(GtkObjectInitFunc) gdam_input_selector_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      input_selector_type = gtk_type_unique (parent, &input_selector_info);
    }
  return input_selector_type;
}


static GdamInputSelector* latest = NULL;
static guint             latest_destroy_id = 0;

static void set_latest_input_selector(GdamInputSelector* selector)
{
  if (latest == selector)
    return;
  if (latest != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (latest), latest_destroy_id);
      latest = NULL;
    }
  if (selector != NULL)
    {
      /* This is certainly too tricky and clever... */
      latest_destroy_id =
	gtk_signal_connect_object (GTK_OBJECT (selector),
				   "destroy",
				   GTK_SIGNAL_FUNC
				   (set_latest_input_selector), NULL);
      latest = selector;
    }
}

static gboolean gdam_input_selector_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GDAM_INPUT_SELECTOR (skin)->entry = GTK_ENTRY (object);
  return TRUE;
}

static gboolean gdam_input_selector_skin_tree_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GDAM_INPUT_SELECTOR (skin)->tree = GTK_TREE (object);
  return TRUE;
}

static void input_selector_open(GdamInputSelector* skin)
{
  char *device_name;
  GdamInputDevice *input_device;
  g_return_if_fail (skin->entry != NULL);
  device_name = gtk_entry_get_text (skin->entry);
  input_device = gdam_input_device_get_by_name (device_name);
}

typedef struct _SelInfo {
	GdamInputSelector* sel;
	GtkTree* tree;
} SelInfo;

static void add_skin(gpointer ptr, gpointer ptr2);
static void add_packed_skin(gpointer key, gpointer value, gpointer si)
{
  add_skin (value, si);
}

static void add_skin(gpointer ptr, gpointer ptr2)
{
  GdamSkin *skin;
  GdamInputSelector *selector;
  GtkType packer_type;
  GtkWidget *leaf;
  GdamModel *model;
  SelInfo *si;
  const char *skin_type;
  si = (SelInfo *) ptr2;
  skin = GDAM_SKIN (ptr);
  selector = GDAM_INPUT_SELECTOR (si->sel);
  model = skin->model;
  skin_type = gtk_type_name (GTK_OBJECT (skin)->klass->type);
  leaf = gtk_tree_item_new_with_label (skin_type);
  gtk_tree_append (si->tree, leaf);
  packer_type = gtk_type_from_name ("GdamPacker");
  if (model != NULL)
    {
      GtkWidget *new_leaf;
      GtkWidget *subtree;
      const char *model_type;
      model_type = gtk_type_name (GTK_OBJECT (model)->klass->type);
      subtree = gtk_tree_new ();
      gtk_tree_item_set_subtree (GTK_TREE_ITEM (leaf), subtree);
      new_leaf = gtk_tree_item_new_with_label (model_type);
      gtk_tree_append (GTK_TREE (subtree), new_leaf);
      gtk_widget_show (subtree);
      gtk_widget_show (new_leaf);
    }
  else if (gtk_type_is_a (GTK_OBJECT (skin)->klass->type, packer_type))
    {
      GHashTable *named_skins;
      GtkWidget *subtree;
      GtkTree *oldtree;
      named_skins = GDAM_PACKER (skin)->named_skins;
      subtree = gtk_tree_new ();
      gtk_tree_item_set_subtree (GTK_TREE_ITEM (leaf), subtree);
      gtk_widget_show (subtree);
      oldtree = si->tree;
      si->tree = GTK_TREE (subtree);
      g_hash_table_foreach (named_skins, add_packed_skin, si);
      si->tree = oldtree;
    }
  gtk_widget_show (leaf);
}

static void refresh_skins(GdamInputSelector* selector)
{
  GList *skin_list;
  SelInfo si;
  skin_list = gdam_skin_get_global_skin_list ();
  g_return_if_fail (selector->tree != NULL);
  g_return_if_fail (skin_list != NULL);
  gtk_tree_clear_items (selector->tree, 0, -1);
  si.sel = selector;
  si.tree = selector->tree;
  g_list_foreach (skin_list, add_skin, &si);
}

static gboolean gdam_input_selector_refresh_skins_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (refresh_skins),
			     (GtkObject *) skin);
  return TRUE;
}

static gboolean gdam_input_selector_open_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  set_latest_input_selector (GDAM_INPUT_SELECTOR (skin));
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (input_selector_open),
			     (GtkObject *) skin);
  return TRUE;
}

/* FIXME:  the columns are hardcoded.
 *             0 is the patch name,
 *             1 is the description,
 *             2 is the destination.
 */

static void handle_new_trap(GdamInputSelector* selector, GdamInputTrap* trap)
{
  /* XXX: destination hasn't been dealt with. */
  const char *description;
  //const char* device_name = trap->device->filename;
  char *trap_id;
  char *strings[3];

  int index;

  if (selector->clist == NULL)
    return;

  description = gdam_input_trap_get_name (trap);
  trap_id = gdam_input_event_stringify (trap->event);
  strings[0] = trap_id;
  strings[1] = (char *) description;
  strings[2] = "";
  index = gtk_clist_append (selector->clist, strings);
  gtk_clist_set_row_data (selector->clist, index, trap);
  g_free (trap_id);
}
static gboolean clist_emission_hook(GtkObject*       object,
                                    guint            signal_id,
				    guint            n_params,
				    GtkArg*          params,
				    gpointer         skin_as_data)
{
  GdamInputSelector *skin = GDAM_INPUT_SELECTOR (skin_as_data);
  GdamInputTrap *trap;
  g_return_val_if_fail (skin != NULL, TRUE);
  g_return_val_if_fail (n_params == 1, TRUE);
  g_return_val_if_fail (params[0].type == GTK_TYPE_POINTER, TRUE);
  trap = (GdamInputTrap *) GTK_VALUE_POINTER (params[0]);
  handle_new_trap (skin, trap);
  return TRUE;
}

static void handle_input_iterate(GdamInputTrap* trap, gpointer user_data)
{
  handle_new_trap (GDAM_INPUT_SELECTOR (user_data), trap);
}

/* Attach this handler in the xml file to register this
 * clist with the skin. */

static void patch_selected(GtkWidget*      clist,
                           gint            row,
			   gint            col,
			   GdkEventButton* event,
                           GdamInputSelector* selector)
{
  GdamInputTrap *trap;
  trap = gtk_clist_get_row_data (selector->clist, row);

  selector->selected_trap = trap;
  gdam_input_trap_set_default (trap);
}

static gboolean gdam_input_selector_clist_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamInputSelector *sel_skin = GDAM_INPUT_SELECTOR (skin);
  GtkCList *clist = GTK_CLIST (object);
  g_return_val_if_fail (sel_skin->clist == NULL, TRUE);
  g_return_val_if_fail (clist != NULL, TRUE);
  if (gdam_new_trap_signal_id == 0)
    {
      gdam_new_trap_signal_id = gtk_signal_lookup ("new-trap",
						   GDAM_TYPE_INPUT_DEVICE);
      gdam_input_event_signal_id = gtk_signal_lookup ("event",
						      GDAM_TYPE_INPUT_DEVICE);
    }
  g_return_val_if_fail (gdam_new_trap_signal_id != 0, TRUE);
  sel_skin->clist = clist;
  sel_skin->emission_id = gtk_signal_add_emission_hook
    (gdam_new_trap_signal_id, clist_emission_hook, skin);
  gtk_signal_connect (GTK_OBJECT (clist),
		      "select_row",
		      GTK_SIGNAL_FUNC (patch_selected), sel_skin);

  gtk_clist_freeze (clist);
  gdam_input_trap_forall (NULL, handle_input_iterate, skin);
  gtk_clist_thaw (clist);
  return TRUE;
}

typedef struct _GdamInputPatch GdamInputPatch;
struct _GdamInputPatch {
	GtkObject*		object;
	guint                   object_destroy_id;

	GSList*			handle_ids;

	/* Nonzero when groping. */
	guint			grope_emission_hook;
	guint			groped_trap_id;

	/* Function to call to associate a new hook. */
	guint                  (*do_connect)(GdamInputPatch*,
	                                     GdamInputTrap*);

	/* Information about the signal we are
	 * connecting. */
	GdamModel*		model;
	char*			arg_name;
	gdouble			min_value, max_value;
};

static void gdam_input_patch_destroy(GdamInputPatch* input_patch)
{
  GSList *tmp;
  for (tmp = input_patch->handle_ids; tmp != NULL; tmp = tmp->next)
    {
      guint id = GPOINTER_TO_UINT (tmp->data);
      gdam_input_hook_disconnect (id);
    }
  if (input_patch->grope_emission_hook)
    {
      gtk_signal_remove_emission_hook
	(gdam_input_event_signal_id, input_patch->grope_emission_hook);
    }
  g_free (input_patch->arg_name);
  g_free (input_patch);
}

static void object_destroyed(GdamInputPatch* patch)
{
  patch->object = NULL;
  patch->object_destroy_id = 0;
  gdam_input_patch_destroy (patch);
}

GdamInputPatch* gdam_input_patch_force_associated(GtkObject* object)
{
  GdamInputPatch *patch;
  patch = gtk_object_get_data (object, "gdam-input-patch");
  if (patch == NULL)
    {
      patch = g_new0 (GdamInputPatch, 1);
      patch->object = object;
      patch->object_destroy_id
	= gtk_signal_connect_object
	(object, "destroy",
	 GTK_SIGNAL_FUNC (object_destroyed), (GtkObject *) patch);
      gtk_object_set_data (object, "gdam-input-patch", patch);
    }
  return patch;
}

GdamInputPatch* gdam_input_patch_create_associated(GtkObject* object)
{
  GdamInputPatch *patch = gdam_input_patch_force_associated (object);
  g_return_val_if_fail (patch != NULL, NULL);
  g_return_val_if_fail (patch->do_connect == NULL, NULL);
  return patch;
}

static void gdam_input_patch_connect(GdamInputPatch* input_patch,
                                     GdamInputTrap*  trap)
{
  guint id;
  g_return_if_fail (input_patch->do_connect != NULL);
  id = (*input_patch->do_connect) (input_patch, trap);
  input_patch->handle_ids = g_slist_prepend (input_patch->handle_ids,
					     GUINT_TO_POINTER (id));
}

static void gdam_input_patch_stop_groping(GdamInputPatch* input_patch)
{
  if (input_patch->grope_emission_hook != 0)
    {
      guint trap_id = input_patch->groped_trap_id;
      gtk_signal_remove_emission_hook
	(gdam_input_event_signal_id, input_patch->grope_emission_hook);
      if (trap_id != 0)
	{
	  GdamInputTrap *trap;
	  trap = gdam_input_trap_get_by_id (trap_id);
	  g_return_if_fail (trap->trap_id == trap_id);
	  gdam_input_patch_connect (input_patch, trap);
	}
      input_patch->grope_emission_hook = 0;
      input_patch->groped_trap_id = 0;
    }
}
static gboolean grope_emission_hook(GtkObject*      object,
				    guint           signal_id,
				    guint	    n_params,
				    GtkArg*	    params,
				    gpointer        data)
{
  GdamInputTrap *trap;
  GdamInputPatch *patch = data;


  trap = GTK_VALUE_POINTER (params[1]);
  /* event == params[0] */
  gdam_debug ("%s: got trap %d", __FUNCTION__, trap->trap_id);
  patch->groped_trap_id = trap->trap_id;
  return TRUE;
}

static void gdam_input_patch_start_groping(GdamInputPatch* input_patch)
{
  if (input_patch->grope_emission_hook != 0)
    return;
  input_patch->grope_emission_hook
    = gtk_signal_add_emission_hook
    (gdam_input_event_signal_id, grope_emission_hook, input_patch);
}


/* A input grope for a generic skin.
 * (The skin here is *not* a input selector)
 */
static gboolean gdam_input_grope_stop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamInputPatch *input_patch;
  input_patch = gdam_input_patch_force_associated (object);
  gdam_debug ("gdam_input_grope_stop handled.");
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (gdam_input_patch_stop_groping),
			     (GtkObject *) input_patch);
  return TRUE;
}

static void gdam_input_patch_clear(GdamInputPatch* patch)
{
  GSList *tmp;
  for (tmp = patch->handle_ids; tmp != NULL; tmp = tmp->next)
    {
      guint hook_id = GPOINTER_TO_UINT (tmp->data);
      gdam_input_hook_disconnect (hook_id);
    }
  g_slist_free (patch->handle_ids);
  patch->handle_ids = NULL;
}


static guint gdam_input_patch_range_connect(GdamInputPatch* patch,
                                            GdamInputTrap*  trap)
{
  gdam_debug ("doing connect: trap %d, to %s", trap->trap_id,
	      patch->arg_name);
  return gdam_input_hook_connect_range (trap, GTK_OBJECT (patch->model),
					patch->arg_name, patch->min_value,
					patch->max_value);
}


/* A input grope for a generic skin.
 * (The skin here is *not* a input selector)
 */
static gboolean gdam_input_patch_clear_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamInputPatch *input_patch;
  if (num_handler_names != 1)
    {
      g_warning ("usage: gdam_input_patch_clear");
    }

  input_patch = gdam_input_patch_force_associated (object);
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (gdam_input_patch_clear),
			     (GtkObject *) input_patch);
  gdam_debug ("gdam_input_patch_clear handled.");
  return TRUE;
}

/* A input grope for a generic skin.
 * (The skin here is *not* a input selector)
 */
static gboolean gdam_input_grope_start_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  char *arg_name;
  char *desc;
  char *endp;
  gdouble min_val, max_val;
  GdamInputPatch *input_patch;
  if (num_handler_names != 5)
    {
      g_warning ("usage: gdam_input_grope:argname:min:max:description");
      return TRUE;
    }
  arg_name = g_strdup (handler_names[1]);
  min_val = strtod (handler_names[2], &endp);
  if (*endp != '\0')
    {
      g_warning ("input_patch: error parsing double");
      return TRUE;
    }
  max_val = strtod (handler_names[3], &endp);
  if (*endp != '\0')
    {
      g_warning ("input_patch: error parsing double");
      return TRUE;
    }
  desc = g_strdup (handler_names[4]);

  input_patch = gdam_input_patch_force_associated (object);
  g_return_val_if_fail (input_patch != NULL, TRUE);
  input_patch->model = skin->model;
  input_patch->arg_name = arg_name;
  input_patch->min_value = min_val;
  input_patch->max_value = max_val;
  input_patch->grope_emission_hook = 0;
  input_patch->do_connect = gdam_input_patch_range_connect;
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (gdam_input_patch_start_groping),
			     (GtkObject *) input_patch);
  gdam_debug ("gdam_input_grope_start handled.");
  return TRUE;
}

static void gdam_input_patch_connect_default(GdamInputPatch* input_patch)
{
  GdamInputTrap *trap;
  trap = gdam_input_trap_get_default ();
  g_return_if_fail (trap != NULL);
  gdam_input_patch_connect (input_patch, trap);
}

/* A input patch for a generic skin.
 * when signalled, attaches to current device and event
 * (The skin here is *not* a input selector)
 *
 * Arguments:
 *       argname            : the name of the double-argument to change
 *       min_val            : the minimum value for that param
 *       max_val            : the maximum value for that param
 *       description        : input hook description
 */
static gboolean gdam_input_patch_range_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  char *arg_name;
  char *desc;
  char *endp;
  gdouble min_val, max_val;
  GdamInputPatch *input_patch;
  if (num_handler_names != 5)
    {
      g_warning ("usage: gdam_input_patch:argname:min:max:description");
      return TRUE;
    }
  arg_name = g_strdup (handler_names[1]);
  min_val = strtod (handler_names[2], &endp);
  if (*endp != '\0')
    {
      g_warning ("input_patch: error parsing double");
      return TRUE;
    }
  max_val = strtod (handler_names[3], &endp);
  if (*endp != '\0')
    {
      g_warning ("input_patch: error parsing double");
      return TRUE;
    }
  desc = g_strdup (handler_names[4]);

  input_patch = gdam_input_patch_create_associated (object);
  g_return_val_if_fail (input_patch != NULL, TRUE);
  input_patch->model = skin->model;
  input_patch->arg_name = arg_name;
  input_patch->min_value = min_val;
  input_patch->max_value = max_val;
  input_patch->do_connect = gdam_input_patch_range_connect;
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC
			     (gdam_input_patch_connect_default),
			     (GtkObject *) input_patch);
  return TRUE;
}
static gboolean gdam_input_patch_default_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamInputPatch *input_patch;
  if (num_handler_names != 1)
    {
      g_warning ("usage: gdam_input_patch_default");
    }

  input_patch = gdam_input_patch_force_associated (object);
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC
			     (gdam_input_patch_connect_default),
			     (GtkObject *) input_patch);
  gdam_debug ("gdam_input_patch_clear handled.");
  return TRUE;
}


static GdamSkinGenericConnect connect_infos[] =
{ 	{ "gdam_input_selector_open",
	   gdam_input_selector_open_handler },
  	{ "gdam_input_selector_refresh_skins",
	   gdam_input_selector_refresh_skins_handler },
	{ "gdam_input_selector_entry",
	   gdam_input_selector_entry_handler },
	{ "gdam_input_selector_skin_tree",
	   gdam_input_selector_skin_tree_handler },
	{ "gdam_input_selector_clist",
	   gdam_input_selector_clist_handler }, };

static GdamSkinGenericConnect skin_connect_info[] =
{	{ "gdam_input_patch_range",
           gdam_input_patch_range_handler },
 	{ "gdam_input_patch_default",
           gdam_input_patch_default_handler },
 	{ "gdam_input_grope_stop",
           gdam_input_grope_stop_handler },
 	{ "gdam_input_grope_start",
           gdam_input_grope_start_handler },
 	{ "gdam_input_patch_clear",
           gdam_input_patch_clear_handler } };

static void register_skin_connect_infos()
{
  GdamSkinClass *skin_class;
  skin_class = GDAM_SKIN_CLASS (gtk_type_class (GDAM_TYPE_SKIN));
  gdam_skin_class_register_generic_handlers
    (skin_class, G_N_ELEMENTS (skin_connect_info), skin_connect_info);
}

static void gdam_input_selector_class_init (GdamSkinClass* skin_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (skin_class);
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  register_skin_connect_infos ();
  object_class->destroy = gdam_input_selector_destroy;
}

#if 0
static gboolean grope_input_event_handler(GtkObject*       object,
                                         guint            signal_id,
				         guint            n_params,
				         GtkArg*          params,
				         gpointer         skin_as_data)
{
	GdamInputTrap* trap;
	GdamInputEvent* event;
	event = (GdamInputEvent*) GTK_VALUE_POINTER(params[0]);
	trap = gdam_input_trap_find(event);
	g_return_val_if_fail(trap != NULL, TRUE);

	return TRUE;
}
#endif
