/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamskinbindingfile.h"
#include "gdampacker.h"
#include <component/gdaminputdevice.h>
#include <component/gdambindingfile.h>
#include <component/gdammodel.h>
#include <stdlib.h>
#include <string.h>

/*                     The binding interface documentation :)

<Gdam-Binding-Interface>
  <constant type="GdamGain">
    <argument>muted</argument>
    <value>1</value>
    <GdamMidiEventSysex>
      <sysex>0x01 0x02 0x03 0x04</sysex>
      <device>/dev/midi</device>
    </GdamMidiEventSysex>
  </constant>
  <range type="GdamGain">
    <argument>volume</argument>
    <GdamMidiEventPitchWheel>
      <input_id>3</input_id>
      <device>/dev/midi</device>
    </GdamMidiEventPitchWheel>
    <min_value>0.0</min_value>
    <max_value>2.0</max_value>
  </range>
</Gdam-Binding-Interface>

*/


static const char* get_string_content(xmlNode* node)
{
  if (node->content)
    return node->content;
  if (node->childs == NULL)
    return NULL;
  return node->childs->content;
}

static gboolean parse_signal_xml_node (xmlNode*       const_node,
                                       GdamSkin*      skin)
{
  xmlNode *node;
  xmlNode *children = const_node->childs;
  const char *signal_name = NULL;
  char *handlers[2];
  xmlNode *value_node = NULL;
  guint input_hook;
  const char *property;
  handlers[0] = NULL;
  handlers[1] = NULL;
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "signal") == 0)
	{
	  signal_name = get_string_content (node);
	  continue;
	}
      if (strcmp (node->name, "handler") == 0)
	{
	  handlers[0] = g_strdup (get_string_content (node));
	  continue;
	}
      g_warning ("couldn't parse input event from "
		 "node named `%s'", node->name);
      return FALSE;
    }
  if (signal_name == NULL)
    {
      g_warning ("<signal> was missing <signal>");
      return FALSE;
    }
  if (handlers[0] == NULL)
    {
      g_warning ("<signal> was missing <handler>");
      return FALSE;
    }
  if (!gdam_skin_do_connect (skin, GTK_OBJECT (skin), signal_name, 1,
			     handlers))
    {
      g_warning ("error connecting signal=%s handler=%s", signal_name,
		 handlers[0]);
    }
  return TRUE;
}


static gboolean parse_arg(GdamArg*    to_parse,
                          GtkObject*  object,
                          const char* arg_name,
			  xmlNode*    value_node)
{
  int type;
  const char *content;

  type = gdam_arg_query_type (object->klass->type, arg_name);
  if (type == 0)
    {
      g_warning ("no argument %s was found in %s",
		 arg_name, gtk_type_name (object->klass->type));
      return FALSE;
    }
  to_parse->name = g_strdup (arg_name);
  to_parse->type = type;
  if (value_node == NULL)
    {
      GDAM_ARG_DOUBLE (to_parse) = 0.0;
      return TRUE;
    }
  content = get_string_content (value_node);
  if (content == NULL)
    {
      g_warning ("<value> for <%s> had no string content", arg_name);
      return FALSE;
    }
  return gdam_arg_parse_val_from_string (to_parse, content) > 0;
}


static GdamSkin* type_property_to_skin(GdamSkin* skin, const char* property)
{
  GdamModel *model;
  GtkType allow_type;		/* type this tag allows */
  GtkType model_type;
  GtkType skin_type;
  const char *c;

  if (property == NULL)
    return NULL;
  allow_type = gtk_type_from_name (property);
  skin_type = GTK_OBJECT (skin)->klass->type;
  if (gtk_type_is_a (skin_type, allow_type))
    {
      return skin;
    }
  return NULL;
}



	  /* We apply tags only to skins/models whose types match. */
static GdamModel* type_property_to_model(GdamSkin* skin, const char* property)
{
  GdamModel *model;
  GtkType allow_type;		/* type this tag allows */
  GtkType model_type;
  GtkType skin_type;
  const char *c;

  if (property == NULL)
    return NULL;

  model = skin->model;
  skin_type = GTK_OBJECT (skin)->klass->type;
  c = strchr (property, ':');
  /*  If property is colonated list, parse list and handle. */
  if (c != NULL)
    {
      char **tokens = g_strsplit (property, ":", 0);
      int i;
      allow_type = gtk_type_from_name (tokens[0]);
      model = NULL;
      if (tokens[0] && tokens[1]
	  && strcmp (tokens[0], "GdamPacker") == 0
	  && gtk_type_is_a (skin_type, allow_type))
	{
	  model = gdam_packer_get_model (GDAM_PACKER (skin), tokens[1]);
	  if (!model)
	    g_warning ("packer had no %s", property);
	}
      else
	{
	  g_warning ("binding parse error: %s", property);
	}
      for (i = 0; tokens[i] != NULL; i++)
	{
	  g_free (tokens[i]);
	}
      g_free (tokens);
      return model;
    }
  /*  Else 'type' property must match type of model.  */
  else
    {
      if (model == NULL)
	return NULL;
      model_type = GTK_OBJECT (model)->klass->type;
      allow_type = gtk_type_from_name (property);
      if (allow_type == 0)
	return NULL;
      if (!gtk_type_is_a (model_type, allow_type))
	{
	  gdam_debug ("binding: skipping `%s': not a `%s'",
		      gtk_type_name (model_type), property);
	  return NULL;
	}
      return model;
    }
}


gboolean gdam_skin_binding_file_process      (xmlNode*       node_list, 
                                              GdamSkin*      skin)
{
  xmlNode *node;
  int processed = 0;

  for (node = node_list; node != NULL; node = node->next)
    {
      const char *name;
      const char *property;
      GdamModel *model;
      GdamSkin *newskin;
      newskin = NULL;

      name = node->name;

      property = xmlGetProp (node, "type");
      model = type_property_to_model (skin, property);
      if (model == NULL)
	{
	  newskin = type_property_to_skin (skin, property);
	}

      if (strcmp (name, "constant") == 0 && model)
	{
	  if (!gdam_binding_parse_constant_xml_node (node, model))
	    ;
	  else
	    processed++;
	}
      else if (strcmp (name, "range") == 0 && model)
	{
	  if (!gdam_binding_parse_range_xml_node (node, model))
	    ;
	  else
	    processed++;
	}
      else if (strcmp (name, "signal") == 0 && newskin)
	{
	  if (!parse_signal_xml_node (node, newskin))
	    ;
	  else
	    processed++;
	}
      else
	{
	  g_warning ("bad binding file node!!");
	  return FALSE;
	}
    }
  if (processed > 0)
    {
      gdam_debug ("Got %d bindings", processed);
    }
  return TRUE;
}

