#include "gdamfirfilterskin.h"
#include <skin/gdamskinutil.h>
#include <gtk/gtkdrawingarea.h>
#include <string.h>
#include <component/plugins/gdamfirfilter.h>
#include <common/gdamsample.h>
#include <gtk/gtksignal.h>
#include "../gdamskinutil.h"
#include <common/gdamplugin.h>

static GtkObjectClass* parent_class = NULL;

static void gdam_fir_filter_skin_init(GdamFirFilterSkin* fir_filter_skin)
{
  fir_filter_skin->minimum_gain = 0.0;
  fir_filter_skin->maximum_gain = 1.0;
  fir_filter_skin->which_dragging = -1;
  fir_filter_skin->circle_radius = 4;
  fir_filter_skin->area_length_in_grid = 32;
}
static void gdam_fir_filter_skin_class_init(GdamSkinClass* skin_class);

GtkType gdam_fir_filter_skin_get_type()
{
  static GtkType fir_filter_skin_type = 0;
  if (!fir_filter_skin_type)
    {
      static const GtkTypeInfo fir_filter_skin_info = {
	"GdamFirFilterSkin",
	sizeof (GdamFirFilterSkin),
	sizeof (GdamFirFilterSkinClass),
	(GtkClassInitFunc) gdam_fir_filter_skin_class_init,
	(GtkObjectInitFunc) gdam_fir_filter_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      parent_class = gtk_type_class (parent);
      fir_filter_skin_type = gtk_type_unique (parent, &fir_filter_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return fir_filter_skin_type;
}

static void alloc_colors(GdamFirFilterSkin* skin)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
  GdkGC *gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (gc, util_alloc_color_cached (65535, 0, 0));
  skin->circle_gc = gc;
  skin->background = skin->drawing_area->style->bg_gc[state];
  skin->line_gc = skin->drawing_area->style->fg_gc[state];
}


static void copy_backing_area(GdamFirFilterSkin* skin, const GdkRectangle* area)
{
  gdk_draw_pixmap (skin->drawing_area->window,
		   skin->drawing_area->
		   style->fg_gc[GTK_WIDGET_STATE (skin->drawing_area)],
		   skin->backing_pixmap, area->x, area->y, area->x, area->y,
		   area->width, area->height);
}

static GdamFirFilter* get_fir_filter(GdamFirFilterSkin* skin)
{
  return GDAM_FIR_FILTER (GDAM_SKIN (skin)->model);
}

static void get_echo_position(GdamFirFilterSkin* skin, int index, int* x, int* y)
{
  gdouble gain_frac;
  GdamFirFilterElement *element;
  gdouble line_index;
  gdouble pix_per_grid;
  GdamFirFilter *fir_filter = get_fir_filter (skin);
  g_return_if_fail (0 <= index && index < fir_filter->num_elements);
  element = fir_filter->elements + index;
  line_index = (gdouble) element->offset;
  pix_per_grid = skin->draw_width / skin->area_length_in_grid;
  gain_frac = (element->volume - skin->minimum_gain)
    / (skin->maximum_gain - skin->minimum_gain);
  *x = (int) (line_index * pix_per_grid);
  *y = (1.0 - gain_frac) * skin->draw_height;
}

static void get_echo_info_from_position(GdamFirFilterSkin* skin, 
                                        int            press_x,
					int            press_y,
					gdouble*       delay,
					gdouble*       gain)
{
  gdouble tmp;
  gdouble pix_per_grid = skin->draw_width / skin->area_length_in_grid;

  tmp = press_x;
  tmp /= pix_per_grid;
  *delay = tmp;

  tmp = skin->draw_height - 1 - press_y;
  tmp /= skin->draw_height;
  tmp *= (skin->maximum_gain - skin->minimum_gain);
  tmp += skin->minimum_gain;
  *gain = tmp;
}

static void redraw_area(GdamFirFilterSkin* skin, const GdkRectangle* area)
{
  int num_points, i;
  int circle_radius = skin->circle_radius;
  gdouble pix_per_grid = skin->draw_width / skin->area_length_in_grid;
  GdkRectangle rect;
  GdamFirFilter *fir_filter = get_fir_filter (skin);
  if (area == NULL)
    {
      rect.width = skin->drawing_area->allocation.width;
      rect.height = skin->drawing_area->allocation.height;
      rect.x = rect.y = 0;
      area = &rect;
    }
  if (!skin->circle_gc)
    alloc_colors (skin);

  gdk_draw_rectangle (skin->backing_pixmap,
		      skin->background,
		      TRUE, area->x, area->y, area->width, area->height);
  num_points = fir_filter->num_elements;

  /* Draw the lines. */
  for (i = 0; i < skin->area_length_in_grid; i++)
    {
      int x = (gdouble) i * pix_per_grid;
      if (area->x > x || area->x + area->width <= x)
	continue;
      gdk_draw_line (skin->backing_pixmap,
		     skin->line_gc, x, 0, x, skin->draw_height);
    }

  /* Draw the echoes. */
  for (i = 0; i < num_points; i++)
    {
      int x, y;
      get_echo_position (skin, i, &x, &y);
      gdk_draw_arc (skin->backing_pixmap,
		    skin->circle_gc,
		    TRUE,
		    x - circle_radius,
		    y - circle_radius,
		    circle_radius * 2, circle_radius * 2, 0, 360 * 64);
    }

  copy_backing_area (skin, area);
}

static int find_keypoint(GdamFirFilterSkin* skin,
                         int            screen_x,
			 int            screen_y)
{
  /* search through elements for one matching...  */
  int radius = skin->circle_radius;
  int skin_radius_squared = radius * radius;
  int i;
  GdamFirFilter *fir_filter = get_fir_filter (skin);
  for (i = 0; i < fir_filter->num_elements; i++)
    {
      int x, y;
      int radius_squared;
      get_echo_position (skin, i, &x, &y);
      x -= screen_x;
      y -= screen_y;
      radius_squared = x * x + y * y;
      if (radius_squared <= skin_radius_squared)
	{
	  return i;
	}
    }
  return -1;
}

static void update_echo_array(GdamFirFilterSkin*     skin,
                              gboolean           update_server)
{
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE
				     (GDAM_SKIN (skin)->model));
}

static void delete_keypoint(GdamFirFilterSkin* skin,
                            int            index)
{
  int x, y;
  int radius = skin->circle_radius;
  GdkRectangle rectangle;
  GdamFirFilter *fir_filter = get_fir_filter (skin);
  g_return_if_fail (0 <= index && index < fir_filter->num_elements);
  get_echo_position (skin, index, &x, &y);
  memmove (fir_filter->elements + index,
	   fir_filter->elements + index + 1,
	   sizeof (GdamFirFilterElement)
	   * (fir_filter->num_elements - index - 1));
  fir_filter->num_elements--;

  rectangle.x = x - radius;
  rectangle.y = y - radius;
  rectangle.width = rectangle.height = 2 * radius + 1;
  redraw_area (skin, &rectangle);
  update_echo_array (skin, TRUE);
}


static void add_fir_filter_point(GdamFirFilterSkin* skin,
                                 gdouble        delay,
			         gdouble        gain,
				 gboolean       update_server)
/*TODO : update_server is ignored ... fix that... */
{
  GdamFirFilter *fir_filter = get_fir_filter (skin);
  int num_echoes = fir_filter->num_elements;
  int x, y, radius;
  GdkRectangle rectangle;

  gdam_fir_filter_add (fir_filter, gain, delay);
  get_echo_position (skin, num_echoes, &x, &y);
  radius = skin->circle_radius;
  rectangle.x = x - radius;
  rectangle.y = y - radius;
  rectangle.width = rectangle.height = 2 * radius + 1;
  redraw_area (skin, &rectangle);
}

static void change_fir_filter_keypoint(GdamFirFilterSkin*   skin, 
				  int              index,
				  gdouble          offset,
				  gdouble          volume,
				  gboolean         redraw,
				  gboolean         update_server)
{
  int old_x, old_y, new_x, new_y;
  int radius;
  GdamFirFilter *fir_filter = get_fir_filter (skin);
  g_return_if_fail (GDAM_FIR_FILTER_SKIN (skin) != NULL);
  g_return_if_fail (0 <= index && index < fir_filter->num_elements);
  get_echo_position (skin, index, &old_x, &old_y);
  radius = skin->circle_radius;
  fir_filter->elements[index].volume = volume;
  fir_filter->elements[index].offset = offset;
  update_echo_array (skin, update_server);
  get_echo_position (skin, index, &new_x, &new_y);
  if (redraw)
    {
      GdkRectangle rectangle;
      rectangle.width = rectangle.height = 2 * radius + 1;

      rectangle.x = old_x - radius;
      rectangle.y = old_y - radius;
      redraw_area (skin, &rectangle);

      rectangle.x = new_x - radius;
      rectangle.y = new_y - radius;
      redraw_area (skin, &rectangle);
    }
}

static gint
fir_filter_expose(GtkDrawingArea* area,
             GdkEventExpose* expose_event,
             GdamFirFilterSkin*  skin)
{
  redraw_area (skin, &expose_event->area);
  return 0;
}

static void queue_redraw(GdamFirFilterSkin* skin)
{
  GtkWidget *widget = skin->drawing_area;
  if (widget == NULL)
    return;
  gtk_widget_queue_draw (widget);
}

static gint
fir_filter_button_press(GtkDrawingArea* area,
                   GdkEventButton* button_event,
		   GdamFirFilterSkin* skin)
{
  int press_x = button_event->x;
  int press_y = button_event->y;
  int width, height;
  int echo_index;
  gdouble gain;
  gdouble delay;
  if (skin->which_dragging > 0)
    {
      if (button_event->button == 3)
	{
	  delete_keypoint (skin, skin->which_dragging);
	  skin->which_dragging = -1;
	}
      return 0;
    }
  g_return_val_if_fail (skin->which_dragging < 0, 0);
  if ((echo_index = find_keypoint (skin, press_x, press_y)) >= 0)
    {
      int pos_x, pos_y;
      if (button_event->button == 3)
	{
	  delete_keypoint (skin, echo_index);
	  return 0;
	}
      get_echo_position (skin, echo_index, &pos_x, &pos_y);
      skin->drag_click_offset_x = press_x - pos_x;
      skin->drag_click_offset_y = press_y - pos_y;
      skin->which_dragging = echo_index;
      return 0;
    }

  if (button_event->button == 3)
    return 0;

  width = skin->draw_width;
  height = skin->draw_height;
  get_echo_info_from_position (skin, press_x, press_y, &delay, &gain);
  gdam_debug ("add_fir_filter_point: time=%4.4g, gain=%4.4g", delay, gain);
  add_fir_filter_point (skin, delay, gain, TRUE);
  return 0;
}

static gint
fir_filter_button_release(GtkDrawingArea* area,
                     GdkEventButton* button_event,
		             GdamFirFilterSkin* skin)
{
  int release_x;
  int release_y;
  gdouble time;
  gdouble volume;
  int dragging = skin->which_dragging;
  if (dragging < 0)
    return 0;
  skin->which_dragging = -1;
  release_x = button_event->x - skin->drag_click_offset_x;
  release_y = button_event->y - skin->drag_click_offset_y;
  get_echo_info_from_position (skin, release_x, release_y, &time, &volume);
  change_fir_filter_keypoint (skin, dragging, time, volume, TRUE, TRUE);
  queue_redraw (skin);
  return 0;
}

static gint
fir_filter_button_motion(GtkDrawingArea* area,
                     GdkEventMotion* motion_event,
		     GdamFirFilterSkin* skin)
{
  int x, y;
  int release_x, release_y;
  gdouble time;
  gdouble volume;
  guint state;
  int dragging = skin->which_dragging;
  if (dragging < 0)
    return 0;

  if (motion_event->is_hint)
    gdk_window_get_pointer (motion_event->window, &x, &y, &state);
  else
    {
      x = motion_event->x;
      y = motion_event->y;
      state = motion_event->state;
    }

  x = motion_event->x;
  y = motion_event->y;

  release_x = x - skin->drag_click_offset_x;
  release_y = y - skin->drag_click_offset_y;
  get_echo_info_from_position (skin, release_x, release_y, &time, &volume);
  change_fir_filter_keypoint (skin, dragging, time, volume, TRUE, FALSE);
  return 0;
}

static gint
fir_filter_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        GdamFirFilterSkin* skin)
{
  if (skin->backing_pixmap)
    gdk_pixmap_unref (skin->backing_pixmap);
  skin->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  return TRUE;
}

static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamFirFilterSkin* skin)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 800, 300);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_press_event",
		      GTK_SIGNAL_FUNC (fir_filter_button_press), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_release_event",
		      GTK_SIGNAL_FUNC (fir_filter_button_release), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (fir_filter_button_motion), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event",
		      GTK_SIGNAL_FUNC (fir_filter_expose), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (fir_filter_configure), skin);
  gtk_widget_show (GTK_WIDGET (drawing_area));
  gtk_widget_add_events (GTK_WIDGET (drawing_area),
			 GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);
}


/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */


static gboolean gdam_fir_filter_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
      return TRUE;
    }
  connect_drawing_area (GTK_DRAWING_AREA (object),
			GDAM_FIR_FILTER_SKIN (skin));
  return TRUE;
}

static void on_line_width(GtkAdjustment* adj, GdamFirFilterSkin* skin)
{
  skin->area_length_in_grid = adj->value;
  queue_redraw (skin);
  /* TODO: scale existing points. */
}

static gboolean gdam_fir_filter_line_width_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj), signal_name, on_line_width, skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_fir_filter_drawing_area",
	   gdam_fir_filter_drawing_area_handler },
	{ "gdam_fir_filter_line_width_adjustment",
	   gdam_fir_filter_line_width_adjustment_handler },
};

static void gdam_fir_filter_skin_class_init(GdamSkinClass* skin_class)
{
  skin_class->model_type = GDAM_TYPE_FIR_FILTER;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}
