/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamladspaskin.h"
#include "../gdamskinutil.h"
#include "../gdambuilderskin.h"
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtklabel.h>
#include <gtk/gtksignal.h>
#include <component/plugins/gdamladspafilter.h>
#include <math.h>

static void gdam_ladspa_skin_init(GdamLadspaSkin* ladspa_skin)
{
}

static gboolean gdam_ladspa_build_new_model(GdamSkin*      skin,
					    GdamChannel*   channel,
					    int            num_options,
	                                    char**         options)
{
  if (num_options == 0)
    {
      g_warning ("LADSPA plugin really *needs* argument " "for plugintype");
      return FALSE;
    }
  skin->model = GDAM_MODEL (gdam_ladspa_filter_new (options[0]));
  if (skin->model == NULL)
    {
      g_warning ("couldn't load LADSPA plugin");
      return FALSE;
    }
  skin->model->channel = channel;
  return TRUE;
}

typedef struct _DoAdjustmentInfo DoAdjustmentInfo;
struct _DoAdjustmentInfo {
	GdamLadspaFilter*	filter;
	int			param_index;

	unsigned		is_integer : 1,
				is_logarithmic : 1,
				is_sample_rate_ratio : 1;
};

static void
handle_toggle_event(GtkToggleButton* toggle, DoAdjustmentInfo* info)
{
  gdouble value;
  if (gtk_toggle_button_get_active (toggle))
    value = +1.0;
  else
    value = -1.0;
  gdam_ladspa_filter_set_control (info->filter,
				  info->param_index, value, value);
}

static void
handle_adj_event(GtkAdjustment* adjustment, DoAdjustmentInfo* info)
{
  gdouble value;
  value = adjustment->value;
  if (info->is_logarithmic)
    value = exp (value);
  if (info->is_sample_rate_ratio)
    value *= 44100.0;
  if (info->is_integer)
    value = (int) value;

  gdam_ladspa_filter_set_control (info->filter,
				  info->param_index, value, value);
}

static void gdam_ladspa_post_connection_handler(GdamSkin*       skin)
{
  GdamLadspaSkin *ladspa_skin;
  GdamLadspaPluginType *plugin_type;
  GdamLadspaFilter *ladspa_filter;
  int i;

  ladspa_skin = GDAM_LADSPA_SKIN (skin);
  g_return_if_fail (ladspa_skin->interface_box != NULL);

  ladspa_filter = GDAM_LADSPA_FILTER (skin->model);
  plugin_type = ladspa_filter->plugin_type;
  g_return_if_fail (plugin_type != NULL);

  for (i = 0; i < plugin_type->num_controls; i++)
    {
      GdamLadspaParameterType *parameter = plugin_type->controls + i;
      GtkWidget *control_widget = NULL;
      DoAdjustmentInfo *info;

      info = g_new (DoAdjustmentInfo, 1);
      info->filter = ladspa_filter;
      info->param_index = i;
      info->is_integer = parameter->is_integer;
      info->is_logarithmic = parameter->is_logarithmic;
      info->is_sample_rate_ratio = parameter->is_sample_rate_ratio;

      if (parameter->is_toggle)
	{
	  control_widget = gtk_toggle_button_new ();
	  gtk_signal_connect_full
	    (GTK_OBJECT (control_widget),
	     "toggled",
	     GTK_SIGNAL_FUNC (handle_toggle_event),
	     NULL, info, g_free, FALSE, FALSE);
	}
      else if (parameter->has_min_value && parameter->has_max_value)
	{
	  GtkAdjustment *adjustment;
	  gdouble min_value = parameter->min_value;
	  gdouble max_value = parameter->max_value;
	  if (parameter->is_logarithmic)
	    {
	      if (min_value == 0)
		{
		  min_value = -30.0;
		}
	      else if (min_value > 0)
		min_value = log (min_value);
	      else
		g_warning ("minimum was " "negative but logarithmic");
	      if (min_value == 0)
		{
		  min_value = -30.0;
		}
	      else if (max_value > 0)
		max_value = log (max_value);
	      else
		g_warning ("maximum was " "negative but logarithmic");
	    }
	  adjustment = (GtkAdjustment *)
	    gtk_adjustment_new (min_value,
				min_value, max_value, 0.01, 0.01, 0.01);
	  control_widget = gtk_hscrollbar_new (adjustment);
	  gtk_signal_connect_full
	    (GTK_OBJECT (adjustment),
	     "value_changed",
	     GTK_SIGNAL_FUNC (handle_adj_event),
	     NULL, info, g_free, FALSE, FALSE);
	}
      else
	{
	  GtkAdjustment *adjustment;
	  g_warning ("using range 0 .. 2 since the plugin "
		     "gives no hint...");
	  if (info->is_logarithmic)
	    {
	      g_warning ("   also, cannot support logarithmic"
			 " with no range hints...");
	      info->is_logarithmic = FALSE;
	    }
	  adjustment = (GtkAdjustment *)
	    gtk_adjustment_new (0.0, 0.0, 2.0, 0.01, 0.01, 0.01);
	  control_widget = gtk_hscrollbar_new (adjustment);
	  gtk_signal_connect_full
	    (GTK_OBJECT (adjustment),
	     "value_changed",
	     GTK_SIGNAL_FUNC (handle_adj_event),
	     NULL, info, g_free, FALSE, FALSE);
	}
      {
	GtkWidget *label;
	label = gtk_label_new (parameter->name);
	gtk_widget_show (label);
	gtk_box_pack_start_defaults (ladspa_skin->interface_box, label);
      }
      gtk_widget_show (control_widget);
      gtk_box_pack_start_defaults (ladspa_skin->interface_box,
				   control_widget);
    }
}

static gboolean gdam_ladspa_interface_box_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  g_return_val_if_fail (GTK_IS_BOX (object), TRUE);
  GDAM_LADSPA_SKIN (skin)->interface_box = GTK_BOX (object);
  return TRUE;
}

static gboolean gdam_ladspa_plugin_name_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamLadspaFilter *filter;
  g_return_val_if_fail (GTK_IS_LABEL (object), TRUE);
  filter = GDAM_LADSPA_FILTER (skin->model);
  gtk_label_set_text (GTK_LABEL (object), filter->plugin_type->database_name);
  //GDAM_LADSPA_SKIN(skin)->name_label = GTK_LABEL(object);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_ladspa_interface_box",
	   gdam_ladspa_interface_box_handler },
	{ "gdam_ladspa_plugin_name",
	   gdam_ladspa_plugin_name_handler },
};

static void gdam_ladspa_skin_class_init(GdamSkinClass* skin_class)
{
  char **names;
  skin_class->post_connecting_handler = gdam_ladspa_post_connection_handler;
  skin_class->build_new_model = gdam_ladspa_build_new_model;

  (void) GDAM_TYPE_LADSPA_FILTER;
  (void) GDAM_TYPE_BUILDER_SKIN;

  /* Iterate the existing filters and add them
   * to the builder skin. */
  names = gdam_ladspa_plugin_type_get_names ();
  if (names != NULL)
    {
      while (*names)
	{
	  const char *name = *names++;
	  const char *options[2];
	  const char *label_name = name;	/* TODO: change this? */
	  options[0] = name;
	  options[1] = NULL;
	  gdam_builder_skin_class_add_extra
	    ("GdamLadspaSkin", "ladspa_skin.xml", label_name, options);
	}
    }

  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}

GtkType gdam_ladspa_skin_get_type()
{
  static GtkType ladspa_skin_type = 0;
  if (!ladspa_skin_type)
    {
      static const GtkTypeInfo ladspa_skin_info = {
	"GdamLadspaSkin",
	sizeof (GdamLadspaSkin),
	sizeof (GdamLadspaSkinClass),
	(GtkClassInitFunc) gdam_ladspa_skin_class_init,
	(GtkObjectInitFunc) gdam_ladspa_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      ladspa_skin_type = gtk_type_unique (parent, &ladspa_skin_info);
    }
  return ladspa_skin_type;
}
