/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <component/gdamutil.h>
#include <glade/glade.h>
//#include "../gdamfactory.h"
#include "../gdamskin.h"
#include "../gdamskinutil.h"
#include "../../component/plugins/gdamspatialstereo.h"
#include "gdamspatialstereoskin.h"
#include <gtk/gtk.h>
#include <common/gdambuffer.h>
#include <client/gdamchannel.h>
#include <common/gdamplugin.h>


#define DRAGGING_NONE GDAM_SPATIAL_STEREO_DRAGGING_NONE
#define DRAGGING_CENTER GDAM_SPATIAL_STEREO_DRAGGING_CENTER
#define DRAGGING_LEFT GDAM_SPATIAL_STEREO_DRAGGING_LEFT
#define DRAGGING_RIGHT GDAM_SPATIAL_STEREO_DRAGGING_RIGHT
#define DRAGGING_HEAD_CENTER GDAM_SPATIAL_STEREO_DRAGGING_HEAD_CENTER
#define DRAGGING_HEAD_LEFT GDAM_SPATIAL_STEREO_DRAGGING_HEAD_LEFT
#define DRAGGING_HEAD_RIGHT GDAM_SPATIAL_STEREO_DRAGGING_HEAD_RIGHT


static gboolean gdam_spatial_stereo_skin_build_new_model
                                                 (GdamSkin*    skin,
                                                  GdamChannel* channel,
						  int 	       num_options,
					          char**        options);

static void reset_spatial_parameters(GdamSpatialStereoSkin* skin)
{
  GdamSkin *this_skin = GDAM_SKIN (skin);
  GdamSpatialStereo *spatial = GDAM_SPATIAL_STEREO (this_skin->model);
  gdam_spatial_stereo_positions_change (spatial);
}

static void world_to_screen_position(GdamSpatialStereoSkin* skin,
                                     const gdouble*     world,
				     int*               x_out,
				     int*               y_out);

static void gdam_spatial_stereo_skin_init(GdamSpatialStereoSkin* skin)
{
  skin->head_center_radius = 3;
  skin->head_side_radius = 3;
  skin->output_center_radius = 7;
  skin->output_side_radius = 10;
  skin->drag_mode = DRAGGING_NONE;

  skin->world_left = -10.0;
  skin->world_top = -10.0;
  skin->world_width = skin->world_height = 20.0;
}
static void gdam_spatial_stereo_skin_class_init(GdamSkinClass* skin_class);
GtkType gdam_spatial_stereo_skin_get_type()
{
  static GtkType spatial_stereo_skin_type = 0;
  if (!spatial_stereo_skin_type)
    {
      static const GtkTypeInfo spatial_stereo_skin_info = {
	"GdamSpatialStereoSkin",
	sizeof (GdamSpatialStereoSkin),
	sizeof (GdamSpatialStereoSkinClass),
	(GtkClassInitFunc) gdam_spatial_stereo_skin_class_init,
	(GtkObjectInitFunc) gdam_spatial_stereo_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      spatial_stereo_skin_type = gtk_type_unique (parent,
						  &spatial_stereo_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return spatial_stereo_skin_type;
}


/*=====================================================================
 * Functions to render the draw area. 
 */
static void render_foreground(GdamSpatialStereoSkin* skin,
                              const double* left_pos,
                              const double* right_pos,
			      int           little_radius,
			      int           big_radius)
{
  GdkPixmap *pixmap = skin->backing_pixmap;
  int x, y;
  gdouble center_pos[3];

  world_to_screen_position (skin, left_pos, &x, &y);
  gdam_debug ("draw_arc: %d,%d (radius %d)", x, y, little_radius);
  gdk_draw_arc (pixmap,
		skin->circle_gc,
		TRUE,
		x - little_radius + 1,
		y - little_radius + 1,
		little_radius * 2 - 1, little_radius * 2 - 1, 0, 360 * 64);

  world_to_screen_position (skin, right_pos, &x, &y);
  gdam_debug ("draw_arc: %d,%d (radius %d)", x, y, little_radius);
  gdk_draw_arc (pixmap,
		skin->circle_gc,
		TRUE,
		x - little_radius + 1,
		y - little_radius + 1,
		little_radius * 2 - 1, little_radius * 2 - 1, 0, 360 * 64);

  center_pos[0] = (left_pos[0] + right_pos[0]) * 0.5;
  center_pos[1] = (left_pos[1] + right_pos[1]) * 0.5;
  center_pos[2] = (left_pos[2] + right_pos[2]) * 0.5;
  world_to_screen_position (skin, center_pos, &x, &y);
  gdam_debug ("draw_arc: %d,%d (radius %d)", x, y, big_radius);
  gdk_draw_arc (pixmap,
		skin->circle_gc,
		TRUE,
		x - big_radius + 1,
		y - big_radius + 1,
		big_radius * 2 - 1, big_radius * 2 - 1, 0, 360 * 64);
}
static gint
configure_event(GtkWidget* widget,
                GdkEventConfigure* configure_event,
		GdamSpatialStereoSkin* skin)
{
  if (skin->backing_pixmap)
    gdk_pixmap_unref (skin->backing_pixmap);
  skin->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  return TRUE;
}

static void alloc_colors(GdamSpatialStereoSkin* skin)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
	/** HACK!!! */
  skin->circle_gc = skin->drawing_area->style->bg_gc[state];
  skin->background = skin->drawing_area->style->fg_gc[state];
}

static void copy_backing_area(GdamSpatialStereoSkin* skin, int x, int y, int w, int h)
{
  gdk_draw_pixmap (skin->drawing_area->window,
		   skin->drawing_area->
		   style->fg_gc[GTK_WIDGET_STATE (skin->drawing_area)],
		   skin->backing_pixmap, x, y, x, y, w, h);
}

static void redraw_area(GdamSpatialStereoSkin* skin, int x, int y, int w, int h)
{
  GdamSpatialStereo *spatial;
  GdamSkin *this_skin = GDAM_SKIN (skin);
  if (!skin->circle_gc)
    alloc_colors (skin);
  gdam_debug ("redraw_area: %d,%d x %d,%d", x, y, w, h);
  gdk_draw_rectangle (skin->backing_pixmap,
		      skin->background, TRUE, x, y, w, h);
  spatial = GDAM_SPATIAL_STEREO (this_skin->model);
  render_foreground (skin,
		     spatial->left_channel_pos,
		     spatial->right_channel_pos,
		     skin->output_side_radius, skin->output_center_radius);
  render_foreground (skin,
		     spatial->left_ear_pos,
		     spatial->right_ear_pos,
		     skin->head_center_radius, skin->head_side_radius);
  copy_backing_area (skin, x, y, w, h);
}

static gint
expose_event(GtkDrawingArea* area,
             GdkEventExpose* expose_event,
             GdamSpatialStereoSkin* skin)
{
  redraw_area (skin,
	       expose_event->area.x, expose_event->area.y,
	       expose_event->area.width, expose_event->area.height);
  return 0;
}

static int world_to_screen_x(GdamSpatialStereoSkin* skin, gdouble x)
{
  return (x - skin->world_left) * skin->draw_width / skin->world_width;
}

static int world_to_screen_y(GdamSpatialStereoSkin* skin, gdouble y)
{
  return (y - skin->world_top) * skin->draw_height / skin->world_height;
}
static void world_to_screen_position(GdamSpatialStereoSkin* skin,
                                     const gdouble*     world,
				     int*               x_out,
				     int*               y_out)
{
  *x_out = world_to_screen_x (skin, world[0]);
  *y_out = world_to_screen_y (skin, world[1]);
}

static void screen_to_world(GdamSpatialStereoSkin* skin,
                            int x, int y,
			    gdouble* position_out)
{
  gdouble world_x, world_y;

  world_x = x;
  world_x /= skin->draw_width;
  world_x *= skin->world_width;
  world_x += skin->world_left;
  position_out[0] = world_x;

  world_y = y;
  world_y /= skin->draw_height;
  world_y *= skin->world_height;
  world_y += skin->world_top;
  position_out[1] = world_y;

  position_out[2] = 0.0;
}

static inline void circle_area(int cx, int cy, int rad, GdkRectangle* out)
{
  out->x = cx - rad;
  out->y = cy - rad;
  out->width = out->height = rad * 2;
}
static void get_rendered_area(GdamSpatialStereoSkin* skin,
                              const double* left_pos,
                              const double* right_pos,
			      int           little_radius,
			      int           big_radius,
			      GdkRectangle* out)
{
  GdkRectangle left, center, right, tmp;
  int left_x, left_y, right_x, right_y;
  world_to_screen_position (skin, left_pos, &left_x, &left_y);
  world_to_screen_position (skin, right_pos, &right_x, &right_y);
  circle_area (left_x, left_y, little_radius, &left);
  circle_area (right_x, right_y, little_radius, &right);
  circle_area ((left_x + right_x) / 2,
	       (left_y + right_y) / 2, big_radius, &center);
  gdk_rectangle_union (&left, &center, &tmp);
  gdk_rectangle_union (&tmp, &right, out);
}

static inline gboolean
is_in_radius(int x1, int y1, int x2, int y2, int maxdist)
{
  int dist_squared = (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2);
  int max_dist_squared = (maxdist + 1) * (maxdist + 1);
  return dist_squared <= max_dist_squared;
}
/*=====================================================================
 * Dragging the knobs.
 */
static int mouse_down_event(GtkWidget*        drawing_area,
                            GdkEventButton*   event,
			    GdamSpatialStereoSkin*         skin)
{
  int obj_x, obj_y;
  int mouse_x = event->x, mouse_y = event->y;
  int left_x, left_y, right_x, right_y;
  GdamSkin *this_skin = GDAM_SKIN (skin);
  GdamSpatialStereo *spatial;
  spatial = GDAM_SPATIAL_STEREO (this_skin->model);
  if (skin->drag_mode != DRAGGING_NONE)
    return TRUE;

  /*
   * Are they dragging the output device?
   */

  world_to_screen_position (skin,
			    spatial->left_channel_pos, &left_x, &left_y);
  world_to_screen_position (skin,
			    spatial->right_channel_pos, &right_x, &right_y);

  obj_x = (left_x + right_x) / 2;
  obj_y = (left_y + right_y) / 2;
  if (is_in_radius (mouse_x, mouse_y,
		    obj_x, obj_y, skin->output_center_radius))
    {
      skin->drag_mode = DRAGGING_CENTER;
      goto found_object;
    }

  obj_x = left_x;
  obj_y = left_y;
  if (is_in_radius (mouse_x, mouse_y, obj_x, obj_y, skin->output_side_radius))
    {
      skin->drag_mode = DRAGGING_LEFT;
      goto found_object;
    }

  obj_x = right_x;
  obj_y = right_y;
  if (is_in_radius (mouse_x, mouse_y, obj_x, obj_y, skin->output_side_radius))
    {
      skin->drag_mode = DRAGGING_RIGHT;
      goto found_object;
    }

  /*
   * Are they dragging the head?
   */

  world_to_screen_position (skin, spatial->left_ear_pos, &left_x, &left_y);
  world_to_screen_position (skin, spatial->right_ear_pos, &right_x, &right_y);

  obj_x = left_x;
  obj_y = left_y;
  if (is_in_radius (mouse_x, mouse_y, obj_x, obj_y, skin->head_side_radius))
    {
      skin->drag_mode = DRAGGING_HEAD_LEFT;
      goto found_object;
    }

  obj_x = right_x;
  obj_y = right_y;
  if (is_in_radius (mouse_x, mouse_y, obj_x, obj_y, skin->head_side_radius))
    {
      skin->drag_mode = DRAGGING_HEAD_RIGHT;
      goto found_object;
    }

  obj_x = (left_x + right_x) / 2;
  obj_y = (left_y + right_y) / 2;
  if (is_in_radius (mouse_x, mouse_y, obj_x, obj_y, skin->head_center_radius))
    {
      skin->drag_mode = DRAGGING_HEAD_CENTER;
      goto found_object;
    }

  return TRUE;

found_object:
  skin->drag_offset_x = obj_x - mouse_x;
  skin->drag_offset_y = obj_y - mouse_y;
  return TRUE;
}

static int mouse_up_event(GtkWidget*        drawing_area,
                          GdkEventButton*   event,
			  GdamSpatialStereoSkin*         skin)
{
  skin->drag_mode = DRAGGING_NONE;
  return FALSE;
}

static void update_side_position(gdouble* dragged, gdouble* other,
                                 gboolean anchor_other,
				 gdouble* new_pos)
{
  int i;
  if (anchor_other)
    for (i = 0; i < 3; i++)
      dragged[i] = new_pos[i];
  else
    for (i = 0; i < 3; i++)
      {
	other[i] -= (new_pos[i] - dragged[i]);
	dragged[i] = new_pos[i];
      }
}

static int mouse_move_event(GtkWidget*        drawing_area,
                            GdkEventMotion*   event,
			    GdamSpatialStereoSkin*         skin)
{
  GdkRectangle old_channel_area, old_head_area;
  GdkRectangle channel_area, head_area;
  GdkRectangle total_area;
  gdouble world_position[3];
  GdamSkin *this_skin = GDAM_SKIN (skin);
  GdamSpatialStereo *spatial;
  guint state;
  int x, y;
  gboolean redraw_output = FALSE, redraw_head = FALSE;

  spatial = GDAM_SPATIAL_STEREO (this_skin->model);

  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      x = event->x;
      y = event->y;
      state = event->state;
    }

  /* We aren't interested in this event afterall. */
  if ((state & GDK_BUTTON1_MASK) != GDK_BUTTON1_MASK)
    return TRUE;

  /* They are holding down the button, but not dragging anything. */
  if (skin->drag_mode == DRAGGING_NONE)
    return TRUE;

  /* Get the area we need to clear: */
  get_rendered_area (skin,
		     spatial->left_channel_pos,
		     spatial->right_channel_pos,
		     skin->output_side_radius,
		     skin->output_center_radius, &old_channel_area);
  get_rendered_area (skin,
		     spatial->left_ear_pos,
		     spatial->right_ear_pos,
		     skin->head_center_radius,
		     skin->head_side_radius, &old_head_area);

  /* Something gets a new position */
  screen_to_world (skin,
		   x + skin->drag_offset_x,
		   y + skin->drag_offset_y, world_position);

  switch (skin->drag_mode)
    {
    case DRAGGING_LEFT:
      {
	update_side_position (spatial->left_channel_pos,
			      spatial->right_channel_pos,
			      skin->is_shift_dragging, world_position);
	redraw_output = 1;
	break;
      }
    case DRAGGING_RIGHT:
      {
	update_side_position (spatial->right_channel_pos,
			      spatial->left_channel_pos,
			      skin->is_shift_dragging, world_position);
	redraw_output = 1;
	break;
      }
    case DRAGGING_CENTER:
      {
	int i;
	for (i = 0; i < 3; i++)
	  {
	    gdouble delta, old_pos;
	    old_pos = (spatial->right_channel_pos[i]
		       + spatial->left_channel_pos[i]) / 2;
	    delta = world_position[i] - old_pos;
	    spatial->left_channel_pos[i] += delta;
	    spatial->right_channel_pos[i] += delta;
	  }
	redraw_output = 1;
	break;
      }
    case DRAGGING_HEAD_LEFT:
      {
	update_side_position (spatial->left_ear_pos,
			      spatial->right_ear_pos,
			      skin->is_shift_dragging, world_position);
	redraw_head = 1;
	break;
      }
    case DRAGGING_HEAD_RIGHT:
      {
	update_side_position (spatial->right_ear_pos,
			      spatial->left_ear_pos,
			      skin->is_shift_dragging, world_position);
	redraw_head = 1;
	break;
      }
    case DRAGGING_HEAD_CENTER:
      {
	int i;
	for (i = 0; i < 3; i++)
	  {
	    gdouble delta, old_pos;
	    old_pos = (spatial->right_ear_pos[i]
		       + spatial->left_ear_pos[i]) / 2;
	    delta = world_position[i] - old_pos;
	    spatial->left_ear_pos[i] += delta;
	    spatial->right_ear_pos[i] += delta;
	  }
	redraw_head = 1;
	break;
      }
    default:
      g_assert_not_reached ();
    }

  reset_spatial_parameters (skin);

  if (redraw_head == FALSE && redraw_output == FALSE)
    return TRUE;

  if (redraw_output)
    {
      GdkRectangle tmp;
      get_rendered_area (skin,
			 spatial->left_channel_pos,
			 spatial->right_channel_pos,
			 skin->output_side_radius,
			 skin->output_center_radius, &tmp);
      gdk_rectangle_union (&tmp, &old_channel_area, &channel_area);
    }
  if (redraw_head)
    {
      GdkRectangle tmp;
      get_rendered_area (skin,
			 spatial->left_ear_pos,
			 spatial->right_ear_pos,
			 skin->head_center_radius,
			 skin->head_side_radius, &tmp);
      gdk_rectangle_union (&tmp, &old_head_area, &head_area);
    }
  if (redraw_head && redraw_output)
    gdk_rectangle_union (&head_area, &channel_area, &total_area);
  else if (redraw_head)
    total_area = head_area;
  else if (redraw_output)
    total_area = channel_area;
  else
    g_assert_not_reached ();
  redraw_area (skin,
	       total_area.x,
	       total_area.y, total_area.width, total_area.height);
  return TRUE;
}


static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamSpatialStereoSkin* skin)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 250, 250);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_press_event",
		      GTK_SIGNAL_FUNC (mouse_down_event), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_release_event",
		      GTK_SIGNAL_FUNC (mouse_up_event), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (mouse_move_event), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event", GTK_SIGNAL_FUNC (expose_event), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (configure_event), skin);
  gtk_widget_add_events (GTK_WIDGET (drawing_area),
			 GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);
}


void spatial_stereo_connect_func(const gchar*     handler,
                                 GtkObject*       object,
			         const gchar*     signal_name,
				 const gchar*     signal_data,
			         GtkObject*       connect_object,
			         gboolean         after,
			         gpointer         user_data);

/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */



static gboolean gdam_spatial_stereo_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
      return TRUE;
    }
  connect_drawing_area (GTK_DRAWING_AREA (object),
			GDAM_SPATIAL_STEREO_SKIN (skin));
  return TRUE;
}

static gboolean gdam_spatial_stereo_skin_build_new_model
                                             (GdamSkin*      skin,
                                              GdamChannel*   channel,
					      int 	     num_options,
					      char**         options)
{
  GdamModel *model;
  model = GDAM_MODEL (gtk_type_new (GDAM_TYPE_SPATIAL_STEREO));
  model->channel = channel;
  skin->model = model;
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_spatial_stereo_drawing_area",
           gdam_spatial_stereo_drawing_area_handler		}	};
static void gdam_spatial_stereo_skin_class_init(GdamSkinClass* skin_class)
{
  skin_class->build_new_model = gdam_spatial_stereo_skin_build_new_model;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}

