/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamvolumemeterskin.h"
#include "../gdamskinutil.h"
#include <gtk/gtk.h>
#include <common/gdamsample.h>
#include <component/plugins/gdammeter.h>
#include <common/gdamplugin.h>


static void gdam_volume_meter_skin_init(GdamVolumeMeterSkin* volume_meter_skin)
{
}
static void gdam_volume_meter_skin_class_init(
			GdamVolumeMeterSkinClass* volume_meter_skin_class);

GtkType gdam_volume_meter_skin_get_type()
{
  static GtkType volume_meter_skin_type = 0;
  if (!volume_meter_skin_type)
    {
      static const GtkTypeInfo volume_meter_skin_info = {
	"GdamVolumeMeterSkin",
	sizeof (GdamVolumeMeterSkin),
	sizeof (GdamVolumeMeterSkinClass),
	(GtkClassInitFunc) gdam_volume_meter_skin_class_init,
	(GtkObjectInitFunc) gdam_volume_meter_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      volume_meter_skin_type = gtk_type_unique (parent,
						&volume_meter_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return volume_meter_skin_type;
}
 
static void got_args(int num_args, GdamArg* args, gpointer user_data)
{
  GdamVolumeMeterSkin *skin;
  int min;
  int max;
  int length;
  int *array;
  float fmin;
  float fmax;
  g_return_if_fail (num_args > 0);
  g_return_if_fail (num_args < 2);
  length = GDAM_ARG_INT_ARRAY (args).length;
  g_return_if_fail (length == 2 || length == 4);
  array = GDAM_ARG_INT_ARRAY (args).array;
  if (length == 4)
    {
      min = MIN (array[0], array[2]);
      max = MAX (array[1], array[3]);
    }
  else
    {
      min = array[0];
      max = array[1];
    }
  fmin = (float) min / GdamFixed_ONE;
  fmax = (float) max / GdamFixed_ONE;
  skin = GDAM_VOLUME_METER_SKIN (user_data);
  skin->last_vol = skin->target_vol;
  skin->target_vol = fmax - fmin;
  gdam_arg_set (GTK_OBJECT (GDAM_SKIN (skin)->model), "meterlevel",
		skin->target_vol);
  if (skin->progress_bar)
    {
      gtk_progress_bar_update (skin->progress_bar,
			       CLAMP (skin->target_vol, 0.0, 1.0));
    }
  if (skin->left_progress_bar && length == 4)
    {
      gfloat left_vol = (array[1] - array[0]);
      left_vol /= GdamFixed_ONE;
      gtk_progress_bar_update (skin->left_progress_bar,
			       CLAMP (left_vol, 0.0, 1.0));
    }
  if (skin->right_progress_bar && length == 4)
    {
      gfloat right_vol = (array[3] - array[2]);
      right_vol /= GdamFixed_ONE;
      gtk_progress_bar_update (skin->right_progress_bar,
			       CLAMP (right_vol, 0.0, 1.0));
    }

  skin->request_pending = 0;
  gtk_object_unref (GTK_OBJECT (skin));
}

/* XXX: if we set ``request_pending'' to FALSE, it would be more accurate
        but we also don't have a good way of flagging that error condition. */
static void didnt_get_args(gpointer user_data)
{
  g_warning ("error getting volume from volume meter");
  gtk_object_unref (GTK_OBJECT (user_data));
}

static void zero_progress_bars(GdamVolumeMeterSkin* volume_skin)
{
  //g_message("zeroing progress bars.");
  gdam_arg_set (GTK_OBJECT (GDAM_SKIN (volume_skin)->model), "meterlevel",
		0.0);
  if (volume_skin->progress_bar)
    gtk_progress_bar_update (volume_skin->progress_bar, 0);
  if (volume_skin->left_progress_bar)
    gtk_progress_bar_update (volume_skin->left_progress_bar, 0);
  if (volume_skin->right_progress_bar)
    gtk_progress_bar_update (volume_skin->right_progress_bar, 0);
}

static gboolean meter_timer_callback(gpointer data)
{
  GdamRequest *request;
  GdamSkin *skin = GDAM_SKIN (data);
  GdamVolumeMeterSkin *vskin = GDAM_VOLUME_METER_SKIN (data);
  int num_args = 1;
  int source_id;
  int cycle_length = 4;
  const char *name = "range";
  if (!skin->model)
    return TRUE;
  source_id = skin->model->source_id;
  if (!source_id)
    {
      if (vskin->had_source_id)
	{
	  vskin->had_source_id = FALSE;
	  zero_progress_bars (vskin);
	}
      vskin->timer_count = 0;
      return TRUE;
    }
  vskin->had_source_id = TRUE;
  vskin->timer_count++;
  if (vskin->timer_count == cycle_length)
    vskin->timer_count = 0;
  if (vskin->timer_count == 0 && !vskin->request_pending)
    {
      vskin->request_pending = TRUE;
      gtk_object_ref (GTK_OBJECT (vskin));
      request = gdam_request_get_args (source_id,
				       num_args,
				       &name,
				       got_args, didnt_get_args, vskin);
      gdam_channel_append_request (skin->model->channel, request);
    }
  else
    {
      float vol = vskin->last_vol;
      vol += (vskin->target_vol - vskin->last_vol) * vskin->timer_count
	/ cycle_length;
      if (vskin->progress_bar)
	gtk_progress_bar_update (vskin->progress_bar, CLAMP (vol, 0.0, 1.0));
      gdam_arg_set (GTK_OBJECT (skin->model), "meterlevel", vol);
    }
  return TRUE;
}

static void do_add_timeout(GdamVolumeMeterSkin* skin)
{
  if (skin->timer_id != 0)
    return;

  skin->timer_id = gtk_timeout_add (200,
				    (GtkFunction) meter_timer_callback, skin);
}

static gboolean gdam_volume_meter_add_timeout_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  do_add_timeout (GDAM_VOLUME_METER_SKIN (skin));
  return TRUE;
}

static gboolean gdam_volume_meter_vertical_progress_bar_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_progress_bar_set_orientation (GTK_PROGRESS_BAR (object),
				    GTK_PROGRESS_BOTTOM_TO_TOP);
  if (GDAM_VOLUME_METER_SKIN (skin)->progress_bar)
    {
      g_warning ("Sorry, only one progres bar per meter for now...");
      return TRUE;
    }
  GDAM_VOLUME_METER_SKIN (skin)->progress_bar = GTK_PROGRESS_BAR (object);
  do_add_timeout (GDAM_VOLUME_METER_SKIN (skin));
  return TRUE;
}

static gboolean gdam_volume_meter_progress_bar_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkProgressBar *bar = GTK_PROGRESS_BAR (object);
  char *name = handler_names[0];
  GdamVolumeMeterSkin *vol_skin = GDAM_VOLUME_METER_SKIN (skin);
  GtkProgressBar **dest;

  if (strcmp (name, "gdam_volume_meter_left_progress_bar") == 0)
    {
      dest = &vol_skin->left_progress_bar;
    }
  else if (strcmp (name, "gdam_volume_meter_right_progress_bar") == 0)
    {
      dest = &vol_skin->right_progress_bar;
    }
  else if (strcmp (name, "gdam_volume_meter_progress_bar") == 0)
    {
      dest = &vol_skin->progress_bar;
    }
  else
    {
      g_message ("progress_bar handler mismatch [%s]", name);
      return TRUE;
    }

  if (*dest != NULL)
    {
      g_warning ("Sorry, only one progres bar per meter for now...");
      return TRUE;
    }

  g_return_val_if_fail (bar != NULL, TRUE);

  *dest = bar;
  do_add_timeout (GDAM_VOLUME_METER_SKIN (skin));
  return TRUE;
}


static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_volume_meter_progress_bar",
	   gdam_volume_meter_progress_bar_handler},
	{ "gdam_volume_meter_left_progress_bar",
	   gdam_volume_meter_progress_bar_handler},
	{ "gdam_volume_meter_right_progress_bar",
	   gdam_volume_meter_progress_bar_handler},
	{ "gdam_volume_meter_vertical_progress_bar",
	   gdam_volume_meter_vertical_progress_bar_handler},
	{ "gdam_volume_meter_add_timeout",
	   gdam_volume_meter_add_timeout_handler}
};

static void gdam_volume_meter_skin_destroy(GtkObject* object)
{
  GdamVolumeMeterSkin *skin = GDAM_VOLUME_METER_SKIN (object);
  if (skin->timer_id)
    gtk_timeout_remove (skin->timer_id);
}

static void gdam_volume_meter_skin_class_init(GdamVolumeMeterSkinClass* klass)
{
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  skin_class->model_type = GDAM_TYPE_METER;
  GTK_OBJECT_CLASS (skin_class)->destroy = gdam_volume_meter_skin_destroy;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}
