/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamxmmsfilterskin.h"
#include "../../component/plugins/gdamxmmsfilter.h"
#include <common/gdamscandir.h>
#include <gtk/gtkitem.h>
#include <gtk/gtksignal.h>
#include <gtk/gtklistitem.h>
#include <gtk/gtkentry.h>
#include <gtk/gtklabel.h>
#include <gtk/gtklist.h>
#include <gtk/gtkcombo.h>
#include <skin/gdamskinutil.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#define VISPLUGIN_BASE	"lib/xmms/Visualization"

static GSList* make_slist_of_xmms_plugin_filenames()
{
  static GSList *list;
  GSList *tmp;
  static int has_inited = FALSE;
  if (has_inited)
    return list;

  list = NULL;
  tmp = gdam_scandir_for_shared_libs ("/usr/local/" VISPLUGIN_BASE);
  list = g_slist_concat (list, tmp);
  tmp = gdam_scandir_for_shared_libs ("/usr/" VISPLUGIN_BASE);
  list = g_slist_concat (list, tmp);

  return list;
}

static GList* make_list_of_xmms_plugins_as_items()
{
  GSList *filenames;
  GList *rv = NULL;
  for (filenames = make_slist_of_xmms_plugin_filenames ();
       filenames != NULL; filenames = filenames->next)
    {
      GtkWidget *item;
      const char *filename = filenames->data;
      item = gtk_list_item_new_with_label (filename);
      gtk_widget_show (item);
      rv = g_list_prepend (rv, item);
    }
  return g_list_reverse (rv);
}

static gboolean do_timed_connect(gpointer data)
{
  gpointer *tmp = data;
  GdamXmmsFilter *filter = GDAM_XMMS_FILTER (tmp[0]);
  char *path = tmp[1];
  gdam_xmms_filter_connect (filter, path);
  gtk_object_unref (GTK_OBJECT (filter));
  g_free (path);
  return FALSE;
}

static void chose_plugin(GtkEntry* entry, GdamXmmsFilterSkin* xmms_skin)
{
  static int seqno = 1;
  char *path;
  GdamXmmsFilter *xmms_filter;
  GdamSkin *skin;
  int pid;
  char *text;
  int pipe_fds[2];
  gpointer *tmp;
  skin = GDAM_SKIN (xmms_skin);
  text = g_strdup (gtk_entry_get_text (entry));

  xmms_filter = GDAM_XMMS_FILTER (skin->model);
  if (xmms_filter->is_connecting || xmms_filter->is_connected)
    return;

  /* make a temporary unix-domain socket name to rendevous at */
  path = g_strdup_printf ("/tmp/xmmstmp-%d-%d", getpid (), seqno++);

  if (pipe (pipe_fds) < 0)
    {
      g_warning ("couldn't make necessary pipe");
      return;
    }

  /* spawn gdamxmmspluginhelper */
  pid = fork ();
  if (pid < 0)
    {
      g_warning ("xmms_filter: couldn't fork: %s", strerror (errno));
      return;
    }
  if (pid == 0)
    {
      char read_fd_str[128];
      sprintf (read_fd_str, "%d", pipe_fds[0]);
      close (pipe_fds[1]);
      fcntl (pipe_fds[0], F_SETFL, 0);
      /* child process */
      execlp ("gdamxmmsvishelper", "gdamxmmsvishelper",
	      "--unix", path,
	      "--plugin", text, "--control-fd", read_fd_str, NULL);
      g_warning ("couldn't find gdamxmmsvishelper on path");
      _exit (127);
    }

  /* do the connect */
  tmp = g_new (gpointer, 2);
  tmp[0] = xmms_filter;
  tmp[1] = path;
  gtk_object_ref (GTK_OBJECT (xmms_filter));
  g_timeout_add (1000, do_timed_connect, tmp);
  xmms_skin->control_fd = pipe_fds[1];

  /* clean up */
  g_free (text);
}

static void do_launch(GdamXmmsFilterSkin* xmms_skin){
	g_return_if_fail(xmms_skin->combo != NULL);
	chose_plugin(GTK_ENTRY(xmms_skin->combo->entry), xmms_skin);
}
static gboolean gdam_xmms_filter_launch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_launch), (GtkObject *) skin);
  return TRUE;
}

static void init_box(GdamXmmsFilterSkin* xmms_skin)
{
  GtkWidget *widget;
  GtkBox *box = xmms_skin->box;
  GtkCombo *combo;
  GList *list;
  widget = gtk_label_new ("Select xmms plugin:");
  g_return_if_fail (box != NULL);
  gtk_box_pack_start_defaults (box, widget);
  gtk_widget_show (widget);

  widget = gtk_combo_new ();
  combo = GTK_COMBO (widget);
  xmms_skin->combo = combo;
  list = make_list_of_xmms_plugins_as_items ();
  gtk_list_append_items ((GtkList *) combo->list, list);
  gtk_signal_connect (GTK_OBJECT (combo->entry),
		      "activate", GTK_SIGNAL_FUNC (chose_plugin), xmms_skin);

  gtk_box_pack_start_defaults (box, widget);
  gtk_widget_show (widget);
}

static gboolean gdam_xmms_filter_box_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamXmmsFilterSkin *xmms_skin;
  xmms_skin = GDAM_XMMS_FILTER_SKIN (skin);
  g_return_val_if_fail (GTK_IS_BOX (object), TRUE);
  xmms_skin->box = GTK_BOX (object);
  init_box (xmms_skin);
  return TRUE;
}

static void do_control_generic(GdamXmmsFilterSkin* xmms, char c)
{
  if (xmms->control_fd < 0)
    return;
  while (write (xmms->control_fd, &c, 1) < 0)
    if (errno != EINTR)
      {
	g_warning ("cannot send control signal: %s", strerror (errno));
	break;
      }
}

static void do_about(GdamXmmsFilterSkin* xmms)
{
  do_control_generic (xmms, 'a');
}
static gboolean gdam_xmms_filter_about_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_about), (GtkObject *) skin);
  return TRUE;
}
static void do_configure(GdamXmmsFilterSkin* xmms)
{
  do_control_generic (xmms, 'c');
}
static gboolean gdam_xmms_filter_configure_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_configure),
			     (GtkObject *) skin);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_xmms_filter_box",
	   gdam_xmms_filter_box_handler },
	{ "gdam_xmms_filter_about",
	   gdam_xmms_filter_about_handler },
	{ "gdam_xmms_filter_launch",
	   gdam_xmms_filter_launch_handler },
	{ "gdam_xmms_filter_configure",
	   gdam_xmms_filter_configure_handler },
};

static void gdam_xmms_filter_skin_init(GdamXmmsFilterSkin* xmms_filter_skin)
{
  xmms_filter_skin->control_fd = -1;
}
static void gdam_xmms_filter_skin_class_init(GdamXmmsFilterSkinClass* klass)
{
  GdamSkinClass *skin_class;
  skin_class = GDAM_SKIN_CLASS (klass);
  skin_class->model_type = GDAM_TYPE_XMMS_FILTER;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}

GtkType gdam_xmms_filter_skin_get_type()
{
  static GtkType xmms_filter_skin_type = 0;
  if (!xmms_filter_skin_type)
    {
      static const GtkTypeInfo xmms_filter_skin_info = {
	"GdamXmmsFilterSkin",
	sizeof (GdamXmmsFilterSkin),
	sizeof (GdamXmmsFilterSkinClass),
	(GtkClassInitFunc) gdam_xmms_filter_skin_class_init,
	(GtkObjectInitFunc) gdam_xmms_filter_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      xmms_filter_skin_type = gtk_type_unique (parent,
					       &xmms_filter_skin_info);
    }
  return xmms_filter_skin_type;
}
